Require Import CheneyLemma.
Require Import ScanField.
Require Import Heap.
Require Import HeapDom.


Set Implicit Arguments.


(* code address constants *)
Definition CHENEYALLOC := 10.
Notation CHENEYREAD1 := 11.
Notation CHENEYREAD2 := 12.
Notation CHENEYWRITE1 := 13.
Notation CHENEYWRITE2 := 14.

Definition CHENEYDOALLOC := 15.
Definition CHENEYPOSTGC := 16.

Definition CHENEYLOOP := 17.
Definition CHENEYLOOP2 := 18.
Definition CHENEYENTER := 19.
Definition CHENEYLOOPHEADER := 20.
Definition CHENEYRETURN := 21.

Definition INFLOOP := 22.




(* local register conventions *)
Definition rscan := t3.
Definition rroot := t0.
Definition rgcInfo := s7.
Definition raSave := v1.
Definition raSave2 := t9.


(* code blocks *)
Definition cheneyLoop :=
  beq rscan rfree CHENEYRETURN;;
  addiu a0 rscan 0;
  jal SCANFIELD CHENEYLOOP2.

Definition cheneyLoop2 :=
  addiu a0 rscan WORDSIZE;
  addiu rscan rscan OBJSIZE;
  jal SCANFIELD CHENEYLOOP.

Definition cheneyReturn :=
  addiu ra raSave 0;
  jr ra.

Definition cheneyEnter :=
  (* swap spaces, init rtoStart and rtoEnd *)
  lw rtoStart 0 rgcInfo;
  lw rtemp 8 rgcInfo;
  sw rtemp 0 rgcInfo;
  sw rtoStart 8 rgcInfo;
  lw rtoEnd 4 rgcInfo;
  lw rtemp 12 rgcInfo;
  sw rtemp 4 rgcInfo;
  (* store final space limit later,
     so we can reuse this memory for scanField *)

  (* init rfree and rscan *)
  addiu rfree rtoStart 0;
  addiu rscan rtoStart 0;
  
  (* save ra *)
  addiu raSave ra 0;

  (* scan the root *)
  sw rroot 12 rgcInfo;
  addiu a0 rgcInfo 12;
  jal SCANFIELD CHENEYLOOPHEADER.


Definition cheneyLoopHeader :=
  (* restore updated root *)
  lw rroot 12 rgcInfo;

  (* finish space swap *)
  sw rtoEnd 12 rgcInfo;
  
  j CHENEYLOOP.


Definition cheneyAlloc :=
  addiu raSave2 ra 0;
  lw rfree 16 rgcInfo;
  lw rtemp 12 rgcInfo;
  bne rfree rtemp CHENEYDOALLOC;;
  jal CHENEYENTER CHENEYPOSTGC.


Definition cheneyDoAlloc :=
  addiu v0 zero 1;
  sw v0 0 rfree;
  sw v0 4 rfree;
  addiu v0 rfree 0;
  addiu rfree rfree 8;
  sw rfree 16 rgcInfo;
  addiu ra raSave2 0;
  jr ra.


Definition cheneyPostGC :=
  lw rtemp 12 rgcInfo;
  bne rfree rtemp CHENEYDOALLOC;;
  j INFLOOP.


Definition cheneyRead k :=
  lw v0 k rroot;
  jr ra.

Definition cheneyWrite k :=
  sw a0 k rroot;
  jr ra.

Definition infLoop :=
  j INFLOOP.


(* specs *)

Definition cheneyLoopStOk phi G W M F A s :=
  set_eqv G (rangeObjs (rget s rscan) (rget s rfree)) /\
  rget s rtoStart <= rget s rscan /\
  aligned8 (rget s rscan) (rget s rfree) /\
  scanFieldStOk phi W M F A s.

Definition cheneyLoopPre s :=
  exists phi, exists G, exists W, exists M, exists F,
    cheneyLoopStOk phi G W M F (objHp (set_union W M) G ** top) s.


Notation cheneyLoopRegGuar :=
  (fun s s' => regPres 
    (regNotInList (ra::rscan::rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s' /\
    rget s' ra = rget s raSave).


Definition cheneyLoopGuar s s' :=
  (forall phi G W M F A,
    cheneyLoopStOk phi G W M F (objHp (set_union W M) G ** A) s ->
    exists phi', exists W', exists M', exists F', exists G',
      set_eqv W (set_union M' W') /\
      set_eqv F (set_union G' F') /\
      iso_set phi' M' G' /\
      (forall x', set_In x' G' -> exists x, set_In x G /\ reachable (heapOf s') x x') /\
      aligned8 (rget s rfree) (rget s' rfree) /\
      cheneyLoopStOk (map_union phi phi') empty_set W' (set_union M M') F'
      ((foralls x in G, objRespMap (map_union phi phi') s x) ** 
	(foralls x in G', objIsFwded phi phi' s x) ** A) s') /\
  cheneyLoopRegGuar s s'.

Definition ch2Step s :=
  rset (rset (rset s a0 (4+rget s rscan)) rscan (8+rget s rscan)) ra CHENEYLOOP.

Definition cheneyLoop2Pre s :=
  scanFieldPre (ch2Step s) /\
  (forall s', scanFieldGuar (ch2Step s) s' -> cheneyLoopPre s').

Definition cheneyLoop2Guar s s'' :=
  exists s', 
    scanFieldGuar (ch2Step s) s' /\
    cheneyLoopGuar s' s''.


Definition cheneyReturnPre : spred := fun s => True.

Definition cheneyReturnGuar s s' :=
  s' = rset s ra (rget s raSave).


Definition cheneyLoopHeaderPre s :=
  heapOf s ||- (lexists x, 12 + rget s rgcInfo |-> x) ** 
    (fun h => cheneyLoopPre (h, rfileOf s)).


Definition cheneyLoopHeaderGuar s s'' :=
  (forall x A,
    heapOf s ||- 12 + rget s rgcInfo |-> x ** A ->
      exists s',
	heapOf s' ||- 12 + rget s rgcInfo |-> (rget s rtoEnd) ** A /\
	  rfileEqv s' (rset s rroot x) /\
	  cheneyLoopGuar s' s'') /\
  regPres (regNotInList (rroot::ra::rscan::rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s'' /\
  rget s'' ra = rget s raSave.


Definition cheneyStOk fromSt fromEnd toSt toEnd A free root gcInfo h :=
  let objs := rangeObjs toSt free in
    !(aligned8 toSt free /\
      aligned8 free toEnd /\
      okFieldVal objs root /\
      fromEnd - fromSt = toEnd - toSt /\
      fromSt <= fromEnd) **
    (objHp objs objs && heap_eqv h) ** buffer free toEnd **
    buffer fromSt fromEnd **
    gcInfo |->, (fromSt::fromEnd::toSt::toEnd::nil) ** A.

Definition cheneyEnterPre s :=
  exists fromSt, exists fromEnd, exists toSt, exists toEnd, exists h,
    s |||- cheneyStOk fromSt fromEnd toSt toEnd top 
             (rget s rfree) (rget s rroot) (rget s rgcInfo) h.

Definition cheneyEnterGuar s s' :=
  (forall toSt toEnd fromSt fromEnd A h,
    s |||- cheneyStOk toSt toEnd fromSt fromEnd A 
             (rget s rfree) (rget s rroot) (rget s rgcInfo) h ->
    exists phi, exists objs, exists h',
      s' |||- cheneyStOk fromSt fromEnd toSt toEnd A
                (rget s' rfree) (rget s' rroot) (rget s' rgcInfo) h' /\
      h ||- minObjHp (rget s rroot::nil) objs ** top /\
      iso_set phi objs (rangeObjs toSt (rget s' rfree)) /\
      mapX phi (rget s rroot) = Some (rget s' rroot) /\
      h' ||- (foralls x in rangeObjs toSt (rget s' rfree), objIsFwdedH phi h x)) /\
  calleeSavedOk s s' /\
  rget s raSave2 = rget s' raSave2.


Definition mview vs rf rr :=
  lexists fromSt, lexists fromEnd, lexists toSt, lexists toEnd, lexists free,
  let objs := rangeObjs toSt free in
    !(~In rgcInfo rr /\
      aligned8 toSt free /\
      aligned8 free toEnd /\
      (forall r, In r rr -> rget vs r = RFile.get rf r /\ okFieldVal objs (rget vs r)) /\
      fromEnd - fromSt = toEnd - toSt /\
      fromSt <= fromEnd) **
    (objHp objs objs && heap_eqv (heapOf vs)) ** 
    buffer free toEnd ** buffer fromSt fromEnd **
    RFile.get rf rgcInfo |->, (fromSt::fromEnd::toSt::toEnd::free::nil).


Require Import VGC.

Module CheneyGCGuar <: BASIC_MGI_ARG.
  Definition gcGuar (s s':state) (rr:list rnum) := 
    forall objs h rr',
      incl rr' rr ->
      s |||- (minObjHp (List.map (rget s) rr') objs && heap_eqv h) ** top ->
	exists h',
	  s' |||- heap_eqv h' ** top /\
	  isoState rr' (h, rfileOf s) (h', rfileOf s').

End CheneyGCGuar.

Module CMGI := BasicMGI CheneyGCGuar.


Definition cheneyDoAllocPre s := 
  s |||- rget s rfree |->__ ** 16 + rget s rgcInfo |->_ ** top.

Definition cheneyDoAllocGuar s s' :=
  (forall A,
    s |||- rget s rfree |->__ ** 16 + rget s rgcInfo |->_ ** A ->
      s' |||- rget s' v0 |->, (1::1::nil) ** 
	16 + rget s rgcInfo |-> 8 + rget s rfree ** A) /\
  rfileEqv s' (rset (rset (rset s v0 (rget s rfree)) rfree (8 + rget s rfree)) ra (rget s raSave2)).


Definition cheneyPostGCPre s :=
  exists toEnd,
    s |||- 12 + rget s rgcInfo |-> toEnd ** top /\
    (rget s rfree <> toEnd -> cheneyDoAllocPre s).


Definition cheneyPostGCGuar s s' := 
  forall toEnd,
    s |||- 12 + rget s rgcInfo |-> toEnd ** top ->
      if beq_nat (rget s rfree) toEnd then
	False
      else
	cheneyDoAllocGuar (rset s rtemp toEnd) s'.


Definition cheneyAllocPre : state -> Prop := VGC.allocPre mview rroot.

Definition cheneyAllocGuar : state -> state -> Prop :=
  VGC.allocGuar mview rroot CMGI.gcGuar CMGI.presReg 1.


Definition cheneyReadPre : nat -> state -> Prop := VGC.readPre mview rroot.
Definition cheneyReadGuar : nat -> state -> state -> Prop := 
  VGC.readGuar mview rroot CMGI.gcGuar CMGI.presReg.
Notation cheneyReadSpec := (fun k => bspeci (cheneyReadPre k) (cheneyReadGuar k)).

Definition cheneyWritePre : nat -> state -> Prop := VGC.writePre mview rroot.
Definition cheneyWriteGuar : nat -> state -> state -> Prop := 
  VGC.writeGuar mview rroot CMGI.gcGuar CMGI.presReg.
Notation cheneyWriteSpec := (fun k => bspeci (cheneyWritePre k) (cheneyWriteGuar k)).

(* verifying the implementations *)

Definition P : chspec := fun l => 
  if beq_nat l SCANFIELD then
    SomeT (bspeci scanFieldPre scanFieldGuar)
  else if beq_nat l CHENEYLOOP then
    SomeT (bspeci cheneyLoopPre cheneyLoopGuar)
  else if beq_nat l CHENEYLOOP2 then
    SomeT (bspeci cheneyLoop2Pre cheneyLoop2Guar)
  else if beq_nat l CHENEYRETURN then
    SomeT (bspeci cheneyReturnPre cheneyReturnGuar)
  else if beq_nat l CHENEYLOOPHEADER then
    SomeT (bspeci cheneyLoopHeaderPre cheneyLoopHeaderGuar)
  else if beq_nat l CHENEYENTER then
    SomeT (bspeci cheneyEnterPre cheneyEnterGuar)
  else if beq_nat l CHENEYDOALLOC then
    SomeT (bspeci cheneyDoAllocPre cheneyDoAllocGuar)
  else if beq_nat l CHENEYALLOC then
    SomeT (bspeci cheneyAllocPre cheneyAllocGuar)
  else if beq_nat l CHENEYPOSTGC then
    SomeT (bspeci cheneyPostGCPre cheneyPostGCGuar)
  else if beq_nat l INFLOOP then
    SomeT (bspeci (fun s => True) (fun s s' => False))
  else if beq_nat l CHENEYREAD1 then
    SomeT (cheneyReadSpec 0)
  else if beq_nat l CHENEYREAD2 then
    SomeT (cheneyReadSpec 4)
  else if beq_nat l CHENEYWRITE1 then
    SomeT (cheneyWriteSpec 0)
  else if beq_nat l CHENEYWRITE2 then
    SomeT (cheneyWriteSpec 4)
  else
    NoneT _.

Ltac Psimpl := 
  unfold P; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.


(* showing the blocks match their specs *)

Lemma cheneyEnterGuarPresRA : forall s s' : state,
   cheneyEnterGuar s s' -> rget s ra = rget s' ra.
Proof.
  intros s s' [_ [CSO _]].
  apply CSO.
  unfold calleeSaved; simpl; auto.
Qed.


Lemma gcGuarRefl : forall vs rr,
  CMGI.gcGuar vs vs rr.
Proof.
  intros vs rr objs h rr' II Vhp.
  exists h; constructor.
  hypReplace Vhp (sconjDistrib2 Vhp).
  destruct (aconjElim Vhp) as [_ X]; assumption.

  destruct (sconjElim Vhp) as [ha [hb [_ [HA _]]]].
  assert (h ||- minObjHp (List.map (rget vs) rr') objs).
  eapply heapEqvImpl; eauto.
  clear Vhp HA hb ha.
  unfold minObjHp in H; destruct (aconjElim H) as [_ Hp]; clear H.
  unfold okObjHp in Hp; linLogElim Hp.
  eapply isoStateRefl.
  2:instantiate (1 := objs).
  assumption.
  intros r I.
  hypReplace H1 (H1 (rget vs r)).
  apply H1.
  clear H1 II Hp h objs.
  induction rr'; simpl; simpl in I.
  assumption.
  destruct I as [I | I]; [left | right].
  subst a; reflexivity.
  apply IHrr'; assumption.

Qed.


Ltac sllImpl :=
  let h := fresh in
  let x := fresh in
  let Hp := fresh in
    (unfold objRespMap; unfold fieldRespMap;
     unfold objIsFwded; unfold fieldIsFwded;
     cbv beta; intros h x Hp; linLogElim Hp; 
     linLogIntro; sconjMatch).


Lemma objsHeapEqvEqv : forall h S h1, 
  h ||- ((foralls x in S, x |->__) && heap_eqv h1) ** top -> forall h2,
  h ||- ((foralls x in S, x |->__) && heap_eqv h2) ** top ->
  heap_eqv h1 h2.
Proof.
  intros h S h1 Hp h2 Hp'.
  eapply heap_domEqvElim.
  instantiate (1 := h).
  destruct (aconjElim (sconjDistrib2 Hp)); eapply heap_eqvSubHeapL; eauto.
  destruct (aconjElim (sconjDistrib2 Hp')); eapply heap_eqvSubHeapL; eauto.
  assert (Hp1 := heapEqvImplL Hp); clear Hp.
  assert (Hp2 := heapEqvImplL Hp'); clear Hp'.
  
  assert (set_eqv (heap_dom h2)
    (set_mmap (fun x => set_union (set_add x empty_set) (set_add (4 + x) empty_set)) S)).
  eapply fasDom; eauto.
  cbv beta; intros x h' Hp.
  eapply sconjDom; eauto.
  intros; eapply contDom'; eauto.
  intros; eapply contDom'; eauto.

  assert (set_eqv (heap_dom h1)
    (set_mmap (fun x => set_union (set_add x empty_set) (set_add (4 + x) empty_set)) S)).
  eapply fasDom; eauto.
  cbv beta; intros x h' Hp.
  eapply sconjDom; eauto.
  intros; eapply contDom'; eauto.
  intros; eapply contDom'; eauto.

  eapply set_eqvTrans; eauto.
  apply set_eqvSymm; assumption.

Qed.


Lemma infLoopOk : iseqOkS P (fun _ : state => True) (fun _ _ : state => False) infLoop.
Proof.
  unfold infLoop.
  iseqWeakSolver Psimpl.
  auto.
Qed.


Lemma cheneyReadOk : forall k,
  k = 0 \/ k = 4 ->
  iseqOkS P (cheneyReadPre k) (cheneyReadGuar k) (cheneyRead k).
Proof.
  intros k K.
  unfold cheneyRead.
  iseqWeakSolver fail.
  unfold cheneyReadPre; unfold readPre; unfold mview.
  intros s [vs [Hp Vhp]].
  lle Hp; lle Vhp.
  destruct H3 as [_ [_ [_ [Req _]]]].
  
  assert (Req' : rget vs rroot = rget s rroot /\
    okFieldVal (rangeObjs x1 x3) (rget vs rroot)).
  replace (rget s) with (RFile.get (rfileOf s)); [idtac | reflexivity].
  apply Req; left; reflexivity.

  clear Req; destruct Req' as [Req Rok].
  rewrite Req in Vhp.
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp'].
  hypReplace Hp' (sconjInsert Hp' Vhp); lle Hp'.
  rewrite (lwStep Hp'); try reflexivity.
  constructor.
  trivial.

  intros s' E; subst s'.
  unfold cheneyReadGuar; unfold readGuar.
  constructor.

  rename vs into vs0; rename Hp into Hp0.
  clear Hp' Req Rok.
  unfold mview.
  intros vs A rr Hp RI NI.

  lle Hp; destruct H3 as [NI' [al1 [al2 [Req Eok]]]].
  exists (rset vs v0 x4); constructor.

  (* show the guar holds *)
  intros objs h rr' I Vhp'.

  assert (X : h ||- minObjHp (List.map (rget vs) rr') objs).
  eapply heapEqvImplL; eauto.

  exists h; stSimpl; constructor.
  sconjMatch.
  destruct (aconjElim Hp1); assumption.
  eapply isoStateRefl'.
  instantiate (1 := objs).
  unfold minObjHp in X; destruct (aconjElim X).
  unfold okObjHp in H0; lle H0; assumption.

  intros r I'.
  destruct (regCompare r v0).
  subst r; elimtype False; assert (Y := NI _ (I _ I')); calleeSavedSolver Y.
  destruct (Req _ (I _ I')) as [X' Y].
  replace (rget (h, rfileOf vs) r) with (rget vs r); [idtac | reflexivity].
  replace (rget (h, rfileOf (rset vs v0 x4)) r) with
    (rget (rset vs v0 x4) r); [idtac | reflexivity].
  stSimpl.
  constructor.
  2:reflexivity.
  unfold minObjHp in X; destruct (aconjElim X) as [_ HOK].
  unfold okObjHp in HOK; lle HOK.
  apply H2.
  apply rrlistMapIntro; assumption.

  (* show the rest of the thing *)

  assert (x = x5).
  sconjLift4 Hp0; sconjLift4 Hp; exact (contContEq Hp0 Hp).
  assert (x0 = x6).
  sconjLift5 Hp0; sconjLift5 Hp; exact (contContEq Hp0 Hp).
  assert (x1 = x7).
  sconjLift6 Hp0; sconjLift6 Hp; exact (contContEq Hp0 Hp).
  assert (x2 = x8).
  sconjLift7 Hp0; sconjLift7 Hp; exact (contContEq Hp0 Hp).
  assert (x3 = x9).
  sconjLift8 Hp0; sconjLift8 Hp; exact (contContEq Hp0 Hp).
  subst x5 x6 x7 x8 x9.

  constructor.

  stSimpl; lli.
  constructor.
  simpl; intros [I | I].
  discriminate I.
  apply NI'; assumption.
  constructor.
  apply al1.
  constructor.
  apply al2.
  constructor.
  2:apply Eok.
  replace (RFile.get (rfileOf (rset s v0 x4))) with (rget (rset s v0 x4));
    [idtac | reflexivity].
  simpl; intros r [I | I].

  subst; stSimpl.
  constructor.
  reflexivity.
  destruct (Req _ RI) as [Q OkR]; clear Req.
  assert (Vhp' : vs0 |||- objHp (rangeObjs x1 x3) (rangeObjs x1 x3)).
  eapply heapEqvImplL; eauto.
  destruct OkR.
  elimtype False; eapply not_even_and_odd.
  2:apply H.
  rewrite Q.
  destruct K; subst k.
  simpl in Vhp.
  apply ptrPeven; eapply contElim1L; eauto.
  assert (PP : ptrP (4 + RFile.get (rfileOf s) rroot)).
  eapply contElim1L; eauto.
  apply ptrPeven; simpl in PP; inversion_clear PP; assumption.
  hypReplace Vhp' (objHpElim H Vhp'); unfold okObjAt in Vhp';
    unfold okFieldAt in Vhp'; lle Vhp'.
  destruct K; subst k.
  rewrite Q in Vhp'; simpl in Vhp; assert (U := contContEq Vhp' Vhp); subst x4; assumption.
  sconjLift2 Vhp';
  rewrite Q in Vhp'; simpl in Vhp; assert (U := contContEq Vhp' Vhp); subst x4; assumption.

  destruct (regCompare r v0).
  subst r; elimtype False; assert (Y := NI _ I); calleeSavedSolver Y.
  stSimpl.
  destruct (Req _ I) as [Req1 OkF].
  constructor; assumption.
  assumption.

  stSimpl.
  destruct (Req _ RI) as [Req' _].
  rewrite Req'; clear Req'.
  assert (heap_eqv (heapOf vs0) (heapOf vs)).
  eapply objsHeapEqvEqv; eauto.
  instantiate (1 := rangeObjs x1 x3).
  instantiate (1 := heapOf s).
  clear Hp; sconjMatch.
  destruct (aconjElim Hp); apply aconjIntro.
  eapply objHpIsObjs; eauto.
  assumption.
  sconjMatch.
  destruct (aconjElim Hp); apply aconjIntro.
  eapply objHpIsObjs; eauto.
  assumption.
  eapply heapEqvTrue; eauto.
  apply heap_eqvSymm; assumption.

  intros r I.
  destruct I; stSimpl.
  reflexivity.

Qed.


Lemma cheneyWriteOk : forall k,
  k = 0 \/ k = 4 ->
  iseqOkS P (cheneyWritePre k) (cheneyWriteGuar k) (cheneyWrite k).
Proof.
  intros k K.
  unfold cheneyWrite.
  iseqWeakSolver fail.
  intros s [vs [Hp Vhp]].
  unfold mview in Hp.
  lle Hp; lle Vhp.
  destruct H3 as [_ [_ [_ [Req _]]]].
  
  assert (Req' : rget vs rroot = rget s rroot /\
    okFieldVal (rangeObjs x1 x3) (rget vs rroot)).
  replace (rget s) with (RFile.get (rfileOf s)); [idtac | reflexivity].
  apply Req; left; reflexivity.

  clear Req; destruct Req' as [Req Rok].
  rewrite Req in Vhp.
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp'].
  hypReplace Hp' (sconjInsert Hp' Vhp); lle Hp'.
  rewrite (swStep Hp'); try reflexivity.
  constructor.
  trivial.

  intros s' E; subst s'.
  unfold cheneyWriteGuar; unfold writeGuar.
  constructor.

  rename vs into vs0; rename Hp into Hp0.
  clear Hp' Req Rok.
  unfold mview.
  intros vs A rr Hp Incl.

  lle Hp; destruct H3 as [NI' [al1 [al2 [Req Eok]]]].

  assert (X : rget vs rroot = rget s rroot /\ okFieldVal (rangeObjs x7 x9) (rget s rroot)).
  assert (II : In rroot rr).
  apply Incl; left; reflexivity.
  destruct (Req _ II) as [ReqX Q]; constructor.
  assumption.
  rewrite ReqX in Q; exact Q.

  assert (Y : rget vs a0 = rget s a0 /\ okFieldVal (rangeObjs x7 x9) (rget s a0)).
  assert (II : In a0 rr).
  apply Incl; right; left; reflexivity.
  destruct (Req _ II) as [ReqX Q]; constructor.
  assumption.
  rewrite ReqX in Q; exact Q.

  destruct X as [ReqR OkR].
  destruct Y as [ReqA OkA].
  exists (uhset vs (k + rget s rroot) (rget s a0)); constructor.

  rewrite ReqR; rewrite ReqA; apply gcGuarRefl.

  lli.
  constructor.
  assumption.
  constructor.
  apply al1.
  constructor.
  apply al2.
  constructor.
  2:eauto.
  intros r II.
  replace (RFile.get (rfileOf (uhset s (k + rget s rroot) (rget s a0))))
    with (rget (uhset s (k + rget s rroot) (rget s a0))); [stSimpl | reflexivity].
  exact (Req _ II).
  assert (HpX := heapEqvImplL Hp).
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp']; clear Hp.

  destruct OkR as [OkR | OkR].

  (* case: root is atomic.  proof by contradiction. *)
  elimtype False; eapply not_even_and_odd.
  2:apply OkR.
  apply ptrPeven.
  assert (X := contElim1L Vhp).
  destruct K; subst k; simpl in X; [idtac | inversion X]; assumption.

  (* case: root is not atomic *)
  hypReplace HpX (objHpElim OkR HpX).
  unfold okObjAt in HpX; unfold okFieldAt in HpX; lle HpX.
  rename x10 into v0; rename x11 into v1; rename H3 into v0Ok; rename H6 into v1Ok.

  assert (HE : Heap.set (heapOf vs) (k + rget s rroot) (rget s a0) = 
    Some (heapOf (uhset vs (k + rget s rroot) (rget s a0)))).
  destruct vs; simpl.
  destruct K; subst k; [idtac | sconjLift2 HpX]; apply (contSetEqL HpX).
  hypReplace Hp' (sconjHpSet Hp' _ _ HE).
  sconjMatch.
  apply aconjIntro.
  2:assumption.
  eapply heapEqvTrue.
  apply heap_eqvSymm; eapply heap_eqvTrans.
  apply Hp.
  apply heap_eqvSymm; apply HeapMap.map_canonEqv.  
  assert (HpX' : uhset vs (k + rget s rroot) (rget s a0)
    |||- objHp (rangeObjs x7 x9) (set_add (rget s rroot) 
      (set_remove (rget s rroot) (rangeObjs x7 x9)))).
  destruct K; subst k; [idtac | sconjLift2 HpX]; 
    hypReplace HpX (contUpdateSt HpX (rget s a0));
    [idtac | sconjLift2 HpX];
    hypReplace HpX (sconjAssocR HpX);
    apply objHpIntro; sconjMatch; unfold okObjAt; unfold okFieldAt; lli.
  2:sconjMatch.
  4:sconjMatch.
  assumption.
  assumption.
  assumption.
  assumption.
  eapply objHpWeaken.
  unfold hpredTrue; rewrite <- (addRemoveEqv OkR); apply HpX'.
  intros y I; exact I.

  intros r I; stSimpl; reflexivity.

Qed.


Lemma cheneyPostGCOk : iseqOkS P cheneyPostGCPre cheneyPostGCGuar cheneyPostGC.
Proof.
  unfold cheneyPostGC.
  iseqWeakSolver Psimpl.
  
  intros s [toEnd [Hp Pre]].
  rewrite (lwStep Hp); try reflexivity; stSimpl.
  assert (BT := doBtestSound btne (rget s rfree) toEnd).
  destruct (doBtest btne (rget s rfree) toEnd); simpl in BT.

  (* case: GC freed up more space *)
  hypReplace Pre (Pre BT).
  constructor.
  clear Hp.
  unfold cheneyDoAllocPre in * |- * ; stSimpl; assumption.
  intros s' Guar toEnd' Hp'.
  assert (X := contContEq Hp Hp'); clear Hp'; subst toEnd'.
  rewrite (beq_false BT).
  assumption.

  (* case: still no free space: abort *)
  constructor.
  trivial.
  intros; elimtype False; assumption.

Qed.

    
Lemma cheneyEnterOk : iseqOkS P cheneyEnterPre cheneyEnterGuar cheneyEnter.
Proof.
  unfold cheneyEnter.
  iseqWeakSolver Psimpl.
  apply scanFieldGuarPresRa.
  
  (* first, step through the block *)
  unfold cheneyEnterPre.
  intros s [toSt [toEnd [fromSt [fromEnd [hx Hp]]]]].
  unfold cheneyStOk in Hp; linLogElim Hp.
  destruct H1 as [alSt [alFr [OkR [sizeOk frLe]]]].
  assert (Hp0 := Hp).
  sconjLift4of4 Hp0; simpl in Hp0; linLogElim Hp0.
  rewrite (lwStep Hp0); try reflexivity.
  assert (Hp1 := Hp0); sconjLift3 Hp1.
  rewrite (lwStep Hp1); try reflexivity.
  clear Hp1; stSimpl.
  rewrite (swStep Hp0); try reflexivity.
  hypReplace Hp0 (contUpdateSt Hp0 fromSt).
  sconjLift3 Hp0.
  stSimpl.
  rewrite (swStep Hp0); try reflexivity.
  hypReplace Hp0 (contUpdateSt Hp0 toSt).
  hypReplace Hp0 (sconjComm Hp0); linLogElim Hp0.
  sconjLift2 Hp0.
  rewrite (lwStep Hp0); try reflexivity.
  sconjLift3 Hp0.
  rewrite (lwStep Hp0); try reflexivity.
  sconjLift2 Hp0.
  rewrite (swStep Hp0); try reflexivity.
  hypReplace Hp0 (contUpdateSt Hp0 fromEnd).
  stSimpl.
  rewrite addiuStep0; stSimpl.
  rewrite addiuStep0; stSimpl.
  rewrite addiuStep0; stSimpl.
  sconjLift2 Hp0.
  rewrite (swStep Hp0); try reflexivity.
  stSimpl.
  hypReplace Hp0 (contUpdateSt Hp0 (rget s rroot)).
  rewrite addiuStep; stSimpl.

  assert (alFrSE := aligned8Trans alSt alFr).

  assert (alToSE : aligned8 toSt toEnd).
  destruct alFrSE as [k E].
  exists k.
  omega.

  assert (LEff : rget s rfree <= fromEnd).
  destruct alFr as [y E].
  omega.

  set (objs := rangeObjs fromSt (rget s rfree)).

  assert (Oeq : set_eqv (set_union objs empty_set) objs).
  setoid_rewrite (set_eqvUnionSymm objs empty_set).
  setoid_rewrite <- (set_eqvUnionEmptyUnit objs).
  apply set_eqvRefl.

  assert (HH : exists s1, s1 =
    rset (rset (uhset
      (rset (rset
	(rset (uhset
	  (rset (rset
	    (uhset (uhset
	      (rset (rset s rtoStart toSt) rtemp fromSt) (rget s rgcInfo) fromSt)
	    (8 + rget s rgcInfo) toSt) rtoEnd toEnd) rtemp fromEnd)
	  (S (S (S (S (rget s rgcInfo))))) fromEnd) rfree
	toSt) rscan toSt) raSave 
      (rget s ra))
      (12 + rget s rgcInfo)
      (rget s rroot)) a0 (12 + rget s rgcInfo)) ra CHENEYLOOPHEADER).
  econstructor; reflexivity.

  destruct HH as [s1 S1eq].

  assert (Hp0Ok : 
    scanFieldStOk (fun n:nat => None (A:=nat))
    objs empty_set (rangeObjs toSt toEnd)
    (okFieldAt (set_union objs empty_set) (rget s1 a0) ** 
      rget s rgcInfo |-> fromSt ** 4 + rget s rgcInfo |-> fromEnd **
	8 + rget s rgcInfo |-> toSt ** buffer (rget s rfree) fromEnd ** top)
    s1).

  rewrite S1eq.
  unfold scanFieldStOk; stSimpl.
  constructor.
  subst objs.
  assert (HH : 8 * set_card (rangeObjs fromSt (rget s rfree)) <=
    8 * set_card (rangeObjs toSt toEnd)).
  rewrite (rangeObjsCard alSt).
  rewrite (rangeObjsCard alToSE).
  rewrite sizeOk.
  omega.
  omega.

  constructor.
  clear Hp0.
  setoid_rewrite Oeq.
  sconjLift3 Hp; sconjLift2 Hp.

  assert (Hp' : heapOf s ||- 
    buffer fromSt (rget s rfree) ** buffer toSt toEnd ** top).
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  eapply objHpIsBuffer.
  destruct (aconjElim Hp); eauto.
  assumption.
  apply (sconjImplR Hp); intros; linLogIntro.

  clear Hp.

  assert (PT : ptrP toSt \/ toSt = toEnd).
  compare toSt toEnd; intro R.
  right; assumption.
  left.
  sconjLift2 Hp'.
  assert (heapOf s ||- ((!(ptrP toSt) ** top) ** top) ** top).
  apply (sconjImpl Hp'); intros h H; clear Hp'.
  assert (Al4 : aligned4 toSt toEnd).
  destruct alToSE as [q E]; exists (2 * q); omega.
  destruct Al4 as [k4 Al4].
  unfold buffer in H.
  assert (I : set_In toSt (addrRange toSt toEnd)).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0); auto.
  hypReplace H (fasElim I H).
  apply (sconjImpl H); clear h H; intros h H.
  linLogIntro; linLogElim H.
  exact (contElim1 H).
  linLogIntro.
  linLogIntro.
  linLogElim H.
  assumption.

  assert (PF : ptrP fromSt \/ fromSt = rget s rfree).
  compare fromSt (rget s rfree); intro R.
  right; assumption.
  left.
  assert (heapOf s ||- ((!(ptrP fromSt) ** top) ** top) ** top).
  apply (sconjImpl Hp'); intros h H; clear Hp'.
  assert (Al4 : aligned4 fromSt (rget s rfree)).
  destruct alSt as [q E]; exists (2 * q); omega.
  destruct Al4 as [k4 Al4].
  unfold buffer in H.
  assert (I : set_In fromSt (addrRange fromSt (rget s rfree))).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0); auto.
  hypReplace H (fasElim I H).
  apply (sconjImpl H); clear h H; intros h H.
  linLogIntro; linLogElim H.
  exact (contElim1 H).
  linLogIntro.
  linLogIntro.
  linLogElim H.
  assumption.

  destruct PT as [PT | Teq].
  destruct PF as [PF | Feq].

  assert (D : set_disj (addrRange fromSt (rget s rfree))
    (addrRange toSt toEnd)).
  refine (fasFasDisj Hp' _ _).
  intros h x Hp; linLogElim Hp; linLogIntro.
  assert (Q := unitIR Hp); sconjMatch.
  intros h x Hp; linLogElim Hp; linLogIntro.
  assert (Q := unitIR Hp); sconjMatch.

  subst objs.
  refine (set_disjSub _ (objRangeAddrSetSubset fromSt (rget s rfree))).
  refine (addrAddrSetDisj _ _ _ _ D).

  assumption.
  assumption.

  destruct (ptrP_elim PF) as [k E].
  destruct alSt as [k' E'].
  rewrite E'; rewrite E.
  apply (ptrP_intro (k + 2 * k')).
  omega.

  destruct (ptrP_elim PT) as [k E].
  destruct alToSE as [k' E'].
  rewrite E'; rewrite E.
  apply (ptrP_intro (k + 2 * k')).
  omega.

  subst objs.
  rewrite <- Feq.
  setoid_rewrite <- (set_eqvEmptyRangeObjs fromSt).
  apply set_disjSymm; apply set_disjEmptyUnit.

  rewrite Teq.
  setoid_rewrite <- (set_eqvEmptySetRange toEnd).
  apply set_disjEmptyUnit.



  constructor.
  simpl; apply set_eqvRefl.
  constructor.
  discriminate 2.
  constructor.
  omega.

  hypReplace Hp0 (unitIL Hp0).
  apply (sconjImpl Hp0); clear Hp Hp0; intros h Hp.
  apply mapHpIntroE; assumption.
  sconjLift4 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold hpredTrue; setoid_rewrite Oeq; destruct (aconjElim Hp); assumption.

  sconjLift5 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  apply bufferIsObjs; assumption.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold okFieldAt; existsl (rget s rroot).
  linLogIntro.
  rewrite Oeq; assumption.
  assumption.

  sconjLift4 Hp.
  sconjLift2 Hp.
  sconjLift3 Hp.
  assumption.

  apply callChain.

  (* scanFieldPre *)
  constructor.
  unfold scanFieldPre; stSimpl.
  exists (fun n:nat => None (A:=nat)).
  exists objs; exists empty_set; exists (rangeObjs toSt toEnd).
  rewrite S1eq in Hp0Ok.
  eapply scanFieldStOkImpl.
  rewrite S1eq; apply Hp0Ok.
  stSimpl.
  intros h H.
  sconjMatch.


  (* post condition *)
  intros s' [Guar1 Rp'].

  subst.
  assert (X := Guar1 _ _ _ _ _ Hp0Ok).
  destruct X as [M' [W' [F' [G' [phi' X]]]]].
  generalize dependent X.
  replace (S (S (S (S (S (S (S (S (S (S (S (S (rget s rgcInfo)))))))))))))
    with (12 + rget s rgcInfo).
  2:reflexivity.
  replace (S (S (S (S (S (S (S (S ((rget s rgcInfo))))))))))
    with (8 + rget s rgcInfo).
  2:reflexivity.
  replace (S (S (S (S ((rget s rgcInfo))))))
    with (4 + rget s rgcInfo).
  2:reflexivity.
  stSimpl.
  unfold fieldRespMap; stSimpl.
  unfold cheneyLoopHeaderPre; unfold cheneyLoopPre.
  unfold cheneyLoopStOk; unfold scanFieldStOk.
  rewrite (regPresEq Rp' rtoStart); [stSimpl | regNotInListSolver].
  rewrite (regPresEq Rp' rtoEnd); [stSimpl | regNotInListSolver].
  intro X.
  destruct X as [objsEq' [GFeq' [iso' [reach' [al8' [IWok SCok]]]]]].

  rewrite (regPresEq Rp'); [stSimpl | regNotInListSolver].
  destruct SCok as [WFlt' [WMdisj' [Feq' [phiOk' [stLt' Hp']]]]].
  sconjLift4 Hp'.

  assert (HH : heapOf s' ||-
    (foralls x in objs, x |->__) **
    rget s rgcInfo |-> fromSt **
      4 + rget s rgcInfo |-> fromEnd ** 8 + rget s rgcInfo |-> toSt ** 
	12 + rget s rgcInfo |->_ ** top).
  assert (HH : heapOf s' ||- 12 + rget s rgcInfo |->_ **
    mapHp (set_union empty_set M')
    (map_union (fun _ : nat => None) phi') **
    objHp (set_union W' (set_union empty_set M')) W' **
    top ** top ** rget s rgcInfo |-> fromSt **
      4 + rget s rgcInfo |-> fromEnd ** 8 + rget s rgcInfo |-> toSt ** 
	top).
  apply (sconjImpl Hp'); clear Hp'; intros h Hp'.
  linLogElim Hp'.
  exists x; assumption.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  apply (sconjImpl Hp'); clear h Hp'; intros h Hp'.
  linLogIntro.
  apply (sconjImpl Hp'); clear h Hp'; intros h Hp'.
  linLogIntro.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  linLogIntro.
  sconjLift3 HH; sconjLift3 HH; hypReplace HH (sconjAssocR HH).
  apply (sconjImpl HH); clear HH Hp'; intros h Hp'.
  unfold hpredTrue; setoid_rewrite objsEq'.
  rewrite <- hpredTrueEq; apply fasMerge.
  apply (sconjImpl Hp'); clear h Hp'; intros h Hp'.
  unfold mapHp in Hp'; linLogElim Hp'.
  unfold hpredTrue; setoid_rewrite (set_eqvUnionEmptyUnit M');
    rewrite <- hpredTrueEq.
  eapply fasImpl.
  apply Hp'.
  simpl; intros x h0 H; linLogElim H; linLogIntro.
  eauto.
  
  unfold objHp in Hp'; unfold okObjAt in Hp'; unfold okFieldAt in Hp'.
  eapply fasImpl; eauto.
  simpl; intros x h0 Hpp; linLogElim Hpp.
  linLogIntro; eauto.

  (* set_disj M' W' *)
  unfold mapHp in Hp'; unfold objHp in Hp'.
  linLogElim Hp'.
  eapply fasFasDisj.
  instantiate (1 := unit).
  linLogIntro.
  apply Hp'.
  simpl; intros h0 x HP; linLogElim HP.
  linLogIntro; sconjMatch.
  simpl; unfold okObjAt; unfold okFieldAt; intros h0 x HP; linLogElim HP.
  linLogIntro; sconjMatch.

  sconjLift4 Hp'.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  sconjLift4 Hp'.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  sconjLift4 Hp'.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  linLogIntro.

  assert (infoNotInObjs :
    forall x, set_In x objs -> forall y, 
      y = 0 \/ y = 4 \/ y = 8 \/ y = 12 ->
      x <> (y + rget s rgcInfo) /\ 4 + x <> (y + rget s rgcInfo)).
  intros x IO y Yeq.
  assert (HH' : heapOf s' ||- (x |->_ ** 4 + x |->_ ** top) ** 
    y + rget s rgcInfo |->_ ** top).
  apply (sconjImpl HH); clear HH Hp'; intros h Hp'.
  hypReplace Hp' (fasElim IO Hp').
  simpl in Hp'; hypReplace Hp' (sconjAssocL Hp').
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  apply (sconjImplR Hp'); clear h Hp'; intros; linLogIntro.
  destruct Yeq as [R | [R | [R | R]]]; rewrite R.
  linLogIntro; sconjMatch.
  sconjLift2 Hp'; linLogIntro; sconjMatch.
  sconjLift3 Hp'; linLogIntro; sconjMatch.
  linLogElim Hp'; sconjLift4 Hp'; linLogIntro; sconjMatch.

  clear Yeq Hp' HH IWok Rp' Hp Hp0.
  assert (QQ : ~(x = y + rget s rgcInfo \/ 4 + x = y + rget s rgcInfo)).
  assert (QQ : x = y + rget s rgcInfo \/ 4 + x = y + rget s rgcInfo ->
    heapOf s' ||- (!False ** top) ** top).
  intros R.
  assert (QQ : 
    heapOf s' ||- (y + rget s rgcInfo |->_ ** y + rget s rgcInfo |->_) ** top).
  apply sconjAssocR.
  sconjLift2 HH'; apply (sconjImplR HH'); clear HH'; intros h HH'.
  hypReplace HH' (sconjAssocL HH').
  destruct R as [R | R]; rewrite R in HH'.
  apply (sconjImplR HH'); intros; linLogIntro.
  sconjLift2 HH'; hypReplace HH' (sconjAssocL HH').
  apply (sconjImplR HH'); intros; linLogIntro.

  apply (sconjImpl QQ); clear QQ; intros h Hp'.
  linLogElim Hp'; linLogIntro; eapply sconjDisj; eauto.
  assumption.
  intro Z; assert (Y := QQ Z); linLogElim Y; assumption.
  intuition.

  assert (Pgc0 : ptrP (rget s rgcInfo)).
  assert (Pgc : s |||- (!ptrP (rget s rgcInfo) ** top) ** top).
  sconjLift4of4 Hp; simpl in Hp; linLogElim Hp.
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  linLogIntro; exact (contElim1 Hp).  
  linLogIntro.
  linLogElim Pgc; assumption.

  assert (Pgc4 : ptrP (4 + rget s rgcInfo)).
  assert (Pgc : s |||- (!ptrP (4 + rget s rgcInfo) ** top) ** top).
  sconjLift4of4 Hp; simpl in Hp; linLogElim Hp; sconjLift2 Hp.
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  linLogIntro; exact (contElim1 Hp).  
  linLogIntro.
  linLogElim Pgc; assumption.

  assert (Pgc8 : ptrP (8 + rget s rgcInfo)).
  assert (Pgc : s |||- (!ptrP (8 + rget s rgcInfo) ** top) ** top).
  sconjLift4of4 Hp; simpl in Hp; linLogElim Hp; sconjLift3 Hp.
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  linLogIntro; exact (contElim1 Hp).  
  linLogIntro.
  linLogElim Pgc; assumption.

  assert (Pgc12 : ptrP (12 + rget s rgcInfo)).
  assert (Pgc : s |||- (!ptrP (12 + rget s rgcInfo) ** top) ** top).
  sconjLift4of4 Hp; simpl in Hp; linLogElim Hp; sconjLift4 Hp.
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  linLogIntro; exact (contElim1 Hp).  
  linLogIntro.
  linLogElim Pgc; assumption.

  (* show it is okay to enter the loop *)
  constructor.
  apply (sconjImpl Hp'); clear Hp'; intros h Hp'.
  linLogElim Hp'.
  exists x; assumption.

  unfold hpredTrue.
  exists (map_union (fun _ : nat => None) phi').
  exists G'; exists W'; exists (set_union empty_set M'); exists F'.
  replace (rget (heap_canon h, rfileOf s')) with (rget s').
  2:reflexivity.
  rewrite (regPresEq Rp' rtoStart); [stSimpl | regNotInListSolver].
  rewrite (regPresEq Rp' rtoEnd); [stSimpl | regNotInListSolver].
  rewrite (regPresEq Rp' rscan); [stSimpl | regNotInListSolver].

  constructor.
  simpl.
  eapply rangeSplit.
  instantiate (1 := toEnd).
  setoid_rewrite <- Feq'.
  sconjLift3 Hp'; sconjLift4 Hp'.
  eapply fasFasDisj.
  apply Hp'.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; stSimpl; simpl.
  intros h' x H; linLogElim H; linLogIntro; sconjMatch.
  simpl; intros; apply (sconjImplR H); intros; linLogIntro.
  omega.
  setoid_rewrite <- Feq'; apply set_eqvSymm; assumption.
  assumption.

  constructor.
  omega.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.

  simpl.
  unfold heap_canon.
  rewrite (HeapMap.map_canonPEq (HeapMap.map_canonCanonP h)).
  rewrite <- hpredTrueEq.
  apply (sconjImpl Hp'); clear h Hp'; intros h Hp'.
  unfold hpredTrue; setoid_rewrite (set_eqvUnionEmptyUnit M'); assumption.
  apply (sconjImpl Hp'); clear h Hp'; intros h Hp'.
  unfold hpredTrue; setoid_rewrite (set_eqvUnionEmptyUnit M'); assumption.
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp'.
  apply (sconjImpl Hp'); clear h Hp'; intros h Hp'.
  clear Rp' Hp0 IWok.
  unfold objHp; eapply fasImplI.
  apply Hp'.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied.
  clear Hp'; simpl; intros x h' IG' H; linLogElim H.
  unfold okObjAt; unfold okFieldAt.

  assert (IO : set_In x0 objs).
  setoid_rewrite objsEq'; apply set_union_intro1.
  destruct iso' as [I1 _].
  destruct (I1 x0) as [_ X].
  apply X; rewrite H2; discriminate.

  hypReplace infoNotInObjs (infoNotInObjs x0 IO).

  assert (Hp2 : heapOf s ||- (okObjAt objs x0 ** top) ** top).
  sconjMatch.
  destruct (aconjElim Hp) as [Hp' _]; hypReplace Hp (objHpElim IO Hp').
  sconjMatch.

  unfold okObjAt in Hp2; unfold okFieldAt in Hp2; linLogElim Hp2.

  clear IO HH; rename H7 into HH.
  unfold POnat.POnat.A in * |- * .

  assert (Eq1 : x3 = x1).
  assert (H' : 
    uset (uset (uset 
      (uset (heapOf s) (rget s rgcInfo) fromSt) (8 + rget s rgcInfo) toSt)
    (4 + rget s rgcInfo) fromEnd) (12 + rget s rgcInfo) (rget s rroot) ||- 
    x0 |-> x3 ** top).
  refine (contUpdateNeqTop _ _ _ Pgc12).
  refine (contUpdateNeqTop _ _ _ Pgc4).
  refine (contUpdateNeqTop _ _ _ Pgc8).
  refine (contUpdateNeqTop _ _ _ Pgc0).
  apply (sconjImplR Hp2); intros; linLogIntro.
  destruct (infoNotInObjs 0); auto.
  destruct (infoNotInObjs 8); [auto | omega].
  destruct (infoNotInObjs 4); [auto | omega].
  destruct (infoNotInObjs 12); [auto | omega].
  exact (contContEq H' HH).

  assert (Eq2 : x4 = x2).
  assert (H' : uset
    (uset (uset (uset (heapOf s) (rget s rgcInfo) fromSt)
      (8 + rget s rgcInfo) toSt) (4 + rget s rgcInfo) fromEnd)
    (12 + rget s rgcInfo) (rget s rroot) ||- 4 + x0 |-> x4 ** top).
  refine (contUpdateNeqTop _ _ _ Pgc12).
  refine (contUpdateNeqTop _ _ _ Pgc4).
  refine (contUpdateNeqTop _ _ _ Pgc8).
  refine (contUpdateNeqTop _ _ _ Pgc0).
  sconjLift2 Hp2; apply (sconjImplR Hp2); intros; linLogIntro.
  destruct (infoNotInObjs 0); auto.
  destruct (infoNotInObjs 8); [auto | omega].
  destruct (infoNotInObjs 4); [auto | omega].
  destruct (infoNotInObjs 12); [auto | omega].
  exact (contContEq H' H8).

  subst.
  linLogIntro; eauto.
  setoid_rewrite (set_eqvUnionSymm W' M').
  setoid_rewrite <- objsEq'; assumption.
  setoid_rewrite (set_eqvUnionSymm W' M').
  setoid_rewrite <- objsEq'; assumption.
  linLogIntro.

  (* showing guar for entire function holds *)
  intros s3 Guar2.
  constructor.

  unfold cheneyStOk; intros toSt0 toEnd0 fromSt0 fromEnd0 A h0 Hp00.
  linLogElim Hp00.
  clear H1.
  assert (toSt = toSt0).
  sconjLift4 Hp; sconjLift4 Hp00; exact (contContEq Hp Hp00).
  assert (toEnd = toEnd0).
  sconjLift5 Hp; sconjLift5 Hp00; exact (contContEq Hp Hp00).
  assert (fromSt = fromSt0).
  sconjLift6 Hp; sconjLift6 Hp00; exact (contContEq Hp Hp00).
  assert (fromEnd = fromEnd0).
  sconjLift7 Hp; sconjLift7 Hp00; exact (contContEq Hp Hp00).
  subst toSt0 toEnd0 fromSt0 fromEnd0.
  clear Hp; rename Hp00 into Hp.
  clear Hp0 Hp0Ok M' W' F' G' phi' objsEq' GFeq' iso' IWok WFlt' reach'
    WMdisj' Feq' phiOk' Hp'.

  destruct (Guar1 (fun n:nat => None (A:=nat)) objs empty_set (rangeObjs toSt toEnd)
    (rget s rgcInfo |-> fromSt ** 4 + rget s rgcInfo |-> fromEnd **
	8 + rget s rgcInfo |-> toSt ** buffer (rget s rfree) fromEnd ** A)) as 
    [M [W [F [G [phi X]]]]]; clear Guar1.
  (* a lot of this is just cut and paste from before *)
  unfold scanFieldStOk; stSimpl.
  constructor.
  subst objs.
  assert (HH' : 8 * set_card (rangeObjs fromSt (rget s rfree)) <=
    8 * set_card (rangeObjs toSt toEnd)).
  rewrite (rangeObjsCard alSt).
  rewrite (rangeObjsCard alToSE).
  rewrite sizeOk.
  omega.
  omega.

  constructor.
  setoid_rewrite Oeq.
  sconjLift3 Hp; sconjLift2 Hp.

  assert (Hp' : heapOf s ||- 
    buffer fromSt (rget s rfree) ** buffer toSt toEnd ** top).
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  eapply objHpIsBuffer.
  destruct (aconjElim Hp); eauto.
  assumption.
  apply (sconjImplR Hp); intros; linLogIntro.
  clear Hp.

  assert (PT : ptrP toSt \/ toSt = toEnd).
  compare toSt toEnd; intro R.
  right; assumption.
  left.
  sconjLift2 Hp'.
  assert (heapOf s ||- ((!(ptrP toSt) ** top) ** top) ** top).
  apply (sconjImpl Hp'); intros h H; clear Hp'.
  assert (Al4 : aligned4 toSt toEnd).
  destruct alToSE as [q E]; exists (2 * q); omega.
  destruct Al4 as [k4 Al4].
  unfold buffer in H.
  assert (I : set_In toSt (addrRange toSt toEnd)).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0); auto.
  hypReplace H (fasElim I H).
  apply (sconjImpl H); clear h H; intros h H.
  linLogIntro; linLogElim H.
  exact (contElim1 H).
  linLogIntro.
  linLogIntro.
  linLogElim H.
  assumption.

  assert (PF : ptrP fromSt \/ fromSt = rget s rfree).
  compare fromSt (rget s rfree); intro R.
  right; assumption.
  left.
  assert (heapOf s ||- ((!(ptrP fromSt) ** top) ** top) ** top).
  apply (sconjImpl Hp'); intros h H; clear Hp'.
  assert (Al4 : aligned4 fromSt (rget s rfree)).
  destruct alSt as [q E]; exists (2 * q); omega.
  destruct Al4 as [k4 Al4].
  unfold buffer in H.
  assert (I : set_In fromSt (addrRange fromSt (rget s rfree))).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0); auto.
  hypReplace H (fasElim I H).
  apply (sconjImpl H); clear h H; intros h H.
  linLogIntro; linLogElim H.
  exact (contElim1 H).
  linLogIntro.
  linLogIntro.
  linLogElim H.
  assumption.

  destruct PT as [PT | Teq].
  destruct PF as [PF | Feq].

  assert (D : set_disj (addrRange fromSt (rget s rfree))
    (addrRange toSt toEnd)).
  refine (fasFasDisj Hp' _ _).
  intros h x Hp; linLogElim Hp; linLogIntro.
  assert (Q := unitIR Hp); sconjMatch.
  intros h x Hp; linLogElim Hp; linLogIntro.
  assert (Q := unitIR Hp); sconjMatch.

  subst objs.
  refine (set_disjSub _ (objRangeAddrSetSubset fromSt (rget s rfree))).
  refine (addrAddrSetDisj _ _ _ _ D).

  assumption.
  assumption.

  destruct (ptrP_elim PF) as [k E].
  destruct alSt as [k' E'].
  rewrite E'; rewrite E.
  apply (ptrP_intro (k + 2 * k')).
  omega.

  destruct (ptrP_elim PT) as [k E].
  destruct alToSE as [k' E'].
  rewrite E'; rewrite E.
  apply (ptrP_intro (k + 2 * k')).
  omega.

  subst objs.
  rewrite <- Feq.
  setoid_rewrite <- (set_eqvEmptyRangeObjs fromSt).
  apply set_disjSymm; apply set_disjEmptyUnit.

  rewrite Teq.
  setoid_rewrite <- (set_eqvEmptySetRange toEnd).
  apply set_disjEmptyUnit.

  constructor.
  simpl; apply set_eqvRefl.
  constructor.
  discriminate 2.
  constructor.
  omega.

  sconjLift4 Hp; hypReplace Hp (contUpdateSt Hp fromSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp toSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp fromEnd).
  sconjLift7 Hp; hypReplace Hp (contUpdateSt Hp (rget s rroot)).
  hypReplace Hp (unitIL Hp).
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  apply mapHpIntroE; assumption.
  sconjLift5 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold hpredTrue; setoid_rewrite Oeq; destruct (aconjElim Hp); assumption.

  sconjLift6 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  apply bufferIsObjs; assumption.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold okFieldAt; existsl (rget s rroot).
  linLogIntro.
  rewrite Oeq; assumption.
  assumption.

  sconjLift3 Hp; assumption.

  generalize dependent X; unfold fieldRespMap; stSimpl; intro.
  destruct X as [objsEq [toEq [Iso [reach [_ [WcopyOk X]]]]]].
  destruct X as [leWF [WMdisj [Feq [NSP [_ Hp']]]]].
  destruct Guar2 as [Guar2 [Req2 RaEq2]].

  sconjLift4 Hp'; linLogElim Hp'.
  assert (GIeq : rget s' rgcInfo = rget s rgcInfo).
  rewrite <- Rp'; stSimpl; [reflexivity | regNotInListSolver].
  rewrite GIeq in Guar2.
  sconjLift3 Hp'; destruct (sconjExtract Hp') as [hobjs [Hp'2 hobjsOk]]; clear Hp'.
  rename Hp'2 into Hp'; sconjLift3 Hp'; sconjLift3 Hp'.
  destruct (Guar2 _ _ Hp') as [s'' [Hp'' [Req'' Guar3]]]; clear Guar2; rename Hp' into Hp0'Y.
  destruct Guar3 as [Guar3 [Req3 _]].

  clear Req2; rename Req3 into Req'''; rename Rp' into Req'.
  destruct (Guar3 phi G W M F 
    (rget s rgcInfo |->, (fromSt::fromEnd::toSt::toEnd::nil) **
      buffer (rget s rfree) fromEnd ** A)) as [phi' [W' [M' [F' [G' X]]]]]; clear Guar3.
  unfold cheneyLoopStOk; unfold scanFieldStOk.
  repeat (rewrite Req''); stSimpl.
  rewrite <- (Req' rscan); [stSimpl | regNotInListSolver].
  rewrite <- (Req' rtoStart); [stSimpl | regNotInListSolver].
  rewrite <- (Req' rtoEnd); [stSimpl | regNotInListSolver].

  assert (GFdisj : set_disj G F).
  sconjLift4 Hp''; sconjLift5 Hp''; eapply fasFasDisj; eauto.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; stSimpl;
    intros h x0 HpX; linLogElim HpX; linLogIntro; sconjMatch.
  simpl; intros h x0 HpX; linLogElim HpX; linLogIntro; sconjMatch.

  assert (ToE : rget s' rtoEnd = toEnd).
  rewrite <- (Req' rtoEnd); [stSimpl; reflexivity | regNotInListSolver].

  assert (ToS : rget s' rtoStart = toSt).
  rewrite <- (Req' rtoStart); [stSimpl; reflexivity | regNotInListSolver].

  rewrite ToE in WMdisj; rewrite ToE in Feq; rewrite ToE in NSP;
    rewrite ToE in Hp''.
  rewrite ToS in WMdisj; rewrite ToS in NSP.

  assert (Geqv : set_eqv G (rangeObjs toSt (rget s' rfree))).
  eapply rangeSplit.
  instantiate (1 := toEnd).
  rewrite <- Feq; assumption.
  destruct stLt'; assumption.
  apply set_eqvSymm; rewrite <- Feq; assumption.
  assumption.

  constructor.
  assumption.
  constructor.
  auto.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  exact WMdisj.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  linLogIntro.
  sconjLift3 Hp''; hypReplace Hp'' (sconjInsert Hp'' hobjsOk); sconjLift3 Hp''; sconjLift3 Hp''.
  sconjLift8 Hp''; sconjLift8 Hp''; sconjLift8 Hp'';
  sconjLift8 Hp''; sconjLift8 Hp''; sconjLift8 Hp''; sconjLift8 Hp''.
  sconjMatch.
  generalize dependent Hp0; unfold objCopiedPhi; unfold objCopied; unfold fieldCopied;
    stSimpl; intro.
  unfold objHp; eapply fasImplI.
  eauto.
  cbv beta; intros y h' I HpX; linLogElim HpX.

  assert (I' : set_In x0 objs).
  rewrite objsEq; apply set_union_intro1.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ X]; apply X.
  rewrite H1; discriminate.
  sconjLift4 Hp; hypReplace Hp (contUpdateSt Hp fromSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp toSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp fromEnd).
  sconjLift7 Hp; hypReplace Hp (contUpdateSt Hp (rget s rroot)).
  sconjLift5 Hp; destruct (aconjElim (sconjDistrib2 Hp)) as [HpH _]; clear Hp.

  assert (Q : uset
          (uset
             (uset (uset (heapOf s) (rget s rgcInfo) fromSt)
                (4 + (4 + rget s rgcInfo)) toSt) (4 + rget s rgcInfo) fromEnd)
          (4 + (4 + (4 + rget s rgcInfo))) (rget s rroot) 
	  ||- (okObjAt (rangeObjs fromSt (rget s rfree)) x0 ** top) ** top).
  clear H7 H8; sconjMatch.
  hypReplace Hp (objHpElim I' Hp); sconjMatch.
  unfold okObjAt in Q; unfold okFieldAt in Q; linLogElim Q.
  assert (X := contContEq H7 Q).
  sconjLift2 Q; assert (Y := contContEq H8 Q).
  subst x3 x4; clear Q HpH.
  unfold okObjAt; unfold okFieldAt; linLogIntro.
  2:eauto.
  rewrite (set_eqvUnionSymm W M); rewrite <- objsEq; assumption.
  rewrite (set_eqvUnionSymm W M); rewrite <- objsEq; assumption.

  (* have the final state *)
  destruct X as [Weqv [Feqv [Iso' [reach' [Al' X]]]]].
  destruct X as [scFrEq [scLt [al3 SF]]].
  unfold scanFieldStOk in SF.
  assert (E1 : rget s3 rtoStart = toSt).
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  rewrite <- Req'; [stSimpl; reflexivity | regNotInListSolver].
  assert (E2 : rget s3 rtoEnd = toEnd).
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  rewrite <- Req'; [stSimpl; reflexivity | regNotInListSolver].
  rewrite E1 in SF; rewrite E2 in SF.
  destruct SF as [WFle' [WMdisj' [F'eq [NSP' [Le' Hp30]]]]].

  assert (reach0 : forall x, set_In x (G |_| G') -> 
    rootReachable (heapOf s3) (rget s3 rroot::nil) x).
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  assert (reachG : forall z, set_In z G -> reachable (heapOf s3) x z).
  intros z I.
  hypReplace reach (reach _ I).
  assert (X := contGetEqL Hp0'Y).
  rewrite reach in X; injection X; intro; subst z; constructor.
  intros z I0.
  exists x; constructor.
  left; reflexivity.
  destruct (set_union_elim I0) as [I | I]; clear I0.
  apply reachG; assumption.
  destruct (reach' _ I) as [z' [I' R']].
  hypReplace reachG (reachG _ I').

  eapply reachableTrans; eauto.
  clear reach reach'; rename reach0 into reach.
  (* done proving the reachability we want *)

  sconjLift5 Hp30; sconjLift5 Hp30; hypReplace Hp30 (sconjAssocR Hp30).
  destruct (sconjExtract Hp30) as [h3 [Hp3 H3ok]]; clear Hp30.

  assert (WMdisj2 : set_disj W M).
  assert (HpX : s'' |||- (foralls x in W, x |->__) ** (foralls x in M, x |->__) ** top).
  sconjLift2 Hp''; sconjLift3 Hp''; sconjMatch.
  eapply objHpIsObjs; eapply heapEqvTrue.
  apply heap_eqvSymm; apply eqvPredEqv; eauto.
  eauto.
  simpl in Hp0; hypReplace Hp0 (mapHpIsObjs Hp0); assumption.
  eapply fasFasDisj; eauto.
  sllImpl.
  sllImpl.

  assert (MM'disj : set_disj M M').
  apply set_disjSymm; eapply set_unionDisjElimL.
  instantiate (1 := W').
  rewrite <- Weqv.
  assumption.

  assert (Iso'' : iso_set (map_union phi phi') (M |_| M') (G |_| G')).
  apply iso_union.
  assumption.
  assumption.
  assumption.
  eapply fasFasDisj.
  instantiate (1 := unit).
  linLogIntro; eauto.
  sllImpl.
  sllImpl.

  assert (GGdisj : set_disj G G').
  eapply fasFasDisj.
  instantiate (1 := unit).
  instantiate (3 := h3).
  linLogIntro; eauto.
  sllImpl.
  sllImpl.

  assert (MW'disj : set_disj M' W').
  assert (X : s3 |||- ((foralls x in M, x |->__) ** (foralls x in M', x |->__)) ** 
    (foralls x in W', x |->__) ** top).
  sconjLift3 Hp3; sconjLift3 Hp3; sconjMatch.
  apply fasSplit.
  eapply mapHpIsObjs; eauto.
  assumption.
  eapply objHpIsObjs; eauto.
  linLogElim X; sconjLift3 X; sconjLift3 X; eapply fasFasDisj; eauto.
  sllImpl.
  sllImpl.

  set (s1 :=
    rset (rset (uhset
      (rset (rset
	(rset (uhset
	  (rset (rset
	    (uhset (uhset
	      (rset (rset s rtoStart toSt) rtemp fromSt) (rget s rgcInfo) fromSt)
	    (8 + rget s rgcInfo) toSt) rtoEnd toEnd) rtemp fromEnd)
	  (4 + rget s rgcInfo) fromEnd) rfree
	toSt) rscan toSt) raSave 
      (rget s ra))
      (12 + rget s rgcInfo)
      (rget s rroot)) a0 (12 + rget s rgcInfo)) ra CHENEYLOOPHEADER).

  assert (Hp3F : h3 ||- foralls g in (G |_| G'), 
    objIsFwded (fun x:nat => None) (map_union phi phi') s1 g).
  apply fasMerge.
  2:assumption.
  sconjMatch.
  eapply fasImplI.
  apply Hp0.
  cbv beta; intros y hh I HpH.
  eapply copyMapFwded.
  2:unfold map_union; eauto.
  instantiate (1 := top ** top); apply sconjAssocL; clear HpH.
  sconjLift5 Hp''; sconjMatch.
  hypReplace Hp1 (fasElim I Hp1); sconjMatch.
  generalize dependent Hp1; unfold objCopiedPhi; unfold objCopied; unfold fieldCopied.
  intro Q; linLogElim Q; linLogIntro.
  3:apply Q.
  instantiate (1 := x0); unfold map_union; rewrite H1; reflexivity.
  apply H7.
  apply H8.
  apply objIsFwdShift.
  eapply objCopiedFwded.
  eauto.
  2:eauto.

  rename Hp0'Y into Hp'; sconjLift3 Hp'; sconjLift3 Hp''.
  assert (HX := objHpIsObjs hobjsOk).
  assert (Q := eqvObjCopied Hp' Hp'' _ HX).

  assert (HpX : s'' |||- ((foralls x in M', objCopied x s' x) ** 
    (foralls x in W', objCopied x s' x)) ** top).
  sconjMatch.
  apply fasSplit.
  unfold hpredTrue; rewrite <- Weqv; assumption.
  assumption.
  linLogElim HpX; sconjMatch.
  eapply fasImplI.
  eauto.
  cbv beta; intros x0 h4 I HpX.

  eapply objCopiedTrans2.
  2:eauto.
  apply WcopyOk.
  rewrite Weqv; apply set_union_intro1; assumption.
  intros x0 y0 I.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Iso1]; clear Iso.
  destruct (phi x0).
  2:reflexivity.
  elimtype False; apply (set_disj_elim MM'disj x0).
  apply Iso1; discriminate.
  destruct Iso' as [Iso' _]; destruct (Iso' x0) as [_ Q]; apply Q.
  rewrite I; discriminate.
  (* done showing forwarding of G |_| G' *)
  clear NSP NSP'.
  
  unfold map_union in H3.

  assert (PhiR : mapX phi (rget s rroot) = Some x).
  destruct H3 as [v0 [HQ Xeq]].
  unfold mapX in Xeq.
  assert (v0 = rget s rroot).
  sconjLift4 Hp; hypReplace Hp (contUpdateSt Hp fromSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp toSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp fromEnd).
  sconjLift7 Hp; hypReplace Hp (contUpdateSt Hp (rget s rroot)).
  sconjLift5 Hp; destruct (aconjElim (sconjDistrib2 Hp)) as [HpH _]; clear Hp.
  sconjLift2 HpH; exact (contContEq HQ HpH).
  subst v0; assumption.

  clear H3.

  assert (GF'disj : set_disj G' F').
  hypReplace Hp3 (sconjInsert Hp3 H3ok); linLogElim Hp3.
  sconjLift5 Hp3; sconjLift3 Hp3.
  eapply fasFasDisj; eauto.
  cbv beta; unfold objIsFwded; unfold fieldIsFwded; intros z h Z; linLogElim Z;
    linLogIntro; sconjMatch.
  cbv beta; intros z h Z; linLogElim Z; linLogIntro; sconjMatch.

  assert (RE : rget s' rtoEnd = toEnd).
  rewrite <- Req'; [stSimpl; reflexivity | regNotInListSolver].

  rewrite RE in Feq.

  assert (G'eq : set_eqv G' (rangeObjs (rget s' rfree) (rget s3 rfree))).
  eapply rangeSplit.
  instantiate (1 := toEnd).
  rewrite <- F'eq.
  assumption.
  destruct Le'; assumption.
  apply set_eqvSymm.
  rewrite <- Feq; rewrite <- F'eq; assumption.
  replace (rget s' rfree) with (rget s'' rfree).
  assumption.
  rewrite Req''; stSimpl; reflexivity.

  assert (GFdisj : set_disj G F).
  sconjLift4 Hp''; sconjLift5 Hp''; eapply fasFasDisj; eauto.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; stSimpl;
    intros h x0 HpX; linLogElim HpX; linLogIntro; sconjMatch.
  simpl; intros h x0 HpX; linLogElim HpX; linLogIntro; sconjMatch.

  assert (Geqv : set_eqv G (rangeObjs toSt (rget s' rfree))).
  eapply rangeSplit.
  instantiate (1 := toEnd).
  rewrite <- Feq.
  assumption.
  destruct stLt'; assumption.
  apply set_eqvSymm; rewrite <- Feq; assumption.
  assumption.

  assert (GGeqv : set_eqv (G |_| G') (rangeObjs toSt (rget s3 rfree))).
  rewrite Geqv; rewrite G'eq; apply rangeObjsTransEqv.
  assumption.
  destruct Al' as [k E]; rewrite E; rewrite Req''; stSimpl; omega.


 (* show the final heap is well-formed *)
  exists (map_union phi phi'); exists (M |_| M'); exists h3.
  constructor.
  linLogIntro.

  assert (Al3 : aligned8 toSt (rget s3 rfree)).
  simpl in al8'; eapply aligned8Trans.
  apply al8'.
  eapply aligned8Trans.
  2:apply Al'.
  rewrite Req''; stSimpl; apply aligned8Refl.

  constructor.
  assumption.
  constructor.
  destruct alToSE as [k R]; rewrite R.
  destruct Al3 as [k' R']; rewrite R'.
  exists (k - k'); omega.
  constructor.
  rewrite <- (Req''' rroot); [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  rewrite <- GGeqv.
  assert (X := isOddOk (rget s rroot)).
  unfold mapX in PhiR; destruct (isOdd (rget s rroot)); [left | right].
  injection PhiR; intro; subst x; assumption.
  apply set_union_intro1; destruct Iso as [_ [Iso _]]; eapply Iso; eauto.

  constructor.
  auto.
  destruct alFrSE; omega.

  sconjLift6 Hp3; linLogElim Hp3.
  sconjLift4 Hp3; sconjLift4 Hp3.
  hypReplace Hp3 (sconjAssocR Hp3); hypReplace Hp3 (sconjAssocR Hp3).
  sconjLift3 Hp3; sconjLift3 Hp3.

  rewrite <- (Req''' rgcInfo); [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  rewrite <- Req'; [stSimpl | regNotInListSolver].
  sconjMatch.

  (* showing the objs heap is okay *)
  apply aconjIntro.
  2:assumption.
  eapply heapEqvTrue.
  apply heap_eqvSymm; apply (eqvPredEqv Hp0).
  unfold hpredTrue; rewrite <- GGeqv.
  rewrite <- hpredTrueEq; eapply objHpIsFwded; eauto.

  apply objsIsBuffer.
  unfold hpredTrue; rewrite <- F'eq; assumption.

  (* aligned8 (rget s3 rfree) toEnd *)
  destruct Al' as [k E]; rewrite E; rewrite Req''; stSimpl.
  destruct al8' as [k' E']; rewrite E'; destruct alToSE as [k'' E'']; rewrite E''.
  exists (k'' - (k + k')).
  cut (k + k' <= k'').
  intro.
  omega.
  destruct Le' as [_ Le']; 
  replace (rget s3 rfree) with (toSt + 8 * k' + 8 * k) in Le'.
  rewrite E'' in Le'.
  omega.
  rewrite <- E'.
  rewrite E; rewrite Req''; stSimpl; reflexivity.

  apply objsIsBuffer.
  2:assumption.
  unfold hpredTrue; rewrite <- (rangeObjsTransEqv alSt LEff).
  apply fasMerge; sconjMatch.
  subst objs; unfold hpredTrue; rewrite objsEq; rewrite Weqv.
  rewrite <- (set_eqvUnionAssoc M M' W').
  apply fasMerge; sconjMatch.
  eapply mapHpIsObjs; eauto.
  eapply objHpIsObjs; eauto.
  (* set_disj (M |_| M') W' *)
  apply set_unionDisjIntro.
  apply set_disjSymm; eapply set_unionDisjElimR; instantiate (1 := M').
  rewrite <- Weqv; assumption.
  assumption.
  apply bufferIsObjs; assumption.

  apply rangeSplitDisj.

  rewrite <- (Req''' rroot); [stSimpl | regNotInListSolver].
  rewrite (Req'' rroot); stSimpl.

  (* (h0 ||- minObjHp (rget s rroot :: nil) (M |_| M') ** top *)
  constructor.
  unfold minObjHp.
  assert (HpL : h0 ||- objHp objs (M |_| M') ** top).
  assert (HpL : h0 ||- objHp objs (M |_| M') ** objHp objs W').
  apply objHpSplit.
  unfold hpredTrue; rewrite (set_eqvUnionAssoc M M' W'); subst objs.
  rewrite <- hpredTrueEq; eapply heapEqvImplL; eauto.
  instantiate (2 := heapOf s); sconjMatch.
  destruct (aconjElim Hp) as [Hp1 Hp2]; apply aconjIntro.
  unfold hpredTrue; rewrite <- Weqv; rewrite <- objsEq; assumption.
  assumption.

  apply set_unionDisjIntro.
  apply set_disjSymm; eapply set_unionDisjElimR; instantiate (1 := M').
  rewrite <- Weqv; assumption.  

  assumption.

  sconjMatch.

  apply (sconjImplS HpL).
  2:intros; linLogIntro.
  intros h S Hp'.
  assert (h ||- objHp (M |_| M') (M |_| M')).

  eapply objHpUnforward; eauto.
  eapply subHeapTrans; eauto.
  destruct (sconjDistrib2 Hp) as [_ Hp1].
  sconjLift4 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 toSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromEnd).
  sconjLift7 Hp1; hypReplace Hp1 (contUpdateSt Hp1 (rget s rroot)).
  sconjLift5 Hp1; eapply sconjEqSubHeap; apply Hp1.

  assert (OK : okFieldVal (M |_| M') (rget s rroot)).
  assert (X := isOddOk (rget s rroot)).
  unfold mapX in PhiR; destruct (isOdd (rget s rroot)); [left | right].
  assumption.
  apply set_union_intro1; destruct Iso as [Iso _]; destruct (Iso (rget s rroot)) as [_ Q];
    apply Q; rewrite PhiR; discriminate.

  apply aconjIntro.
  intros z I.

  assert (rootReachable (heapOf s1) (rget s rroot::nil) z).
  assert (OO := isOddOk (rget s rroot)).
  unfold mapX in PhiR; destruct (isOdd (rget s rroot)).
  
  (* case: root is atomic *)
  elimtype False.
  destruct Iso'' as [Iso'' ISO]; destruct (Iso'' z) as [X _].
  hypReplace X (X I); clear I Iso''.
  destruct ISO as [ISO _]; hypReplace ISO (ISO z).
  destruct (map_union phi phi' z).
  assert (II : set_In a (G |_| G')).
  eauto.
  hypReplace reach (reach _ II).
  replace (rget s3 rroot) with x in reach.
  injection PhiR; intro U; subst x; clear PhiR X.
  destruct reach as [u [V X]].
  simpl in V; destruct V.
  subst u.
  apply (not_even_and_odd (rget s rroot)).
  2:assumption.
  apply ptrPeven.
  inversion X.
  subst x a.
  hypReplace Hp3F (fasElim II Hp3F).
  cbv beta in Hp3F; unfold objIsFwded in Hp3F; unfold fieldIsFwded in Hp3F; linLogElim Hp3F.
  apply (contElim1L Hp3F).
  unfold Heap.get in H0; assert (Q := isPtrOk (rget s rroot)).
  destruct (isPtr (rget s rroot)); [assumption | discriminate H0].
  unfold Heap.get in H0; assert (Q := isPtrOk (4 + rget s rroot)).
  destruct (isPtr (4 + rget s rroot)); [idtac | discriminate H0].
  inversion Q; assumption.
  assumption.
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl; reflexivity.
  elimtype False; apply X; reflexivity.

  (* case: root is a pointer *)
  eapply rootReachForward.
  apply Iso''.
  assert (X := sconjInsert Hp3 Hp3F).
  instantiate (1 := s3); sconjMatch.
  apply reach.
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  unfold map_union; rewrite PhiR; reflexivity.
  assumption.

  eapply rootReachableSubheap.
  eapply heapEqvTrue.
  2:apply H.
  apply Heap.heap_eqvSymm; apply Heap.HeapMap.map_canonEqv.
  3:apply H0.
  2:assumption.
  eapply Heap.subHeapTrans.
  instantiate (1 := h).
  intros q r EE; rewrite Heap.heap_canonEqv; assumption.
  eapply Heap.subHeapTrans.
  eauto.
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp1].
  sconjLift4 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 toSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromEnd).
  sconjLift7 Hp1; hypReplace Hp1 (contUpdateSt Hp1 (rget s rroot)).
  sconjLift5 Hp1; eapply sconjEqSubHeap; eauto.

  (* show that M |_| M' is a closed object heap *)
  unfold okObjHp; linLogIntro.
  simpl; intros v I; destruct I as [I | I].
  subst v; assumption.
  elimtype False; assumption.
  assumption.

  constructor.
  rewrite <- GGeqv; assumption.
  constructor.
  unfold mapX in PhiR |- *; destruct (isOdd (rget s rroot)).
  assumption.
  unfold map_union; rewrite PhiR; reflexivity.

  (* show that the objects are forwarded properly *)
  assert (Hp1 := Hp).
  sconjLift4 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 toSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromEnd).
  sconjLift7 Hp1; hypReplace Hp1 (contUpdateSt Hp1 (rget s rroot)).
  sconjLift5 Hp1.

  unfold hpredTrue; rewrite <- GGeqv; rewrite <- hpredTrueEq; eapply fasImpl.
  apply Hp3F.
  cbv beta; unfold objIsFwded; unfold fieldIsFwded; unfold objIsFwdedH;
    intros z h HpZ; linLogElim HpZ.
  linLogIntro.
  2:simpl; eauto.
  destruct H6 as [v0 [Hpv0 M0]]; destruct H7 as [v1 [Hpv1 M1]].
  exists x0; exists v0; exists v1; constructor.
  assumption.
  constructor.
  simpl.
  assert (Sub : Heap.subHeap h0 (heapOf s1)).
  destruct (aconjElim (sconjDistrib2 Hp1)) as [Hp1a Hp1b].
  eapply sconjEqSubHeap; eauto.
  linLogIntro.
  assert (Hp0 := heapEqvImplL Hp1).
  assert (I : set_In x0 (rangeObjs fromSt (rget s rfree))).
  subst objs; rewrite objsEq; rewrite Weqv.
  rewrite <- (set_eqvUnionAssoc M M' W').
  apply set_union_intro1.
  destruct Iso'' as [Iso'' _]; destruct (Iso'' x0) as [_ MM]; apply MM;
    rewrite H1; discriminate.
  hypReplace Hp0 (fasElim I Hp0);
  cbv beta in Hp0; unfold okObjAt in Hp0; unfold okFieldAt in Hp0; linLogElim Hp0.
  assert (E11 := subHeapContEq Sub Hpv0 Hp0); subst x3.
  sconjLift2 Hp0; assert (E12 := subHeapContEq Sub Hpv1 Hp0); subst x4.
  sconjLift2 Hp0; sconjMatch.

  (* showing the mapping matches up *)
  unfold map_union in M0, M1 |- .
  unfold mapX in M0, M1 |- * .
  constructor.
  destruct (isOdd v0); assumption.
  destruct (isOdd v1); assumption.


  (* show the registers are okay *)
  destruct Guar2 as [_ [Req'' RaEq]].
  rewrite <- Req''; [stSimpl | regNotInListSolver].
  rewrite <- Rp'; [stSimpl | regNotInListSolver].
  constructor.
  2:reflexivity.
  intros r X; unfold calleeSaved in X; simpl in X.
  destruct X as [X | [X | [X | [X | [X | [X | [X | [X | [X | X]]]]]]]]];
    try (subst r; 
      first [rewrite RaEq |
	rewrite <- Req''; [stSimpl | regNotInListSolver]];
      rewrite <- Rp'; [stSimpl | regNotInListSolver]; reflexivity).
  elimtype False; assumption.

Qed.


Lemma cheneyIsoOk' : forall s s' objs objs' phi h h' rr,
  okFieldVal objs (rget s rroot) ->
  iso_set phi objs objs' ->
  mapX phi (rget s rroot) = Some (rget s' rroot) ->

  h ||- minObjHp (List.map (rget s) (rroot :: rr)) objs ->

  h' ||- objHp objs' objs' ->
  h' ||- (foralls x in objs', objIsFwdedH phi h x) ->

  (forall r' : rnum, In r' (rroot :: rr) -> r' = rroot) ->

  (h' ||- Heap.heap_eqv h' ** top) /\
  isoState (rroot :: rr) (h, rfileOf s) (h', rfileOf s').

Proof.
  intros s s' objs objs' phi h h' rr rootOk Iso rootMap.
  intros Hp Hp' HpFwd RRok.

  constructor.
  assert (UU : h' ||- Heap.heap_eqv h').
  apply eqvPredEqv'.
  apply Heap.heap_eqvRefl.
  hypReplace UU (unitIR UU).
  sconjMatch.

  unfold isoState.
  exists objs; exists objs'; exists phi.
  constructor.
  destruct (aconjElim Hp) as [_ Hp2].
  unfold okObjHp in Hp2; linLogElim Hp2.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  2:assumption.
  intros r I.
  hypReplace RRok (RRok _ I); subst r.
  apply rootMap.
Qed.
  

Lemma cheneyIsoOk : forall s s' vs objs objs' phi h h'  rr,
  rget vs rroot = rget s rroot ->
  okFieldVal objs (rget vs rroot) ->
  mapX phi (rget s rroot) = Some (rget s' rroot) ->
  vs |||- (minObjHp (List.map (rget vs) (rroot :: rr)) objs && Heap.heap_eqv h) ** top ->
  (forall r' : rnum, In r' (rroot :: rr) -> r' = rroot) ->
  iso_set phi objs objs' ->
  h' ||- (foralls x in objs', objIsFwdedH phi (heapOf vs) x) ->
  okFieldVal objs' (rget s' rroot) ->
  s |||- Heap.heap_eqv (heapOf vs) ** top ->
  s' |||- (objHp objs' objs' && Heap.heap_eqv h') ** top ->
    (h' ||- Heap.heap_eqv h' ** top) /\ isoState (rroot :: rr) (h, rfileOf vs) (h', rfileOf s').
Proof.
  intros s s' vs objs objs' phi h h' rr rootEq rootOk rootMap Hp rphi iso 
    Hp' rootOk' Hp0 Hp0'.
  eapply cheneyIsoOk'.
  2:eauto.
  assumption.
  rewrite rootEq; assumption.
  eapply heapEqvImplL; eauto.
  eapply heapEqvImplL; eauto.
  2:assumption.

  clear rootEq rootOk rootOk' rootMap rphi Hp0 Hp0'.
  apply (fasImplI Hp'); clear Hp'.
  unfold objIsFwdedH; intros x h0 I Hp0; linLogElim Hp0.
  destruct H1 as [x' [v0 [v1 [Peq [Hp' [Meq0 Meq1]]]]]].
  linLogIntro.
  2:eauto.
  exists x'; exists v0; exists v1; intuition.
  assert (S : Heap.subHeap h (heapOf vs)).
  destruct (aconjElim (sconjDistrib2 Hp)) as [Hp1a Hp1b].
  eapply sconjEqSubHeap; eauto.

  hypReplace Hp (heapEqvImplL Hp).
  destruct (aconjElim Hp) as [_ Hp1]; clear Hp.
  unfold okObjHp in Hp1; linLogElim Hp1; clear H1.

  assert (I' : set_In x' objs).
  destruct iso.
  destruct (H x') as [_ Q]; apply Q; rewrite Peq; discriminate.

  hypReplace Hp1 (objHpElim I' Hp1); unfold okObjAt in Hp1; unfold okFieldAt in Hp1;
    linLogElim Hp1.
  linLogIntro.
  linLogElim Hp'.
  
  assert (E1 := subHeapContEq S Hp' Hp1).
  sconjLift2 Hp'; sconjLift2 Hp1.
  assert (E2 := subHeapContEq S Hp' Hp1).
  subst x2 x3.
  sconjLift2 Hp1; sconjMatch.

Qed.


Lemma cheneyAllocOk : iseqOkS P cheneyAllocPre cheneyAllocGuar cheneyAlloc.
Proof.
  unfold cheneyAlloc.
  iseqWeakSolver Psimpl.
  
  intros; apply cheneyEnterGuarPresRA; assumption.

  unfold cheneyAllocPre.
  intros s [vs Hp].
  unfold mview in Hp; linLogElim Hp.
  rename x into toSt; rename x0 into toEnd; 
    rename x1 into frSt; rename x2 into frEnd; rename x3 into free.
  destruct H3 as [_ [al1 [al2 [rootOk0 [sizeOk Le]]]]].
  assert (rootOk : rget vs rroot = RFile.get (rfileOf s) rroot /\ 
    okFieldVal (rangeObjs frSt free) (rget vs rroot)).
  apply rootOk0; simpl; auto.
  clear rootOk0; destruct rootOk as [rootEq rootOk].
  assert (Hp' := Hp).
  sconjLift8 Hp'.
  rewrite addiuStep0; stSimpl.
  rewrite (lwStep Hp'); try reflexivity.
  sconjLift8 Hp'.
  rewrite (lwStep Hp'); try reflexivity.
  stSimpl.
  assert (BT := doBtestSound btne free frEnd).
  clear Hp'.
  destruct (doBtest btne free frEnd); simpl in BT.

  (* case: do allocation *)
  unfold cheneyDoAllocPre; unfold cheneyDoAllocGuar; stSimpl.

  assert (I : set_In free (rangeObjs free frEnd)).
  destruct al2 as [k E]; subst frEnd.
  destruct k.
  elimtype False; auto.
  eapply rangeObjsIntro.
  2:instantiate (1 := 0).
  omega.
  omega.
  constructor.
  
  sconjLift8 Hp; sconjLift3 Hp.
  assert (Hp' : s |||- (free |->__ ** top) ** 16 + rget s rgcInfo |->_ ** top).
  sconjMatch.
  hypReplace Hp (bufferIsObjs _ al2 Hp).
  hypReplace Hp (fasElim I Hp).
  sconjMatch.
  linLogIntro; eauto.
  linLogElim Hp'; linLogIntro; eauto.

  intros s' [G Req].
  constructor.

  rename Hp into Hp0; rename vs into vs0; clear rootEq rootOk.
  intros vs A Hp.
  unfold mview in Hp; linLogElim Hp.
  assert (x = toSt).
  sconjLift4 Hp0; sconjLift4 Hp; exact (contContEq Hp Hp0).
  assert (x0 = toEnd).
  sconjLift5 Hp0; sconjLift5 Hp; exact (contContEq Hp Hp0).
  assert (x1 = frSt).
  sconjLift6 Hp0; sconjLift6 Hp; exact (contContEq Hp Hp0).
  assert (x2 = frEnd).
  sconjLift7 Hp0; sconjLift7 Hp; exact (contContEq Hp Hp0).
  assert (x3 = free).
  sconjLift8 Hp0; sconjLift8 Hp; exact (contContEq Hp Hp0).
  subst x x0 x1 x2 x3.
  destruct H3 as [_ [_ [_ [rootOk0 _]]]].
  assert (rootOk : rget vs rroot = RFile.get (rfileOf s) rroot /\
    okFieldVal (rangeObjs frSt free) (rget vs rroot)).
  apply rootOk0; simpl; auto.
  clear rootOk0 Hp0 vs0.
  destruct rootOk as [rootEq rootOk].
  exists vs; exists free.
  constructor.
  apply gcGuarRefl.
  
  unfold mview.
  assert (Hp' : s |||- (free |->__ ** buffer (8 + free) frEnd) ** 
    (objHp (rangeObjs frSt free) (rangeObjs frSt free) && Heap.heap_eqv (heapOf vs)) **
    buffer toSt toEnd **
    RFile.get (rfileOf s) rgcInfo |->, (toSt::toEnd::frSt::frEnd::nil) ** 
    16 + rget s rgcInfo |->_ ** A).
  sconjLift2 Hp.
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  hypReplace Hp (bufferIsObjs _ al2 Hp).
  hypReplace Hp (fasElim I Hp).
  sconjMatch.
  apply objsIsBuffer.
  unfold hpredTrue; rewrite <- (rangeRemoveEqv free frEnd); assumption.
  destruct al2 as [k E]; subst frEnd.
  destruct k.
  elimtype False; auto.
  exists k; omega.
  linLogIntro; eauto.

  constructor.
  hypReplace Hp' (sconjAssocL Hp').
  sconjLift6 Hp'; sconjLift2 Hp'.
  hypReplace Hp' (G _ Hp'); clear G Hp.
  replace (RFile.get (rfileOf s) rgcInfo) with (rget s rgcInfo) in Hp'.
  2:reflexivity.
  replace (rget s rgcInfo) with (rget s' rgcInfo) in Hp'.
  2:rewrite (Req rgcInfo); stSimpl; reflexivity.
  apply sconjComm; sconjLift7of7 Hp'; sconjMatch.
  existsl toSt; existsl toEnd; existsl frSt; existsl frEnd; existsl (8 + free).
  linLogIntro.

  constructor.
  simpl; intros [D | [D | D]]; auto; discriminate D.
  constructor.
  destruct al1 as [k E]; exists (S k); omega.
  constructor.
  destruct al2 as [k E]; destruct k.
  elimtype False; omega.
  exists k; omega.
  constructor.
  2:intuition.
  intros r Ir.
  replace (RFile.get (rfileOf s') r) with (rget s' r); [idtac | reflexivity].
  simpl in Ir; destruct Ir as [D | [D | D]].
  subst r.
  rewrite Req; stSimpl.
  constructor.
  reflexivity.
  right.
  
  destruct (rangeObjsElim I) as [X _].
  destruct al1.
  eapply rangeObjsIntro.
  auto.
  instantiate (1 := x); assumption.

  subst r.
  stSimpl.
  rewrite Req; stSimpl.
  constructor.
  assumption.
  destruct rootOk; [left | right].
  assumption.
  2:elimtype False; assumption.

  destruct (rangeObjsElim H) as [X [k Y]].
  eapply rangeObjsIntro.
  omega.
  eauto.

  (* showing the heap is okay *)
  stSimpl.
  replace (rget s' v0) with free in Hp.
  2:rewrite Req; stSimpl; reflexivity.

  assert (SE : set_eqv (rangeObjs frSt (8 + free)) (set_add free (rangeObjs frSt free))).
  apply set_eqvSymm; apply rangeObjsEndAdd; assumption.
  sconjLift4 Hp; hypReplace Hp (sconjAssocR Hp).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  destruct (aconjElim (sconjDistrib2 Hp)) as [Hp1 Hp2]; clear Hp.
  apply aconjIntro.
  unfold hpredTrue; rewrite SE.
  apply objHpIntro.
  apply sconjComm; clear Hp2; sconjMatch.
  eapply objHpWeaken; eauto.
  intros x IX; apply set_add_intro2; assumption.
  simpl in Hp; unfold okObjAt; unfold okFieldAt; linLogIntro; eauto;
    left; constructor; constructor.
  clear Hp1; rename Hp2 into Hp.
  simpl in Hp; sconjLift2 Hp.
  simpl; apply unitER.
  apply heapEqvAdd.
  apply sconjGLift2'.
  apply heapEqvAdd.
  linLogIntro; assumption.
  sconjLift3 Hp; sconjLift3 Hp.
  sconjMatch.
  sconjLift2of2 Hp.
  linLogElim Hp; apply unitER; linLogIntro.
  sconjMatch.
  linLogIntro; assumption.

  (* showing that the object is fresh *)
  intros B Vhp; stSimpl.
  clear Hp G Req; rename Hp' into Hp.
  simpl; replace (S (S (S (S free)))) with (4 + free); [idtac | reflexivity].
  linLogElim Hp.

  assert (X : s |||- (!(ptrP free) ** top) ** top).
  sconjMatch; linLogIntro; eapply contElim1; eauto.

  linLogElim X; clear X; rename H1 into freeP.
  apply sconjComm; linLogIntro; apply sconjComm; linLogIntro.
  sconjLift4 Hp; hypReplace Hp (sconjDistrib2 Hp); destruct (aconjElim Hp) as [_ Hp'];
    clear Hp; rename Hp' into Hp.
  apply heapAddOk.
  apply sconjComm; apply heapAddOk.
  assumption.
  assumption.
  intro Vhp'; linLogElim Vhp'.
  hypReplace Hp (sconjInsert Hp Vhp'); linLogElim Hp.
  destruct (ptrP_elim freeP) as [k E]; apply (ptrP_intro (S k)).
  omega.
  intro Vhp'; linLogElim Vhp'.
  sconjLift2 Hp; hypReplace Hp (contUpdateSt Hp 1).
  hypReplace Hp (heapEqvAdd Hp).
  hypReplace Hp (sconjInsert Hp Vhp'); linLogElim Hp.

  clear Hp G.
  unfold CMGI.presReg; unfold calleeSaved.
  intros r Z.
  simpl in Z.
  rewrite Req.
  intuition; subst; stSimpl; reflexivity.

  (* case: call GC before allocation *)
  replace (RFile.get (rfileOf s) rgcInfo) with (rget s rgcInfo) in Hp; 
    [idtac | reflexivity].
  subst free.
  apply callChain.
  replace (RFile.get (rfileOf s) rroot) with (rget s rroot) in rootEq; [idtac | reflexivity].

  constructor.
  unfold cheneyEnterPre; stSimpl.
  exists toSt; exists toEnd; exists frSt; exists frEnd; exists (heapOf vs).
  unfold cheneyStOk; linLogIntro.
  intuition.
  rewrite <- rootEq; assumption.
  sconjMatch.

  unfold cheneyEnterGuar; stSimpl.
  intros s' [Guar [Req RaEq]].
  constructor.
  destruct (Guar toSt toEnd frSt frEnd (16 + rget s rgcInfo |-> frEnd ** top) (heapOf vs)) as
    [phi [objs [h' X]]].
  unfold cheneyStOk; linLogIntro.
  intuition.
  rewrite <- rootEq; assumption.
  sconjMatch.
  destruct X as [Hp' [Iso [RootOk Vhp]]].
  exists toEnd.
  constructor.
  unfold cheneyStOk in Hp'; lle Hp'; sconjLift7 Hp'; sconjMatch.
  intros Ne.
  unfold cheneyDoAllocPre.
  rewrite <- (Req rgcInfo).
  stSimpl.
  unfold cheneyStOk in Hp'; linLogElim Hp'.
  assert (HH : s' |||- (rget s' rfree |->__ ** top) ** 16 + rget s rgcInfo |->_ ** top).
  sconjLift8 Hp'; sconjLift3 Hp'; sconjMatch.
  2:linLogIntro; eauto.

  destruct H1 as [_ [X _]].
  hypReplace Hp0 (bufferObjElim _ Hp0 X Ne).
  sconjMatch.
  linLogElim HH; linLogIntro; eauto.
  unfold calleeSaved; unfold rgcInfo.
  right; right; right; right; right; right;
    right; right; left; reflexivity.


  (* show the guarantee holds together *)
  intros s'' Guar2; constructor.

  rename vs into vs0; rename Hp into Hp0.
  clear rootOk.
  unfold mview; intros vs A Hp.
  linLogElim Hp.
  replace (RFile.get (rfileOf s) rgcInfo) with (rget s rgcInfo) in Hp;
    [idtac | reflexivity].
  destruct H3 as [_ [_ [_ [rootOk _]]]].

  assert (x = toSt).
  sconjLift4 Hp0; sconjLift4 Hp; exact (contContEq Hp Hp0).
  assert (x0 = toEnd).
  sconjLift5 Hp0; sconjLift5 Hp; exact (contContEq Hp Hp0).
  assert (x1 = frSt).
  sconjLift6 Hp0; sconjLift6 Hp; exact (contContEq Hp Hp0).
  assert (x2 = frEnd).
  sconjLift7 Hp0; sconjLift7 Hp; exact (contContEq Hp Hp0).
  assert (x3 = frEnd).
  sconjLift8 Hp0; sconjLift8 Hp; exact (contContEq Hp Hp0).
  subst x x0 x1 x2 x3.
  clear Hp0 vs0 rootEq.

  hypReplace rootOk (rootOk rroot).
  assert (Z : rget vs rroot = rget s rroot /\ okFieldVal (rangeObjs frSt frEnd) (rget vs rroot)).
  clear Guar Hp.
  replace (RFile.get (rfileOf s) rroot) with (rget s rroot) in rootOk; [idtac | reflexivity].
  apply rootOk; simpl; auto.

  clear rootOk; destruct Z as [rootEq rootOk].
  destruct (Guar toSt toEnd frSt frEnd (16 + rget s rgcInfo |-> frEnd ** A) (heapOf vs)) as
    [phi [objs [h' X]]]; clear Guar.

  unfold cheneyStOk; linLogIntro.
  intuition.
  rewrite <- rootEq; assumption.
  sconjMatch.

  destruct X as [Hp' [MHp [Iso [rootIso Vhp]]]].
  unfold cheneyPostGCGuar in Guar2.
  unfold cheneyStOk in Hp'; linLogElim Hp'.
  destruct H1 as [Al1' [Al2' [rootOk' _]]].

  assert (Hp'2 : s' |||- 12 + rget s' rgcInfo |-> toEnd ** top).
  sconjLift7 Hp'; sconjMatch.
  hypReplace Guar2 (Guar2 _ Hp'2); clear Hp'2.
  compare (rget s' rfree) toEnd; intro NE.
  
  subst toEnd; rewrite beq_true in Guar2; elimtype False; assumption.

  rewrite (beq_false NE) in Guar2.
  destruct Guar2 as [Guar' Req'].

  assert (GIeq : rget s' rgcInfo = rget s rgcInfo).
  rewrite <- Req.
  stSimpl; reflexivity.
  unfold calleeSaved; unfold rgcInfo;
    right; right; right; right; right; right;
      right; right; left; reflexivity.
  generalize dependent Guar'; stSimpl; intro.
  rewrite GIeq in Guar'; rewrite GIeq in Hp'.
  sconjLift8 Hp'; sconjLift3 Hp'.
  assert (Hp'2 : s' |||- (rget s' rfree |->__ ** buffer (rget s' rfree + 8) toEnd) **
    16 + rget s rgcInfo |->_ ** (objHp (rangeObjs toSt (rget s' rfree))
      (rangeObjs toSt (rget s' rfree)) && Heap.heap_eqv h') **
    buffer frSt frEnd ** rget s rgcInfo |-> frSt ** 4 + rget s rgcInfo |-> frEnd **
      4 + (4 + rget s rgcInfo) |-> toSt ** 4 + (4 + (4 + rget s rgcInfo)) |-> toEnd ** A).
  sconjMatch; clear Guar' Hp.
  hypReplace Hp0 (bufferObjElim _ Hp0 Al2' NE).
  assumption.
  linLogIntro; eauto.

  hypReplace Hp'2 (sconjAssocL Hp'2).
  sconjLift3 Hp'2; sconjLift2 Hp'2.
  hypReplace Guar' (Guar' _ Hp'2); clear Hp' Hp'2.
  rename Guar' into Hp''.

  sconjLift4 Hp''.
  generalize dependent Hp''; rewrite (Req' v0); stSimpl; intro.
  replace (rget s' rfree |->, (1 :: 1 :: nil)) with
    (rget s' rfree |-> 1 ** 4 + rget s' rfree |-> 1) in Hp''; [idtac | reflexivity].
  exists (h', rfileOf s''); exists (rget s' rfree).
  stSimpl.

  assert (Hp''2 : s'' |||- (objHp (rangeObjs toSt (rget s' rfree))
                 (rangeObjs toSt (rget s' rfree)) && 
               Heap.heap_eqv h') **
              rget s' rfree |-> 1 ** 4 + rget s' rfree |-> 1 **
              buffer (rget s' rfree + 8) toEnd **
              buffer frSt frEnd **
              rget s rgcInfo |->, (frSt::frEnd::toSt::toEnd::8+rget s' rfree::nil) ** A).
  clear Hp Req.
  sconjLift5 Hp''; sconjLift5 Hp''.
  sconjLift4 Hp''; sconjLift4 Hp''.
  linLogElim Hp''.
  apply (sconjImplR Hp''); clear Hp''; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  assumption.

  clear Hp''; rename Hp''2 into Hp''.
  constructor.
  unfold CMGI.gcGuar.
  unfold CheneyGCGuar.gcGuar.
  intros objs0 h rr' I Vhp'.
  assert (D := inclEq I); clear I.

  destruct rr' as [ | r rr'].

  (* case: no roots *)
  exists Heap.empty_heap.

  constructor.
  assert (HpX : h' ||- unit ** top).
  linLogIntro.
  sconjMatch.
  assert (H'eqv : heap_eqv h Heap.empty_heap).
  apply unitElim.
  destruct (sconjElim Vhp') as [ha [hb [_ [HA _]]]].
  eapply noRootsEmptyMinObjHp; eauto.

  assert (SE : state_eqv (h, rfileOf vs) (Heap.empty_heap, rfileOf vs)).
  constructor.
  assumption.
  intro x; reflexivity.

  eapply isoStateMorph.
  apply state_eqvSymm; apply SE.
  apply state_eqvRefl.
  apply isoStateEmpty.

  (* case: just one root *)

  (* first a little cleanup, including dumping s' *)
  stSimpl.
  assert (r = rroot).
  apply D; simpl; left; reflexivity.
  subst r.

  generalize dependent (rget s' rfree); intro free'; intros.

  assert (RootEq : rget s' rroot = rget s'' rroot).
  rewrite Req'; stSimpl; reflexivity.

  rewrite RootEq in rootIso; rewrite RootEq in rootOk'.
  clear s' Req GIeq Req' RootEq RaEq.
  rename s'' into s'.

  assert (OE : set_eqv objs objs0).
  eapply minObjHpEqv; eauto.
  destruct (aconjElim (sconjDistrib2 Vhp')); eauto.
  constructor; intro I.
  destruct I.
  left; rewrite rootEq; assumption.
  elimtype False; assumption.
  rewrite <- rootEq.
  replace (rget vs rroot :: nil) with (set_add (rget vs rroot) empty_set).
  rewrite <- (regListMapEqv (rget vs) D).
  assumption.
  reflexivity.

  exists h'.
  eapply cheneyIsoOk.
  6:apply Iso.
  eauto.
  assert (vs |||- (!(okFieldVal objs (rget vs rroot)) ** top) ** top).
  clear Vhp'; sconjMatch.
  linLogIntro.
  eapply minObjHpRootOk; eauto.
  rewrite rootEq; left; reflexivity.
  linLogElim H; assumption.
  assumption.
  sconjMatch.
  destruct (aconjElim Hp0) as [HpA HpB]; apply aconjIntro.
  unfold hpredTrue; rewrite OE; assumption.
  assumption.
  assumption.
  assumption.
  assumption.
  sconjMatch; destruct (aconjElim Hp); assumption.
  sconjMatch.

  constructor.
  sconjLift7of7 Hp''; apply sconjComm; sconjMatch.
  existsl frSt; existsl frEnd; existsl toSt; existsl toEnd; existsl (8 + rget s' rfree).
  linLogIntro.
  constructor.
  simpl; intros [X | [X | X]]; auto; discriminate X.
  constructor.
  destruct Al1' as [k Al1']; rewrite Al1'; exists (S k); omega.
  constructor.
  destruct Al2' as [k Al2']; rewrite Al2'.
  destruct k; [elimtype False | exists k]; omega.

  constructor.
  intros r I; simpl in I; destruct I as [I | [I | I]].
  3:elimtype False; assumption.
  subst r; stSimpl.
  replace (RFile.get (rfileOf s'') v0) with (rget s'' v0); [rewrite Req'; stSimpl | reflexivity].
  constructor.
  reflexivity.
  destruct Al1' as [k Al1'].
  right; eapply rangeObjsIntro.
  auto.
  rewrite Al1'; reflexivity.

  subst r; stSimpl.
  constructor.
  reflexivity.
  replace (rget (h', rfileOf s'') rroot) with (rget s'' rroot); 
    [rewrite Req' ; stSimpl | reflexivity].
  destruct rootOk' as [H | H]; [left | right].
  assumption.
  destruct (rangeObjsElim H) as [B [k C]].
  rewrite C; eapply rangeObjsIntro.
  2:reflexivity.
  omega.
  constructor.
  auto.
  destruct al1; omega.
  
  clear Hp; rename Hp0 into Hp.
  hypReplace Hp (sconjAssocR Hp).
  hypReplace Hp (sconjAssocR Hp).
  sconjMatch.
  linLogElim Hp.
  destruct (aconjElim (sconjDistrib2 Hp)) as [Hp1 Hp2]; apply aconjIntro; clear Hp;
    [clear Hp2 | clear Hp1].
  unfold hpredTrue; rewrite <- (rangeObjsEndAdd Al1').
  apply objHpIntro.
  apply sconjComm; sconjMatch.
  apply (objHpWeaken Hp).
  intros x I; apply set_add_intro2; assumption.
  unfold okObjAt; unfold okFieldAt; linLogIntro.
  2:eauto.
  left; constructor; constructor.
  left; constructor; constructor.
  sconjLift2 Hp2; hypReplace Hp2 (heapEqvAdd Hp2).
  sconjLift2of2 Hp2.
  apply unitER.
  replace (heapOf (uhset (uhset (h', rfileOf s'') (rget s' rfree) 1) (4 + rget s' rfree) 1)) with
    (Heap.uset (Heap.uset h' (rget s' rfree) 1) (4 + rget s' rfree) 1). 
  2:reflexivity.
  apply heapEqvAdd.
  linLogIntro; assumption.
  replace (8 + rget s' rfree) with (rget s' rfree + 8).
  assumption.
  omega.
  replace (RFile.get (rfileOf s'') rgcInfo) with (rget s'' rgcInfo); [idtac | reflexivity].
  rewrite Req'; stSimpl; rewrite <- Req; stSimpl.
  assumption.
  right; right; right; right; right; right; right; right; left; reflexivity.

  assert (X : s'' |||- (!(ptrP (rget s' rfree)) ** top) ** top).
  sconjLift2 Hp''.
  sconjMatch.
  linLogIntro.
  eapply contElim1; eauto.
  linLogElim X; clear X; rename H1 into frP.

  assert (Hp' : s'' |||- Heap.heap_eqv h' ** rget s' rfree |-> 1 **
              4 + rget s' rfree |-> 1 ** top).
  sconjMatch.
  destruct (aconjElim Hp0); assumption.

  clear Hp Hp''; intros B Hp.
  apply sconjComm; linLogIntro; apply sconjComm; linLogIntro.
  replace (heapOf (uhset (uhset (h', rfileOf s'') (rget s' rfree) 1) (4 + rget s' rfree) 1)) with
    (Heap.uset (Heap.uset h' (rget s' rfree) 1) (4 + rget s' rfree) 1). 
  2:reflexivity.

  apply heapAddOk.
  apply sconjComm; apply heapAddOk.
  assumption.
  assumption.
  intro Vhp'; linLogElim Vhp'.
  hypReplace Hp (sconjInsert Hp' Vhp'); linLogElim Hp.
  destruct (ptrP_elim frP) as [k E]; apply (ptrP_intro (S k)).
  omega.
  intro Vhp'; linLogElim Vhp'.
  sconjLift2 Hp'; hypReplace Hp' (contUpdateSt Hp' 1).
  hypReplace Hp' (heapEqvAdd Hp').
  hypReplace Hp' (sconjInsert Hp' Vhp'); linLogElim Hp'.

  (* show the registers are preserved *)
  unfold regPres; unfold CMGI.presReg.
  intros r C.
  destruct (Guar toSt toEnd frSt frEnd top (heapOf vs)) as [phi [h' X]].
  unfold cheneyStOk.
  linLogIntro.
  rewrite <- rootEq.
  intuition.
  sconjMatch.

  clear Hp Guar.
  destruct X as [h'' [X _]].
  assert (Hp' : s' |||- 12 + rget s' rgcInfo |-> toEnd ** top).
  unfold cheneyStOk in X; linLogElim X.
  sconjLift7 X; sconjMatch.

  clear X; assert (X := Guar2 _ Hp'); clear Guar2 Hp'.
  destruct (beq_nat (rget s' rfree) toEnd).
  elimtype False; assumption.
  destruct X as [_ X].
  rewrite X; clear X sizeOk Le rootEq al1 al2 rootOk.
  assert (C' := C).
  unfold calleeSaved in C'; simpl in C'.
  destruct C' as [X | [X | [X | [X | [X | [X | [X | [X | [X | X]]]]]]]]]; 
    try (elimtype False; assumption); subst r; stSimpl; try (rewrite <- Req; stSimpl; auto; fail).

  simpl in RaEq; rewrite RaEq; reflexivity.

Qed.


Lemma cheneyDoAllocOk : iseqOkS P cheneyDoAllocPre cheneyDoAllocGuar cheneyDoAlloc.
Proof.
  unfold cheneyDoAlloc.
  iseqWeakSolver Psimpl.
  unfold cheneyDoAllocPre.
  intros s Hp; linLogElim Hp.
  rewrite addiuStep; stSimpl.
  replace (1 + 0) with 1; [idtac | reflexivity].
  rewrite (swStep Hp); try reflexivity.
  hypReplace Hp (contUpdateSt Hp 1).
  sconjLift2 Hp; rewrite (swStep Hp); try reflexivity.
  hypReplace Hp (contUpdateSt Hp 1).
  rewrite addiuStep0; rewrite addiuStep; stSimpl.
  sconjLift3 Hp; rewrite (swStep Hp); try reflexivity; stSimpl.
  rewrite addiuStep0; stSimpl.
  constructor.
  trivial.

  intros s' X; subst s'.
  unfold cheneyDoAllocGuar.
  constructor.

  rename Hp into Hp0; intros A Hp.
  stSimpl.
  linLogElim Hp.
  linLogIntro.
  hypReplace Hp (contUpdateSt Hp 1).
  sconjLift2 Hp; hypReplace Hp (contUpdateSt Hp 1).
  sconjLift3 Hp; hypReplace Hp (contUpdateSt Hp (8 + rget s rfree)).
  sconjLift2 Hp; sconjLift3 Hp; assumption.

  intro r.
  destruct (regCompare r v0).
  subst r.
  stSimpl; reflexivity.
  destruct (regCompare r rfree).
  subst; stSimpl; reflexivity.
  destruct (regCompare r ra); subst; stSimpl; reflexivity.

Qed.


Lemma cheneyLoopHeaderOk : 
  iseqOkS P cheneyLoopHeaderPre cheneyLoopHeaderGuar cheneyLoopHeader.
Proof.
  unfold cheneyLoopHeader.
  iseqWeakSolver Psimpl.
  unfold cheneyLoopHeaderPre.
  intros s Hp.

  assert (Hp2 : exists x,
    s |||- 12 + rget s rgcInfo |-> x **
      (fun h : Heap.heap => cheneyLoopPre (h, rfileOf s))).
  destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]].
  linLogElim HA.
  exists x.
  eapply sconjIntro; eauto.

  clear Hp; destruct Hp2 as [x Hp].
  rewrite (lwStep Hp); try reflexivity.
  rewrite (swStep Hp); try reflexivity.
  stSimpl.

  assert (Hp' : cheneyLoopPre
     (uhset (rset s rroot x) (12 + rget s rgcInfo) (rget s rtoEnd))).
  unfold cheneyLoopPre in Hp |- * ; unfold cheneyLoopStOk in Hp |- * .
  stSimpl.
  destruct (sconjElim Hp) as [h1 [h2 [J [Hp1 Hp2]]]]; clear Hp.
  destruct Hp2 as [phi [G [W [M [F [Geq [StLt [Al Hp2]]]]]]]].
  exists phi; exists G; exists W; exists M; exists F.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  clear Geq StLt Al.
  unfold scanFieldStOk in Hp2 |- * .
  destruct Hp2 as [WF [D [Feq [Imp [LeLe Hp2]]]]].
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.
  assumption.
  clear WF D Feq Imp LeLe.
  unfold hpredTrue in Hp2.
  replace (Heap.heap_canon (heapOf (Heap.heap_canon h2, rfileOf s))) with
    (Heap.heap_canon h2) in Hp2.
  rewrite <- hpredTrueEq in Hp2.
  assert (Hp := sconjIntro Hp1 Hp2 J); clear J Hp1 Hp2.
  destruct s as [h r].
  replace (heapOf (uhset (rset (h, r) rroot x) 
    (12 + rget (h, r) rgcInfo) (rget (h, r) rtoEnd))) with
    (Heap.uset h (12 + rget (h, r) rgcInfo) (rget (h, r) rtoEnd)).
  2:reflexivity.
  replace (heapOf (h, r)) with h in Hp.
  2:reflexivity.
  hypReplace Hp (contUpdateSt Hp (rget (h, r) rtoEnd)).
  hypReplace Hp (sconjComm Hp); linLogElim Hp.
  apply (sconjImplR Hp); clear Hp; intros h' Hp.
  apply (sconjImplR Hp); clear h' Hp; intros h' Hp.
  apply (sconjImplR Hp); clear h' Hp; intros h' Hp.
  apply (sconjImplR Hp); clear h' Hp; intros h' Hp.
  linLogIntro.
  simpl.
  unfold Heap.heap_canon.
  exact (Heap.HeapMap.map_canonEqvEq 
    (Heap.HeapMap.map_canonEqv h2)).

  constructor.
  assumption.

  destruct Hp' as [phi [G [W [M [F Hp']]]]].
  intros s' Guar'.

  constructor.
  intros y A Hp2.
  assert (R := contContEq Hp2 Hp); subst.
  econstructor.
  constructor.
  2:constructor.
  3:eauto.
  hypReplace Hp2 (contUpdateSt Hp2 (rget s rtoEnd)).
  assumption.
  intro r; stSimpl; reflexivity.

  destruct Guar' as [_ [Req RA]].
  constructor.
  2:rewrite RA; stSimpl; reflexivity.
  intros r I; simpl in I.
  rewrite <- Req.
  destruct I; stSimpl; reflexivity.
  intuition; regNotInListSolver.

Qed.


Lemma cheneyReturnOk : iseqOkS P cheneyReturnPre cheneyReturnGuar cheneyReturn.
Proof.
  unfold cheneyReturn.
  iseqWeakSolver Psimpl.
  unfold cheneyReturnPre; unfold cheneyReturnGuar.
  intros s _.
  rewrite addiuStep.
  intuition.
Qed.


Lemma cheneyLoop2Ok : iseqOkS P cheneyLoop2Pre cheneyLoop2Guar cheneyLoop2.
Proof.
  unfold cheneyLoop2.
  iseqWeakSolver Psimpl.

  apply scanFieldGuarPresRa.

  intros s [H1 H2].
  rewrite addiuStep.
  rewrite addiuStep.
  stSimpl.
  constructor.
  constructor.
  apply H1.
  apply H2.

  intros s''' [s' [H' H''']].
  hypReplace H2 (H2 s' H').
  unfold cheneyLoop2Guar.
  exists s'.
  constructor; assumption.
Qed.


Lemma cheneyLoopOk : iseqOkS P cheneyLoopPre cheneyLoopGuar cheneyLoop.
Proof.
  unfold cheneyLoop.
  iseqWeakSolver Psimpl.

  intros; apply scanFieldGuarPresRa; auto.

  unfold cheneyLoopPre.
  intros s [phi [G [W [M [F [Geq [LtSt [Al H0]]]]]]]].
  compare (rget s rscan) (rget s rfree); intro Eq.

  (* return case *)
  rewrite Eq; rewrite bteqEq; rewrite beq_true.
  constructor.
  unfold cheneyReturnPre; trivial.
  unfold cheneyReturnGuar.
  intros s' R; rewrite R; clear R s'.
  clear phi G W M F Geq LtSt H0.
  unfold cheneyLoopGuar; stSimpl.
  
  constructor.
  intros phi G W M F A H.
  exists (fun x:nat => None (A:=nat)).
  exists W.
  exists empty_set.
  exists F.
  exists empty_set.
  constructor.
  apply set_eqvUnionEmptyUnit.
  constructor.
  apply set_eqvUnionEmptyUnit.

  constructor.
  unfold iso_set; simpl.
  constructor.
  intuition.
  constructor.
  discriminate 2.
  constructor.
  discriminate 1.
  intros; elimtype False; assumption.

  clear Al.
  unfold cheneyLoopStOk in H |- * .
  destruct H as [Geq [LtSt [Al H]]].
  stSimpl.

  rewrite Eq; rewrite Eq in Geq; rewrite Eq in LtSt; clear Eq.

  assert (Geq' : set_eqv G empty_set).
  setoid_rewrite (set_eqvEmptyRangeObjs (rget s rfree)); assumption.
  clear Geq; rename Geq' into Geq.

  constructor.
  intros x' I; simpl in I; elimtype False; assumption.
  constructor.
  apply aligned8Refl.
  constructor.
  setoid_rewrite <- (set_eqvEmptyRangeObjs (rget s rfree)); apply set_eqvRefl.
  constructor.
  assumption.
  constructor.
  apply aligned8Refl.
  destruct H as [WFlt [WMdisj [Feq [PhiOk [StLt H]]]]].
  unfold scanFieldStOk.

  assert (Meq : set_eqv (set_union M empty_set) M).
  setoid_replace (set_union M empty_set) with (set_union empty_set M).
  apply set_eqvUnionSymm.
  apply set_eqvSymm; apply set_eqvUnionEmptyUnit.

  constructor.
  assumption.
  constructor.
  setoid_rewrite Meq; assumption.
  constructor.
  assumption.

  constructor.
  unfold map_union.
  intros x y; assert (X := PhiOk x y).
  destruct (phi x); assumption.
  constructor.
  assumption.

  clear LtSt WFlt WMdisj Feq StLt.
  eapply sconjImpl.
  apply H.
  clear H; intros h H.
  unfold hpredTrue.
  setoid_rewrite Meq.
  rewrite <- hpredTrueEq.
  unfold mapHp.
  unfold mapHp in H.
  linLogElim H.
  linLogIntro.
  unfold map_union.
  intros x y Z.
  assert (Zz := H2 x y).
  destruct (phi x); auto.
  apply (fasImpl H).
(*  refine (fasImpl (fun x => lexists a, !(phi x = Some a) ** x |-> a ** 4 + x |->_) 
    (fun x => lexists a,
      !(map_union phi (fun _ : nat => None) x = Some a) ** x |-> a ** 4 + x |->_)
    _ M h H).*)
  clear h H; intros x h H.
  linLogElim H.
  existsl x0.
  linLogIntro.
  unfold map_union.
  rewrite H3.
  reflexivity.
  eauto.

  clear H; intros h H.
  eapply sconjImpl.
  apply H.
  clear h H; intros h H.
  unfold hpredTrue.
  setoid_rewrite Meq.
  assumption.

  clear h H; intros h H.
  eapply sconjImplR.
  apply H.
  clear h H; intros h H.
  apply (sconjImpl H); clear h H; intros h H.
  unfold hpredTrue.
  setoid_rewrite Geq.
  rewrite <- hpredTrueEq.
  apply fasIntroE.
  unfold objHp in H.
  eapply fasElimE.
  unfold hpredTrue.
  setoid_rewrite <- Geq.
  rewrite <- hpredTrueEq.
  apply H.

  assert (h ||- unit ** A).
  linLogIntro; assumption.
  apply (sconjImplL H0); clear h H H0; intros h H.
  apply fasIntroE; assumption.

  unfold regPres; simpl; intuition.
  stSimpl; reflexivity.


  (* loop body case *)
  rewrite bteqEq; rewrite beq_false.
  2:assumption.
  rewrite addiuStep.
  simpl.

  assert (IG : set_In (rget s rscan) G).
  apply (set_eqv_elim2 Geq).
  destruct Al as [k Al].
  eapply rangeObjsIntro.
  rewrite Al.
  rewrite Al in Eq.
  omega.
  assert (rget s rscan = rget s rscan + 8 * 0).
  omega.
  eauto.
  

  assert (scanFieldPre (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) /\
    forall s' : state,
      scanFieldGuar (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) s' ->
      (scanFieldPre (ch2Step s') /\
	(forall s'0 : state, scanFieldGuar (ch2Step s') s'0 -> cheneyLoopPre s'0 /\
	  (forall s'' : state, cheneyLoopGuar s'0 s'' ->
	    cheneyLoopGuar s s'')))).

  assert (scanFieldStOk phi W M F 
    (okFieldAt (set_union W M) (rget s rscan) **
      okFieldAt (set_union W M) (4 + rget s rscan) **
      objHp (set_union W M) (set_remove (rget s rscan) G) ** top)
    (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2)).
  destruct H0 as [WFlt [WMdisj [Feq [PhiOk [[StLt StLt'] H0]]]]].
  unfold scanFieldStOk.
  stSimpl; intuition.
  rename H0 into H.
  apply (sconjImplR H); clear H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocL.
  apply sconjAssocL.
  apply (sconjImplL H); clear h H; intros h H.
  replace (okFieldAt (set_union W M) (rget s rscan) **
    okFieldAt (set_union W M) (4 + rget s rscan)) with
  (okObjAt (set_union W M) (rget s rscan)).
  apply objHpElim; assumption.
  reflexivity.

  constructor.
  unfold scanFieldPre.
  exists phi; exists W; exists M; exists F.
  stSimpl.
  eapply scanFieldStOkTop; eauto.
  
  unfold scanFieldGuar.
  stSimpl.
  intros s' [H' Rp].
  hypReplace H (H' _ _ _ _ _ H); rename H' into QQ.

  assert (Feq0 : set_eqv F (rangeObjs (rget s rfree) (rget s rtoEnd))).
  clear QQ Rp H.
  unfold scanFieldStOk in H0; intuition.

  rename H0 into Horig.
  destruct H as [M'' [W' [F' [G' [phi' [Weq [Feq [Piso H]]]]]]]].

  rename QQ into QQQ.

  assert (Req' : rget s' rtoEnd = rget s rtoEnd).
  rewrite (regPresEq Rp).
  stSimpl; reflexivity.
  regNotInListSolver.

  assert (Alf' : aligned8 (rget s rfree) (rget s' rfree)).
  unfold scanFieldStOk in H.
  intuition.

  assert (Lt' : rget s' rfree <= rget s rtoEnd).
  rewrite <- Req'.
  unfold scanFieldStOk in H.
  intuition.

  assert (WMeqv : set_eqv (set_union W M) (set_union W' (set_union M M''))).
  setoid_replace (set_union W' (set_union M M'')) with (set_union W M).
  setoid_rewrite Weq.
  setoid_replace (set_union M'' W') with (set_union W' M'').
  apply set_eqvUnionSymm.
  setoid_replace (set_union M M'') with (set_union M'' M).
  apply set_eqvUnionSymm.
  apply set_eqvUnionAssoc.
  apply set_eqvRefl.

  assert (GFdisj' : set_disj G' F').
  apply set_disj_intro.
  intros x I1 I2.
  clear Geq LtSt Al Eq IG Rp Weq Feq.
  unfold scanFieldStOk in H.
  intuition.
  clear H0 H H1 H2; rename H8 into H.
  assert (heapOf s' ||- (foralls x in F', x |->__) ** 
    (foralls x in G',
        objCopiedPhi phi' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x) **
    top).
  sconjLift3 H.
  apply (sconjImplR H); clear H; intros h H.
  sconjLift4 H.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.
  clear H; rename H0 into H.
  assert (heapOf s' ||- (lexists y, x |-> y ** top) ** (lexists y, x |-> y ** top) ** top).
  apply (sconjImpl H); clear H; intros h H.
  hypReplace H (fasElim I2 H); simpl in H; linLogElim H.
  exists x0; apply (sconjImplR H); intros; linLogIntro.
  apply (sconjImpl H); clear h H; intros h H.
  assert (IM : exists q, set_In q M'' /\ phi' q = Some x).
  unfold objCopiedPhi in H.
  hypReplace H (fasElim I1 H); simpl in H; linLogElim H.
  exists x0.
  constructor.
  unfold iso_set in Piso.
  intuition.
  destruct (H0 x0) as [_ Q]; clear H0; apply Q.
  rewrite H8; discriminate 1.
  assumption.
  unfold objCopiedPhi in H.
  hypReplace H (fasElim I1 H); simpl in H; linLogElim H.
  unfold objCopied in H; unfold fieldCopied in H.
  linLogElim H.
  exists x1.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.
  assumption.
  clear H; rename H0 into H.
  linLogElim H.

  assert (scanFieldStOk (map_union phi phi') W' (set_union M M'') F'
        (okFieldAt (set_union W' (set_union M M'')) (rget (ch2Step s') a0) **
	  fieldRespMap (map_union phi phi')
          (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) 
          (rget s rscan) **
          (foralls x in G',
            (lexists x', !(phi' x' = Some x) **
              objCopied x' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x)) **
          objHp (set_union W' (set_union M M'')) (set_remove (rget s rscan) G) ** top) (ch2Step s')).
  unfold ch2Step.
  stSimpl.
  rewrite (regPresEq Rp).
  2:regNotInListSolver.
  clear Rp.
  stSimpl.
  unfold scanFieldStOk in H |- * .
  intuition.
  stSimpl.
  apply (sconjImplR H8); clear H H8; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  sconjLift3 H.
  apply (sconjImpl H); clear h H; intros h H.
  unfold hpredTrue; setoid_rewrite <- WMeqv; assumption.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplL H); clear h H; intros h H.
  unfold hpredTrue; setoid_rewrite <- WMeqv; assumption.
  clear H; rename H0 into H.

  constructor.
  unfold scanFieldPre.

  exists (map_union phi phi'); exists W'; exists (set_union M M''); exists F'.
  apply (scanFieldStOkImpl H).
  clear H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.

  intro s''.
  unfold scanFieldGuar.
  intros [X Rp'].
  
  assert (F'eq : set_eqv F' (rangeObjs (rget s' rfree) (rget s rtoEnd))).
  unfold scanFieldStOk in H; unfold ch2Step in H.
  generalize dependent H; stSimpl.
  rewrite (regPresEq Rp rtoEnd).
  2:regNotInListSolver.
  stSimpl.
  intuition.

  assert (Y := X _ _ _ _ _ H); rename H into Horig'; rename X into XQ.

  destruct Y as [M3 [W3 [F3 [G3 [phi3 [Weq3 [Feq3 [Iso3 [Conn [Alf H]]]]]]]]]].
  unfold cheneyLoopPre.

  assert (GFdisj3 : set_disj G3 F3).
  apply set_disj_intro.
  intros x I1 I2.
  clear Geq LtSt Al Eq IG Rp Weq Feq.
  unfold scanFieldStOk in H.
  intuition.
  clear H0 H H1 H2; rename H6 into H.
  assert (heapOf s'' ||- (foralls x in F3, x |->__) ** 
    (foralls x in G3, (lexists x', !(phi3 x' = Some x) **
         objCopied x' (rset
               (rset (rset s' a0 (4 + rget s rscan)) rscan (8 + rget s rscan))
               ra CHENEYLOOP) x)) ** top).
  sconjLift3 H.
  apply (sconjImplR H); clear H; intros h H.
  sconjLift4 H.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.
  clear H; rename H0 into H.

  assert (heapOf s'' ||- (lexists y, x |-> y ** top) ** (lexists y, x |-> y ** top) ** top).
  apply (sconjImpl H); clear H; intros h H.
  hypReplace H (fasElim I2 H); simpl in H; linLogElim H.
  exists x0; apply (sconjImplR H); intros; linLogIntro.
  apply (sconjImpl H); clear h H; intros h H.
  assert (IM : exists q, set_In q M3 /\ phi3 q = Some x).
  hypReplace H (fasElim I1 H); simpl in H; linLogElim H.
  exists x0.
  constructor.
  unfold iso_set in Iso3.
  intuition.
  destruct (H0 x0) as [_ Q]; clear H0; apply Q.
  rewrite H6; discriminate 1.
  assumption.
  hypReplace H (fasElim I1 H); simpl in H; linLogElim H.
  unfold objCopied in H; unfold fieldCopied in H.
  linLogElim H.
  exists x1.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.
  assumption.
  clear H; rename H0 into H.
  linLogElim H.

  constructor.
  (* precondition for next to last step (?) *)
  exists (map_union (map_union phi phi') phi3); 
    exists (set_union (set_remove (rget s rscan) G) (set_union G' G3));
    exists W3; exists (set_union (set_union M M'') M3); exists F3.
  unfold cheneyLoopStOk.
  generalize dependent H.
  rewrite (regPresEq Rp' rscan); [idtac | regNotInListSolver].
  rewrite (regPresEq Rp' rtoStart); [idtac | regNotInListSolver].
  unfold ch2Step; stSimpl.
  rewrite (regPresEq Rp rscan); [stSimpl | regNotInListSolver].
  rewrite (regPresEq Rp rtoStart); [stSimpl | regNotInListSolver].
  intro.

  constructor.
  unfold scanFieldStOk in H.
  intuition.
  clear H Iso3 Weq3.

  setoid_rewrite Geq; clear Geq.
  assert (Req'' : rget s'' rtoEnd = rget s rtoEnd).
  rewrite (regPresEq Rp').
  unfold ch2Step; stSimpl; rewrite Req'; reflexivity.
  regNotInListSolver.

  clear Rp Rp'.
  setoid_rewrite (rangeRemoveEqv (rget s rscan) (rget s rfree)).
  
  setoid_replace (set_union G' G3) with (rangeObjs (rget s rfree) (rget s'' rfree)).
  setoid_replace G' with (rangeObjs (rget s rfree) (rget s' rfree)).
  apply set_eqvSymm; eapply rangeSplit.
  instantiate (1 := rget s rtoEnd).
  setoid_rewrite <- F'eq.
  assumption.
  assumption.

  setoid_rewrite <- Feq0.
  setoid_rewrite Feq.
  setoid_rewrite <- F'eq.
  apply set_eqvRefl.
  assumption.

  setoid_replace G3 with (rangeObjs (rget s' rfree) (rget s'' rfree)).
  apply set_eqvSymm; eapply rangeSplit.
  instantiate (1 := rget s rtoEnd).
  rewrite Req'' in H2.
  setoid_rewrite <- H2.
  assumption.
  rewrite Req'' in H7; assumption.

  rewrite Req'' in H2.
  setoid_rewrite <- H2.
  setoid_rewrite <- Feq3.
  assumption.
  assumption.

  apply set_eqvSymm.
  eapply rangeObjsTransEqv.
  assumption.
  destruct Alf as [x R].
  rewrite R; stSimpl.
  unfold ch2Step; stSimpl.
  omega.
  
  eapply rangeObjsTransEqv.
  destruct Al as [k R].
  destruct k.
  elimtype False; apply Eq; omega.
  rewrite R.
  exists k; omega.
  destruct Alf as [k R]; rewrite R; stSimpl.
  unfold ch2Step; stSimpl.
  destruct Alf' as [k' R']; rewrite R'; stSimpl.
  omega.

  constructor.
  omega.

  constructor.
  eapply aligned8Trans.
  2:apply Alf.
  unfold ch2Step; stSimpl.
  eapply aligned8Trans.
  2:apply Alf'.
  destruct Al as [k R].
  destruct k.
  elimtype False; apply Eq; omega.
  rewrite R; exists k; omega.

  unfold scanFieldStOk in H |- * .
  intuition.
  apply (sconjImplR H6); clear H H6; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.

  assert (H' : h ||-
    ((foralls x in G3, (lexists x', !(phi3 x' = Some x) **
      objCopied x' (rset (rset (rset s' a0 (4 + rget s rscan)) rscan (8 + rget s rscan))
        ra CHENEYLOOP) x)) **
    (foralls x in G', (lexists x', !(phi' x' = Some x) **
      objCopied x' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x)) **
    objHp (set_union W' (set_union M M'')) (set_remove (rget s rscan) G)) **
       top).
  apply sconjAssocR.
  sconjLift2 H.
  apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR.
  sconjLift3 H.
  apply (sconjImplR H); clear h H; intros h H.
  sconjLift3 H.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.
  
  clear H.
  apply (sconjImpl H'); clear h H'; intros h H.
  2:linLogIntro.
  unfold hpredTrue.
  setoid_replace (set_union W3 (set_union (set_union M M'') M3)) with
    (set_union W' (set_union M M'')).
  setoid_replace (set_union (set_union M M'') M3) with
    (set_union M3 (set_union M M'')).
  apply set_eqvUnionSymm.
  setoid_rewrite <- (set_eqvUnionAssoc W3 M3 (set_union M M'')).
  setoid_rewrite Weq3.
  setoid_replace (set_union M3 W3) with (set_union W3 M3).
  apply set_eqvUnionSymm.
  apply set_eqvRefl.

  sconjLift3of3 H.
  rewrite <- hpredTrueEq.
  apply objHpMerge.
  apply (sconjImplR H); clear h H; intros h H.
  hypReplace H (sconjComm H).
  apply objHpMerge.
  clear Rp Rp' F'eq.
  apply (sconjImpl H); clear h H; intros h H.
  
  unfold objHp.
  apply (fasImpl H); clear h H.
  intros x h H.
  linLogElim H.
  unfold okObjAt.
  unfold objCopied in H.

  assert (H00 : heapOf s ||- okObjAt (set_union W M) x0 ** top).
    assert (set_In x0 W).
    setoid_rewrite Weq.
    apply set_union_intro1.
    destruct Piso.
    destruct (H4 x0) as [_ U].
    apply U; rewrite H8; discriminate.
    assert (HH : heapOf s ||- objHp (set_union W M) W ** top).
    unfold scanFieldStOk in Horig; intuition.
    sconjLift2 H14; apply (sconjImplR H14); intros y X; linLogIntro.
  assert (heapOf s ||- (okObjAt (set_union W M) x0 ** top) ** top).
  apply (sconjImplL HH); clear H HH; intros hh H.
  apply (sconjImplR (objHpElim H4 H)); intros; linLogIntro.
  linLogElim H6; apply (sconjImplR H6); intros; linLogIntro.
  unfold okObjAt in H00; unfold okFieldAt in H00.
  linLogElim H00.
  unfold okFieldAt; unfold fieldCopied in H; simpl in H.
  linLogElim H.
  assert (x1 = x3).
  apply (contContEq H00 H12).
  assert (x2 = x4).
  sconjLift2 H00; apply (contContEq H00 H13).
  subst.

  linLogIntro.
  2:eauto.
  setoid_rewrite <- WMeqv; assumption.
  setoid_rewrite <- WMeqv; assumption.

  unfold objHp.
  apply (fasImpl H); clear h H.
  intros x h H.
  linLogElim H.
  unfold okObjAt.
  unfold objCopied in H.

  unfold fieldCopied in H; simpl in H.
  assert (H00 : heapOf s' ||- okObjAt (set_union W' (set_union M M'')) x0 ** top).
    assert (set_In x0 W').
    setoid_rewrite Weq3; apply set_union_intro1.
    destruct Iso3.
    destruct (H4 x0) as [_ U].
    apply U; rewrite H8; discriminate.
    assert (HH : heapOf s' ||- objHp (set_union W' (set_union M M'')) W' ** top).
    unfold scanFieldStOk in Horig'; intuition.
    sconjLift2 H14; apply (sconjImplR H14); intros y X; linLogIntro.
  assert (heapOf s' ||- (okObjAt (set_union W' (set_union M M'')) x0 ** top) ** top).
  apply (sconjImplL HH); clear H HH; intros hh H.
  apply (sconjImplR (objHpElim H4 H)); intros; linLogIntro.
  linLogElim H6; apply (sconjImplR H6); intros; linLogIntro.
  unfold okObjAt in H00; unfold okFieldAt in H00.
  linLogElim H00.
  unfold okFieldAt; unfold fieldCopied in H; simpl in H.
  linLogElim H.
  assert (x1 = x3).
  apply (contContEq H00 H12).
  assert (x2 = x4).
  sconjLift2 H00; apply (contContEq H00 H13).
  subst.
  linLogIntro.
  2:eauto.
  assumption.
  assumption.

  (* adding a cheneyLoopGuar implies the whole thing is cheneyLoopGuar *)
  unfold cheneyLoopGuar.
  clear Horig Horig' H.
  clear W M F Geq LtSt Conn Al IG M'' W' F' G' phi' Weq Feq Piso Feq0
    WMeqv GFdisj' F'eq M3 W3 F3 G3 phi3 Weq3 Feq3 Iso3 GFdisj3 phi G.

  intros s3 X.
  destruct X as [X RpL].
  constructor.
  intros phi G W M F A.

  (* in hindsight, combining everything together might not actually be
  that useful, as you end up having to reapply the guarantees each time
  anyways, leaving little useful information from other "branches" of the proof *)
  destruct 1 as [Geq [LtSt [Al H]]].

  assert (IG : set_In (rget s rscan) G).
  apply (set_eqv_elim2 Geq).
  destruct Al as [k Al].
  eapply rangeObjsIntro.
  rewrite Al.
  rewrite Al in Eq.
  omega.
  assert (rget s rscan = rget s rscan + 8 * 0).
  omega.
  eauto.

  assert (scanFieldStOk phi W M F 
    (okFieldAt (set_union W M) (rget s rscan) **
      okFieldAt (set_union W M) (4 + rget s rscan) **
      objHp (set_union W M) (set_remove (rget s rscan) G) ** A)
    (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2)).

  destruct H as [WFlt [WMdisj [Feq [PhiOk [[StLt StLt'] H0]]]]].
  unfold scanFieldStOk; stSimpl.
  intuition.
  rename H into RP1; rename H1 into RP2.
  rename H0 into H.
  apply (sconjImplR H); clear H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocL.
  apply sconjAssocL.
  apply (sconjImpl H); clear h H; intros h H.
  replace (okFieldAt (set_union W M) (rget s rscan) **
    okFieldAt (set_union W M) (4 + rget s rscan)) with
  (okObjAt (set_union W M) (rget s rscan)).
  apply objHpElim; assumption.
  reflexivity.
  assumption.

  destruct (scanFieldStOkExtract H0) as [h00 [H0' Hok]].
  clear H0; rename H0' into H0.

  assert (Feq0 : set_eqv F (rangeObjs (rget s rfree) (rget s rtoEnd))).
  unfold scanFieldStOk in H0; intuition.

  rename H0 into Horig; clear H.
  assert (H := QQQ _ _ _ _ _ Horig); clear QQQ.

  destruct H as [M'' [W' [F' [G' [phi' [Weq [Feq [Piso [Al' H]]]]]]]]].

  assert (HX0 := H).

  assert (WMeqv : set_eqv (set_union W M) (set_union W' (set_union M M''))).
  setoid_replace (set_union W' (set_union M M'')) with (set_union W M).
  setoid_rewrite Weq.
  setoid_replace (set_union M'' W') with (set_union W' M'').
  apply set_eqvUnionSymm.
  setoid_replace (set_union M M'') with (set_union M'' M).
  apply set_eqvUnionSymm.
  apply set_eqvUnionAssoc.
  apply set_eqvRefl.

  assert (GFdisj' : set_disj G' F').
  apply set_disj_intro.
  intros x I1 I2.
  clear Geq LtSt Al Eq IG Rp Weq Feq.
  unfold scanFieldStOk in H.
  intuition.
  clear H0 H1 H H3 H4 H6 H9 H11; rename H12 into H.
  assert (heapOf s' ||- (foralls x in F', x |->__) ** 
    (foralls x in G', (lexists x', !(phi' x' = Some x) **
         objCopied x' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x)) ** top).
  sconjLift3 H.
  apply (sconjImplR H); clear H; intros h H.
  sconjLift4 H.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.
  clear H; rename H0 into H.
  assert (heapOf s' ||- (lexists y, x |-> y ** top) ** (lexists y, x |-> y ** top) ** top).
  apply (sconjImpl H); clear H; intros h H.
  hypReplace H (fasElim I2 H); simpl in H; linLogElim H.
  exists x0; apply (sconjImplR H); intros; linLogIntro.
  apply (sconjImpl H); clear h H; intros h H.
  assert (IM : exists q, set_In q M'' /\ phi' q = Some x).
  hypReplace H (fasElim I1 H); simpl in H; linLogElim H.
  exists x0.
  constructor.
  unfold iso_set in Piso.
  intuition.
  destruct (H0 x0) as [_ Q]; clear H0; apply Q.
  rewrite H6; discriminate 1.
  assumption.
  hypReplace H (fasElim I1 H); simpl in H; linLogElim H.
  unfold objCopied in H; unfold fieldCopied in H.
  linLogElim H.
  exists x1.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.
  assumption.
  clear H; rename H0 into H.
  linLogElim H.

  destruct (sconjExtract (sconjComm Hok)) as [h00' [Hok' h00Ok']].

  assert (H' : exists h003,
    scanFieldStOk (map_union phi phi') W' (set_union M M'') F'
    (okFieldAt (set_union W' (set_union M M'')) (rget (ch2Step s') a0) **
      heap_eqv h003 ** heap_eqv h00') (ch2Step s') /\
    h003 ||- fieldRespMap (map_union phi phi')
      (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) (rget s rscan) **
      (foralls x in G', (lexists x', !(phi' x' = Some x) **
           objCopied x' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x))).
  assert (Hm : scanFieldStOk (map_union phi phi') W' (set_union M M'') F'
    (heap_eqv h00 **
      (fieldRespMap (map_union phi phi')
	(rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) 
	(rget s rscan) **
         (foralls x in G',
          (lexists x', !(phi' x' = Some x) **
           objCopied x' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x)))) s').
  destruct H as [HHX [Wcopy H]].
  apply (scanFieldStOkImpl H).
  clear H; intros h H; apply sconjComm; apply sconjAssocR; assumption.
  clear H.
  destruct (scanFieldStOkExtract Hm) as [h003 [Hm' h003Ok]].
  exists h003.
  constructor; [clear h003Ok | assumption].
  unfold ch2Step; stSimpl; clear Hm.
  unfold scanFieldStOk in Hm' |- * ; stSimpl.
  intuition.
  apply (sconjImplR H10); clear H H10; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  hypReplace H (sconjInsert H Hok').
  hypReplace H (sconjAssocL H).
  sconjLift3of3 H; sconjLift3of3 H.
  apply (sconjImplL H); clear h H; intros h H.
  rewrite (regPresEq Rp).
  2:regNotInListSolver.
  stSimpl.
  unfold hpredTrue; setoid_rewrite <- WMeqv; exact H.

  destruct H' as [h003 [H' h003Ok]].

  assert (F'eq : set_eqv F' (rangeObjs (rget s' rfree) (rget s rtoEnd))).
  unfold scanFieldStOk in H; unfold ch2Step in H.
  generalize dependent H; stSimpl.
  rewrite (regPresEq Rp rtoEnd).
  2:regNotInListSolver.
  stSimpl.
  intuition.

  rename H' into Horig'.
  assert (Y := XQ _ _ _ _ _ Horig'); clear XQ H.

  destruct Y as [M3 [W3 [F3 [G3 [phi3 [Weq3 [Feq3 [Iso3 [Comm3 [Alf3 [WFr H]]]]]]]]]]].
  assert (HX0'' := H).
  unfold cheneyLoopPre.
  unfold cheneyLoopStOk.

  assert (GFdisj3 : set_disj G3 F3).
  apply set_disj_intro.
  intros x I1 I2.
  clear Geq LtSt Al Eq IG Rp Weq Feq.
  unfold scanFieldStOk in H.
  intuition.
  clear H0 H; rename H10 into H.
  assert (heapOf s'' ||- (foralls x in F3, x |->__) ** 
    (foralls x in G3, (lexists x', !(phi3 x' = Some x) **
         objCopied x' (rset
               (rset (rset s' a0 (4 + rget s rscan)) rscan (8 + rget s rscan))
               ra CHENEYLOOP) x)) ** top).
  sconjLift3 H.
  apply (sconjImplR H); clear H; intros h H.
  sconjLift4 H.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.
  clear H; rename H0 into H.
  assert (heapOf s'' ||- (lexists y, x |-> y ** top) ** (lexists y, x |-> y ** top) ** top).
  apply (sconjImpl H); clear H; intros h H.
  hypReplace H (fasElim I2 H); simpl in H; linLogElim H.
  exists x0; apply (sconjImplR H); intros; linLogIntro.
  apply (sconjImpl H); clear h H; intros h H.
  assert (IM : exists q, set_In q M3 /\ phi3 q = Some x).
  hypReplace H (fasElim I1 H); simpl in H; linLogElim H.
  exists x0.
  constructor.
  unfold iso_set in Iso3.
  intuition.
  destruct (H0 x0) as [_ Q]; clear H0; apply Q.
  rewrite H13; discriminate 1.
  assumption.
  hypReplace H (fasElim I1 H); simpl in H; linLogElim H.
  unfold objCopied in H; unfold fieldCopied in H.
  linLogElim H.
  exists x1.
  apply (sconjImplR H); clear h H; intros h H.
  linLogIntro.
  assumption.
  clear H; rename H0 into H.
  linLogElim H.

  assert (Q : scanFieldStOk (map_union (map_union phi phi') phi3) W3
    (set_union (set_union M M'') M3) F3
    ((heap_eqv h00' ** heap_eqv h003) **
      (fieldRespMap (map_union (map_union phi phi') phi3) 
	(ch2Step s') (rget (ch2Step s') a0) **
      (foralls x in G3,
        (lexists x', !(phi3 x' = Some x) ** objCopied x' (ch2Step s') x)))) s'').
  apply (scanFieldStOkImpl H); clear H; intros h H.
  apply sconjComm; linLogIntro.
  apply (sconjImplR H); intros.
  apply (sconjImplR H0); intros.
  apply sconjComm; assumption.

  clear H; destruct (scanFieldStOkExtract Q) as [h004 [H h004Ok]]; clear Q.

  assert (Q : scanFieldStOk (map_union (map_union phi phi') phi3) W3
    (set_union (set_union M M'') M3) F3
    (objHp (set_union W' (set_union M M'')) (set_remove (rget s rscan) G) ** A **
      Heap.heap_eqv h003 ** Heap.heap_eqv h004) s'').
  apply (scanFieldStOkImpl H).
  clear H; intros h H.
  linLogElim H; hypReplace H (sconjInsert H h00Ok'); linLogElim H.
  apply (sconjImplL H); clear h H; intros h H.
  unfold hpredTrue; setoid_rewrite <- WMeqv; assumption.

  clear H; rename Q into H.

  destruct (sconjExtract h003Ok) as 
    [h003' [h003Ok1 h003Ok2]].
  clear h003Ok.
  destruct (sconjExtract h004Ok) as 
    [h004' [h004Ok1 h004Ok2]].
  clear h004Ok.
  
  assert (H' : scanFieldStOk (map_union (map_union phi phi') phi3) W3
        (set_union (set_union M M'') M3) F3
        (objHp (set_union W' (set_union M M'')) (set_remove (rget s rscan) G) **
         A ** Heap.heap_eqv h003' **
         (foralls x in G', (lexists x', !(phi' x' = Some x) **
	   objCopied x' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x)) **
	 (Heap.heap_eqv h004' ** (foralls x in G3,
              (lexists x', !(phi3 x' = Some x) ** objCopied x' (ch2Step s') x)))) s'').
  apply (scanFieldStOkImpl H); clear H.
  intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocL.
  refine (sconjInsert _ h003Ok1).
  apply sconjComm.
  refine (sconjInsert _ h004Ok1).
  apply sconjComm; assumption.

  clear H; rename H' into H.
  assert (Q : cheneyLoopStOk (map_union (map_union phi phi') phi3)
     (set_union (set_remove (rget s rscan) G) (set_union G' G3)) W3
     (set_union (set_union M M'') M3) F3
     (objHp (set_union W3 (set_union (set_union M M'') M3))
        (set_union (set_remove (rget s rscan) G) (set_union G' G3)) **
	Heap.heap_eqv h003' ** Heap.heap_eqv h004' ** A) s'').

  unfold cheneyLoopStOk.
  generalize dependent H.
  rewrite (regPresEq Rp').
  2:regNotInListSolver.
  unfold ch2Step; stSimpl.
  rewrite (regPresEq Rp).
  2:regNotInListSolver.
  stSimpl.

  intro.

  constructor.
  unfold scanFieldStOk in H.
  intuition.
  clear H6 H4 H H0 Iso3 Weq3.

  setoid_rewrite Geq; clear Geq.
  assert (Req'' : rget s'' rtoEnd = rget s rtoEnd).
  rewrite (regPresEq Rp').
  unfold ch2Step; stSimpl; rewrite Req'; reflexivity.
  regNotInListSolver.

  rewrite Req'' in H5.

  clear Rp Rp'.
  setoid_rewrite (rangeRemoveEqv (rget s rscan) (rget s rfree)).
  
  setoid_replace (set_union G' G3) with (rangeObjs (rget s rfree) (rget s'' rfree)).
  setoid_replace G' with (rangeObjs (rget s rfree) (rget s' rfree)).
  apply set_eqvSymm; eapply rangeSplit.
  instantiate (1 := rget s rtoEnd).
  setoid_rewrite <- F'eq.
  assumption.
  assumption.

  setoid_rewrite <- Feq0.
  setoid_rewrite Feq.
  setoid_rewrite <- F'eq.
  apply set_eqvRefl.
  assumption.

  setoid_replace G3 with (rangeObjs (rget s' rfree) (rget s'' rfree)).
  apply set_eqvSymm; eapply rangeSplit.
  instantiate (1 := rget s rtoEnd).
  setoid_rewrite <- H5.
  assumption.
  rewrite Req'' in H11; assumption.

  setoid_rewrite <- H5.
  setoid_rewrite <- F'eq.
  setoid_rewrite <- Feq3; apply set_eqvRefl.
  assumption.

  apply set_eqvSymm.
  eapply rangeObjsTransEqv.
  assumption.
  destruct Alf as [x R].
  rewrite R; stSimpl.
  unfold ch2Step; stSimpl.
  omega.
  
  eapply rangeObjsTransEqv.
  destruct Al as [k R].
  destruct k.
  elimtype False; apply Eq; omega.
  rewrite R.
  exists k; omega.
  destruct Alf as [k R]; rewrite R; stSimpl.
  unfold ch2Step; stSimpl.
  destruct Alf' as [k' R']; rewrite R'; stSimpl.
  omega.

  constructor.
  rewrite (regPresEq Rp'); unfold ch2Step; stSimpl.
  rewrite (regPresEq Rp); stSimpl.
  omega.
  regNotInListSolver.
  regNotInListSolver.

  constructor.
  eapply aligned8Trans.
  2:apply Alf.
  unfold ch2Step; stSimpl.
  eapply aligned8Trans.
  2:apply Alf'.
  destruct Al as [k R].
  destruct k.
  elimtype False; apply Eq; omega.
  rewrite R; exists k; omega.

  unfold scanFieldStOk in H |- * .
  intuition.
  apply (sconjImplR H10); clear H H10; intros h H.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplR H); clear h H; intros h H.

  sconjLift3 H; apply sconjComm; linLogIntro.
  apply (sconjImplR H); clear h H; intros h H.
  sconjLift4 H.
  apply (sconjImplR H); clear h H; intros h H.
  sconjLift2 H. 
  apply (sconjImplR H); clear h H; intros h H.

  unfold hpredTrue.
  setoid_replace (set_union W3 (set_union (set_union M M'') M3)) with
    (set_union W' (set_union M M'')).
  setoid_replace (set_union (set_union M M'') M3) with
    (set_union M3 (set_union M M'')).
  apply set_eqvUnionSymm.
  setoid_rewrite <- (set_eqvUnionAssoc W3 M3 (set_union M M'')).
  setoid_rewrite Weq3.
  setoid_replace (set_union M3 W3) with (set_union W3 M3).
  apply set_eqvUnionSymm.
  apply set_eqvRefl.

  rewrite <- hpredTrueEq.
  apply objHpMerge.
  apply (sconjImplR H); clear h H; intros h H.
  apply objHpMerge.
  clear Rp Rp' F'eq.
  apply (sconjImpl H); clear h H; intros h H.
  
  unfold objHp.
  apply (fasImpl H); clear h H.
  intros x h H.
  linLogElim H.
  unfold okObjAt.
  unfold objCopied in H.

  unfold fieldCopied in H; simpl in H.
  assert (H00 : heapOf s ||- okObjAt (set_union W M) x0 ** top).
    assert (IW : set_In x0 W).
    setoid_rewrite Weq.
    apply set_union_intro1.
    destruct Piso as [Piso1 Piso2].
    destruct (Piso1 x0) as [_ U].
    apply U; rewrite H12; discriminate.
    assert (HH : heapOf s ||- objHp (set_union W M) W ** top).
    unfold scanFieldStOk in Horig; intuition.
    sconjLift2 H17; apply (sconjImplR H17); intros y Z; linLogIntro.
  assert (heapOf s ||- (okObjAt (set_union W M) x0 ** top) ** top).
  apply (sconjImplL HH); clear H HH; intros hh H.
  apply (sconjImplR (objHpElim IW H)); intros; linLogIntro.
  linLogElim H8; apply (sconjImplR H8); intros; linLogIntro.
  unfold okObjAt in H00; unfold okFieldAt in H00.
  linLogElim H00.
  unfold okFieldAt; unfold fieldCopied in H; simpl in H.
  linLogElim H.
  assert (x1 = x3).
  apply (contContEq H00 H16).
  assert (x2 = x4).
  sconjLift2 H00; apply (contContEq H00 H17).
  subst.
  linLogIntro.
  2:eauto.
  setoid_rewrite <- WMeqv; assumption.
  setoid_rewrite <- WMeqv; assumption.

  unfold objHp.
  apply (fasImpl H); clear h H.
  intros x h H.
  linLogElim H.
  unfold okObjAt.
  unfold objCopied in H.

  unfold fieldCopied in H; simpl in H.
  assert (H00 : heapOf s' ||- okObjAt (set_union W' (set_union M M'')) x0 ** top).
    assert (IW : set_In x0 W').
    setoid_rewrite Weq3; apply set_union_intro1.
    destruct Iso3 as [Iso3 Iso3'].
    destruct (Iso3 x0) as [_ U].
    apply U; rewrite H12; discriminate.
    assert (HH : heapOf s' ||- objHp (set_union W' (set_union M M'')) W' ** top).
    unfold scanFieldStOk in Horig'; intuition.
    sconjLift2 H17; apply (sconjImplR H17); intros y Z; linLogIntro.
  assert (heapOf s' ||- (okObjAt (set_union W' (set_union M M'')) x0 ** top) ** top).
  apply (sconjImplL HH); clear H HH; intros hh H.
  apply (sconjImplR (objHpElim IW H)); intros; linLogIntro.
  linLogElim H8; apply (sconjImplR H8); intros; linLogIntro.
  unfold okObjAt in H00; unfold okFieldAt in H00.
  linLogElim H00.
  unfold okFieldAt; unfold fieldCopied in H; simpl in H.
  linLogElim H.
  assert (x1 = x3).
  apply (contContEq H00 H16).
  assert (x2 = x4).
  sconjLift2 H00; apply (contContEq H00 H17).
  subst.
  linLogIntro.
  2:eauto.
  assumption.
  assumption.

  clear H.
  rename Q into Horig3.
  assert (H := X _ _ _ _ _ _ Horig3); clear X.

  destruct H as [phi4 [W4 [M4 [F4 [G4 [Weq4 [Feq4 [Iso4 [Comm4 H]]]]]]]]].
  unfold cheneyLoopStOk in H.
  destruct H as [Eeq [StEq4 [Al4 H]]].

  exists (map_union (map_union phi' phi3) phi4).
  exists W4; exists (set_union M'' (set_union M3 M4)).
  exists F4; exists (set_union G' (set_union G3 G4)).

  constructor.
  setoid_rewrite Weq; setoid_rewrite Weq3; setoid_rewrite Weq4.
  setoid_rewrite <- (set_eqvUnionAssoc M3 M4 W4).
  apply set_eqvSymm; apply set_eqvUnionAssoc.

  constructor.
  setoid_rewrite Feq; setoid_rewrite Feq3; setoid_rewrite Feq4.
  setoid_rewrite <- (set_eqvUnionAssoc G3 G4 F4).
  apply set_eqvSymm; apply set_eqvUnionAssoc.

  constructor.
  clear Eq Rp Req' Alf' Lt' Rp' Alf LtSt Al F'eq.
  clear GFdisj' Alf3 GFdisj3 Eeq StEq4 Al4.
  setoid_rewrite <- (set_eqvUnionAssoc M'' M3 M4).
  setoid_rewrite <- (set_eqvUnionAssoc G' G3 G4).
  apply iso_union.
  apply iso_union.
  assumption.
  assumption.
  3:assumption.
  assert (D : set_disj (set_union M3 W3) (set_union M M'')).
  setoid_rewrite <- Weq3.
  apply (scanFieldStOkWMDisj Horig').
  eapply set_unionDisjElimR.
  apply set_disjSymm.
  eapply set_unionDisjElimL.
  apply D.

  apply set_disjSymm.
  eapply set_unionDisjElimL.
  instantiate (1 := F3).
  setoid_rewrite <- Feq3.
  assert (Horig'' : scanFieldStOk (map_union phi phi') W' (set_union M M'') F'
    ((foralls x in G',
      (lexists x',
        !(phi' x' = Some x) **
        objCopied x' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2)
        x)) ** top) (ch2Step s')).
  apply (scanFieldStOkImpl Horig').
  intros h U; sconjLift2 U.
  hypReplace U (sconjInsert U h003Ok1).
  linLogElim U; sconjLift2 U.
  apply (sconjImplR U); intros; linLogIntro.
  clear Horig' H Horig.
  apply (scanFieldStOkFGDisj Horig'').
  unfold objCopied; unfold fieldCopied; simpl.
  intros h x H; linLogElim H; linLogIntro.
  sconjMatch.

  clear H Horig Horig'.
  eapply set_unionDisjElimR.
  instantiate (1 := M).
  setoid_rewrite <- (set_eqvUnionAssoc M M'' M3).
  apply set_disjSymm.
  eapply set_unionDisjElimL.
  instantiate (1 := W4).
  setoid_rewrite <- Weq4.
  unfold cheneyLoopStOk in Horig3; intuition.
  apply (scanFieldStOkWMDisj H8).

  clear H Horig Horig'.
  eapply set_unionDisjElimR.
  instantiate (1 := (set_remove (rget s rscan) G)).
  apply set_disjSymm.
  eapply set_unionDisjElimL.
  instantiate (1 := F4).
  setoid_rewrite <- Feq4.
  unfold cheneyLoopStOk in Horig3; intuition.
  apply (scanFieldStOkFGDisj H8).
  unfold okObjAt; unfold okFieldAt; intros h x U; linLogElim U.
  linLogIntro; sconjMatch.

  (* showing that the new objects are reachable from the objects we are scanning *)
  constructor.
  assert (Comm' : forall x', set_In x' (G' |_| G3) ->
    exists x, set_In x G /\ reachable (heapOf s3) x x').

  intros x' I; destruct (set_union_elim I) as [I' | I']; clear I.
  hypReplace Al' (Al' _ I').
  exists (rget s rscan).
  constructor.
  assumption.
  econstructor 2.
  instantiate (1 := x').
  2:constructor.
  unfold fieldRespMap in h003Ok2; linLogElim h003Ok2.
  replace (Heap.get (heapOf s') (rget s rscan)) with (Some x) in Al'.
  injection Al'; clear Al'; intro; subst x'.
  unfold scanFieldStOk in H; intuition.
  sconjLift6 H12; assert (HX := sconjInsert H12 h003Ok2); eapply contGetEqL; eauto.
  symmetry; unfold scanFieldStOk in Horig'; intuition.
  sconjLift5 H13; assert (HX := sconjInsert H13 h003Ok1); linLogElim HX;
    hypReplace HX (sconjInsert HX h003Ok2); eapply contGetEqL; eauto.
  
  hypReplace Comm3 (Comm3 _ I').
  generalize dependent Comm3; generalize dependent h004Ok2; generalize dependent HX0''; 
    unfold ch2Step; stSimpl; rewrite <- Rp; [stSimpl | regNotInListSolver]; intros.
  exists (rget s rscan); constructor.
  assumption.
  econstructor 3.
  instantiate (1 := x').
  2:constructor.
  unfold fieldRespMap in h004Ok2; linLogElim h004Ok2.
  replace (Heap.get (heapOf s'') (4 + rget s rscan)) with (Some x) in Comm3.
  injection Comm3; clear Comm3; intro; subst x'.
  unfold scanFieldStOk in H; intuition.
  sconjLift7 H12; assert (HX := sconjInsert H12 h004Ok2); eapply contGetEqL; eauto.
  unfold scanFieldStOk in HX0''; intuition.
  sconjLift4 H13; unfold fieldRespMap in H13; linLogElim H13.
  destruct H18 as [v0 [V0eq M0]]; assert (EE : x0 = x).
  destruct H2 as [v1 [V1eq M1]].
  assert (Q := contContEq V0eq V1eq).
  subst v0; rewrite M0 in M1; injection M1; intro; assumption.
  subst x0; symmetry; eapply contGetEqL; eauto.

  intros x' I; destruct (set_union_elim I) as [I' | I']; clear I.
  apply Comm'; apply set_union_intro1; assumption.
  destruct (set_union_elim I') as [I | I]; clear I'.
  apply Comm'; apply set_union_intro2; assumption.
  destruct (Comm4 _ I) as [x [I' R]]; clear I.
  destruct (set_union_elim I') as [I | I]; clear I'.
  exists x; constructor.
  destruct (set_remove_elim _ _ _ I) as [_ X]; assumption.
  assumption.
  destruct (Comm' _ I) as [y [I' R']].
  exists y; constructor.
  assumption.
  eapply reachableTrans; eauto.
  (* finished the reachability *)

  constructor.
  eapply aligned8Trans; eauto.
  eapply aligned8Trans; eauto.
  constructor.
  assumption.
  constructor.
  assumption.
  destruct H as [al2 H].
  constructor.
  assumption.
  
  setoid_replace (set_union M (set_union M'' (set_union M3 M4))) with
    (set_union (set_union (set_union M M'') M3) M4).
  setoid_rewrite <- (set_eqvUnionAssoc M M'' (set_union M3 M4)).
  setoid_rewrite <- (set_eqvUnionAssoc (set_union M M'') M3 M4).
  apply set_eqvRefl.
  
  assert (PhiEq : 
    mapEqv (map_union (map_union (map_union phi phi') phi3) phi4)
    (map_union phi (map_union (map_union phi' phi3) phi4))).
  unfold map_union.
  intros x.
  destruct (phi x).
  reflexivity.
  destruct (phi' x).
  reflexivity.
  destruct (phi3 x); reflexivity.

  assert (DG : set_disj (set_remove (rget s rscan) G) (set_union G' G3)).
  eapply set_unionDisjElimR.
  instantiate (1 := (set_add (rget s rscan) empty_set)).
  replace (set_union (set_add (rget s rscan) empty_set)
    (set_remove (rget s rscan) G)) with
    (set_add (rget s rscan) (set_remove (rget s rscan) G)).
  2:reflexivity.
  setoid_rewrite (addRemoveEqv IG).
  apply set_disjSymm.
  eapply set_unionDisjElimL.
  instantiate (1 := F3).
  setoid_rewrite (set_eqvUnionAssoc G' G3 F3).
  setoid_rewrite <- Feq3.
  setoid_rewrite <- Feq.
  assert (HorigI : scanFieldStOk phi W M F
    (objHp (set_union W M) G ** A) (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2)).
  apply (scanFieldStOkImpl Horig).
  intros h U; hypReplace U (sconjComm U).
  hypReplace U (sconjInsert U Hok).
  apply sconjComm; linLogElim U; sconjLift3 U.
  apply (sconjImplR U); clear h U; intros h U.
  unfold hpredTrue.
  setoid_rewrite <- (addRemoveEqv IG).
  sconjLift3of3 U.
  hypReplace U (sconjAssocR U).
  apply (objHpIntro _ U).
  apply (scanFieldStOkFGDisj HorigI).
  unfold okObjAt; unfold okFieldAt; intros h x U; linLogElim U.
  linLogIntro; sconjMatch.

  refine (scanFieldStOkPhiEqv _ PhiEq).
  apply (scanFieldStOkImpl H); clear H.
  intros h H.
  sconjLift5of5 H.
  sconjGLift3.
  apply (sconjImplR H); clear h H; intros h H.

  assert (h003Ok2' : h003' ||- fieldRespMap 
    (map_union phi (map_union (map_union phi' phi3) phi4)) s (rget s rscan)).
  generalize dependent h003Ok2.
  unfold fieldRespMap; stSimpl; unfold mapX.
  intros U0; destruct (lexE U0) as [v0' U]; clear U0.
  existsl v0'; linLogElim U.
  destruct H2 as [v0 [V X]]; linLogIntro.
  exists v0; constructor.
  assumption.
  destruct (isOdd v0).
  assumption.
  rewrite <- PhiEq.
  apply mapUnionEq; apply mapUnionEq; assumption.
  regNotInListSolver.
  
  assert (h004Ok2' : h004' ||- fieldRespMap
    (map_union phi (map_union (map_union phi' phi3) phi4)) s (4 + rget s rscan)).
  generalize dependent h004Ok2.
  unfold fieldRespMap; unfold ch2Step; stSimpl.
  rewrite (regPresEq Rp).
  stSimpl; clear H; intro H.
  rename H into H0.
  destruct (lexE H0) as [v0' H]; clear H0; linLogElim H.
  existsl v0'; destruct H2 as [v0 [H2 H2']].
  linLogIntro.
  2:assumption.
  exists v0; constructor.
  assert (HX : h00 ||- 4 + rget s rscan |-> v0 ** top).
  assert (HX' : h00 ||- lexists v, 4 + rget s rscan |-> v ** top).
  hypReplace Hok' (sconjComm Hok').
  unfold okFieldAt in Hok'; linLogElim Hok'.
  exists x; apply (sconjImplR Hok'); intros; linLogIntro.
  destruct HX' as [v1 HX'].
  unfold scanFieldStOk in HX0; intuition.
  clear H0 H5 .
  sconjLift6of6 H11.
  hypReplace H11 (sconjInsert H11 HX'); linLogElim H11.
  assert (v0 = v1).
  eapply contContEq.
  apply H2.
  generalize dependent H10; stSimpl; intro; apply H11.
  subst; assumption.
  unfold scanFieldStOk in Horig; intuition.
  clear H8 H9 H11.
  sconjLift5of5 H12.
  hypReplace H12 (sconjInsert H12 HX); linLogElim H12.
  apply (sconjImplR H12); intros; linLogIntro.
  unfold mapX in H2' |- * .
  destruct (isOdd v0).
  assumption.
  rewrite <- PhiEq.
  apply mapUnionEq; assumption.
  regNotInListSolver.

  assert (Q : h ||-
    ((foralls x in set_remove (rget s rscan) G,
      objRespMap (map_union (map_union (map_union phi phi') phi3) phi4) s'' x) **
    Heap.heap_eqv h003' ** Heap.heap_eqv h004') **
    (foralls x in set_union G' G3,
      objRespMap (map_union (map_union (map_union phi phi') phi3) phi4) s'' x) **
    (foralls x in G4,
      objIsFwded (map_union (map_union phi phi') phi3) phi4 s'' x)).
  linLogIntro; sconjGLift2; sconjLift3 H; 
  apply (sconjImplR H); clear h H; intros h H.
  sconjLift3of3 H; apply sconjAssocR.
  apply (sconjImplR H); clear h H; intros h H.
  hypReplace H (sconjComm H).
  linLogIntro; sconjGLift2; apply (sconjImplR H); clear h H; intros h H.
  apply fasSplit; assumption.

  clear H Horig' Horig3; rename Q into H.
  apply (sconjImpl H); clear h H; intros h H.
  unfold hpredTrue; setoid_rewrite <- (addRemoveEqv IG).
  rewrite <- hpredTrueEq; apply fasIntro.
  hypReplace H (sconjComm H).
  apply (sconjImpl H); clear h H; intros h H.
  hypReplace H (sconjInsert H h003Ok2').
  hypReplace H (sconjComm (sconjInsert (sconjComm H) h004Ok2')).
  assumption.

  clear h003Ok2' h004Ok2' DG Al4 StEq4 Eeq Iso4 Feq4 Weq4 W4 M4 F4 G4 Comm4.
  clear h003Ok1 h003Ok2 h003' h004 GFdisj3 h004Ok1 h004Ok2.
  clear Req' Alf' Lt' Eq Alf Geq LtSt Al IG Feq0 Weq Feq Piso Al' WMeqv GFdisj'.
  clear F'eq Weq3 Feq3 Iso3 Alf3.

  assert (IIG : forall x, set_In x (set_remove (rget s rscan) G) -> 
    heapOf s'' ||- objCopied x s x ** top).
  intros x I.

  assert (Step1 : heapOf s ||- heap_eqv h00' ** top).
  unfold scanFieldStOk in Horig.
  intuition.
  sconjLift5of5 H11.
  hypReplace H11 (sconjInsert H11 Hok'); linLogElim H11.
  apply (sconjImplR H11); intros; linLogIntro.

  assert (Step3 : heapOf s'' ||- heap_eqv h00' ** top).
  unfold scanFieldStOk in HX0''.
  intuition.
  sconjLift7of7 H11.
  apply (sconjImplR H11); intros; linLogIntro.
  
  assert (InH : h00' ||- x |->__ ** top).
  assert (InH : h00' ||- (x |->__ ** top) ** top).
  apply (sconjImpl h00Ok'); clear h H; intros h H.
  hypReplace H (objHpElim I H).
  apply (sconjImpl H); clear h H; intros h H.
  unfold okObjAt in H; unfold okFieldAt in H.
  linLogElim H; linLogIntro; eauto.
  linLogIntro.
  linLogIntro.
  hypReplace InH (sconjAssocL InH); apply (sconjImpl InH); intros;
    [assumption | linLogIntro].

  refine (hpHpObjCopied _ Step1 _ Step3 InH).

  apply (fasImplI H).

  unfold objRespMap; unfold fieldRespMap.
  clear h H; intros x h IG H.
  assert (X := IIG x IG); clear IIG IG.
  linLogElim H.
  unfold objCopied in X; unfold fieldCopied in X; linLogElim X.
  destruct H4 as [v0 [Iv0 M0]]; destruct H5 as [v0' [Ivo' M0']].
  assert (R1 := contContEq X Iv0).
  assert (x3 = v0').
  sconjLift2 X; exact (contContEq X Ivo').
  subst.
  linLogIntro.
  2:eauto.
  exists v0; constructor.
  assumption.
  refine (mapXSub _ M0 _).
  apply mapEqvSub.
  assumption.

  exists v0'; constructor.
  assumption.
  refine (mapXSub _ M0' _).
  apply mapEqvSub.
  assumption.

  intro R; apply (set_remove_elim2 _ R); reflexivity.

  clear h003Ok2' h004Ok2' Al4 StEq4 Eeq Alf3.
  clear Alf Geq LtSt Al IG DG Lt' Alf' Req'.

  assert (WMdisj := scanFieldStOkWMDisj Horig).
  assert (phiM := scanPhiDom Horig).

  unfold hpredTrue; setoid_rewrite <- (set_eqvUnionAssoc G' G3 G4); 
    rewrite <- hpredTrueEq.
  apply fasMerge.
  apply (sconjImpl H); clear h H; intros h H.

  assert (D : set_disj G' G3).
  apply set_disjSymm.
  eapply set_unionDisjElimL.
  instantiate (1 := F3).
  setoid_rewrite <- Feq3.
  apply set_disjSymm; assumption.
  
  apply fasMerge.
  hypReplace H (fasSplit _ _ H D).
  clear D.
  apply (sconjImpl H); clear h H; intros h H.

  assert (CpOk : forall x, set_In x G' ->
    exists x', phi' x' = Some x /\ heapOf s'' ||- objCopied x' s x ** top).
  clear h H; intros x I.
  assert (HH : heapOf s'' ||-
    ((lexists x', !(phi' x' = Some x) ** 
      objCopied x' (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x) ** top) ** top).
  unfold scanFieldStOk in HX0''.
  intuition.
  sconjLift6 H10.
  hypReplace H10 (sconjInsert H10 h003Ok1); linLogElim H10; sconjLift2 H10.
  apply (sconjImpl H10); clear H; intros h H.
  hypReplace H (fasElim I H); apply (sconjImplR H); intros; linLogIntro.
  linLogIntro.
  hypReplace HH (sconjAssocL HH).
  destruct (sconjElim HH) as [h1 [h2 [J [HA HB]]]].
  linLogElim HA.
  exists x0; constructor.
  assumption.
  hypReplace HH (sconjIntro HA HB J).
  apply (sconjImpl HH).
  unfold objCopied; unfold fieldCopied; stSimpl; auto.
  intros; linLogIntro.

  apply (fasImplI H); clear h H; intros x h I H.
  destruct (CpOk _ I) as [x' [Peq H']]; clear CpOk I.
  unfold objIsFwded; existsl x'; linLogIntro.

  unfold map_union.
  replace (phi x') with (None (A := nat)).
  rewrite Peq; reflexivity.
  symmetry; apply phiM.
  intro Q; refine (set_disj_elim WMdisj _ _ Q).
  unfold iso_set in Piso; intuition.
  setoid_rewrite Weq.
  apply set_union_intro1.
  destruct (H2 x') as [_ RR]; apply RR; rewrite Peq; discriminate.

  unfold fieldIsFwded; unfold objRespMap in H; unfold fieldRespMap in H;
    unfold objCopied in H'; unfold fieldCopied in H'.
  linLogElim H'; linLogElim H.
  destruct H5 as [v01 [H5 H5']]; destruct H6 as [v02 [H6 H6']].
  assert (Eq2 := contContEq H5 H').
  assert (x1 = v02).
  sconjLift2 H'; apply (contContEq H' H6).
  subst.
  linLogIntro.
  2:eauto.
  exists x0; constructor.
  assumption.
  eapply mapXSub.
  apply H5'.
  apply mapEqvSub.
  assumption.
  exists v02; constructor.
  assumption.
  eapply mapXSub.
  apply H6'.
  apply mapEqvSub.
  assumption.

  assert (CpOk : forall x, set_In x G3 ->
    exists x', phi3 x' = Some x /\ heapOf s'' ||- objCopied x' s x ** top).
  clear h H; intros x I.
  assert (HH : heapOf s'' ||-
    ((lexists x', !(phi3 x' = Some x) ** 
      objCopied x' (ch2Step s') x) ** top) ** top).
  unfold scanFieldStOk in HX0''.
  intuition.
  sconjLift5 H10.
  apply (sconjImpl H10); clear H; intros h H.
  hypReplace H (fasElim I H); apply (sconjImplR H); intros; linLogIntro.
  linLogIntro.
  hypReplace HH (sconjAssocL HH).
  destruct (sconjElim HH) as [h1 [h2 [J [HA HB]]]].
  linLogElim HA.
  exists x0; constructor.
  assumption.
  hypReplace HH (sconjIntro HA HB J).
  apply (sconjImpl HH).
  unfold ch2Step; unfold objCopied; unfold fieldCopied; stSimpl.
  intros h H; linLogElim H.

  assert (IM3 : set_In x0 M3).
  unfold iso_set in Iso3; intuition.
  destruct (H7 x0) as [_ Q]; apply Q; rewrite H1; discriminate.
  destruct HX0 as [AlX [X0 SC0]].
  assert (OC : heapOf s' ||- 
    objCopied x0 (rset (rset s a0 (rget s rscan)) ra CHENEYLOOP2) x0 ** top).
  apply X0.
  setoid_rewrite Weq3; apply set_union_intro1; assumption.

  unfold objCopied in OC; unfold fieldCopied in OC.
  generalize dependent OC; stSimpl; intro.
  linLogElim OC.
  assert (R1 := contContEq H5 OC).
  assert (x2 = x4).
  sconjLift2 OC; exact (contContEq H6 OC).
  subst.
  linLogIntro; eauto.
  intros; linLogIntro.

  destruct HX0 as [AlX [X0 HX0]].
  apply (fasImplI H); clear h H; intros x h I H.
  destruct (CpOk _ I) as [x' [Peq H']]; clear CpOk I.
  unfold objIsFwded; existsl x'; linLogIntro.
  unfold map_union.
  replace (phi' x') with (None (A := nat)).
  rewrite Peq; reflexivity.
  assert (map_union phi phi' x' = None).
  eapply scanPhiDom; eauto.
  intro IMM.
  assert (IM3 : set_In x' W').
  unfold iso_set in Iso3; intuition.
  setoid_rewrite Weq3; apply set_union_intro1.
  destruct (H2 x') as [_ RR]; apply RR; rewrite Peq; discriminate.
  assert (X := scanFieldStOkWMDisj HX0).
  eapply set_disj_elim; eauto.
  unfold map_union in H0; destruct (phi x').
  discriminate H0.
  auto.

  unfold fieldIsFwded; unfold objRespMap in H; unfold fieldRespMap in H;
    unfold objCopied in H'; unfold fieldCopied in H'.
  linLogElim H'; linLogElim H.
  destruct H5 as [v01 [H5 H5']]; destruct H6 as [v02 [H6 H6']].
  assert (Eq2 := contContEq H5 H').
  assert (x1 = v02).
  sconjLift2 H'; apply (contContEq H' H6).
  subst.
  linLogIntro; eauto.
  exists x0; constructor.
  assumption.
  eapply mapXSub.
  apply H5'.
  apply mapEqvSub.
  assumption.

  exists v02; constructor.
  assumption.
  eapply mapXSub.
  apply H6'.
  apply mapEqvSub.
  assumption.
  assumption.

  assert (CpOk : forall x, set_In x M4 ->
    heapOf s'' ||- objCopied x s x ** top).
  clear h H; intros x IM4.

  assert (Step2 : heapOf s'' ||- objCopied x s' x ** top).
  assert (IW3 : set_In x W3).
  setoid_rewrite Weq4; apply set_union_intro1; assumption.
  assert (Q := WFr _ IW3).
  apply (sconjImplL Q).
  unfold objCopied; unfold fieldCopied; stSimpl; auto.

  destruct HX0 as [AlX [X0 _]].
  assert (Step1 : heapOf s' ||- objCopied x s x ** top).
  assert (IW' : set_In x W').
  setoid_rewrite Weq3; apply set_union_intro2;
    setoid_rewrite Weq4; apply set_union_intro1; assumption.
  assert (Q := X0 _ IW').
  apply (sconjImplL Q).
  unfold objCopied; unfold fieldCopied; stSimpl; auto.
  eapply objCopiedTrans; eauto.

  apply (fasImplI H); clear h H; intros x h I.
  unfold objIsFwded; unfold fieldIsFwded.
  intros H; linLogElim H; existsl x0.
  assert (IM4 : set_In x0 M4).
  clear H7 H8 x1 x2 H Rp Horig Hok.

  unfold iso_set in Iso4; intuition.
  destruct (H4 x0) as [_ Q]; apply Q; rewrite H2; discriminate.
  destruct H7 as [v0 [H7a H7b]]; destruct H8 as [v0' [H8a H8b]].
  unfold map_union; replace (phi' x0) with (None (A := nat)).
  replace (phi3 x0) with (None (A := nat)).

  assert (QQ := CpOk _ IM4).
  unfold objCopied in QQ; unfold fieldCopied in QQ; linLogElim QQ.
  assert (Eq2 := contContEq H7a QQ).
  assert (x4 = v0').
  sconjLift2 QQ; apply (contContEq QQ H8a).
  subst.

  linLogIntro.
  2:eauto.
  exists x3; constructor.
  assumption.
  eapply mapXSub.
  eauto.
  apply mapEqvSub.
  assumption.
  exists v0'; constructor.
  assumption.
  eapply mapXSub.
  eauto.
  apply mapEqvSub.
  assumption.

  assert (X : phi3 x0 = None \/ set_In x0 M3).
  unfold iso_set in Iso3; intuition.
  destruct (H5 x0) as [_ RR].
  destruct (phi3 x0).
  right; apply RR; discriminate.
  left; reflexivity.
  destruct X.
  symmetry; assumption.
  elimtype False.
  assert (set_In x0 W3).
  setoid_rewrite Weq4; apply set_union_intro1; assumption.
  assert (X := scanFieldStOkWMDisj HX0'').
  eapply set_disj_elim; eauto.
  apply set_union_intro2; assumption.

  assert (X : phi' x0 = None \/ set_In x0 M'').
  unfold iso_set in Piso; intuition.
  destruct (H3 x0) as [_ RR].
  destruct (phi' x0).
  right; apply RR; discriminate.
  left; reflexivity.
  destruct X.
  symmetry; assumption.
  elimtype False.
  assert (set_In x0 W').
  setoid_rewrite Weq3; apply set_union_intro2.
  setoid_rewrite Weq4; apply set_union_intro1; assumption.
  destruct HX0 as [_ [_ HX0]].
  assert (X := scanFieldStOkWMDisj HX0).
  eapply set_disj_elim; eauto.
  apply set_union_intro2; assumption.

  apply set_disjSymm.
  apply (set_unionDisjElimL G4 F4).
  setoid_rewrite <- Feq4.
  apply set_disjSymm.
  apply set_unionDisjIntro.
  apply set_disjSymm.
  apply (set_unionDisjElimR G3 F3).
  setoid_rewrite <- Feq3.
  apply set_disjSymm; assumption.
  assumption.

  unfold regPres; simpl; intuition.
  rewrite (regPresEq H).
  rewrite (regPresEq Rp'); unfold ch2Step; stSimpl.
  rewrite (regPresEq Rp); stSimpl.
  reflexivity.
  regNotInListSolver.
  regNotInListSolver.
  regNotInListSolver.
  rewrite H0; rewrite (regPresEq Rp'); unfold ch2Step; stSimpl.
  rewrite (regPresEq Rp); stSimpl.
  reflexivity.
  regNotInListSolver.
  regNotInListSolver.


  (* putting it together *)
  destruct H as [X Y].
  constructor.
  constructor.
  assumption.
  intros s' H; destruct (Y _ H) as [Q1 Q2].
  clear X Y.
  unfold cheneyLoop2Pre.
  constructor.
  assumption.
  intros s'' H''; destruct (Q2 _ H'') as [AA BB].
  assumption.

  clear X; intros s''' [s' [G1 G2]].
  destruct (Y _ G1) as [Y1 Y2]; clear Y G1.
  unfold cheneyLoop2Guar in G2.
  destruct G2 as [s'' [G1 G2]].
  destruct (Y2 _ G1) as [X Y]; clear Y2 G1.
  exact (Y _ G2).

Qed.


