Require Import StateTheory.
Require Import VGC.
Require Import List.
Require Import GC.
Require Import ScanField.
Require Import Cheney.
Require CheneyLemma.
Require Import SCAP.


Set Implicit Arguments.


Module CheneyGCMGI := BasicMGI (CheneyGCGuar).


Module CheneyGCImpl <: VIEW_GC with Module I := CheneyGCMGI.

  Module I := CheneyGCMGI.

  Definition mview := mview.


  Lemma atomRegEq : forall h vs rf rr A r,
    h ||- mview vs rf rr ** A ->
    In r rr ->
    I.atomP (RFile.get rf r) \/ I.atomP (rget vs r) ->
    RFile.get rf r = rget vs r.
  Proof.
    intros h vs rf rr A r Hp I D.
    unfold mview in Hp; unfold Cheney.mview in Hp; linLogElim Hp.
    destruct H3 as [_ [_ [objs [Req Vhp]]]].
    destruct (Req _ I) as [X _].
    auto.
  Qed.

  Lemma setNonRootOk : forall h vs rf rr r v,
    h ||- mview vs rf rr ->
    ~In r (rgcInfo::rr) ->
    h ||- mview vs (RFile.set rf r v) rr.
  Proof.
    unfold mview.
    unfold Cheney.mview.
    intros h vs rf rr r v Hp I.
    linLogElim Hp.
    destruct H1 as [rrOk [alOk1 [alOk2 [rootsOk [endOk endOk']]]]].
    linLogIntro.
    constructor.
    assumption.
    constructor.
    apply alOk1.
    constructor.
    apply alOk2.
    constructor.
    intros r0 I0; destruct (rootsOk _ I0) as [X Y].
    constructor.
    replace (RFile.get (RFile.set rf r v) r0) with (rget (rset (h, rf) r v) r0).
    2:reflexivity.
    rewrite X.
    replace (RFile.get rf r0) with (rget (h, rf) r0).
    2:reflexivity.
    destruct (regCompare r0 r).
    subst r; elimtype False; apply I; right; assumption.
    stSimpl; reflexivity.
    assumption.
    constructor; eauto.
    sconjMatch.
    destruct (regCompare r rgcInfo).
    subst r.
    elimtype False; apply I; left; reflexivity.
    replace (RFile.get (RFile.set rf r v) rgcInfo) with (rget (rset (h, rf) r v) rgcInfo).
    stSimpl.
    assumption.
    reflexivity.
  Qed.

  
  Lemma rootCopyOk : forall h vs rf rr rs rd,
    h ||- mview vs rf rr ->
    In rs rr -> 
    ~In rd (zero::rgcInfo::nil) ->
    h ||- mview (rset vs rd (rget vs rs)) (RFile.set rf rd (RFile.get rf rs)) (rd::rr).
  Proof.
    unfold mview; unfold Cheney.mview.
    intros h vs rf rr rs rd Hp I0 Nz.
    assert (NE1 : rd <> zero).
    intro E; subst rd; intuition.
    assert (NE2 : rd <> rgcInfo).
    intro E; subst rd; intuition.
    linLogElim Hp.
    destruct H1 as [rrOk [al1 [al2 [rootOk boundsOk]]]].
    linLogIntro.

    constructor.
    simpl; intro X; destruct X.
    subst rd; intuition.
    auto.
    constructor.
    apply al1.
    constructor.
    eauto.
    constructor.
    2:eauto.
    simpl; intros r I1.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) r) with
      (rget (rset (h, rf) rd (rget (h, rf) rs)) r).
    2:reflexivity.
    destruct (regCompare r rd).
    subst r; stSimpl.
    destruct (rootOk _ I0) as [X Y].
    rewrite X; constructor.
    reflexivity.
    rewrite X in Y; assumption.
    destruct I1 as [I | I].
    elimtype False; auto.
    stSimpl.
    destruct (rootOk _ I) as [R Ok].
    rewrite R; rewrite R in Ok; constructor.
    reflexivity.
    assumption.

    stSimpl; sconjMatch.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) rgcInfo) with
      (rget (rset (h, rf) rd (rget (h, rf) rs)) rgcInfo).
    stSimpl; assumption.
    reflexivity.
  Qed.


  Lemma rootWeaken : forall h vs rf rr rr',
    h ||- mview vs rf rr ->
    incl rr' rr ->
    h ||- mview vs rf rr'.
  Proof.
    unfold mview; unfold Cheney.mview.
    intros h vs rf rr rr' Hp IR.
    linLogElim Hp.
    destruct H1 as [rrOk [al1 [al2 [rootOk endOk]]]].
    linLogIntro.
    constructor.
    auto.
    constructor.
    apply al1.
    constructor.
    eauto.
    constructor.
    2:eauto.
    intros r I.
    apply rootOk.
    auto.

    assumption.

  Qed.


  Lemma addAtomRoot : forall h vs rf rr A r,
    h ||- mview vs rf rr ** A ->
    ~In r (zero::I.rgcInfo::rr) ->
    I.atomP (RFile.get rf r) ->
    h ||- mview (rset vs r (RFile.get rf r)) rf (r::rr) ** A.
  Proof.
    unfold mview; unfold Cheney.mview.
    intros h vs rf rr A r Hp NI Atom.
    assert (NE1 : r <> zero).
    intro E; subst r; intuition.
    assert (NE2 : r <> rgcInfo).
    intro E; subst r; apply NI; right; left; reflexivity.
    linLogElim Hp.
    destruct H3 as [rrOk [al1 [al2 [rootOk endOk]]]].
    linLogIntro.

    constructor.
    simpl; intros [X | X]; auto.
    constructor.
    apply al1.
    constructor.
    apply al2.
    constructor.
    simpl; intros r0 I.
    replace (RFile.get rf) with (rget (h, rf)); [idtac | reflexivity].
    destruct I.
    subst r0; stSimpl.
    constructor.
    reflexivity.
    left; assumption.
    destruct (rootOk _ H) as [Req OkF].
    replace (RFile.get rf) with (rget (h, rf)) in Req; [idtac | reflexivity].
    rewrite <- Req; clear Req.
    destruct (regCompare r0 r).
    subst r0; elimtype False; apply NI; right; right; assumption.
    stSimpl.
    intuition.
    eauto.
    assumption.
  Qed.


  Definition phi n := 
    match Cheney.P n with
      | SomeT x => SomeT x
      | NoneT => ScanField.Pscanf n
    end.


  Ltac phiStep Ne f k :=
    destruct (natEqDec f k) as [R | R];
      [subst f; rewrite beq_true in Ne; apply setRangeIntro; omega |
       rewrite (beq_false R) in Ne; clear R].

  Lemma gcPhiDomOk : forall f,
    phi f <> NoneT _ -> set_In f I.gcPhiDom.
  Proof.
    unfold I.gcPhiDom; unfold phi; unfold P; unfold Pscanf.
    unfold SCANFIELD; unfold CHENEYLOOP; unfold CHENEYLOOP2; unfold CHENEYRETURN;
      unfold CHENEYLOOPHEADER; unfold CHENEYENTER; unfold CHENEYDOALLOC; unfold CHENEYALLOC;
      unfold CHENEYPOSTGC; unfold INFLOOP.
    unfold RETURN; unfold SCANPOINTER; unfold NOTTOSPACEPTR; 
      unfold SCANCOPIED; unfold SCANNOCOPY;
	unfold FWDOBJ.

    intros f Ne.

    phiStep Ne f 30.
    phiStep Ne f 17.
    phiStep Ne f 18.
    phiStep Ne f 21.
    phiStep Ne f 20.
    phiStep Ne f 19.
    phiStep Ne f 15.
    phiStep Ne f 10.
    phiStep Ne f 16.
    phiStep Ne f 22.
    phiStep Ne f 11.
    phiStep Ne f 12.
    phiStep Ne f 13.
    phiStep Ne f 14.
    phiStep Ne f 37.
    phiStep Ne f 34.
    phiStep Ne f 32.
    phiStep Ne f 36.
    phiStep Ne f 35.
    phiStep Ne f 31.

    elimtype False; apply Ne; reflexivity.

  Qed.


  Ltac PhiOkSolver :=
    unfold phi; unfold P; rewrite beq_true;
    repeat (rewrite beq_false; [idtac | discriminate]);
    simpl; intuition.

  Lemma read1PhiOk :
    optPredT (phi I.read1Lbl)
      (specWeaker (read1Spec mview I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma read2PhiOk :
    optPredT (phi I.read2Lbl)
      (specWeaker (read2Spec mview I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write1PhiOk : 
    optPredT (phi I.write1Lbl)
      (specWeaker (write1Spec mview I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write2PhiOk : 
    optPredT (phi I.write2Lbl)
      (specWeaker (write2Spec mview I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma allocPhiOk : 
    optPredT (phi I.allocLbl)
      (specWeaker (allocSpec mview I.rroot I.gcGuar I.presReg I.NULL)).
  Proof.
    PhiOkSolver.
  Qed.


  Definition GCC' f :=
    if beq_nat f CHENEYREAD1 then
      Some (cheneyRead 0)
    else if beq_nat f CHENEYREAD2 then
      Some (cheneyRead 4)
    else if beq_nat f CHENEYWRITE1 then
      Some (cheneyWrite 0)
    else if beq_nat f CHENEYWRITE2 then
      Some (cheneyWrite 4)
    else if beq_nat f RETURN then
      Some CheneyLemma.returnBlock
    else if beq_nat f SCANPOINTER then
      Some scanPointer
    else if beq_nat f NOTTOSPACEPTR then
      Some isNotToSpacePtr
    else if beq_nat f SCANCOPIED then
      Some scanCopied
    else if beq_nat f SCANNOCOPY then
      Some scanNoCopy
    else if beq_nat f FWDOBJ then
      Some fwdObj
    else
      None.


  Definition GCC f :=
    if beq_nat f SCANFIELD then
      Some scanField
    else if beq_nat f CHENEYLOOP then
      Some cheneyLoop
    else if beq_nat f CHENEYLOOP2 then
      Some cheneyLoop2
    else if beq_nat f CHENEYRETURN then
      Some cheneyReturn
    else if beq_nat f CHENEYLOOPHEADER then
      Some cheneyLoopHeader
    else if beq_nat f CHENEYENTER then
      Some cheneyEnter
    else if beq_nat f CHENEYDOALLOC then
      Some cheneyDoAlloc
    else if beq_nat f CHENEYALLOC then
      Some cheneyAlloc
    else if beq_nat f CHENEYPOSTGC then
      Some cheneyPostGC
    else if beq_nat f INFLOOP then
      Some infLoop
    else 
      GCC' f.


  Ltac weakerStep f f' H :=
    destruct (natEqDec f f') as [R | R];
      [subst f; rewrite beq_true;
	injection 1; clear H; intros; subst p g;
	  econstructor; econstructor; constructor;
	    [unfold P; repeat (rewrite beq_false; [idtac | discriminate]) | 
	      constructor; [reflexivity | intros s s' Guar; apply Guar]] |
       rewrite (beq_false R); clear R].

  Lemma phiWeakerPscanf : forall f p0 g0,
    ScanField.Pscanf f = SomeT (bspeci p0 g0) ->
    exists p1, exists g1, phi f = SomeT (bspeci p1 g1) /\
      p0 = p1 /\ forall s s', g1 s s' -> g0 s s'.
  Proof.
    unfold Pscanf; unfold phi.
    intros f p g.

    weakerStep f RETURN H.
    reflexivity.
    weakerStep f SCANPOINTER H.
    reflexivity.
    weakerStep f NOTTOSPACEPTR H.
    reflexivity.
    weakerStep f SCANCOPIED H.
    reflexivity.
    weakerStep f SCANNOCOPY H.
    reflexivity.
    weakerStep f FWDOBJ H.
    reflexivity.
    weakerStep f SCANFIELD H;
      [rewrite beq_true; reflexivity |
	idtac].
    discriminate 1.

  Qed.


  Lemma phiWeakerP : forall f p g,
    P f = SomeT (bspeci p g) ->
    exists p', exists g', phi f = SomeT (bspeci p' g') /\
      p = p' /\ forall s s', g' s s' -> g s s'.
  Proof.
    unfold phi.
    intros f p g Peq; destruct (P f).
    injection Peq; intro; subst b; clear Peq.
    econstructor; econstructor; constructor.
    reflexivity.
    intuition.
    discriminate.
  Qed.


  Ltac GCCokStep' Peq f f' R0 :=
    destruct (natEqDec f f') as [R | R];
      [subst f; rewrite beq_true; injection 1; clear Peq; intros; subst p g |
	rewrite (beq_false R); rename R into R0].

  Ltac GCCokStep Peq f f' :=
    destruct (natEqDec f f') as [R | R];
      [subst f; rewrite beq_true; injection 1; clear Peq; intros; subst p g |
	rewrite (beq_false R); clear R].
    

  Lemma GCCok : cdHeapOkSC phi GCC phi.
  Proof.
    intros f [p g] Peq.
    unfold phi in Peq; unfold GCC; generalize dependent Peq; unfold P.
    
    GCCokStep' Peq f SCANFIELD RZ.
    2:GCCokStep Peq f CHENEYLOOP.
    3:GCCokStep Peq f CHENEYLOOP2.
    4:GCCokStep Peq f CHENEYRETURN.
    5:GCCokStep Peq f CHENEYLOOPHEADER.
    6:GCCokStep Peq f CHENEYENTER.
    7:GCCokStep Peq f CHENEYDOALLOC.
    8:GCCokStep Peq f CHENEYALLOC.
    9:GCCokStep Peq f CHENEYPOSTGC.
    10:GCCokStep Peq f INFLOOP.
    11:unfold GCC'; GCCokStep Peq f 11.

    apply (iseqOkWeakP _ _ _ _ scanFieldOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ cheneyLoopOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyLoop2Ok _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyReturnOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyLoopHeaderOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyEnterOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyDoAllocOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyAllocOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ cheneyPostGCOk _ phiWeakerP).
    apply (iseqOkWeakP _ _ _ _ infLoopOk _ phiWeakerP).
    eapply iseqOkWeakP; [apply cheneyReadOk; auto | exact phiWeakerP].
    GCCokStep Peq f 12.
    2:GCCokStep Peq f 13.
    3:unfold Pscanf; GCCokStep Peq f 14.
    eapply iseqOkWeakP; [apply cheneyReadOk; auto | exact phiWeakerP].
    eapply iseqOkWeakP; [apply cheneyWriteOk; auto | exact phiWeakerP].
    eapply iseqOkWeakP; [apply cheneyWriteOk; auto | exact phiWeakerP].

    GCCokStep Peq f RETURN.
    2:GCCokStep Peq f SCANPOINTER.
    3:GCCokStep Peq f NOTTOSPACEPTR.
    4:GCCokStep Peq f SCANCOPIED.
    5:GCCokStep Peq f SCANNOCOPY.
    6:GCCokStep Peq f FWDOBJ.

    eapply iseqOkWeakP; [apply CheneyLemma.returnOk | discriminate 1].
    apply (iseqOkWeakP _ _ _ _ scanPointerOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ isNotToSpacePtrOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ scanCopiedOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ scanNoCopyOk _ phiWeakerPscanf).
    apply (iseqOkWeakP _ _ _ _ fwdObjOk _ phiWeakerPscanf).

    GCCokStep Peq f SCANFIELD.
    auto.
    discriminate 1.

  Qed.


End CheneyGCImpl.
