Require Export WeakSCAP.
Require Export LinLog.
Require Export List.
Require Export SpecUtil.
Require Export GC.
Require Export StateTheory.
Require Export FnMap.


Set Implicit Arguments.


(* various defs used in the specs *)
Definition notSpacePtrP x sp0 sp1 := atomP x \/ ~(set_In x (setRange sp0 sp1)).

Definition fieldRespMap phi s x : hpred :=
  lexists v0', 
    !(exists v0, heapOf s ||- x |-> v0 ** top /\ mapX phi v0 = Some v0') **
    x |-> v0'.

Definition objRespMap phi s x : hpred :=
  (fieldRespMap phi s x) ** (fieldRespMap phi s (4+x)).


Definition fieldIsFwded x0 phi s x : hpred :=
  lexists v0',
  !(exists v0, heapOf s ||- x0 |-> v0 ** top /\ mapX phi v0 = Some v0') **
  x |-> v0'.

Definition objIsFwded phi phi' s x : hpred :=
  lexists x0, !(phi' x0 = Some x) **
  fieldIsFwded x0 (map_union phi phi') s x ** 
  fieldIsFwded (4 + x0) (map_union phi phi') s (4 + x).

(* object copied from some location to x, respecting phi *)
Definition objCopiedPhi phi s x :=
  (lexists x', !(phi x' = Some x) ** objCopied x' s x).



Definition rfileEqv s s' := forall r, rget s r = rget s' r.




Lemma objHpIso : forall h objs' phi s,
  h ||- (foralls x in objs', objCopiedPhi phi s x) ->
    forall objs0 objs A,
      s |||- objHp objs0 objs ** A ->
	iso_set phi objs objs' ->
	h ||- objHp objs0 objs'.
Proof.
  intros h objs' phi s Hp objs0 objs A Hp' iso.
  unfold objHp.
  eapply fasImpl.
  apply Hp.
  simpl; unfold objCopiedPhi; unfold objCopied.
  clear h Hp; intros x h Hp.
  linLogElim Hp.

  assert (Hp'' : s |||- (okObjAt objs0 x0 ** top) ** top).
  assert (I : set_In x0 objs).
  destruct iso as [iso _].
  destruct (iso x0) as [_ X].
  apply X; rewrite H1; discriminate.
  apply (sconjImpl Hp'); clear h Hp; intros h Hp.
  hypReplace Hp (objHpElim I Hp).
  sconjMatch.
  linLogIntro.
  
  unfold okObjAt in Hp''.
  linLogElim Hp''; clear Hp'; rename Hp'' into Hp'.
  unfold okObjAt; unfold okFieldAt in Hp'; linLogElim Hp'.
  unfold fieldCopied in Hp.

  linLogElim Hp.
  unfold okFieldAt; sconjMatch.
  apply lexI with (x := x3); linLogIntro.
  assert (E := contContEq Hp' H5).
  subst x1; assumption.
  assumption.
  apply lexI with (x := x4); linLogIntro.
  sconjLift2 Hp'.
  assert (E := contContEq Hp' H6).
  subst x2; assumption.
  assumption.
Qed.


Lemma objHpIsFwded : forall h W phi s,
  h ||- (foralls x in W, objIsFwded (fun _ => None) phi s x) ->
    forall s0 s0', iso_set phi s0 s0' -> 
      h ||- objHp s0' W.
Proof.
  intros h W phi s Hp s0 s0' iso0.
  unfold objHp.
  eapply fasImpl.
  apply Hp.
  unfold objIsFwded; unfold okObjAt.
  clear h Hp; intros x h Hp; linLogElim Hp.

  apply (sconjImpl Hp); clear h Hp; intros h Hp;
      unfold fieldIsFwded in Hp; unfold okFieldAt;
	linLogElim Hp;
	apply lexI with (x := x1);
	destruct H2 as [v0 [_ M]];
	  unfold map_union in M; unfold mapX in M;
	    assert (IO := isOddOk v0).
  
  linLogIntro; [idtac | assumption].
  destruct (isOdd v0); [left | right].
  injection M; intro R; subst; assumption.
  destruct iso0 as [_ [X0 _]]; hypReplace X0 (X0 v0).
  destruct (phi v0); eauto.

  linLogIntro; [idtac | assumption].
  destruct (isOdd v0); [left | right].
  injection M; intro R; subst; assumption.
  destruct iso0 as [_ [X0 _]]; hypReplace X0 (X0 v0).
  destruct (phi v0); eauto.

Qed.


Lemma copyMapFwded : forall s' phi' s x A,
  s' |||- objCopiedPhi phi' s x ** A -> (* was phi *)
    forall h phi,
      h ||- objRespMap (map_union phi phi') s' x ->
	h ||- objIsFwded phi phi' s x.
Proof.
  unfold objRespMap; unfold objIsFwded; unfold objCopiedPhi;
      unfold objCopied.
  intros s' phi s x A Hp' h phi' Hp.
  linLogElim Hp'; existsl x0; linLogIntro.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.

  unfold fieldRespMap in Hp; unfold fieldIsFwded.
  linLogElim Hp.
  existsl x1; linLogIntro.
  destruct H1 as [v0 [H1 MX]].
  exists v0; intuition.
  unfold fieldCopied in Hp'; linLogElim Hp'.
  assert (R := contContEq H1 Hp').
  subst; assumption.
  assumption.

  unfold fieldRespMap in Hp; unfold fieldIsFwded.
  linLogElim Hp.
  existsl x1; linLogIntro.
  destruct H1 as [v0 [H1 MX]].
  exists v0; intuition.
  unfold fieldCopied in Hp'; linLogElim Hp'.
  sconjLift2 Hp'.
  assert (R := contContEq H1 Hp').
  subst; assumption.
  assumption.

Qed.


Lemma fwdedOkFieldVal : forall h y phi s z A,
  h ||- fieldIsFwded y phi s z ** A -> forall S S',
    iso_set phi S S' -> forall x,
    s |||- y |-> x ** top ->
      okFieldVal S x.
Proof.
  unfold fieldIsFwded.
  intros h y phi s z A Hp' S S' iso x Hp.
  linLogElim Hp'.
  destruct H3 as [v0 [Hp0 Meq]]; clear Hp'.
  assert (Q := contContEq Hp Hp0); subst v0; clear Hp Hp0.
  unfold mapX in Meq.
  assert (IO := isOddOk x).
  destruct (isOdd x); [left | right].
  assumption.
  destruct iso as [I _]; hypReplace I (I x); destruct I as [_ I].
  apply I; rewrite Meq; discriminate.
Qed.


Lemma hpHpObjCopied : forall s h A,
  heapOf s ||- Heap.heap_eqv h ** A ->
    forall s' B, heapOf s' ||- Heap.heap_eqv h ** B ->
      forall x C, h ||- x |->__ ** C ->
	heapOf s' ||- objCopied x s x ** top.
Proof.
  intros s h A HA s' B HB x C HC.
  linLogElim HC.
  unfold objCopied; unfold fieldCopied.
  hypReplace HA (sconjInsert HA HC); linLogElim HA.
  hypReplace HB (sconjInsert HB HC); linLogElim HB.
  linLogIntro.
  instantiate (1 := x0).
  sconjMatch.
  instantiate (1 := x1).
  sconjMatch.
  sconjLift2 HA; sconjMatch.
Qed.


Lemma objCopiedTrans : forall s' x s y,
  heapOf s' ||- objCopied x s y ** top ->
    forall s'' z,
      heapOf s'' ||- objCopied y s' z ** top ->
	heapOf s'' ||- objCopied x s z ** top.
Proof.
  unfold objCopied; unfold fieldCopied.
  intros s' x s y H s'' z H'.
  linLogElim H.
  linLogElim H'.
  linLogIntro.
  eauto.
  instantiate (1 := x1).
  assert (R1 := contContEq H5 H).
  assert (x3 = x1).
  sconjLift2 H; exact (contContEq H9 H).
  subst.
  assumption.
  assumption.
Qed.


(* code common between the Cheney and Baker collectors *)

Notation map_add := (map_add beq_nat).


(* size constants *)
Definition WORDSIZE := 4.
Definition OBJSIZE := 8.





(* definition and verification of the basic return block *)
Definition RETURN := 1.


Definition returnBlock :=
  jr ra.


Definition returnPre : spred := fun s => True.

Definition returnGuar : guar := fun s s' => s = s'.


Lemma returnOk : iseqOkS (fun _ => NoneT _) returnPre returnGuar returnBlock.
Proof.
  unfold returnBlock.
  iseqWeakSolver PsharedSimpl.
  unfold returnPre; unfold returnGuar.
  intuition.
Qed.


Definition isoState rr s s' :=
  exists objs, exists objs', exists phi,
    s |||- objHp objs objs /\
    s' |||- objHp objs' objs' /\
    iso_set phi objs objs' /\
    (forall r, In r rr -> mapX phi (rget s r) = Some (rget s' r)) /\
    s' |||- (foralls x in objs', objIsFwdedH phi (heapOf s) x).


Lemma isoStateRefl' : forall h rf rf' rr objs,
  h ||- objHp objs objs ->
  (forall r, In r rr -> 
    okFieldVal objs (rget (h, rf) r) /\ rget (h, rf) r = rget (h, rf') r) ->
  isoState rr (h, rf) (h, rf').
Proof.
  intros h rf rf' rr objs Hp Rok.
  unfold isoState.
  exists objs; exists objs.
  exists (fun x => if set_mem x objs then Some x else None).
  constructor.
  assumption.
  constructor.
  assumption.
  constructor.

  clear Hp rr h rf rf' Rok.
  constructor.
  intros x; assert (X := set_memSound x objs); constructor;
    intro I; destruct (set_mem x objs).
  discriminate.
  intro; auto.
  assumption.
  elimtype False; apply I; reflexivity.

  constructor.
  intros x y; assert (X := set_memSound x objs).
  destruct (set_mem x objs).
  injection 1; intro Z; subst y; assumption.
  discriminate 1.

  constructor.
  intros x x' y; assert (X := set_memSound x objs);
    assert (X' := set_memSound x' objs).
  destruct (set_mem x objs).
  destruct (set_mem x' objs).
  injection 1; intro U; injection 1; intro V; subst x x'; reflexivity.
  discriminate 2.
  discriminate 1.

  intros y I; assert (X := set_memSound y objs).
  exists y; destruct (set_mem y objs).
  reflexivity.
  elimtype False; auto.

  constructor.
  intros r I; unfold mapX.
  hypReplace Rok (Rok _ I); clear I.
  destruct Rok as [Rok Req].
  rewrite <- Req; clear Req rf'.
  destruct Rok.

  assert (X := isOddOk (rget (h, rf) r)).
  destruct (isOdd (rget (h, rf) r)).
  reflexivity.
  elimtype False; eapply not_even_and_odd; eauto.
  destruct (isOdd (rget (h, rf) r)).
  reflexivity.
  assert (X := set_memSound (rget (h, rf) r) objs).
  destruct (set_mem (rget (h, rf) r) objs).
  reflexivity.
  elimtype False; auto.

  stSimpl.
  unfold objHp in Hp.
  clear rr Rok rf rf'.
  eapply fasImplIS.
  eauto.
  simpl.
  clear Hp; unfold okObjAt; unfold okFieldAt; intros x h' I S Hp; linLogElim Hp.
  unfold objIsFwdedH; linLogIntro.
  2:simpl; eauto.
  exists x; exists x0; exists x1.
  assert (X := set_memSound x objs).
  destruct (set_mem x objs).
  2:elimtype False; auto.
  constructor.
  reflexivity.
  constructor.
  linLogIntro.
  apply sconjAssocL; eapply subHeapImpl; eauto.
  unfold mapX.
  clear X.
  constructor.

  assert (X0 := isOddOk x0).
  destruct H3; destruct (isOdd x0).
  reflexivity.
  elimtype False; eapply not_even_and_odd; eauto.
  reflexivity.
  assert (X := set_memSound x0 objs).
  destruct (set_mem x0 objs).
  reflexivity.
  elimtype False; auto.

  assert (X0 := isOddOk x1).
  destruct H4; destruct (isOdd x1).
  reflexivity.
  elimtype False; eapply not_even_and_odd; eauto.
  reflexivity.
  assert (X := set_memSound x1 objs).
  destruct (set_mem x1 objs).
  reflexivity.
  elimtype False; auto.

Qed.


Lemma isoStateRefl : forall s rr objs,
  s |||- objHp objs objs ->
  (forall r, In r rr -> okFieldVal objs (rget s r)) ->
  isoState rr s s.
Proof.
  intros [h rf] rr objs Hp Rok.
  eapply isoStateRefl'; eauto.
Qed.


Lemma inclEq' : forall (r:rnum) rr,
  incl rr (r::nil) ->
  rr = nil \/ (forall r', In r' rr -> r' = r).
Proof.
  induction rr.
  simpl; left; reflexivity.
  intros I; right.
  intros r' I'.
  destruct IHrr.
  intros r'' I''.
  apply I; simpl; right; assumption.
  assert (X := I r' I').
  simpl in X; destruct X; [auto | elimtype False; assumption].
  simpl in I'; destruct I'.
  subst a.
  hypReplace I (I r'); simpl in I; destruct I.
  left; reflexivity.
  auto.
  elimtype False; assumption.
  apply H; assumption.
Qed.


Lemma inclEq : forall (r:rnum) rr,
  incl rr (r::nil) ->
  forall r', In r' rr -> r' = r.
Proof.
  intros r rr I r' I'.
  destruct (inclEq' I).
  subst rr; simpl in I'; elimtype False; assumption.
  auto.
Qed.


Lemma isoStateEmpty : forall rf rf',
  isoState nil (Heap.empty_heap, rf) (Heap.empty_heap, rf').
Proof.
  intros rf rf'.
  exists empty_set; exists empty_set; exists (empty_map nat nat).
  stSimpl.

  constructor.
  apply objHpIntroE; apply unitIntro.

  constructor.
  apply objHpIntroE; apply unitIntro.

  constructor.
  apply empty_mapIso.

  constructor.
  simpl; intros; elimtype False; assumption.

  apply fasIntroE; apply unitIntro.
Qed.


Add Morphism isoState : isoStateMorph.
Proof.
  intros l s1 s1' s2 s2' E1 E2 [objs [objs' [phi I]]].
  exists objs; exists objs'; exists phi.
  intuition.
  eapply heapEqvTrue; eauto;
    apply Heap.heap_eqvSymm; apply stateHeapEqv; assumption.
  eapply heapEqvTrue; eauto;
    apply Heap.heap_eqvSymm; apply stateHeapEqv; assumption.
  rewrite <- (rgetEqv E1); rewrite <- (rgetEqv E2); auto.
  eapply heapEqvTrue; eauto.
  apply Heap.heap_eqvSymm; apply stateHeapEqv; eauto.
  
  eapply fasImpl.
  eauto.
  unfold objIsFwdedH; intros x h Hp;
    linLogElim Hp; linLogIntro.
  2:eauto.
  destruct H6 as [x2 [v0 [v1 [Peq [Vhp [Meq Meq']]]]]].
  exists x2; exists v0; exists v1.
  intuition.
  eapply heapEqvTrue; eauto;
    apply Heap.heap_eqvSymm; apply stateHeapEqv; assumption.

Qed.


Lemma regListMapEqv : forall r rr f,
  (forall r' : rnum, In r' (r::rr) -> r' = r) ->
  set_eqv (List.map f (r::rr)) (set_add (f r) empty_set).
Proof.
  intros r rr f Req.
  simpl.
  induction rr.

  simpl; constructor; intro I; auto.

  assert (Q : forall r' : rnum, In r' (r :: rr) -> r' = r).
  intros; apply Req.
  simpl in H; destruct H.
  left; auto.
  right; right; assumption.

  hypReplace IHrr (IHrr Q); clear Q.
  assert (a = r).
  apply Req; right; left; reflexivity.
  subst a.
  rewrite <- IHrr; clear IHrr.
  constructor; simpl; intro I; intuition.
Qed.


Lemma reachForward : forall phi X X',
  iso_set phi X X' -> forall s s',
  s' |||- (foralls x' in X', objIsFwded (fun x:nat => None) phi s x') ** top -> forall v',
  (forall x', set_In x' X' -> reachable (heapOf s') v' x') -> forall v,
  phi v = Some v' ->
  forall x, set_In x X ->
    reachable (heapOf s) v x.
Proof.
  intros phi objs objs' iso s s' Hp v' F v Pv x I.
  assert (exists x', set_In x' objs' /\ phi x = Some x').
  destruct iso as [iso1 [iso2 _]].
  hypReplace iso1 (iso1 x); destruct iso1 as [iso1 _];
    hypReplace iso1 (iso1 I); clear I.
  hypReplace iso2 (iso2 x).
  destruct (phi x).
  exists a; constructor; eauto.
  elimtype False; apply iso1; reflexivity.
  destruct H as [x' [I' Peq]].
  assert (Fx := F _ I').

  assert (Fok : forall y, set_In y objs' -> s' |||- objIsFwded (fun _ : nat => None) phi s y ** top).
  intros y IY.
  assert (Y : s' |||- (objIsFwded (fun _ : nat => None) phi s y ** top) ** top).
  sconjMatch.
  hypReplace Hp (fasElim IY Hp).
  sconjMatch.
  linLogElim Y; assumption.
  clear Hp F.

  generalize dependent x.
  generalize dependent v.

  induction Fx.

  destruct iso as [_ [_ [Q _]]].
  intros; assert (Y := Q _ _ _ Pv Peq); subst v; constructor.

  rename y into y'.
  hypReplace IHFx (IHFx I').
  unfold POnat.POnat.A in * .
  intros v Peq y I Peq'.
  clear I'.
  assert (I' : set_In x objs').
  destruct iso as [_ [iso _]].
  eapply iso; eauto.
  assert (Fok' := Fok _ I').
  unfold objIsFwded in Fok'; unfold fieldIsFwded in Fok'; 
    unfold map_union in Fok'; linLogElim Fok'.
  assert (x' = x1).
  assert (Some x' = Some x1).
  rewrite <- H; eapply contGetEqL.
  eauto.
  injection H0; auto.

  subst x1.
  assert (x0 = v).
  destruct iso as [_ [_ [Q _]]]; eauto.
  subst x0.
  destruct H9 as [v0 [Hp0 M]].
  econstructor 2.
  instantiate (1 := v0).
  eapply contGetEqL; eauto.
  unfold mapX in M.
  clear H10 H4.
  assert (IO := isOddOk v0).
  destruct (isOdd v0).
  injection M; intro U; subst v0; clear M.
  elimtype False; eapply not_even_and_odd; eauto.
  clear IO; apply ptrPeven.
  inversion Fx; subst.
  assert (Y : set_In y' objs').
  destruct iso as [_ [iso _]].
  apply (iso y); rewrite Peq'; reflexivity.
  hypReplace Fok (Fok _ Y).
  unfold objIsFwded in Fok; unfold fieldIsFwded in Fok; 
    unfold map_union in Fok; linLogElim Fok.
  eapply contElim1L; eauto.
  unfold Heap.get in H0.
  assert (U := isPtrOk x'); destruct (isPtr x').
  assumption.
  discriminate H0.
  assert (ptrP (4 + x')).
  unfold Heap.get in H0.
  assert (U := isPtrOk (4 + x')); destruct (isPtr (4 + x')).
  assumption.
  discriminate H0.
  simpl in H2; inversion_clear H2; assumption.
  clear IO.
  apply IHFx; auto.

  (* mostly same as previous case *)
  rename y into y'.
  hypReplace IHFx (IHFx I').
  unfold POnat.POnat.A in * .
  intros v Peq y I Peq'.
  clear I'.
  assert (I' : set_In x objs').
  destruct iso as [_ [iso _]].
  eapply iso; eauto.
  assert (Fok' := Fok _ I').
  unfold objIsFwded in Fok'; unfold fieldIsFwded in Fok'; 
    unfold map_union in Fok'; linLogElim Fok'.
  assert (x' = x2).
  assert (Some x' = Some x2).
  sconjLift2 Fok'; rewrite <- H; eapply contGetEqL; eauto.
  injection H0; auto.

  subst x2.
  assert (x0 = v).
  destruct iso as [_ [_ [Q _]]]; eauto.
  subst x0.
  destruct H10 as [v0 [Hp0 M]].
  econstructor 3.
  instantiate (1 := v0).
  eapply contGetEqL; eauto.
  unfold mapX in M.
  clear H9 H4.
  assert (IO := isOddOk v0).
  destruct (isOdd v0).
  injection M; intro U; subst v0; clear M.
  elimtype False; eapply not_even_and_odd; eauto.
  clear IO; apply ptrPeven.
  inversion Fx; subst.
  assert (Y : set_In y' objs').
  destruct iso as [_ [iso _]].
  apply (iso y); rewrite Peq'; reflexivity.
  hypReplace Fok (Fok _ Y).
  unfold objIsFwded in Fok; unfold fieldIsFwded in Fok; 
    unfold map_union in Fok; linLogElim Fok.
  eapply contElim1L; eauto.
  unfold Heap.get in H0.
  assert (U := isPtrOk x'); destruct (isPtr x').
  assumption.
  discriminate H0.
  assert (ptrP (4 + x')).
  unfold Heap.get in H0.
  assert (U := isPtrOk (4 + x')); destruct (isPtr (4 + x')).
  assumption.
  discriminate H0.
  simpl in H2; inversion_clear H2; assumption.
  clear IO.
  apply IHFx; auto.
Qed.


Lemma rootReachForward : forall phi X X',
  iso_set phi X X' -> forall s s',
  s' |||- (foralls x' in X', objIsFwded (fun x:nat => None) phi s x') ** top -> forall v',
  (forall x', set_In x' X' -> rootReachable (heapOf s') (v'::nil) x') -> forall v,
  phi v = Some v' ->
  forall x, set_In x X ->
    rootReachable (heapOf s) (v::nil) x.
Proof.
  unfold rootReachable.
  intros.
  exists v; constructor.
  left; reflexivity.
  eapply reachForward; eauto.
  intros x' I'.
  destruct (H1 _ I'); intuition.
  simpl in H5; destruct H5.
  subst; assumption.
  elimtype False; assumption.
Qed.


Lemma objIsFwdShift : forall h S phi phi' s,
  h ||- (foralls x in S, objIsFwded phi phi' s x) ->
  (forall x y, phi' x = Some y -> phi x = None) ->
  h ||- foralls x in S, objIsFwded (fun _ : nat => None) (map_union phi phi') s x.
Proof.
  unfold objIsFwded; unfold fieldIsFwded.
  intros h S phi phi' s Hp Peq.
  apply (fasImplI Hp); clear h Hp.
  intros x h I Hp; lle Hp; hypReplace Peq (Peq _ _ H1); lli.
  3:eauto.
  instantiate (1 := x0); unfold map_union; rewrite Peq; assumption.
  assumption.
  assumption.
Qed.


Lemma objCopiedFwded : forall h S phi phi' s',
  h ||- (foralls x in S, objIsFwded phi phi' s' x) -> forall s S',
  s' |||- (foralls x in S', objCopied x s x) ** top -> 
    iso_set phi' S' S ->
  h ||- (foralls x in S, objIsFwded phi phi' s x).
Proof.
  unfold objIsFwded; unfold fieldIsFwded.
  intros h S phi phi' s' Hp s S' Hp' Iso.
  apply (fasImplI Hp); clear h Hp.
  intros x h I Hp; lle Hp.

  assert (Hp'' : s' |||- (objCopied x0 s x0 ** top) ** top).
  sconjMatch.
  assert (I0 : set_In x0 S').
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Q]; apply Q.
  rewrite H1; discriminate.
  hypReplace Hp0 (fasElim I0 Hp0).
  sconjMatch.

  clear Hp'; unfold objCopied in Hp''; unfold fieldCopied in Hp''; lle Hp''.
  destruct H6 as [v0 [Hp0 M0]]; destruct H7 as [v1 [Hp1 M1]].
  lli; eauto.
  exists x3; constructor.
  assumption.
  assert (X := contContEq Hp'' Hp0); subst x3; assumption.
  exists x4; constructor.
  assumption.
  sconjLift2 Hp''; assert (X := contContEq Hp'' Hp1); subst x4; assumption.
Qed.


Lemma eqvObjCopied : forall s h A,
  s |||- Heap.heap_eqv h ** A -> forall s' B,
  s' |||- Heap.heap_eqv h ** B -> forall S,
  h ||- (foralls x in S, x |->__) ->
  s' |||- (foralls x in S, objCopied x s x) ** B.
Proof.
  intros s h A Hp s' B Hp' objs Hp''.
  sconjMatch.
  eapply heapEqvTrue.
  apply Heap.heap_eqvSymm; eapply Heap.heap_eqvTrans.
  apply Hp0.
  apply Heap.heap_eqvSymm; apply Heap.HeapMap.map_canonEqv.
  clear Hp0 h0 s'.
  assert (S : Heap.subHeap h (heapOf s)).
  eapply heap_eqvSubHeapL; eauto.
  clear Hp.
  eapply fasImplIS.
  apply Hp''.
  clear Hp''.
  cbv beta; intros x h' I S' Hp'.
  assert (S'' := Heap.subHeapTrans S' S); clear S' S.
  lle Hp'; unfold objCopied; unfold fieldCopied; lli.
  2:eauto.

  assert (HS : s |||- (x |-> x0 ** top) ** top).
  eapply subHeapImpl; eauto; sconjMatch.
  lle HS; sconjMatch.

  assert (HS : s |||- (4 + x |-> x1 ** top) ** top).
  sconjLift2of2 Hp'; eapply subHeapImpl; eauto; sconjMatch.
  lle HS; sconjMatch.

Qed.
Implicit Arguments eqvObjCopied [s h A s' B].



Lemma objCopiedTrans2 : forall s' x s y,
  s' |||- objCopied x s y ** top -> forall h z,
  h ||- objCopied y s' z ->
  h ||- objCopied x s z.
Proof.
  unfold objCopied; unfold fieldCopied.
  intros s' x s y Hp h z Hp'; lle Hp; lle Hp'.
  assert (Q := contContEq H4 Hp); subst x2; clear H4.
  sconjLift2 Hp; assert (Q := contContEq Hp H5); subst x3; clear H5.
  lli; eauto.
Qed.


Lemma objHpUnforward : forall h s,
  Heap.subHeap h (heapOf s) -> forall objs objs',
  h ||- objHp objs objs' -> forall h'' objs'' phi,
  h'' ||- (foralls x in objs'', objIsFwded (fun x:nat=>None) phi s x) ->
  iso_set phi objs' objs'' ->
  h ||- objHp objs' objs'.
Proof.
  unfold objHp.
  intros h s S0 objs objs' Hp h'' objs'' phi Hp'' Iso.

  eapply fasImplIS.
  eauto.

  clear Hp.
  unfold okObjAt; unfold okFieldAt; cbv beta; intros x h' I S' Hp; lle Hp.
  assert (S := Heap.subHeapTrans S' S0); clear S' S0.

  assert (X : exists y, phi x = Some y /\ set_In y objs'').
  destruct Iso as [X [Y _]].
  destruct (X x) as [Z _].
  hypReplace Y (Y x).
  destruct (phi x).
  exists n; constructor; auto.
  elimtype False; apply Z; auto.

  destruct X as [y [Peq Iy]].

  assert (Hp' : h'' ||- objIsFwded (fun _ : nat => None) phi s y ** top).
  hypReplace Hp'' (fasElim Iy Hp''); sconjMatch.

  clear Hp''; unfold objIsFwded in Hp'; unfold fieldIsFwded in Hp'; 
    unfold map_union in Hp'; lle Hp'.
  destruct H10 as [v0 [Hp0 Meq0]].
  destruct H11 as [v1 [Hp1 Meq1]].

  assert (x = x2).
  destruct Iso as [_ [_ [Iso _]]].
  eapply Iso; eauto.
  subst x2.

  assert (X := subHeapContEq S Hp0 Hp).
  sconjLift2of2 Hp; assert (Y := subHeapContEq S Hp1 Hp).
  subst x0 x1.
  sconjLift2of2 Hp.
  clear s S Hp0 Hp1 H5.
  lli.
  2:sconjMatch.

  unfold mapX in Meq0.
  assert (X := isOddOk v0); destruct (isOdd v0).
  left; assumption.
  destruct H3.
  elimtype False; eapply not_even_and_odd; eauto.
  right.
  destruct Iso as [Iso _]; destruct (Iso v0) as [_ Q]; apply Q;
    rewrite Meq0; discriminate.

  unfold mapX in Meq1.
  assert (X := isOddOk v1); destruct (isOdd v1).
  left; assumption.
  destruct H4.
  elimtype False; eapply not_even_and_odd; eauto.
  right.
  destruct Iso as [Iso _]; destruct (Iso v1) as [_ Q]; apply Q;
    rewrite Meq1; discriminate.

Qed.


Lemma rootReachableSubheap : forall h objs,
  h ||- objHp objs objs -> forall h',
  Heap.subHeap h h' -> forall v,
  okFieldVal objs v -> forall x,
  rootReachable h' (v::nil) x ->
  rootReachable h (v::nil) x.
Proof.
  intros h objs Hp h' S v Vok x [b [I R]].
  simpl in I; destruct I.
  2:elimtype False; assumption.
  subst b.
  exists v; constructor.
  left; reflexivity.

  generalize dependent v.
  induction 2.
  constructor.

  destruct Vok as [Odd | I].

  assert (Q := isPtrOk x); unfold Heap.get in H; destruct (isPtr x).
  elimtype False; eapply not_even_and_odd.
  apply ptrPeven; eauto.
  assumption.
  discriminate H.

  hypReplace Hp (objHpElim I Hp).
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp.
  rewrite (S x x0 (contGetEqL Hp)) in H; injection H; clear H; intro; subst x0.
  econstructor 2.
  instantiate (1 := x').
  eapply contGetEqL; eauto.
  apply IHR.
  assumption.

  destruct Vok as [Odd | I].

  assert (Q := isPtrOk (4 + x)); unfold Heap.get in H; destruct (isPtr (4 + x)).
  elimtype False; eapply not_even_and_odd.
  apply ptrPeven; eauto.
  simpl; constructor; constructor; constructor; constructor; assumption.
  discriminate H.

  hypReplace Hp (objHpElim I Hp).
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp.
  sconjLift2 Hp.
  rewrite (S _ _ (contGetEqL Hp)) in H; injection H; clear H; intro; subst x1.
  econstructor 3.
  instantiate (1 := x').
  eapply contGetEqL; eauto.
  apply IHR.
  assumption.

Qed.


Lemma subHeapRootReachable : forall h h',
    Heap.subHeap h h' -> forall x vv,
    rootReachable h vv x ->
    rootReachable h' vv x.
Proof.
  intros h h' S x vv [y [I R]].
  exists y; constructor.
  assumption.
  eapply subHeapReachable; eauto.
Qed.


Lemma heap_eqvReachable : forall h h',
    Heap.heap_eqv h h' -> forall x y,
    reachable h x y ->
    reachable h' x y.
Proof.
  intros; eapply subHeapReachable; eauto.
  intros a b E.
  unfold Heap.get.
  rewrite <- H.
  assumption.
Qed.


Lemma minObjHpEqv' : forall h vv objs,
  h ||- minObjHp vv objs ** top -> forall objs',
  h ||- minObjHp vv objs' ** top ->
  set_subset objs objs'.
Proof.
  unfold minObjHp.
  intros h vv objs Hp objs' Hp' x I.
  destruct (aconjElim (sconjDistrib2 Hp)) as [Hp1 Hp2]; clear Hp.
  assert (h ||- (fun h => forall x, set_In x objs -> rootReachable h vv x)).
  eapply intuitImpl; eauto.

  unfold hpredTrue; intros h1 h2 S Hp y Iy.
  eapply subHeapRootReachable; eauto.
  intros a b E.
  assert (F := S a b).
  rewrite <- Heap.heap_canonEqv.
  apply F; rewrite Heap.heap_canonEqv; assumption.

  clear Hp1.
  hypReplace H (H _ I).
  destruct (aconjElim (sconjDistrib2 Hp')) as [_ Hp]; clear Hp'.
  destruct H as [y [Iy R]].
  unfold okObjHp in Hp; lle Hp.
  assert (okFieldVal objs' x).
  eapply objHpReachOkFieldL; eauto.
  eapply heap_eqvReachable.
  apply Heap.heap_eqvSymm; apply Heap.HeapMap.map_canonEqv.
  assumption.
  destruct H.
  2:assumption.
  elimtype False; eapply not_even_and_odd.
  2:apply H.
  apply ptrPeven.
  unfold okObjHp in Hp2; lle Hp2.
  assert (Hp0 : h ||- ((x |->_ ** top) ** top) ** top).
  sconjMatch.
  hypReplace Hp0 (objHpElim I Hp0); sconjMatch.
  unfold okObjAt in Hp0; sconjMatch.
  unfold okFieldAt in Hp0; lle Hp0; lli; eauto.

  lle Hp0.
  eapply contElim1L; eauto.

Qed.


Lemma minObjHpEqv : forall h vv objs,
  h ||- minObjHp vv objs ** top -> forall vv' objs',
  h ||- minObjHp vv' objs' ** top ->
  set_eqv vv vv' ->
  set_eqv objs objs'.
Proof.
  intros h vv objs Hp vv' objs' Hp' E.

  assert (Hp'' : h ||- minObjHp vv objs' ** top).
  sconjMatch.
  unfold hpredTrue; rewrite E; assumption.

  clear Hp'; rename Hp'' into Hp'.
  constructor; intro I.
  assert (X := minObjHpEqv' Hp Hp'); eauto.
  assert (X := minObjHpEqv' Hp' Hp); eauto.

Qed.



  Lemma objIsFwdedNarrow : forall h objs,
    h ||- objHp objs objs ** top -> forall phi0 phi,
    mapSub phi phi0 -> forall objs',
    iso_set phi objs objs' -> forall ha,
    Heap.subHeap ha h ->
    ha ||- objHp objs objs -> forall x,
    set_In x objs' -> forall h',
    h' ||- objIsFwdedH phi0 h x ->
    (forall x x' y, phi0 x = Some y -> phi0 x' = Some y -> x = x') ->
    h' ||- objIsFwdedH phi ha x.
  Proof.
    unfold objIsFwdedH.
    intros h objs Hp phi phi' Psub objs' Iso ha Hsub HpA y I h' Hp' Iso0.
    lle Hp'.
    rename x into v0'; rename x0 into v1'.
    destruct H1 as [x [v0 [v1 [Peq [Hp0 [Meq0 Meq1]]]]]].
    lli.
    2:eauto.
    exists x; exists v0; exists v1.

    assert (Peq' : phi' x = Some y).
    assert (Psub1 := Psub x).
    destruct Iso as [_ [_ [_ Iso]]]; destruct (Iso _ I) as [x' E]; clear Iso.
    assert (Psub2 := Psub x' y E).
    hypReplace Iso0 (Iso0 _ _ _ Peq Psub2).
    subst x'; clear Psub2.
    assumption.

    constructor.
    assumption.

    assert (I' : set_In x objs).
    destruct Iso as [Iso _]; destruct (Iso x) as [_ Q]; apply Q.
    rewrite Peq'; discriminate.
    hypReplace HpA (objHpElim I' HpA); unfold okObjAt in HpA; unfold okFieldAt in HpA; lle HpA.
    lle Hp0.
    assert (Q1 := subHeapContEq Hsub Hp0 HpA).
    sconjLift2 Hp0; sconjLift2 HpA.
    assert (Q2 := subHeapContEq Hsub Hp0 HpA).
    subst x0 x1.
    sconjLift2 HpA.

    constructor.
    lli; sconjMatch.

    unfold mapX in * .
    clear Hp0 HpA x Peq Peq' I' Hp'.
    constructor.
    assert (IOO := isOddOk v0).
    destruct (isOdd v0).
    assumption.
    destruct H3.
    elimtype False; apply (not_even_and_odd v0); assumption.
    clear IOO H6 Meq1.
    destruct Iso as [Iso _]; destruct (Iso v0) as [X _].
    hypReplace X (X H); clear H.
    assert (Y := Psub v0).
    rewrite Meq0 in Y; clear Meq0 Iso.
    destruct (phi' v0).
    symmetry; apply Y; reflexivity.
    elimtype False; apply X; reflexivity.

    assert (IOO := isOddOk v1).
    destruct (isOdd v1).
    assumption.
    destruct H6.
    elimtype False; apply (not_even_and_odd v1); assumption.
    clear IOO H3 Meq0 v0.
    destruct Iso as [Iso _]; destruct (Iso v1) as [X _].
    hypReplace X (X H); clear H.
    assert (Y := Psub v1).
    rewrite Meq1 in Y; clear Meq1 Iso.
    destruct (phi' v1).
    symmetry; apply Y; reflexivity.
    elimtype False; apply X; reflexivity.
  Qed.


Lemma rrlistMapIntro : forall (r:rnum) rr,
  In r rr -> forall f,
  set_In (f r) (List.map f rr).
Proof.
  intros r rr I f.
  induction rr.
  exact I.
  simpl; simpl in I; destruct I as [I | I].
  subst a.
  auto.
  right; auto.
Qed.


Ltac llMatch :=
  eauto;
  match goal with
    | Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?B =>
      exact Hp
    | Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?C =>
      apply (sconjImplR Hp); try (clear Hp); try (clear h);
	let h' := fresh "h" in
	let Hp' := fresh "Hp" in
	  (intros h' Hp'; llMatch)
    | Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?B =>
      apply (sconjImplL Hp); try (clear Hp); try (clear h); 
	let h' := fresh "h" in
	let Hp' := fresh "Hp" in
	  (intros h' Hp'; llMatch)
    | Hp : ?h ||- sconj _ _ |- ?h ||- sconj _ _ =>
      apply (sconjImpl Hp); try (clear Hp); try (clear h); 
	let h' := fresh "h" in
	let Hp' := fresh "Hp" in
	  (intros h' Hp'; llMatch)
    | |- _ ||- top => apply topIntro

    | Hp : ?h ||- fas ?A ?X |- ?h ||- fas ?B ?X =>
      apply (fasImplI Hp); try (clear Hp); try (clear h);
        let x := fresh "x" in
        let h' := fresh "h" in
        let I := fresh "I" in
        let Hp' := fresh "Hp" in
           (intros x h' I Hp')
    | |- _ => idtac
  end.


Ltac llm := llMatch.


