Require Import WeakSCAP.
Require Import LinLog.
Require Import Machine.
Require Import List.
Require Import SpecUtil.
Require Import GC.
Require Import NatUtil.
Require Import StateTheory.
Require Import Arith.
Require Import MarkSweepGC.
Require Import MarkSweepLemma.
Require Import BuffArrayStack.
Require Import StackSet.
Require Import IMarkSweepGC.

Theorem imarkImplOK :
  iseqOkS imarkP imarkPre imarkGuar imarkImpl.
Proof.
  unfold imarkImpl.
  iseqWeakSolver PsimplIMark.

  intros s [B [G [W [free [objs [ohpSt [ohpEnd [freePtr [stBot [stBuff [stTop 
    [numfree [Hroot [Hnumfree [Hphase [Hst [Hed [Hobjs [Hset_eq  Hp]]]]]]]]]]]]]]]]]]].

  rewrite addiuStep; stSimpl.
  assert (Hp'  := Hp).
  sconjLift5 Hp'.
  unfold contL in Hp'.

  hypReplace Hp' (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL 
    (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL Hp')))))))))).  
  rewrite (lwStep Hp'); trivial.
  hypReplace Hp' (sconjAssocL (sconjAssocL (sconjComm Hp'))).
  rewrite (lwStep Hp'); trivial.
  hypReplace Hp' (sconjAssocL (sconjAssocL (sconjComm Hp'))).
  rewrite (lwStep Hp'); trivial.
  rewrite addiuStep; stSimpl.

  split.
  exists B; exists G; exists W; exists free; exists objs;
    exists ohpSt; exists ohpEnd; exists freePtr; exists stTop;
      exists numfree.
  repeat (split; trivial).
  stSimpl; auto with arith.
  
  assert (HstTop := contGetEqL Hp').
  hypReplace Hp' (sconjComm (sconjAssocR (sconjAssocR Hp'))).
  assert (HstBuff := contGetEqL Hp').
  hypReplace Hp' (sconjComm (sconjAssocR (sconjAssocR Hp'))).
  assert (HstBot := contGetEqL Hp').

  clear Hp' Hp Hset_eq Hobjs Hed Hst Hphase Hroot numfree free Hnumfree
    freePtr ohpEnd ohpSt objs W G B.
  simpl.
  unfold imarkLoopGuar; unfold imarkLoopInv; stSimpl.
  intros s' [HG [Hcall [Hcst1 [Hcst0 Hcst2]]]].
  split.
  intros B G W free objs ohpSt ohpEnd freePtr stBot0 stBuff0 stTop0
    numfree A [Hroot [Hnumfree [Hphase [Hst [Hed [Hobjs [Hset_eq Hp]]]]]]].
  assert (Hp' := Hp).
  sconjLift5 Hp'.
  unfold contL in Hp'.
  
  hypReplace Hp' (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL 
    (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL Hp')))))))))).  

  assert (HstBot' := contGetEqL Hp').
  rewrite HstBot' in HstBot; inversion HstBot; subst stBot0.
  clear HstBot HstBot'.
  hypReplace Hp' (sconjAssocL (sconjAssocL (sconjComm Hp'))).
  assert (HstBuff' := contGetEqL Hp').
  rewrite HstBuff' in HstBuff; inversion HstBuff; subst stBuff0.
  clear HstBuff' HstBuff.
  hypReplace Hp' (sconjAssocL (sconjAssocL (sconjComm Hp'))).
  assert (HstTop' := contGetEqL Hp').
  rewrite HstTop' in HstTop; inversion HstTop; subst stTop0.
  clear HstTop' HstTop Hp'.
  exists (rget s' rsweepsum).
  assert (exists stTop' : nat, exists B' : set, exists G' : set,
    exists W' : set,
      imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr stBot
      stBuff stTop' numfree A /\
      set_card (set_diff B' B) + 0= rget s' rsweepsum /\
      set_subset B B').

  apply HG with G W stTop; clear HG.
  repeat (split; trivial).
  auto with arith.
  destruct H as [stTop' [B0 [G0 [W0 [Hpost [Hsum HBB]]]]]].
  exists stTop'; exists B0; exists G0; exists W0.
  split; trivial.
  split.
  rewrite <- Hsum; auto.
  split; trivial.
  destruct Hpost as [_ [_ [He _]]].
  assumption.
  split; trivial.
  clear HG.
  apply raAndOthers.
  eapply calleeSavedOkTrans'.
  2: apply Hcall.
  calleeSavedOkSolver'.
  assumption.
  split; assumption.
Qed.

Theorem imarkLoopImplOK :
  iseqOkS imarkP imarkLoopPre imarkLoopGuar imarkLoopImpl.
Proof.
  unfold imarkLoopImpl.
  iseqWeakSolver PsimplIMark.

  unfold isEmptyGuar.
  intros s s' _ [_ [_ Hr]].
  regEqSolver.

  simpl.
  intros s [B [G [W [free [objs [ohpSt [ohpEnd [freePtr [stTop [numfree 
    [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs [Hseq Hp]]]]]]]]]]]]]]]]]]; stSimpl.
  split.
  assert (Hp' := Hp).
  sconjLift4 Hp'.
  destruct emptyStackIsEmptySet as [IE1 IE2].
  split.
  apply IE1; clear IE1 IE2.
  exists G.
  apply (sconjImplR Hp').
  intros; linLogIntro.

  intros s' EG.
  destruct (IE2 _ _ EG) as [HGemp [Hheqss' Hr]]; clear IE1 IE2 EG.
  exists B; exists G; exists W; exists free; exists objs;
    exists ohpSt; exists ohpEnd; exists freePtr;
      exists stTop; exists numfree.
  split.
  eapply HGemp.
  apply Hp'.

  assert (forall r, beq_rnum r v0 = false -> 
    rget s' r = rget (rset s ra 502) r).
  intros;rewrite (Hr r); auto.
  intro R; subst; rewrite beq_rnum_refl in H.
  discriminate H.

  clear Hr; rename H into Hr.
  unfold imarkLoopInv.
  unfold setHp.
  regRewrite1 rtop.
  regRewrite1 rbot.
  regRewrite1 rbuff.
  regRewrite1 rgcInfo.
  regRewrite1 rphase.
  regRewrite1 rsweepsum.
  regRewrite1 rroot.
  stSimpl.
  repeat (split;trivial).
  rewrite <- Hheqss'; trivial.

  clear Hp Hseq Hroot Hobjs Hed Hst Hphase Hnumfree Hmksum stTop
    numfree freePtr ohpEnd ohpSt free objs W G B.

  intros s'' [s' [HempAct [HLBodyAct [HcalleeSav [Hcst1 [Hcst0 Hcst2]]]]]].
  destruct emptyStackIsEmptySet as [IE1 IE2].
  hypReplace HempAct (IE2 _ _ HempAct); clear IE1 IE2.
  split.
  intros B G W free objs ohpSt ohpEnd freePtr stTop numfree A
    [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs [Hset_eq Hp]]]]]]]].
  destruct HempAct as [Hv0 [Hheqss' Hr]].

  assert (forall r, beq_rnum r v0 = false -> 
    rget s' r = rget (rset s ra 502) r).
  intros;rewrite (Hr r);auto.
  intro R; subst; rewrite beq_rnum_refl in H; discriminate H.

  clear Hr; rename H into Hr.
  generalize HLBodyAct; clear HLBodyAct.
  unfold imarkLoopInv; unfold setHp.
  regRewrite1 rtop.
  regRewrite1 rbot.
  regRewrite1 rbuff.
  regRewrite1 rgcInfo.
  regRewrite1 rphase.
  regRewrite1 rsweepsum.
  regRewrite1 rroot.
  stSimpl.
  intro HLBodyAct.
  apply HLBodyAct with G W stTop; clear HLBodyAct.
  split.
  eapply Hv0.
  assert (Hp' := Hp).
  sconjLift4 Hp'.
  apply Hp'.
  repeat (split;trivial).
  rewrite <- Hheqss'; trivial.

  clear HLBodyAct.
  destruct HempAct as [_ [_ Hr]].
  assert (forall r, beq_rnum r v0 = false -> 
    rget s' r = rget (rset s ra 502) r).
  intros;rewrite (Hr r);auto.
  intro R; subst; rewrite beq_rnum_refl in H; discriminate H.

  clear Hr; rename H into Hr.
  split.
  apply calleeSavedOkTrans' with s'; trivial.
  intros r HcalleeSav'.
  destruct HcalleeSav' as [H | [H | [H | [H | [
    H | [H | [H | [H | H]]]]]]]];
  try (subst r;regEqSolver).
  elim H.
  split.
  rewrite <- Hcst1;regEqSolver. 
  split.
  rewrite <- Hcst0;regEqSolver.
  rewrite <- Hcst2;regEqSolver.
Qed.

Theorem imarkLoopBodyImplOK :
  iseqOkS imarkP imarkLoopBodyPre imarkLoopBodyGuar imarkLoopBodyImpl.
Proof.
  unfold imarkLoopBodyImpl.
  iseqWeakSolver PsimplIMark.
  
  unfold popGuar.
  intros s s' _ [_  Hr].
  regEqSolver.
  
  intros s [B [G [W [free [objs [ohpSt [ohpEnd [freePtr [stTop [numfree
    [Hv0 [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs [Hset_eq Hp]]]]]]]]]]]]]]]]]]].

  assert (BGdisj : set_disj B G).
  eapply objHdrsDisj; eauto.

  assert (WGdisj : set_disj W G).
  sconjLift2 Hp; sconjLift3 Hp.
  apply set_unionDisjElimL with free.
  eapply objHdrsDisj; eauto.

  unfold doBtest; stSimpl.
  compare (rget s v0) 0; intro Hev0; 
    [rewrite Hev0; rewrite beq_true | rewrite beq_false]; unfold negb; trivial.
  replace (Bool.negb true) with false.
  2:reflexivity.
  2:replace (Bool.negb false) with true.
  3:reflexivity.

  rewrite addiuStep.
  rewrite subuStep; stSimpl.

  compare (5 + 0 - rget s rmarksum) 0; intro Hemksum;
    [rewrite Hemksum; rewrite beq_true | rewrite beq_false]; trivial.
  split.
  (* DoneSome Mark *)
  unfold imarkDoneSomeMarkPre; unfold imarkDoneSomeMarkPreP.
  stSimpl.
  exists B; exists G; exists W; exists free; exists objs;
    exists ohpSt; exists ohpEnd; exists freePtr; exists stTop;
      exists numfree.
  repeat (split; trivial).
  omega.
  rewrite Hev0 in Hv0; unfold sbool in Hv0; trivial.

  clear Hp Hset_eq Hobjs Hed Hst Hphase Hv0 Hmksum Hroot stTop freePtr
    Hnumfree ohpEnd numfree ohpSt objs free W G B BGdisj WGdisj.

  unfold imarkDoneSomeMarkGuar; unfold imarkLoopBodyGuar;
    unfold imarkDoneSomeMarkPreP; unfold imarkLoopInv.
  stSimpl.
  intros s' [HGDone [HcalleeSav [Hcst1 [Hcst0 [Hcst2 Hss'sum]]]]].
  split.
  intros B G W free objs ohpSt ohpEnd freePtr stTop numfree A [Hsb
    [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs [Hset_eq Hp]]]]]]]]].

  assert (exists stTop', exists B', exists G', exists W',
    imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot)
    (rget s rbuff) stTop' numfree A /\
    set_card (set_diff B' B) = 0 /\ set_subset B B').
  eapply HGDone with G W stTop; clear HGDone.
  repeat (split; trivial).
  omega.
  rewrite Hev0 in Hsb; unfold sbool in Hsb; trivial.

  clear HGDone Hp.
  destruct H as [stTop' [B' [G' [W' [Hpost [Hss Hbb]]]]]].
  exists stTop'; exists B'; exists G'; exists W'.
  split; trivial.
  split; trivial.
  rewrite Hss; auto.
  clear HGDone.
  split.
  eapply calleeSavedOkTrans'.
  2: apply HcalleeSav.
  calleeSavedOkSolver'.
  split; trivial.
  split; trivial.
  (* Mark First *)
  split.
  unfold imarkFirstPre; unfold imarkFirstPreP.
  destruct popStackIsPopSet as [PSp PSg].
  split.
  apply PSp; clear PSp PSg.
  unfold stackPopPre; unfold setHp; stSimpl.

  assert (Hp' := Hp).
  sconjLift4 Hp'.
  exists G.
  split.
  rewrite Hev0 in Hv0; unfold sbool in Hv0; trivial.
  apply (sconjImplR Hp').
  intros; linLogIntro.

  intros s' PG.
  destruct (PSg _ _ PG) as [HGpop Hr]; clear PSp PSg PG.
  rewrite Hev0 in Hv0; unfold sbool in Hv0.
  assert (forall r, beq_rnum r v0 = false ->
    beq_rnum r rtop = false ->
    rget s' r = rget (rset (rset (rset s rtemp (5 + 0)) rtemp (5 + 0 - rget s rsweepsum))
      ra 503) r).    
  intros;rewrite (Hr r);auto.
  intro R; subst r; rewrite beq_rnum_refl in H; discriminate H.
  intro R; subst r; rewrite beq_rnum_refl in H0; discriminate H0.

  clear Hr; rename H into Hr.

  assert (set_In (rget s' v0) G).
  destruct (HGpop _ _ (sconjAssocL (sconjComm 
    (sconjAssocR (sconjAssocR Hp))))); trivial.

  exists B; exists (set_remove (rget s' v0) G); exists W; exists free; 
    exists objs; exists ohpSt; exists ohpEnd; exists freePtr;
      exists stTop; exists numfree.
  generalize HGpop; clear HGpop.
  regRewrite2 rbot; regRewrite2 rbuff; regRewrite2 rphase; regRewrite2 rgcInfo;
  regRewrite2 rsweepsum; regRewrite2 rroot.
  unfold setHp; stSimpl.

  intro HGpop.
  split.
  setoid_replace (set_add (rget s' v0) B |_| set_remove (rget s' v0) G) with
    (G |_| B).
  setoid_rewrite (unionRemoveAddEqv (rget s' v0) G B H).
  apply set_eqvUnionSymm.
  setoid_rewrite (set_eqvUnionSymm G B).
  assumption.
  split; trivial.
  split; trivial.
  split.
  omega.
  split; trivial.
  split; trivial.
  split; trivial.

  assert (Hset_eq' : set_eqv (set_add (rget s' v0)
    ((B |_| set_remove (rget s' v0) G) |_| W) |_| free) objs).
  setoid_rewrite <- Hset_eq.
  split; intro; elimSetIn.
  assert (set_In a (set_remove (rget s' v0) G)).
  apply set_remove_intro; trivial.
  apply set_union_intro1.
  elimSetIn.
  assert (Hsetin' : set_In (rget s' v0) objs).
  elimSetIn.
  split; trivial.

  unfold setHp in Hp.
  sconjLift4 Hp.
  destruct (HGpop _ _ Hp Hv0) as [_ Hp']; clear HGpop Hp.

  sconjLift3 Hp'.
  apply sconjGLift3'.
  apply sconjAssocL.
  apply (sconjImpl Hp'); clear Hp'; intros h Hp.

  apply sconjComm; apply objHdrsElim; assumption.

  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.

  destruct Hp as [_ Hp].
  apply aconjIntro.
  linLogIntro.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  assert (Htmp : set_eqv (B |_| G)
    (set_add (rget s' v0) (B |_| set_remove (rget s' v0) G))).
  split; intro; elimSetIn.
  rewrite hpredTrueEq.
  setoid_rewrite <- Htmp.
  apply Hp.

  apply objHpSplit.
  rewrite  hpredTrueEq; setoid_replace
    ((set_remove (rget s' v0) G |_| W) |_| rget s' v0 :: empty_set)
    with (set_union G W).
  constructor; intro I; elimSetIn.
  setoid_replace (set_add (rget s' v0) ((B |_| set_remove (rget s' v0) G) |_| W))
    with ((B |_| G) |_| W).
  split; intro; elimSetIn.
  apply Hp.

  apply set_disj_intro; intros a I1 I2.
  simpl in I2; destruct I2.
  subst a; destruct (set_union_elim I1).
  destruct (set_remove_elim _ _ _ H0).
  apply H1; reflexivity.
  apply (set_disj_elim WGdisj (rget s' v0)); assumption.
  assumption.

  clear Hp Hset_eq Hobjs Hed Hst Hphase Hv0 Hmksum stTop Hroot Hnumfree
    free freePtr numfree ohpEnd ohpSt objs W G B BGdisj WGdisj.
  
  unfold imarkFirstGuar; unfold imarkFirstPreP; unfold imarkLoopBodyGuar;
    unfold imarkLoopInv; unfold stackHp.
  stSimpl.
  intros s'' [s' [HGpop0 HGMF]].
  destruct popStackIsPopSet as [_ PG].
  destruct (PG _ _ HGpop0) as [HGpop Hr]; clear PG HGpop0.
  destruct HGMF as [HGMF [HcalleeSav [Hcst1 [Hcst0 Hcst2]]]].
  split.
  intros.
  destruct H as [Hsb [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs
    [Hset_eq Hp]]]]]]]]].
  assert (forall r, beq_rnum r v0 = false ->
    beq_rnum r rtop = false ->
    rget s' r = rget (rset (rset (rset s rtemp (5 + 0)) rtemp (5 + 0 - rget s rsweepsum))
      ra 503) r).    
  intros; rewrite (Hr r); trivial.
  apply beq_rnum_false_neq; trivial.
  apply beq_rnum_false_neq; trivial.

  clear Hr; rename H into Hr.
  rewrite Hev0 in Hsb; unfold sbool in Hsb.
  assert (Hin : set_In (rget s' v0) G).
  destruct (HGpop _ _ (sconjAssocL (sconjComm 
    (sconjAssocR (sconjAssocR Hp))))); trivial.

  generalize HGMF; generalize HGpop; clear HGMF; clear HGpop.
  regRewrite2 rbot; regRewrite2 rbuff; regRewrite2 rphase; regRewrite2 rgcInfo; 
  regRewrite2 rsweepsum; regRewrite2 rroot; stSimpl.

  intros HGpop HGMF.
  assert (BGdisj : set_disj B G).
  eapply objHdrsDisj; eauto.

  assert (WGdisj : set_disj W G).
  sconjLift2 Hp; sconjLift3 Hp.
  apply set_unionDisjElimL with free.
  eapply objHdrsDisj; eauto.

  assert (Hset_eq' : set_eqv (set_add (rget s' v0)
    ((B |_| set_remove (rget s' v0) G) |_| W) |_| free) objs).
  setoid_rewrite <- Hset_eq.
  split; intro; elimSetIn.
  assert (set_In a (set_remove (rget s' v0) G)).
  apply set_remove_intro; trivial.
  apply set_union_intro1.
  elimSetIn.
  assert (Hsetin' : set_In (rget s' v0) objs).
  elimSetIn.
  
  assert (Hp' : s' |||- rget s' v0 - 4 |-> 1 ** objHdrs B 1 **
    objHdrs (set_remove (rget s' v0) G) 1 **
    objHdrs (W |_| free) 0 ** setHp (set_remove (rget s' v0) G) s' **
    rget s rgcInfo
    |->, (ohpSt :: ohpEnd :: freePtr :: rget s rbot :: rget s rbuff :: stTop ::
      ohpSt :: numfree :: nil) ** ifreeList freePtr free **
    A &&  objHp (set_add (rget s' v0) (B |_| set_remove (rget s' v0) G)) B **
    objHp (set_add (rget s' v0) ((B |_| set_remove (rget s' v0) G) |_| W))
    (set_remove (rget s' v0) G |_| W) **
    objHp (set_add (rget s' v0) ((B |_| set_remove (rget s' v0) G) |_| W))
    (rget s' v0 :: empty_set)).

  destruct (HGpop G _ (sconjAssocL (sconjComm 
    (sconjAssocR (sconjAssocR Hp)))) Hsb) as 
  [_ Hp'];clear HGpop.
  apply sconjAssocL;apply sconjAssocL;
    apply sconjAssocL;apply sconjComm.
  apply sconjAssocR.
  apply (sconjImplR Hp'); clear Hp';intros h Hp'.
  apply (sconjImpl Hp'); clear Hp' h;intros h Hp'.
  apply (sconjImplR Hp'); clear h Hp';intros h Hp'.
  apply (sconjImplR Hp'); clear h Hp';intros h Hp'.
  apply aconjIntro.
  hypReplace Hp (aconjEL Hp'); trivial.
  hypReplace Hp (aconjER Hp').
  apply (sconjImpl Hp); clear Hp Hp' h; intros h Hp.
  assert (Htmp : set_eqv (set_add (rget s' v0) (B |_| set_remove (rget s' v0) G))
    (B |_| G)).
  split; intro; elimSetIn.
  rewrite hpredTrueEq; setoid_rewrite Htmp.
  apply Hp.
  apply objHpSplit.
  rewrite hpredTrueEq.
  setoid_replace (set_add (rget s' v0) ((B |_| set_remove (rget s' v0) G) |_| W))
    with ((B |_| G) |_| W).
  split; intro; elimSetIn.
  setoid_replace ((set_remove (rget s' v0) G |_| W) |_| rget s' v0 :: empty_set)
    with (G |_| W).
  split; intro; elimSetIn.
  apply Hp.

  apply set_disj_intro.
  intros x Hs1 Hs2.
  destruct Hs2; [subst x | trivial].
  apply (set_disj_elim BGdisj (rget s' v0)); trivial.
  destruct (set_union_elim Hs1) as [He1 | He2].
  assert (rget s' v0 <> rget s' v0).
  apply set_remove_elim2 with G; trivial.
  elimtype False; apply H; trivial.
  elimtype False; apply (set_disj_elim WGdisj (rget s' v0)); trivial.
  apply (sconjImplL Hp').
  clear h Hp';intros h Hp'.
  apply sconjComm; apply sconjAssocL.
  apply (sconjImplL (sconjComm Hp')). 
  clear h Hp'; intros h Hp'.
  apply sconjComm.
  apply objHdrsElim; trivial.

  assert (exists stTop' : nat, exists B' : set, exists G' : set,
    exists W' : set,
      imarkPost s'' B' G' W' free objs ohpSt ohpEnd freePtr
      (rget s rbot) (rget s rbuff) stTop' numfree A /\
      set_card (set_diff B' (rget s' v0 :: B)) + 1 +  rget s rsweepsum  =
      rget s'' rsweepsum /\
      set_subset (rget s' v0 :: B) B').
  apply HGMF with (set_remove (rget s' v0) G) W stTop; clear HGMF.
  repeat (split; trivial).
  setoid_replace (set_add (rget s' v0) B |_| set_remove (rget s' v0) G) with
    (G |_| B).
  setoid_rewrite (unionRemoveAddEqv (rget s' v0) G B Hin).
  apply set_eqvUnionSymm.
  setoid_rewrite (set_eqvUnionSymm G B).
  assumption.
  omega.

  destruct H as [stTop' [B0 [G0 [W0 [Hpost [Hinc HBp]]]]]].
  exists stTop'; exists B0; exists G0; exists W0.
  split; trivial.
  split.
  rewrite <- Hinc.
  assert (set_card (set_diff B0 B) = set_card (set_diff B0 (rget s' v0 :: B)) + 1).
  apply notInPlus; auto.
  apply HBp.
  left; reflexivity.
  clear HGMF HGpop Hp Hpost Hnumfree Hr HBp Hinc stTop' B0 G0 W0.
  
  intro Hin'.
  sconjLift2 Hp'.
  hypReplace Hp' (sconjAssocL (fasElimL Hin' Hp')).
  sconjLift3 Hp'.
  assert (rget s' v0 - 4 <> rget s' v0 -4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp'.
  apply H; reflexivity.
  rewrite H; auto; clear H.
  intros v HvB.
  elimSetIn.

  assert (forall r, beq_rnum r v0 = false ->
    beq_rnum r rtop = false ->
    rget s' r = rget (rset (rset (rset s rtemp (5 + 0)) rtemp (5 + 0 - rget s rsweepsum))
            ra 503) r).    
  intros;rewrite (Hr r);auto.
  apply beq_rnum_false_neq; trivial.
  apply beq_rnum_false_neq; trivial.
  clear Hr; rename H into Hr.
  
  clear HGMF HGpop.
  split.
  apply calleeSavedOkTrans' with s'; trivial.
  intros r HcalleeSav'.
  destruct HcalleeSav' as [H | [H | [H | [H | [
    H | [H | [H | [H | H]]]]]]]];
  try (subst r;regEqSolver).
  elim H.
  split.
  rewrite <- Hcst1;regEqSolver. 
  split.
  rewrite <- Hcst0;regEqSolver.
  rewrite <- Hcst2;regEqSolver.
  
  unfold imarkMarkFinishedPre; unfold imarkMarkFinishedGuar;
    unfold imarkMarkFinishedPreP; unfold imarkLoopBodyGuar;
      unfold imarkLoopInv.

  split.
  exists B; exists G; exists W; exists free; exists objs; exists ohpSt;
    exists ohpEnd; exists freePtr; exists stTop; exists numfree.
  destruct (rget s v0).
  elim Hev0; reflexivity.
  unfold sbool in Hv0.
  subst G.
  unfold setHp in Hp.
  assert (rget s rbot = rget s rtop).
  eapply empty_bfstk_eqv_b_t.
  apply (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL 
	(sconjAssocL (sconjComm (sconjAssocL (sconjComm Hp)))))))).
  rewrite <- H in Hp.
  repeat (split; trivial).

  clear Hp Hset_eq Hobjs Hed Hst Hv0 Hmksum Hphase BGdisj WGdisj Hroot Hnumfree
    numfree stTop freePtr free ohpEnd ohpSt objs W G B.

  intros.
  destruct H as [HG [HcalleeSav [Hcst1 [Hcst0 [Hcst2 Hss'sum]]]]].
  split.
  intros.
  destruct H as [Hsb [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs
    [Hset_eq  Hp]]]]]]]]].
  exists (rget s rbot).
  assert (exists B', exists G', exists W',
    imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot)
    (rget s rbuff) (rget s rbot) numfree A /\
    set_card (set_diff B' B) = 0 /\ set_subset B B').
  apply HG with G W stTop; clear HG.
  destruct (rget s v0).
  elim Hev0; reflexivity.
  unfold sbool in Hsb.
  subst G.
  unfold setHp in Hp.
  assert (rget s rbot = rget s rtop).
  eapply empty_bfstk_eqv_b_t.
  apply (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL 
	(sconjAssocL (sconjComm (sconjAssocL (sconjComm Hp)))))))).
  rewrite <- H in Hp.
  repeat (split; trivial).
  clear HG Hp.
  destruct H as [B' [G' [W' [Hpost [Hss Hbb]]]]].
  exists B'; exists G'; exists W'.
  split; trivial.
  split; trivial.
  rewrite Hss; auto.
  split; trivial.
  split; trivial.
  split; trivial.
Qed.

Theorem imarkMarkFirstImplOK :
  iseqOkS imarkP imarkFirstPre imarkFirstGuar imarkFirstImpl.
Proof.
  unfold imarkFirstImpl.
  iseqWeakSolver PsimplIMark.
  
  unfold markFieldGuar.
  intros s s' _ [_ [_ Hr]].
  regEqSolver.

  intros s [B [G [W [free [objs [ohpSt [ohpEnd [freePtr [stTop [numfree
    [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs [Hset_eq Hp]]]]]]]]]]]]]]]]]].
  assert (Hp' := Hp).
  aconjE2 Hp.
  clear Hp; rename Hh into Hp.
  hypReplace Hp (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hp)))).
  sconjLift6of6 Hp.
  destruct (inObjHpGetSome' (x := rget s v0) (s := s) (objs := (rget s v0)::empty_set) 
    (objs' := set_add (rget s v0) ((B |_| G) |_| W)))
  as [[lchd [XA [HA Hlchd]]] [rchd [XB [HB Hrchd]]]];elimSetIn.
  apply (sconjImplR Hp).
  intros; linLogIntro.

  rewrite (lwStep HA); stSimpl; auto.
  destruct Hlchd as [Hatom | Hlin].

  split.
  split.
  left.
  apply Hatom.
  intros s' [Hg [_ Hr]].
  destruct (Hg (or_introl _ Hatom)) as [Hh Hrtop].
  simpl in Hh;clear Hg.
  exists B; exists G; exists W; exists free; exists objs; exists ohpSt;
    exists ohpEnd; exists freePtr; exists stTop; 
      exists numfree; exists lchd.
  unfold setHp.
  rewrite <- Hrtop.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  regRewrite rsweepsum.
  regRewrite rphase.
  regRewrite v0.
  regRewrite rroot.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  assert (H :=contGetEqL HA).
  rewrite <- Hh in H; apply H.
  left;trivial.

  clear Hrchd Hp' Hp Hp Hset_eq Hobjs Hed Hst Hmksum Hphase Hroot
    Hnumfree stTop freePtr ohpEnd numfree free ohpSt objs W G B.
  intros s'' [s' [[Hg1 [Hg2 Hr]] [Hg' [Hcall [Hcst1 [Hcst0 Hcst2]]]]]].
  split.
  intros.
  destruct H as [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs
    [Hset_eqv Hp]]]]]]]].

  generalize Hg';clear Hg' Hcall.
  unfold imarkSecondPreP.
  unfold setHp.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite v0.
  regRewrite rgcInfo.
  regRewrite rphase.
  regRewrite rmarksum.
  regRewrite rroot.
  stSimpl.
  intro Hg'.
  apply Hg' with G W stTop; clear Hg'.
  exists lchd.
  subst objs.
  destruct (Hg1 (or_introl _ Hatom)) as [Hh Hrtop].
  simpl in Hh;clear Hg1 Hg2 Hr.
  rewrite <- Hrtop; rewrite Hh; stSimpl.
  repeat (split; trivial).
  assert (H := contGetEqL HA).
  rewrite <- Hh in H; apply H.
  left;trivial.
  
  clear Hg' Hg1 Hg2.
  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | [
    H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  inversion H.
  
  clear Hcall.
  split.
  rewrite <- Hcst1; regEqSolver.
  split.
  rewrite <- Hcst0; regEqSolver.
  rewrite <- Hcst2; regEqSolver.

  assert (Hlin' : set_In lchd objs).
  elimSetIn.
  assert (Hna := MSPtrNotAtom _ _ _ _ Hlin' Hst Hobjs).
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (sconjAssocL (sconjAssocL (sconjAssocL Hp)))).
  split.
  split.
  clear Hp'.
  right.
  unfold setHp; stSimpl.
  split.
  eapply MSPtrGtFour.
  apply Hlin'.
  apply Hobjs.
  split.
  eapply MSPtrNotAtom.
  apply Hlin'.
  apply Hst.
  apply Hobjs.
  exists G.

  destruct Hlin as [Ha | Hi].
  subst lchd.
  right.
  split.
  subst objs.
  sconjLift3 Hp.
  intro HinG.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  sconjLift3 Hp.
  assert (rget s v0 - 4 <> rget s v0 - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  linLogIntro.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift2 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.

  destruct (set_union_elim Hi) as [HinBG | HinW]; clear Hi.
  destruct (set_union_elim HinBG) as [HinB | HinG]; clear HinBG.
  right.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  split.
  subst objs.
  sconjLift4 Hp.
  intro HinG.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  sconjLift3 Hp.
  assert (lchd - 4 <> lchd - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  linLogIntro.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp;intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.

  left.
  split;trivial.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp h; intros h Hp.
  linLogIntro.

  right.
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HinW) Hp)).
  assert (Hlv' := contGetEqL Hp).
  split.
  intro HinG.
  sconjLift5 Hp.
  hypReplace Hp ((sconjAssocL (fasElimL HinG Hp))).
  assert (Hlv := contGetEqL Hp).
  rewrite Hlv' in Hlv; inversion_clear Hlv.
  linLogIntro.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.
  
  intros s' [Hg1 [Hg2 Hr]].
  unfold imarkSecondPre; unfold imarkSecondPreP.
  unfold setHp.
  regRewrite rmarksum.
  regRewrite rphase.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  regRewrite v0.
  regRewrite rroot.
  stSimpl.
  destruct Hlin as [Ha | Hi].

  (* lchd = rget s v0 *)
  subst lchd.
  assert ( s |||- rget s v0 - 4 |-> 1 ** top).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct (Hg1 (or_intror _ H)) as [Hh Hrtop];clear H Hg1 Hg2.
  exists B; exists G; exists W; exists free; exists objs; exists ohpSt;
    exists ohpEnd; exists freePtr; exists stTop;
      exists numfree; exists (rget s v0).
  rewrite <- Hrtop;stSimpl.
  simpl in Hh.
  rewrite Hh;clear Hr Hrtop.
  subst objs.
  repeat (split; trivial).
  assert (H := contGetEqL HA).
  rewrite <- Hh in H; apply H.
  right; elimSetIn.

  destruct (set_union_elim Hi) as [HinBG | HinW]; clear Hi.

  (*  lchd \in B \cup G *)
  assert (s |||- lchd - 4 |-> 1 ** top).
  destruct (set_union_elim HinBG) as [HinB | HinG]; clear HinBG.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.

  destruct (Hg1 (or_intror _ H)) as [Hh Hrtop]; clear H Hg1 Hg2.
  exists B; exists G; exists W; exists free; exists objs; exists ohpSt;
    exists ohpEnd; exists freePtr; exists stTop; 
      exists numfree; exists lchd.
  simpl in Hh; unfold setHp.
  rewrite <- Hrtop; stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  assert (H := contGetEqL HA).
  rewrite <- Hh in H; apply H.
  right; elimSetIn.

  (*  lchd \in W *)
  exists B; exists (lchd::G); exists (set_remove lchd W); exists free;
    exists objs; exists ohpSt; exists ohpEnd; exists freePtr; 
      exists stTop; exists numfree; exists lchd.
  generalize Hg2; clear Hg1 Hg2 Hp'.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  intro Hg2.

  assert (Hset_eq' : set_eqv (set_add (rget s v0) ((B |_| lchd :: G) |_| set_remove lchd W))
    (set_add (rget s v0) ((B |_| G) |_| W))).
  split; intro; elimSetIn.
  set (GCINF := rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr :: rget s rbot
    :: rget s rbuff :: stTop :: ohpSt :: numfree :: nil)) in * .
  assert (HnotinF : ~ set_In lchd free).
  intro HinF.
  hypReplace Hp (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hp)))).
  sconjLift4 Hp.
  assert (Hpf := inFreeListVal _ _ _ _ _ Hp HinF).
  linLogElim Hpf.
  sconjLift6of6 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro2 _ G _ HinW) Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  sconjLift2 Hp.
  assert (Hgetlr := contGetEqL Hp).
  sconjLift2 Hpf.
  assert (Hgetlr' := contGetEqL Hpf).
  rewrite Hgetlr' in Hgetlr; inversion Hgetlr.
  subst x1; clear Hgetlr Hgetlr' Hp Hpf.
  destruct H6 as [Ha | Hi].
  inversion_clear Ha.
  inversion_clear H.
  inversion_clear H0.
  assert (~ set_In INFREELIST objs).
  eapply inFreeListnotinMSobjs; eauto.
  apply H.
  elimSetIn.
  sconjLift3 Hp.
  sconjLift5 Hp.
  sconjLift5 Hp.
  sconjLift6 Hp.
  unfold objHp, okObjAt, okFieldAt, fas, fas' in Hp.
  simpl in Hp; fold fas' in Hp.
  linLogElim Hp.
  assert (Hgetv0 := contGetEqL Hp).
  assert (Hgetv0' := contGetEqL HA).
  rewrite Hgetv0' in Hgetv0; inversion Hgetv0.
  subst x; clear Hgetv0 Hgetv0'.
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HinW) Hp)).
  sconjLift5 Hp.
  sconjLift6 Hp.
  sconjLift3 Hp.
  hypReplace Hp (Hg2 _ _ Hna Hp).
  Opaque set_eqv.
  repeat (split; trivial).
  destruct Hroot; [left; trivial | right; elimSetIn].
  setoid_rewrite Hset_eq'.
  apply Hset_eq.
  sconjLift4 Hp.
  apply (contGetEqL Hp).
  right; elimSetIn.
  sconjLift6 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift6 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  rewrite <- (not_In_removeEq _ _ HnotinF).
  setoid_rewrite <- (set_removeUnionDistrib lchd W free).
  apply Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  subst GCINF.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply aconjIntro.
  linLogIntro.
  sconjLift3 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  assert (h ||- objHp (set_add (rget s v0) (B |_| G)) B).
  apply Hp.
  generalize h H; clear h H Hp.
  unfold objHp.
  intros h Hp; apply (fasImpl Hp).
  unfold okObjAt, okFieldAt.
  intros.
  linLogElim H.
  linLogIntro.
  2 : apply H.
  destruct H5 as [Ha | Hi].
  left; trivial.
  right; elimSetIn.
  destruct H7 as [Ha | Hi].
  left; trivial.
  right; elimSetIn.
  sconjLift3of3 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite Hset_eq'.
  setoid_replace ((lchd::G) |_| set_remove lchd W) with
    (G |_| W).
  Transparent set_eqv.
  split; intro; elimSetIn.
  apply Hp.
  rewrite hpredTrueEq.
  setoid_rewrite Hset_eq'.
  unfold objHp, okObjAt, okFieldAt, fas, fas'.
  simpl.
  apply unitIR.
  linLogIntro.
  2 : apply Hp.
  assumption.
  assumption.

  sconjLift6 Hp'.
  unfold contL in Hp'.
  hypReplace Hp' (sconjAssocL Hp').
  assert (HgetOhpSt := contGetEqL Hp').
  sconjLift2 Hp'.
  hypReplace Hp' (sconjAssocL Hp').
  assert (HgetOhpEnd := contGetEqL Hp').
  
  subst objs.
  clear Hp' Hp numfree Hset_eq Hlin' Hed Hst  Hmksum Hphase stTop Hroot
    Hnumfree free Hrchd freePtr Hlin W G B rchd XB HB.
  intros s'' [s' [[Hg1 [Hg2  Hr]] [Hg' [
    Hcall [Hcst1 [Hcst0 Hcst2]]]]]].
  split.
  intros.
  destruct H as [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs [Hset_eq Hp]]]]]]]].
  assert (Hp' := Hp).
  sconjLift6 Hp'.
  hypReplace Hp' (sconjAssocL Hp').
  assert (HgetOhpSt' := contGetEqL Hp').
  rewrite HgetOhpSt' in HgetOhpSt; inversion HgetOhpSt.
  subst ohpSt0; clear HgetOhpSt HgetOhpSt'.
  sconjLift2 Hp'.
  hypReplace Hp' (sconjAssocL Hp').
  assert (HgetOhpEnd' := contGetEqL Hp').
  rewrite HgetOhpEnd' in HgetOhpEnd; inversion HgetOhpEnd.
  subst ohpEnd0; clear Hp' HgetOhpEnd' HgetOhpEnd.

  generalize Hg';clear Hg' Hcall.
  unfold imarkSecondPre; unfold imarkSecondPreP.
  unfold setHp.
  regRewrite rphase.
  regRewrite rmarksum.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  regRewrite v0.
  regRewrite rroot.
  stSimpl.
  intro Hg'.
  subst objs.

  destruct (inObjHpGetSome' (x := rget s v0) (s := s) (objs := (rget s v0)::empty_set) 
    (objs' := set_add (rget s v0) ((B |_| G) |_| W)))
  as [[lchd' [XA' [HA' Hlchd]]] [rchd [XB [HB Hrchd]]]];elimSetIn.
  aconjE2 Hp.
  hypReplace Hp (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hh)))).
  sconjLift6of6 Hp.
  apply (sconjImplR Hp).
  intros; linLogIntro.
  assert (Hgetl := contGetEqL HA).
  assert (Hgetl' := contGetEqL HA').
  rewrite Hgetl' in Hgetl; inversion Hgetl.
  subst lchd'; clear Hgetl Hgetl' XA' HA'.
  
  destruct Hlchd as [Hatom | Hlin].
  elimtype False; apply Hna; trivial.

  destruct Hlin as [Ha | Hi].
  generalize Hg1; clear Hg1 Hg2.
  stSimpl.
  intro Hg1.
  subst lchd.

  (* lchd = rget s v0 *)
  apply Hg' with G W stTop; trivial; clear Hg'.
  exists (rget s v0).
  assert (s |||- rget s v0 - 4 |-> 1 ** top).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct (Hg1 (or_intror _ H)) as 
    [Hh Hrtop];clear H Hg1.
  simpl in Hh.
  rewrite <- Hrtop; stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  assert (Hgetv0 := contGetEqL HA).
  rewrite <- Hh in Hgetv0; apply Hgetv0.
  right; elimSetIn.

  destruct (set_union_elim Hi) as [HlinBG | HlinW]; clear Hi.

  (*  lchd \in B \cup G *)
  apply Hg' with G W stTop; trivial; clear Hg'.
  exists lchd.
  assert (s |||- lchd - 4 |-> 1 ** top).
  destruct (set_union_elim HlinBG) as 
    [HlinB | HlinG]; clear HlinBG.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HlinB Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HlinG Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct (Hg1 (or_intror _ H)) as 
    [Hh Hrtop];clear H Hg1 Hg2.
  simpl in Hh.
  rewrite <- Hrtop; stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  assert (Hgetv0 := contGetEqL HA).
  rewrite <- Hh in Hgetv0; apply Hgetv0.
  right; elimSetIn.

  (*  lchd \in W *)
  generalize Hg2; clear Hg1 Hg2.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  intro Hg2.

  apply Hg' with (lchd::G) (set_remove lchd W) stTop;
    clear Hg'; trivial.
  exists lchd.
  set (GCINF := rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr :: rget s rbot
    :: rget s rbuff :: stTop :: ohpSt :: numfree :: nil)) in * .
  aconjE2 Hp.
  rename Hh into Hp'.
  assert (HnotinF : ~ set_In lchd free).
  intro HinF.
  hypReplace Hp' (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hp')))).
  sconjLift3 Hp'.
  assert (Hpf := inFreeListVal _ _ _ _ _ Hp' HinF).
  linLogElim Hpf.
  sconjLift5 Hp'.
  hypReplace Hp' (sconjAssocL (fasElimL (set_union_intro2 _ G _ HlinW) Hp')).
  unfold okObjAt, okFieldAt in Hp'.
  linLogElim Hp'.
  sconjLift2 Hp'.
  assert (Hgetlr := contGetEqL Hp').
  sconjLift2 Hpf.
  assert (Hgetlr' := contGetEqL Hpf).
  rewrite Hgetlr' in Hgetlr; inversion Hgetlr.
  subst x1; clear Hgetlr Hgetlr' Hp' Hpf.
  destruct H6 as [Ha | Hi].
  inversion_clear Ha.
  inversion_clear H.
  inversion_clear H0.
  assert (~ set_In INFREELIST (rangeMSObjs ohpSt ohpEnd)).
  eapply inFreeListnotinMSobjs; eauto.
  apply H.
  elimSetIn.
  assert (Hset_eq' : set_eqv (set_add (rget s v0) ((B |_| lchd :: G) |_| set_remove lchd W))
    (set_add (rget s v0) ((B |_| G) |_| W))).
  split; intro; elimSetIn.
  assert (Hgets'v0 : hget s' (rget s v0) = Some lchd).
  hypReplace Hp' (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hp')))).
  sconjLift6of6 Hp'.
  unfold objHp, okObjAt, okFieldAt, fas, fas' in Hp'.
  simpl in Hp'; fold fas' in Hp'.
  linLogElim Hp'.
  assert (Hgetv0 := contGetEqL Hp').
  assert (Hgetv0' := contGetEqL HA).
  rewrite Hgetv0' in Hgetv0; inversion Hgetv0.
  subst x; clear Hgetv0 Hgetv0'.
  sconjLift7 Hp'.
  sconjLift6 Hp'.
  sconjLift7 Hp'.
  hypReplace Hp' (sconjAssocL (fasElimL (set_union_intro1 _ _ free HlinW) Hp')).
  sconjLift4 Hp'.
  sconjLift4 Hp'.
  sconjLift3 Hp'.
  hypReplace Hp' (Hg2 _ _ Hna Hp').
  sconjLift4 Hp'.
  apply (contGetEqL Hp').
  Opaque set_eqv.
  repeat (split; trivial).
  destruct Hroot; [left; trivial | right; elimSetIn].
  setoid_rewrite Hset_eq'.
  apply Hset_eq.
  right; elimSetIn.
  clear Hp'.
  smartAssocL Hp.
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HlinW) Hp)).
  sconjLift6 Hp.
  sconjLift6 Hp.
  sconjLift3 Hp.
  hypReplace Hp (Hg2 _ _ Hna Hp).
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  rewrite <- (not_In_removeEq _ _ HnotinF).
  setoid_rewrite <- (set_removeUnionDistrib lchd W free).
  apply Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  subst GCINF.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply aconjIntro.
  apply (aconjEL Hp).
  hypReplace Hp (aconjER Hp).
  apply (sconjImpl Hp); clear h Hp.
  unfold objHp.
  intros h Hp; apply (fasImpl Hp).
  unfold okObjAt, okFieldAt.
  intros.
  linLogElim H.
  linLogIntro.
  2 : apply H.
  destruct H4 as [Ha | Hi].
  left; trivial.
  right; elimSetIn.
  destruct H5 as [Ha | Hi].
  left; trivial.
  right; elimSetIn.
  intros h Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  setoid_rewrite Hset_eq'.
  setoid_replace ((lchd::G) |_| set_remove lchd W) with
    (G |_| W).
  Transparent set_eqv.
  split; intro; elimSetIn.
  apply Hp.
  rewrite hpredTrueEq.
  setoid_rewrite Hset_eq'.
  apply Hp.
  
  clear Hg1 Hg2 Hg'.
  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst2 Hcst0 Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | [
    H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  elimtype False; trivial.
  clear Hcall.
  split.
  rewrite <- Hcst1;regEqSolver. 
  split.
  rewrite <- Hcst0;regEqSolver.
  rewrite <- Hcst2;regEqSolver.
Qed.

Theorem imarkMarkSecondImplOK :
  iseqOkS imarkP imarkSecondPre imarkSecondGuar imarkSecondImpl.
Proof.
  unfold imarkSecondImpl.
  iseqWeakSolver PsimplIMark.
  
  unfold markFieldGuar.
  intros s s' _ [_ [_ Hr]].
  regEqSolver.

  intros s [B [G [W [free [objs [ohpSt [ohpEnd [freePtr [stTop [numfree [lchd
    [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs [Hset_eq [Hnodev
      [HokLchd Hp]]]]]]]]]]]]]]]]]]]]].
  aconjE2 Hp.
  clear Hp; rename Hh into Hp.
  assert (Hp' := sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hp)))).
  sconjLift6of6 Hp'.
  assert (okFieldVal ((rget s v0 :: B) |_| G) lchd).
  destruct HokLchd as [F1 | F1]; 
    [left | right]; elimSetIn. 
  clear HokLchd; rename H into HokLchd.
  set (GCINF := rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr :: rget s rbot
    :: rget s rbuff :: stTop :: ohpSt :: numfree :: nil)) in * .
  unfold objHp, fas, fas', okObjAt, okFieldAt in Hp'; fold fas' in Hp'.
  simpl in Hp'.
  destruct (lexEL (sconjAssocL (sconjAssocL Hp'))) as [x XX].
  destruct (bangEL (sconjAssocL XX)) as [_ XXX];clear XX Hp'.
  assert (hget s (rget s v0) = Some x).
  unfold hget; apply (contGetEqL XXX).
  rewrite Hnodev in H;inversion H.
  subst x; clear H.
  destruct (lexEL (sconjAssocL (sconjComm XXX))) as 
    [rchd XX];clear XXX.
  destruct (bangEL (sconjAssocL XX)) as [Hrchd XXX];clear XX.
  assert (Hp' := sconjAssocR (sconjComm (sconjAssocR XXX))).
  assert (HB : hget s (4 + rget s v0) = Some rchd).
  unfold hget;eapply (contGetEqL XXX).

  rewrite (lwStep XXX); stSimpl; trivial; clear XXX.
  subst GCINF.
  destruct Hrchd as [Hatom | Hlin].

  (* first brach : atomP rchd *)
  split.
  split.
  left;apply Hatom.

  intros s' [Hg [_ Hr]].
  destruct (Hg (or_introl _ Hatom)) as [Hh Hrtop].
  simpl in Hh;clear Hg.
  exists ((rget s v0)::B); exists G; exists W; exists free; exists objs; 
    exists ohpSt; exists ohpEnd; exists freePtr; exists stTop;
      exists numfree.
  unfold imarkSumPlusPreP.
  unfold setHp.
  rewrite <- Hrtop.
  regRewrite rphase.
  regRewrite rmarksum.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  regRewrite rroot.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  subst objs.
  assert (HnotinB : ~ set_In (rget s v0) B).
  intro HinB.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  sconjLift3 Hp.
  assert (rget s v0 - 4 <> rget s v0 - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  repeat (split; trivial).
  clear Hp; rename Hp' into Hp.
  hypReplace Hp (unitEL (sconjAssocL (sconjComm Hp))).
  smartAssocL Hp.
  apply (sconjImpl (sconjAssocR Hp)); clear Hp; intros h Hp.
  apply objHdrsIntro; trivial.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  smartAssocL Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  smartAssocL Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply aconjIntro.
  linLogIntro.
  smartAssocL Hp.
  sconjLift2 Hp.
  apply (sconjImplL (sconjComm Hp)); clear h Hp.
  unfold objHp.
  intros h Hp.
  apply fasIntro; trivial.
  apply (sconjImplL (sconjComm Hp)).
  clear h Hp; unfold okObjAt, okFieldAt, fas, fas'; simpl; intros h Hp.
  linLogIntro.
  2 : apply Hp.
  apply HokLchd.
  left; trivial.

  clear Hp' Hp Hobjs Hed Hst HokLchd Hnodev Hset_eq numfree ohpSt ohpEnd Hmksum
    Hnumfree Hroot Hphase stTop free freePtr objs lchd W G B.
  intros s'' [s' [[Hg1 [Hg2 Hr]] [Hg' [Hcall [Hcst1 [Hcst0 Hcst2]]]]]].
  split.

  intros B G W free objs ohpSt ohpEnd freePtr stTop numfree A 
    [lchd [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs [Hset_eq 
      [Hlc [Hlc' Hp]]]]]]]]]]].
  
  generalize Hg';clear Hg' Hcall.
  unfold imarkSumPlusPreP.
  destruct (Hg1 (or_introl _ Hatom)) as [Hh Hrtop].
  simpl in Hh;clear Hg2 Hg1.
  unfold setHp.
  rewrite <- Hrtop.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rphase.
  regRewrite rgcInfo.
  regRewrite rmarksum.
  regRewrite rroot.
  stSimpl.
  intro Hg'.

  apply Hg' with G W stTop; clear Hg'.
  subst objs.
  assert (HnotinB : ~ set_In (rget s v0) B).
  intro HinB.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  sconjLift3 Hp.
  assert (rget s v0 - 4 <> rget s v0 - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  repeat (split; trivial).
  rewrite Hh.
  destruct (sconjElim (sconjComm (sconjAssocR 
    (sconjAssocR (sconjAssocR (sconjAssocR 
    (sconjAssocR (sconjAssocR Hp)))))))) as
  [h1 [h2 [J [Hp1 Hp2]]]].
  clear Hp.
  repeat (apply sconjAssocL).
  apply sconjComm.
  apply sconjIntro with h1 h2.
  apply aconjIntro.
  apply (aconjEL Hp1).
  hypReplace Hp1 (aconjER Hp1).
  assert (set_In (rget s v0) (rget s v0::empty_set)).
  left; reflexivity.
  unfold objHp, fas, fas' in Hp1; simpl in Hp1; fold fas' in Hp1.
  unfold okObjAt, okFieldAt in Hp1; linLogElim Hp1.
  hypReplace Hp1 (sconjAssocL (sconjComm (sconjAssocR Hp1))).
  assert (Heap.get (heapOf s) (rget s v0) = Some x).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply Hp1.
  assert (Heap.get (heapOf s) (4 + rget s v0) = Some x0).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply (sconjAssocL (sconjComm Hp1)).
  unfold hget in Hlc; rewrite Hlc in H0.
  inversion H0; subst x.
  unfold hget in HB; rewrite HB in H1.
  inversion H1; subst x0.
  clear H8 H9 H1 H0.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hp1))); clear Hp1 Hp2; intros h Hp.

  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt; linLogIntro.
  2 : apply Hp.
  apply Hlc'.
  left; trivial.
  apply (sconjImplL Hp2); clear Hp2 Hp1; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply objHdrsIntro; trivial.
  apply J.

  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst0 Hcst2 Hg1 Hg2 Hg' Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | 
    [H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  elimtype False; trivial.
  clear Hg1 Hg2 Hg' Hcall.
  split.
  rewrite <- Hcst1;regEqSolver. 
  split.
  rewrite <- Hcst0;regEqSolver.
  rewrite <- Hcst2;regEqSolver.
  
  (* second brach, rchd <> atom *)
  assert (Hlin' : set_In rchd objs).
  elimSetIn.
  assert (Hna := MSPtrNotAtom _ _ _ _ Hlin' Hst Hobjs).
  split.
  split.
  right.
  unfold setHp.
  stSimpl.
  split.
  eapply MSPtrGtFour.  
  apply Hlin'.
  apply Hobjs.
  split; trivial.
  clear Hp'.
  destruct Hlin as [Ha | Hlin].
  subst rchd.
  exists G.
  right.
  split.
  intro HinG.
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  sconjLift3 Hp.
  assert (rget s v0 - 4 <> rget s v0 - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  linLogIntro.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift2 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.

  destruct (set_union_elim Hlin) as [HinBG | HinW]; clear Hlin.
  destruct (set_union_elim HinBG) as [HinB | HinG]; clear HinBG.
  exists G.
  right.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  split.
  intro HinG.
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  sconjLift3 Hp.
  assert (rchd - 4 <> rchd - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  linLogIntro.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  linLogIntro.

  exists G.
  left.
  split;trivial.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp).
  intros; linLogIntro.

  exists G.
  right.
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HinW) Hp)).
  assert (Hgetr := contGetEqL Hp).
  split.
  intro HinG.
  sconjLift5 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  assert (Hgetr' := contGetEqL Hp).
  rewrite Hgetr' in Hgetr; inversion_clear Hgetr.
  linLogIntro.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift4 Hp.
  apply (sconjImplR Hp).
  intros; linLogIntro.

  intros s' [Hg1 [Hg2 Hr]].
  unfold imarkSumPlusPre; unfold imarkSumPlusPreP.
  unfold setHp.
  regRewrite rphase.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rmarksum.
  regRewrite rgcInfo.
  regRewrite rroot.
  stSimpl.
  
  destruct Hlin as [Ha | Hin].

  (* rchd = rget s v0 *)
  subst rchd.
  assert (s |||- rget s v0 - 4 |-> 1 ** top). 
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct (Hg1 (or_intror _ H)) as [Hh Hrtop].
  clear H Hg1 Hg2.
  exists (rget s v0::B); exists G; exists W; exists free; exists objs; 
    exists ohpSt; exists ohpEnd; exists freePtr; exists stTop; 
      exists numfree.
  rewrite <- Hrtop; stSimpl.
  simpl in Hh; rewrite Hh; clear Hr Hrtop.
  subst objs.
  assert (HnotinB : ~ set_In (rget s v0) B).
  intro HinB.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  sconjLift3 Hp.
  assert (rget s v0 - 4 <> rget s v0 - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  repeat (split; trivial).
  hypReplace Hp' (sconjAssocL (unitEL (sconjAssocL (sconjComm Hp')))).
  apply sconjAssocL; apply sconjAssocL; apply sconjAssocL.
  apply (sconjImpl Hp'); clear Hp Hp'; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply objHdrsIntro; trivial.
  smartAssocL Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  smartAssocL Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply aconjIntro.
  linLogIntro.
  smartAssocL Hp.
  sconjLift2 Hp.
  apply (sconjImplL (sconjComm Hp)); clear h Hp; intros h Hp.
  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImplL (sconjComm Hp)); clear h Hp;intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2 : apply Hp.
  apply HokLchd.
  right; elimSetIn.

  destruct (set_union_elim Hin) as [HinBG | HinW]; clear Hin.

  (*  rchd \in B \cup G *)
  assert (s |||- rchd - 4 |-> 1 ** top).
  destruct (set_union_elim HinBG) as [HinB | HinG]; clear HinBG.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct (Hg1 (or_intror _ H)) as [Hh Hrtop].
  clear H Hg1 Hg2.
  simpl in Hh.
  exists (rget s v0 ::B); exists G; exists W; exists free; exists objs;
    exists ohpSt; exists ohpEnd; exists freePtr; exists stTop; 
      exists numfree.
  rewrite <- Hrtop; stSimpl.
  rewrite Hh;clear Hr Hrtop.
  subst objs.
  assert (HnotinB : ~ set_In (rget s v0) B).
  intro HinB.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  sconjLift3 Hp.
  assert (rget s v0 - 4 <> rget s v0 - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.
  repeat (split; trivial).
  hypReplace Hp' (sconjAssocL (unitEL (sconjAssocL (sconjComm Hp')))).
  apply sconjAssocL; apply sconjAssocL; apply sconjAssocL.
  apply (sconjImpl Hp'); clear Hp' Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply objHdrsIntro; trivial.
  smartAssocL Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  smartAssocL Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply aconjIntro.
  linLogIntro.
  smartAssocL Hp.
  sconjLift2 Hp.
  apply (sconjImplL (sconjComm Hp)); clear h Hp; intros h Hp.
  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImplL (sconjComm Hp)); clear h Hp;intros h Hp.
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2 : apply Hp.
  apply HokLchd.
  right; elimSetIn.

  (*  rchd \in W *)
  subst objs.
  assert (HnotinB : ~ set_In (rget s v0) B).
  intro HinB.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  sconjLift3 Hp.
  assert (rget s v0 - 4 <> rget s v0 - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.

  exists (rget s v0::B); exists (rchd::G); exists (set_remove rchd W); exists free;
    exists (rangeMSObjs ohpSt ohpEnd); exists ohpSt; exists ohpEnd;
      exists freePtr; exists stTop; exists numfree.
  assert (Hset_eq' : set_eqv (((rget s v0 :: B) |_| rchd :: G) |_| set_remove rchd W)
    (set_add (rget s v0) ((B |_| G) |_| W))).
  split; intro; elimSetIn.
  set (GCINF := rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr :: rget s rbot
    :: rget s rbuff :: stTop :: ohpSt :: numfree :: nil)) in * .
  assert (HnotinF : ~ set_In rchd free).
  intro HinF.
  sconjLift5 Hp'.
  assert (Hpf := inFreeListVal _ _ _ _ _ Hp' HinF).
  linLogElim Hpf.
  sconjLift6of6 Hp'.
  smartAssocL Hp'.
  sconjLift2 Hp'.
  hypReplace Hp' (sconjAssocL (fasElimL (set_union_intro2 _ G _ HinW) Hp')).
  unfold okObjAt, okFieldAt in Hp'.
  linLogElim Hp'.
  sconjLift2 Hp'.
  assert (Hgetr := contGetEqL Hp').
  sconjLift2 Hpf.
  assert (Hgetr' := contGetEqL Hpf).
  simpl in Hgetr'; rewrite Hgetr' in Hgetr; inversion Hgetr.
  subst x1; clear Hgetr Hgetr' Hp' Hpf.
  destruct H6 as [Ha | Hi].
  inversion_clear Ha.
  inversion_clear H.
  inversion_clear H0.
  assert (~ set_In INFREELIST (rangeMSObjs ohpSt ohpEnd)).
  eapply inFreeListnotinMSobjs; eauto.
  apply H.
  elimSetIn.
  Opaque set_eqv.
  repeat (split; trivial).
  destruct Hroot; [left; trivial | right; elimSetIn].
  setoid_rewrite Hset_eq'.
  apply Hset_eq.
  generalize Hg2; clear Hg1 Hg2.
  stSimpl.
  intro Hg2.
  sconjLift3 Hp'.
  smartAssocL Hp'.
  sconjLift4 Hp'.
  hypReplace Hp' (sconjAssocL (fasElimL (set_union_intro1 _ _ free HinW) Hp')).
  sconjLift6 Hp'.
  sconjLift6 Hp'.
  sconjLift3 Hp'.
  hypReplace Hp' (Hg2 _ _ Hna Hp').
  sconjLift5 Hp'.
  sconjLift5 Hp'.
  apply (sconjImpl (sconjAssocR Hp')); clear Hg2 Hp Hp'; intros h Hp.
  apply objHdrsIntro; trivial.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp.
  apply (sconjImpl Hp); clear h Hp.
  intros h Hp.
  rewrite hpredTrueEq.
  rewrite <- (not_In_removeEq _ _ HnotinF).
  setoid_rewrite <- (set_removeUnionDistrib rchd W free).
  apply Hp.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply aconjIntro.
  linLogIntro.
  apply (sconjImpl (sconjAssocR (sconjAssocR Hp))); clear h Hp; intros h Hp.
  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  hypReplace Hp (unitER Hp).
  unfold okObjAt, okFieldAt.
  linLogIntro.
  2 : apply Hp.
  destruct HokLchd; [left; trivial | right; elimSetIn].
  right; elimSetIn.
  assert (h ||- objHp (set_add (rget s v0) (B |_| G)) B).
  apply Hp.
  generalize H; clear Hp H.
  unfold objHp.
  intros Hp; apply (fasImpl Hp); clear Hp.
  unfold okObjAt, okFieldAt.
  intros.
  linLogElim H.
  linLogIntro.
  2 : apply H.
  destruct H4; [left; trivial | right; elimSetIn].
  destruct H5; [left; trivial | right; elimSetIn].
  rewrite hpredTrueEq.
  setoid_rewrite Hset_eq'.
  setoid_replace ((rchd :: G) |_| set_remove rchd W) with
    (G |_| W).
  Transparent set_eqv.
  split; intro; elimSetIn.
  apply Hp.

  clear Hp Hp' Hobjs Hed Hst Hlin' HokLchd Hnodev Hset_eq Hlin numfree
    Hroot Hnumfree Hmksum Hphase ohpSt ohpEnd free objs stTop freePtr lchd W G B.
  intros s'' [s' [[Hg1 [Hg2  Hr]] [Hg' [Hcall [Hcst1 [Hcst0 Hcst2]]]]]].
  split.
  intros.
  destruct H as [lchd [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs
    [Hset_eq [Hlc [Hlc' Hp]]]]]]]]]]].
  generalize Hg';clear Hg' Hcall.
  unfold imarkSumPlusPreP.
  unfold setHp.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rphase.
  regRewrite rgcInfo.
  regRewrite rmarksum.
  regRewrite rroot.
  stSimpl.
  intro Hg'.
  subst objs.
  
  assert (okFieldVal ((rget s v0 :: B) |_| G) lchd).
  destruct Hlc' as [F1 | F1]; [left | right]; elimSetIn.
  clear Hlc'; rename H into HokLchd.

  assert (HnotinB : ~ set_In (rget s v0) B).
  intro HinB.
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  sconjLift3 Hp.
  assert (rget s v0 - 4 <> rget s v0 - 4).
  eapply sconjNeq.
  linLogIntro.
  apply Hp.
  apply H; reflexivity.

  destruct (inObjHpGetSome' (x := rget s v0) (s := s) (objs := (rget s v0)::empty_set) 
    (objs' := set_add (rget s v0) ((B |_| G) |_| W)))
  as [[lchd' [XA' [HA' Hlchd]]] [rchd' [XB [HB' Hrchd]]]];elimSetIn.
  aconjE2 Hp.
  hypReplace Hp (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hh)))).
  sconjLift6of6 Hp.
  apply (sconjImplR Hp).
  intros; linLogIntro.
  assert (Hgetl := contGetEqL HA').
  unfold hget in Hlc; rewrite Hlc in Hgetl; inversion Hgetl.
  subst lchd'; clear Hgetl XA' HA'.
  assert (Hgetr := contGetEqL HB').
  unfold hget in HB; rewrite HB in Hgetr; inversion Hgetr.
  subst rchd'; clear XB HB' Hgetr Hlchd.
  
  destruct Hrchd as [Ha | Hin].
  elimtype False; apply Hna; trivial.
  destruct Hin as [Ha | Hin].

  subst rchd.

  (* rchd = rget s v0 *)
  apply Hg' with G W stTop; trivial; clear Hg'.

  assert (s |||- rget s v0 - 4 |-> 1 ** top).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct (Hg1 (or_intror _ H)) as [Hh Hrtop].
  clear H Hg1 Hg2.
  simpl in Hh.
  rewrite <- Hrtop; stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  destruct (sconjElim (sconjComm (sconjAssocR (sconjAssocR (sconjAssocR 
    (sconjAssocR (sconjAssocR (sconjAssocR Hp)))))))) as
  [h1 [h2 [J [Hh1 Hh2]]]].
  repeat (apply sconjAssocL).
  apply sconjComm.
  apply sconjIntro with h1 h2.
  apply aconjIntro.
  apply (aconjEL Hh1).
  hypReplace Hh1 (aconjER Hh1).
  assert (set_In (rget s v0) (rget s v0::empty_set)).
  left; reflexivity.
  unfold objHp, fas, fas' in Hh1; simpl in Hh1; fold fas' in Hh1.
  unfold okObjAt, okFieldAt in Hh1; linLogElim Hh1.
  hypReplace Hh1 (sconjAssocL (sconjComm (sconjAssocR Hh1))).
  assert (Hlchd : Heap.get (heapOf s) (rget s v0) = Some x).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply Hh1.
  assert (Hrchd : Heap.get (heapOf s) (4 + rget s v0) = Some x0).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply (sconjAssocL (sconjComm Hh1)).
  rewrite Hlc in Hlchd; inversion Hlchd.
  subst x.
  rewrite HB in Hrchd; inversion Hrchd.
  subst x0.
  clear H8 H9 Hlchd Hrchd.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hh1))); 
    clear Hp Hh1 Hh2; intros h Hp.

  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImplL Hp); clear h Hp;intros h Hp.
  unfold okObjAt, okFieldAt; linLogIntro.
  2 : apply Hp.
  apply HokLchd.
  right; elimSetIn.
  apply (sconjImplL Hh2); clear Hp Hh1 Hh2; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply objHdrsIntro; trivial.
  apply J.
  
  destruct (set_union_elim Hin) as [HinBG | HinW]; clear Hin.

  (*  rchd \in B \cup G *)
  assert (s |||-rchd - 4 |-> 1 ** top).
  destruct (set_union_elim HinBG) as [HinB | HinG].
  sconjLift2 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  sconjLift3 Hp.
  hypReplace Hp (sconjAssocL (fasElimL HinG Hp)).
  apply (sconjImplR Hp).
  intros; linLogIntro.
  destruct (Hg1 (or_intror _ H)) as [Hh Hrtop].
  clear Hg1 Hg2 H.
  simpl in Hh.
  apply Hg' with G W stTop; trivial; clear Hg'.
  rewrite <- Hrtop; stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  destruct (sconjElim (sconjComm (sconjAssocR (sconjAssocR (sconjAssocR 
    (sconjAssocR (sconjAssocR (sconjAssocR Hp)))))))) as
  [h1 [h2 [J [Hh1 Hh2]]]].
  repeat (apply sconjAssocL).
  apply sconjComm.
  apply sconjIntro with h1 h2.
  apply aconjIntro.
  apply (aconjEL Hh1).
  hypReplace Hh1 (aconjER Hh1).
  assert (set_In (rget s v0) (rget s v0::empty_set)).
  left; reflexivity.
  unfold objHp, fas, fas' in Hh1; simpl in Hh1; fold fas' in Hh1.
  unfold okObjAt, okFieldAt in Hh1; linLogElim Hh1.
  hypReplace Hh1 (sconjAssocL (sconjComm (sconjAssocR Hh1))).
  assert (Hlchd : Heap.get (heapOf s) (rget s v0) = Some x).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply Hh1.
  assert (Hrchd : Heap.get (heapOf s) (4 + rget s v0) = Some x0).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply (sconjAssocL (sconjComm Hh1)).
  rewrite Hlc in Hlchd; inversion Hlchd.
  subst x.
  rewrite HB in Hrchd; inversion Hrchd.
  subst x0.
  clear H8 H9 Hlchd Hrchd.
  apply (sconjImplL (sconjAssocR (sconjAssocR Hh1)));
    clear Hh1 Hp Hh2; intros h Hp.

  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImplL Hp); clear h Hp;intros h Hp.
  unfold okObjAt, okFieldAt; linLogIntro.
  2: apply Hp.
  apply HokLchd.
  right; elimSetIn.
  apply (sconjImplL Hh2); clear Hp Hh1 Hh2; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply objHdrsIntro; trivial.
  apply J.

  (*  rchd \in W *)
  apply Hg' with (rchd::G) (set_remove rchd W) stTop;
    trivial; clear Hg'.
  assert (Hset_eq' : set_eqv (((rget s v0 :: B) |_| rchd :: G) |_| set_remove rchd W)
    (set_add (rget s v0) ((B |_| G) |_| W))).
  split; intro; elimSetIn.
  set (GCINF := rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr :: rget s rbot
    :: rget s rbuff :: stTop :: ohpSt :: numfree :: nil)) in * .
  assert (HnotinF : ~ set_In rchd free).
  intro HinF.
  aconjE2 Hp.
  sconjLift7 Hh.
  assert (Hpf := inFreeListVal _ _ _ _ _ Hh HinF).
  linLogElim Hpf.
  sconjLift2 Hpf.
  assert (Hgetr := contGetEqL Hpf).
  hypReplace Hh (sconjAssocR (sconjAssocR (sconjAssocR Hh))).
  sconjLift6 Hh.
  hypReplace Hh (sconjAssocL (fasElimL (set_union_intro2 _ G _ HinW) Hh)).
  unfold okObjAt, okFieldAt in Hh.
  linLogElim Hh.
  sconjLift2 Hh.
  assert (Hgetr' := contGetEqL Hh).
  rewrite Hgetr' in Hgetr; inversion Hgetr.
  subst x1; clear Hgetr Hgetr' Hh Hpf.
  destruct H6 as [Ha | Hi].
  inversion_clear Ha.
  inversion_clear H.
  inversion_clear H0.
  assert (~ set_In INFREELIST (rangeMSObjs ohpSt ohpEnd)).
  eapply inFreeListnotinMSobjs; eauto.
  apply H.
  elimSetIn.
  Opaque set_eqv.
  repeat (split; trivial).
  destruct Hroot; [left; trivial | right; elimSetIn].
  setoid_rewrite Hset_eq'.
  apply Hset_eq.
  sconjLift4 Hp.
  hypReplace Hp (sconjAssocL (fasElimL (set_union_intro1 _ _ free HinW) Hp)).
  sconjLift6 Hp.
  sconjLift6 Hp.
  sconjLift3 Hp.
  aconjE2 Hp.
  hypReplace Hh (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR Hh)))).
  sconjLift7of7 Hh.
  unfold objHp, okObjAt, okFieldAt, fas, fas' in Hh.
  simpl in Hh; fold fas' in Hh.
  linLogElim Hh.
  assert (Hgetl := contGetEqL Hh).
  rewrite Hlc in Hgetl; inversion Hgetl.
  subst x; clear Hgetl Hlc.
  sconjLift2 Hh.
  assert (Hgetr := contGetEqL Hh).
  simpl in HB; rewrite HB in Hgetr; inversion Hgetr.
  subst x0; clear Hgetr HB.
  sconjLift5 Hh.
  sconjLift5 Hh.
  sconjLift5 Hh.
  hypReplace Hh (Hg2 _ _ Hna Hh).
  sconjLift4 Hh.
  assert (Hgets'l := contGetEqL Hh).
  sconjLift4 Hh.
  assert (Hgets'r := contGetEqL Hh).
  clear Hh Hg1 H3 H6.
  smartAssocL Hp.
  hypReplace Hp (Hg2 _ _ Hna Hp).
  generalize Hp; clear Hp Hg2.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  intro Hp.
  hypReplace Hp (sconjAssocR (sconjAssocR (sconjAssocR (sconjAssocR 
    (sconjAssocR (sconjAssocR Hp)))))).
  destruct (sconjElim (sconjComm Hp)) as [h1 [h2 [J [Hp1 Hp2]]]].
  repeat (apply sconjAssocL). 
  apply sconjComm.
  apply sconjIntro with h1 h2.
  apply aconjIntro.
  apply (aconjEL Hp1).
  hypReplace Hp1 (aconjER Hp1).
  sconjLift3of3 Hp1.
  unfold objHp, okObjAt, okFieldAt, fas, fas' in Hp1; simpl in Hp1; fold fas' in Hp1.
  linLogElim Hp1.
  assert (Heap.get (heapOf s') (rget s v0) = Some x).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply Hp1.
  rewrite Hgets'l in H; inversion H.
  subst x; clear H.
  assert (Heap.get (heapOf s') (4 + rget s v0) = Some x0).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply (sconjAssocL (sconjComm Hp1)).
  simpl in H; rewrite Hgets'r in H; inversion H.
  subst x0; clear H.
  apply (sconjImpl (sconjAssocR (sconjAssocR Hp1))); clear Hp1 Hp2 Hp; intros h Hp.
  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold okObjAt, okFieldAt, fas, fas'.
  linLogIntro.
  2 : apply Hp.
  destruct HokLchd; [left; trivial | right; elimSetIn].
  right; elimSetIn.
  assert (h ||- objHp (set_add (rget s v0) (B |_| G)) B).
  apply Hp.
  generalize H; clear H Hp.
  unfold objHp.
  intros Hp; apply (fasImpl Hp); trivial.
  unfold okObjAt, okFieldAt.
  intros.
  linLogElim H.
  linLogIntro.
  2: apply H.
  destruct H5; [left; trivial | right; elimSetIn].
  destruct H7; [left; trivial | right; elimSetIn].
  rewrite hpredTrueEq.
  setoid_replace ((rchd :: G) |_| set_remove rchd W) with
    (G |_| W).
  Transparent set_eqv.
  split; intro; elimSetIn.
  setoid_rewrite Hset_eq'.
  apply Hp.
  apply (sconjImplL Hp2); clear Hp Hp1 Hp2; intros h Hp.
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  smartAssocL Hp.
  smartAssocR.
  sconjLift5of5 Hp.
  sconjLift5of5 Hp.
  apply (sconjImpl (sconjAssocR Hp)); clear h Hp; intros h Hp.
  apply objHdrsIntro; trivial.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplL (sconjComm Hp)); clear h Hp; intros h Hp.
  rewrite hpredTrueEq.
  rewrite <- (not_In_removeEq _ _ HnotinF).
  setoid_rewrite <- (set_removeUnionDistrib rchd W free).
  apply Hp.
  apply J.

  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst0 Hcst2 Hg1 Hg2 Hg' Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | 
    [H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  elimtype False;auto.
  clear Hcall Hg1 Hg2 Hg'.
  split.
  rewrite <- Hcst1;regEqSolver. 
  split.
  rewrite <- Hcst0;regEqSolver.
  rewrite <- Hcst2;regEqSolver.
Qed.

Theorem imarkMarkSumPlusOK :
  iseqOkS imarkP imarkSumPlusPre imarkSumPlusGuar imarkSumPlusImpl.
Proof.
  unfold imarkSumPlusImpl.
  iseqWeakSolver PsimplIMark.
  
  intros s [B [G [W [free [objs [ohpSt [ohpEnd [freePtr [stTop [numfree
    [Hroot [Hnumfree [Hphase [Hmksum [Hst [Hed [Hobjs [Hset_eq Hp]]]]]]]]]]]]]]]]]].
  rewrite addiuStep; stSimpl.
  split.
  exists B; exists G; exists W; exists free; exists objs; exists ohpSt;
    exists ohpEnd; exists freePtr; exists stTop; exists numfree.
  repeat (split; trivial).
  clear Hp Hset_eq numfree Hobjs Hed Hst Hmksum Hphase Hroot Hnumfree
    stTop freePtr ohpEnd ohpSt free objs W G B.
  unfold imarkLoopGuar; unfold imarkSumPlusGuar.
  unfold imarkLoopInv; unfold imarkSumPlusPreP.
  stSimpl.
  intros s' HG.
  destruct HG as [HG [Hcall [Hcst1 [Hcst0 Hcst2]]]].
  split.
  intros.
  assert (exists stTop', exists B', exists G', exists W',
    imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot)
    (rget s rbuff) stTop' numfree A /\
    set_card (set_diff B' B) + (1 + rget s rsweepsum) =
    rget s' rsweepsum /\ set_subset B B').
  apply HG with  G W stTop; clear HG.
  destruct H as [Hroot [Hnumfree [Hphase [Hmksum [HohpSt [HohpEnd 
    [Hobjs [Hseteq H]]]]]]]].
  repeat (split; trivial).
  destruct H0 as [stTop' [B' [G' [W' [Hpost [Hss Hbb]]]]]].
  exists stTop'; exists B'; exists G'; exists W'.
  split; trivial.
  split; trivial.
  rewrite <- Hss; auto with arith.
  clear HG.
  split.
  eapply calleeSavedOkTrans'.
  2: apply Hcall.
  calleeSavedOkSolver'.
  split.
  rewrite <- Hcst1; stSimpl; trivial.
  split.
  rewrite <- Hcst0; stSimpl; trivial.
  rewrite <- Hcst2; stSimpl; trivial.
Qed.

Theorem imarkDoneSomeMarkImplOK :
  iseqOkS imarkP imarkDoneSomeMarkPre imarkDoneSomeMarkGuar imarkDoneSomeMarkImpl.
Proof.
  unfold imarkDoneSomeMarkImpl.
  iseqWeakSolver PsimplIMark.

  intros s [B [G [W [free [objs [ohpSt [ohpEnd [freePtr [stTop [numfree
    [Hroot [Hnumfree [Hphase [Hmksum [HGnotempty [Hst [Hed [Hobjs
      [Hset_eq Hp]]]]]]]]]]]]]]]]]]].
  sconjLift5 Hp.
  unfold contL in Hp.

  hypReplace Hp (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL
    (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL Hp)))))))))))))))).  

  rewrite (swStep Hp); stSimpl; trivial.
  rewrite addiuStep; stSimpl.
  split; trivial.
  clear Hp Hset_eq numfree Hobjs Hed Hst HGnotempty Hmksum Hphase Hroot
    Hnumfree stTop freePtr ohpEnd ohpSt free objs W G B.

  intros.
  subst s'.
  constructor.
  intros B G W free objs ohpSt ohpEnd freePtr stTop numfree A
    [Hroot [Hnumfree [Hphase [Hmksum [HGnotempty [HohpSt [HohpEnd
      [Hobjs [Hseteq Hp]]]]]]]]].
  exists (rget s rtop).
  exists B; exists G; exists W.
    
  split.
  split; trivial.
  stSimpl; rewrite Hmksum.
  Opaque set_eqv.
  repeat(split; trivial).
  left; split; trivial.
  split; trivial.
  hypReplace Hp (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL
    (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL
      (sconjComm Hp))))))))))).
  unfold contL in Hp.
  hypReplace Hp (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL
    (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL
      (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL Hp)))))))))))))))).  
  apply sconjComm; apply sconjAssocR; apply sconjComm; apply sconjAssocR; apply sconjAssocR;
    apply sconjComm; apply sconjAssocR; apply sconjAssocR; apply sconjComm; 
      apply sconjAssocR; apply sconjAssocR.
  unfold contL.
  apply sconjAssocR; apply sconjComm; apply sconjAssocR; apply sconjAssocR; apply sconjComm;
    apply sconjAssocR; apply sconjAssocR; apply sconjComm; apply sconjAssocR; 
      apply sconjAssocR; apply sconjComm; apply sconjAssocR; apply sconjAssocR;
	apply sconjComm; apply sconjAssocR; apply sconjAssocR.
  apply uhsetHOS.
  apply contUpdateSt with stTop; trivial.
  split.
  rewrite (set_diffSelf B).
  reflexivity.
  unfold set_subset; trivial.
  split.
  calleeSavedOkSolver'.
  split; trivial.
  split; trivial.
  split; trivial.
Qed.

Theorem imarkMarkFinishedOK :
  iseqOkS imarkP imarkMarkFinishedPre imarkMarkFinishedGuar imarkMarkFinishedImpl.
Proof.
  unfold imarkMarkFinishedImpl.
  iseqWeakSolver PsimplIMark.

  intros s [B [G [W [free [objs [ohpSt [ohpEnd [freePtr [stTop [numfree 
    [Hroot [Hnumfree [Hphase [Hmksum [HGempty [Hbeqt [Hst [Hed
      [Hobjs [Hset_eq Hp]]]]]]]]]]]]]]]]]]]].
  sconjLift5 Hp.
  unfold contL in Hp.
  
  hypReplace Hp (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL
    (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL Hp)))))))))))))))).  
  rewrite addiuStep; stSimpl.
  rewrite (swStep Hp); stSimpl; trivial.
  rewrite addiuStep; stSimpl.
  split; trivial.
  clear Hp Hset_eq numfree Hobjs Hed Hst HGempty Hbeqt Hmksum Hphase Hroot
    Hnumfree stTop freePtr ohpEnd ohpSt free objs W G B.

  intros.
  subst s'.
  constructor.
  intros B G W free objs ohpSt ohpEnd freePtr stTop numfree A [Hroot [Hnumfree
    [Hphase [Hmksum [HGempty [Hbeqt [HohpSt [HohpEnd [Hobjs [Hseteq Hp]]]]]]]]]].
  exists B; exists G; exists W.
  split.
  split; trivial.
  stSimpl.
  repeat(split; trivial).
  right; split; trivial.
  split; trivial.
  hypReplace Hp (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL
    (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjComm Hp))))))))))).
  unfold contL in Hp.
  hypReplace Hp (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL
    (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL (sconjAssocL
      (sconjComm (sconjAssocL (sconjAssocL (sconjComm (sconjAssocL Hp)))))))))))))))).  
  apply sconjComm; apply sconjAssocR; apply sconjComm; apply sconjAssocR;
    apply sconjAssocR; apply sconjComm; apply sconjAssocR; apply sconjAssocR;
      apply sconjComm; apply sconjAssocR; apply sconjAssocR.
  unfold contL.
  apply sconjAssocR; apply sconjComm; apply sconjAssocR; apply sconjAssocR;
    apply sconjComm; apply sconjAssocR; apply sconjAssocR; apply sconjComm;
      apply sconjAssocR; apply sconjAssocR; apply sconjComm; apply sconjAssocR; 
	apply sconjAssocR; apply sconjComm; apply sconjAssocR; apply sconjAssocR.
  rewrite Hbeqt; rewrite Hbeqt in Hp.
  apply uhsetHOS.
  apply contUpdateSt with stTop; trivial.
  split.
  rewrite (set_diffSelf B).
  reflexivity.
  unfold set_subset; trivial.
  split.
  calleeSavedOkSolver'.
  split; trivial.
  split; trivial.
  split; trivial.
Qed.
