Require Import Heap.
Require Import WeakSCAP.
Require Import LinLog.
Require Import Machine.
Require Import List.
Require Import SpecUtil.
Require Import GC.
Require Import Arith.
Require Import NatUtil.
Require Import StateTheory.
Require Import MarkSweepGC.
Require Import MarkSweepLemma.
Require Import BuffArrayStack.
Require Import StackSet.

(* some definitions *)
Notation MARK_PHASE := 0.
Notation SWEEP_PHASE := 1.
Notation IDLE_PHASE := 2.

Notation NUM_MARK := 5.
Notation NUM_SWEEP := 5.
Notation FREE_THRESHOLD := 10.
Notation INFREELIST := 2. (* something not an OkfieldVal *)

Definition NULL := 1. (* NULL pointer *)

(* Registers *)
Notation rnumfree := t7.
Notation rphase := t1.
Notation rmarksum := t9.
Notation rsweepsum := t9.

(* incremetal mark *)
Notation IMARK := 500.
Notation IMARKLOOP := 501.
Notation IMARKLOOPBODY := 502.
Notation IMARKFIRST := 503.
Notation IMARKSECOND := 504.
Notation IMARKSUMPLUS := 505.
Notation IDONESOMEMARK := 506.
Notation IMARKFINISHED := 507.

(* imcremental sweep *)
Notation ISWEEP := 600.
Notation ISWEEPLOOP := 601.
Notation ISWEEPADD := 602.
Notation ISWEEPNEXT := 603.
Notation ISWEEPFIN := 604.
Notation IDONESOMESWEEP :=605.

(* incremetal do_gc *)
Notation DOGC := 700.
Notation GCINITRETURN := 701.
Notation NEWOBJNEW := 702.
Notation NEWOBJRET := 703.

(* freelist in incremental mark-sweep *)
Fixpoint ifreeList (x : nat) (objs : set) {struct objs} : hpred :=
  match objs with
    | nil => !(x = NULL)
    | a::objs' => lexists x', !(x = a) ** !(x <> NULL) **
      (a |-> x' ** (addrOffset + a) |-> INFREELIST)
      ** ifreeList x' objs'
  end.

(* imark() the pointer of sweepAt should point to ohpSt *)
Definition imarkPreP
  s B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree A :=
  okFieldVal (B |_| G) (rget s rroot) /\ numfree = set_card free /\
  rget s rphase = MARK_PHASE /\ ptrP ohpSt /\ ptrP ohpEnd /\ 
  objs = rangeMSObjs ohpSt ohpEnd /\
  set_eqv (((B |_| G) |_| W) |_| free) objs /\ 
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    buffStackSet stBot stTop stBuff G **
    rget s rgcInfo |->,
      (ohpSt::ohpEnd::freePtr::stBot::stBuff::stTop::ohpSt::numfree::nil) **
      ifreeList freePtr free **
      (A && (objHp (B |_| G) B ** objHp ((B |_| G) |_| W) (G |_| W))).

Definition imarkPre := fun s =>
  exists B, exists G, exists W, exists free, exists objs, 
    exists ohpSt, exists ohpEnd, exists freePtr, exists stBot,
      exists stBuff, exists stTop, exists numfree,
	imarkPreP s B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop 
	numfree top.

Definition imarkPost
  s B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree A :=
  okFieldVal (B |_| G) (rget s rroot) /\ numfree = set_card free /\
  rget s rmarksum <= NUM_MARK /\
  ((rget s rphase = MARK_PHASE /\ stTop = rget s rtop /\ G <> empty_set) \/ 
    (rget s rphase = SWEEP_PHASE /\ stTop = stBot /\ G = empty_set)) /\
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  objs = rangeMSObjs ohpSt ohpEnd /\
  set_eqv (((B |_| G) |_| W) |_| free) objs /\ 
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    buffStackSet stBot stTop stBuff G **
    rget s rgcInfo |->,
      (ohpSt::ohpEnd::freePtr::stBot::stBuff::stTop::ohpSt::numfree::nil) **
      ifreeList freePtr free **
      (A && (objHp (B |_| G) B ** objHp ((B |_| G) |_| W) (G |_| W))).

Definition imarkGuar := fun s s' =>
  (forall B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree A,
    imarkPreP s B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree A ->
    exists sum, exists stTop', exists B', exists G', exists W',
      imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr stBot stBuff stTop' numfree A
      (* increasement of Black nodes less than fix number*)
      /\  sum = set_card (set_diff B' B) /\ sum <= NUM_MARK /\
      set_subset B B') /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0.

Definition imarkImpl :=
  addiu rcst2 ra 0;
  lw rbot 12 rgcInfo;
  lw rbuff 16 rgcInfo;
  lw rtop 20 rgcInfo; (* stack reinstall*)
  addiu rmarksum zero 0;
  j IMARKLOOP.

Definition imarkLoopInv 
  s B G W free objs ohpSt ohpEnd freePtr stTop numfree A := 
  okFieldVal (B |_| G) (rget s rroot) /\ numfree = set_card free /\
  rget s rphase = MARK_PHASE /\  rget s rmarksum <= NUM_MARK /\
  ptrP ohpSt /\ ptrP ohpEnd /\ objs = rangeMSObjs ohpSt ohpEnd /\
  set_eqv (((B |_| G) |_| W) |_| free) objs /\ 
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    setHp G s **
    rget s rgcInfo |->,(ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)
      ::stTop::ohpSt::numfree::nil)**
    ifreeList freePtr free **
    (A && (objHp (B |_| G) B ** objHp ((B |_| G) |_| W) (G |_| W))).

Definition imarkLoopPre := fun s =>
  exists B, exists G, exists W, exists free, exists objs,
    exists ohpSt, exists ohpEnd, exists freePtr, exists stTop,
      exists numfree,
	imarkLoopInv s B G W free objs ohpSt ohpEnd freePtr stTop numfree top.

Definition imarkLoopGuar := fun s s' =>
  (forall B G W free objs ohpSt ohpEnd freePtr stTop numfree A,
    imarkLoopInv s B G W free objs ohpSt ohpEnd freePtr stTop numfree A ->
    exists stTop', exists B', exists G', exists W',
      imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot)
      (rget s rbuff) stTop' numfree A /\
      set_card (set_diff B' B) + rget s rmarksum = rget s' rmarksum /\
      set_subset B B') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\
  rget s rcst2 = rget s' ra.

Definition imarkLoopImpl :=
  jal STACKEMPTY IMARKLOOPBODY.

Definition imarkLoopBodyPre := fun s =>
  exists B, exists G, exists W, exists free, exists objs,
    exists ohpSt, exists ohpEnd, exists freePtr, exists stTop,
      exists numfree,
	sbool (rget s v0) (G = empty_set) /\
	imarkLoopInv s B G W free objs ohpSt ohpEnd freePtr stTop numfree top.

Definition imarkLoopBodyGuar := fun s s' =>
  (forall B G W free objs ohpSt ohpEnd freePtr stTop numfree A,
    sbool (rget s v0) (G = empty_set) /\ 
    imarkLoopInv s B G W free objs ohpSt ohpEnd freePtr stTop numfree A ->
    exists stTop', exists B', exists G', exists W',
      imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot) (rget s rbuff)
      stTop' numfree A /\
      set_card (set_diff B' B) + rget s rmarksum = rget s' rmarksum /\
      set_subset B B') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\
  rget s rcst2 = rget s' ra.

Definition imarkLoopBodyImpl :=
  bne v0 zero IMARKFINISHED;;
  addiu rtemp zero NUM_MARK;
  subu rtemp rtemp rmarksum;
  beq rtemp zero IDONESOMEMARK;;
  jal STACKPOP IMARKFIRST.

Definition imarkFirstPreP 
  s B G W free objs ohpSt ohpEnd freePtr stTop numfree A :=
  okFieldVal (set_add (rget s v0) B |_| G) (rget s rroot) /\ numfree = set_card free /\
  rget s rphase = MARK_PHASE /\  rget s rmarksum < NUM_MARK /\
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  objs = rangeMSObjs ohpSt ohpEnd /\
  set_eqv ((set_add (rget s v0) ((B |_| G) |_| W)) |_| free) objs /\ 
  s |||- (rget s v0 -4 |-> MARKED)**
    objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    setHp G s **
    rget s rgcInfo |->,(ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)
      ::stTop::ohpSt::numfree::nil)**
    ifreeList freePtr free **
    (A && (objHp (set_add (rget s v0) (B |_| G)) B ** 
      objHp (set_add (rget s v0) ((B |_| G) |_| W)) (G |_| W) ** 
      objHp (set_add (rget s v0) ((B |_| G) |_| W)) ((rget s v0)::empty_set))).

Definition imarkFirstPre := fun s =>
  exists B, exists G, exists W, exists free, exists objs,
    exists ohpSt, exists ohpEnd, exists freePtr, exists stTop,
      exists numfree,
	imarkFirstPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree top.
	
Definition imarkFirstGuar := fun s s' =>
  (forall B G W free objs ohpSt ohpEnd freePtr stTop numfree A,
    imarkFirstPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree A ->
    exists stTop', exists B', exists G', exists W',
      imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot) (rget s rbuff)
      stTop' numfree A /\
      set_card (set_diff B' ((rget s v0)::B)) + 1 + rget s rmarksum = rget s' rmarksum /\
      set_subset ((rget s v0)::B) B') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\
  rget s rcst2 = rget s' ra.

Definition imarkFirstImpl :=
  lw a0 0 v0;
  jal MARKFIELD IMARKSECOND.

Definition imarkSecondPreP 
  s B G W free objs ohpSt ohpEnd freePtr stTop numfree A :=
  exists lchd,
    okFieldVal (set_add (rget s v0) B |_| G) (rget s rroot) /\ numfree = set_card free /\
    rget s rphase = MARK_PHASE /\  rget s rmarksum < NUM_MARK /\
    ptrP ohpSt /\ ptrP ohpEnd /\ 
    objs = rangeMSObjs ohpSt ohpEnd /\
    set_eqv ((set_add (rget s v0) ((B |_| G) |_| W)) |_| free) objs /\ 
    hget s (rget s v0) = Some lchd /\
    okFieldVal (set_add (rget s v0) (B |_| G)) lchd /\
    s |||- (rget s v0 -4 |-> MARKED)**
      objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
      setHp G s **
      rget s rgcInfo |->,(ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)
	::stTop::ohpSt::numfree::nil)**
      ifreeList freePtr free **
      (A && (objHp (set_add (rget s v0) (B |_| G)) B ** 
	objHp (set_add (rget s v0) ((B |_| G) |_| W)) (G |_| W) ** 
	objHp (set_add (rget s v0) ((B |_| G) |_| W)) ((rget s v0)::empty_set))).

Definition imarkSecondPre := fun s =>
  exists B, exists G, exists W, exists free, exists objs,
    exists ohpSt, exists ohpEnd, exists freePtr, exists stTop,
      exists numfree,
	imarkSecondPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree top.
	
Definition imarkSecondGuar := fun s s' =>
  (forall B G W free objs ohpSt ohpEnd freePtr stTop numfree A,
    imarkSecondPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree A ->
    exists stTop', exists B', exists G', exists W',
      imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot) (rget s rbuff)
      stTop' numfree A /\
      set_card (set_diff B' ((rget s v0)::B)) + 1 + rget s rmarksum = rget s' rmarksum /\
      set_subset ((rget s v0)::B) B') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\
  rget s rcst2 = rget s' ra.

Definition imarkSecondImpl :=
  lw a0 4 v0;
  jal MARKFIELD IMARKSUMPLUS.

Definition imarkSumPlusPreP
  s B G W free objs ohpSt ohpEnd freePtr stTop numfree A := 
  okFieldVal (B |_| G) (rget s rroot) /\ numfree = set_card free /\
  rget s rphase = MARK_PHASE /\  rget s rmarksum < NUM_MARK /\
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  objs = rangeMSObjs ohpSt ohpEnd /\
  set_eqv (((B |_| G) |_| W) |_| free) objs /\ 
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    setHp G s **
    rget s rgcInfo |->,(ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)
      ::stTop::ohpSt::numfree::nil)**
    ifreeList freePtr free **
    (A && (objHp (B |_| G) B ** objHp ((B |_| G) |_| W) (G |_| W))).

Definition imarkSumPlusPre := fun s =>
  exists B, exists G, exists W, exists free, exists objs,
    exists ohpSt, exists ohpEnd, exists freePtr, exists stTop,
      exists numfree,
	imarkSumPlusPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree top.

Definition imarkSumPlusGuar := fun s s' =>
  (forall B G W free objs ohpSt ohpEnd freePtr stTop numfree A,
    imarkSumPlusPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree A ->
    exists stTop', exists B', exists G', exists W',
      imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot) (rget s rbuff) 
      stTop' numfree A /\
      set_card (set_diff B' B) + 1 + rget s rmarksum = rget s' rmarksum /\
      set_subset B B') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\
  rget s rcst2 = rget s' ra.

Definition imarkSumPlusImpl :=
  addiu rmarksum rmarksum 1;
  j IMARKLOOP.

Definition imarkDoneSomeMarkPreP
  s B G W free objs ohpSt ohpEnd freePtr stTop numfree A := 
  okFieldVal (B |_| G) (rget s rroot) /\ numfree = set_card free /\
  rget s rphase = MARK_PHASE /\  rget s rmarksum = NUM_MARK /\
  G <> empty_set /\
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  objs = rangeMSObjs ohpSt ohpEnd /\
  set_eqv (((B |_| G) |_| W) |_| free) objs /\ 
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    setHp G s **
    rget s rgcInfo |->,(ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)
      ::stTop::ohpSt::numfree::nil)**
    ifreeList freePtr free **
    (A && (objHp (B |_| G) B ** objHp ((B |_| G) |_| W) (G |_| W))).

Definition imarkDoneSomeMarkPre := fun s =>
  exists B, exists G, exists W, exists free, exists objs,
    exists ohpSt, exists ohpEnd, exists freePtr, exists stTop,
      exists numfree,
	imarkDoneSomeMarkPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree top.

Definition imarkDoneSomeMarkGuar := fun s s' => 
  (forall B G W free objs ohpSt ohpEnd freePtr stTop numfree A,
    imarkDoneSomeMarkPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree A ->
    exists stTop', exists B', exists G', exists W',
      imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot) (rget s rbuff) 
      stTop' numfree A /\
      set_card (set_diff B' B) = 0 /\
      set_subset B B') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\
  rget s rcst2 = rget s' ra /\ rget s rmarksum = rget s' rmarksum.

Definition imarkDoneSomeMarkImpl :=
  sw rtop 20 rgcInfo; (* update stack top *)
  addiu ra rcst2 0;
  jr ra.

Definition imarkMarkFinishedPreP 
  s B G W free objs ohpSt ohpEnd freePtr stTop numfree A := 
  okFieldVal (B |_| G) (rget s rroot) /\ numfree = set_card free /\
  rget s rphase = MARK_PHASE /\  rget s rmarksum <= NUM_MARK /\
  G = empty_set /\ rget s rbot = rget s rtop /\
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  objs = rangeMSObjs ohpSt ohpEnd /\
  set_eqv (((B |_| G) |_| W) |_| free) objs /\ 
  s |||- objHdrs B MARKED ** objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    buffStackSet (rget s rbot) (rget s rbot) (rget s rbuff) G **
    rget s rgcInfo |->,(ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)
      ::stTop::ohpSt::numfree::nil) **
    ifreeList freePtr free **
    (A && (objHp (B |_| G) B ** objHp ((B |_| G) |_| W) (G |_| W))).

Definition imarkMarkFinishedPre := fun s =>
  exists B, exists G, exists W, exists free, exists objs,
    exists ohpSt, exists ohpEnd, exists freePtr,
      exists stTop, exists numfree,
	imarkMarkFinishedPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree top.

Definition imarkMarkFinishedGuar := fun s s' => 
  (forall B G W free objs ohpSt ohpEnd freePtr stTop numfree A,
    imarkMarkFinishedPreP s B G W free objs ohpSt ohpEnd freePtr stTop numfree A ->
    exists B', exists G', exists W',
      imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr (rget s rbot) (rget s rbuff) 
      (rget s rbot) numfree A /\
      set_card (set_diff B' B) = 0 /\
      set_subset B B') /\
  calleeSavedOk' s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0 /\
  rget s rcst2 = rget s' ra /\ rget s rmarksum = rget s' rmarksum .

Definition imarkMarkFinishedImpl :=
  addiu rphase zero SWEEP_PHASE;
  sw rtop 20 rgcInfo; (* update stack top *)
  addiu ra rcst2 0;
  jr ra.

(* isweep() *)
Definition isweepPreP 
  s B W objs free swept ohpSt ohpEnd freePtr stBot stBuff sweepAt numfree A :=
  okFieldVal B (rget s rroot) /\
  rget s rphase = SWEEP_PHASE /\ numfree = set_card free /\
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  set_diff W swept = W /\
  objs = rangeMSObjs ohpSt ohpEnd /\
  swept = rangeMSObjs ohpSt (11 + sweepAt) /\
  (exists k, sweepAt = ohpSt + 12 * k) /\
  set_eqv (set_union (set_union B W) free) objs /\  
  s |||- bobjHdrs B (4 + sweepAt) UNMARK MARKED ** 
    objHdrs (set_union W free) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, 
      (ohpSt::ohpEnd::freePtr::stBot::stBuff::stBot::sweepAt::numfree::nil) ** 
      objHp objs W ** ifreeList freePtr free ** (A && objHp B B).

Definition isweepPre := fun s =>
  exists B, exists W, exists objs, exists free, exists swept,
    exists ohpSt, exists ohpEnd, exists freePtr, exists stBot, 
      exists stBuff, exists sweepAt, exists numfree,
	isweepPreP s B W objs free swept ohpSt ohpEnd freePtr stBot stBuff 
	sweepAt numfree	top.

Definition isweepPost
  s B W objs free swept ohpSt ohpEnd freePtr stBot stBuff sweepAt numfree A :=
  okFieldVal B (rget s rroot) /\
  rget s rsweepsum <= NUM_SWEEP /\ numfree = set_card free /\
  ptrP ohpSt /\ ptrP ohpEnd /\ 
  set_diff W swept = W /\
  objs = rangeMSObjs ohpSt ohpEnd /\
  swept = rangeMSObjs ohpSt (11 + (rget s rsweep)) /\
  (exists k, rget s rsweep = ohpSt + 12 * k) /\
  set_eqv (set_union (set_union B W) free) objs /\
  ((rget s rphase = IDLE_PHASE /\ 12 + rget s rsweep > ohpEnd /\ 
    sweepAt = ohpSt /\ W = empty_set) \/
    (rget s rphase = SWEEP_PHASE /\ 12 + rget s rsweep <= ohpEnd /\ 
      rget s rsweep = sweepAt))/\
  s |||- bobjHdrs B (4 + rget s rsweep) UNMARK MARKED ** 
    objHdrs (set_union W free) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, 
      (ohpSt::ohpEnd::freePtr::stBot::stBuff::stBot::sweepAt::numfree::nil) ** 
      objHp objs W ** ifreeList freePtr free ** (A && objHp B B).

Definition isweepGuar := fun s s' =>
  (forall B W objs free swept ohpSt ohpEnd freePtr stBot stBuff sweepAt numfree A,
    isweepPreP s B W objs free swept ohpSt ohpEnd freePtr stBot stBuff 
    sweepAt numfree A ->
    exists num, exists W', exists free', exists swept', exists freePtr',
      exists sweepAt', exists numfree',
	isweepPost s' B W' objs free' swept' ohpSt ohpEnd freePtr' stBot stBuff 
	sweepAt' numfree' A /\
	(rget s' rsweep = sweepAt + 12 * num /\ num <= NUM_SWEEP)) /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0.

Definition isweepImpl :=
  lw rhpEnd 4 rgcInfo;
  lw rfree 8 rgcInfo;
  lw rsweep 24 rgcInfo;
  lw rnumfree 28 rgcInfo;
  addiu rsweepsum zero 0;
  j ISWEEPLOOP.

Definition isweepLoopPreP 
  s B W objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A :=
  okFieldVal B (rget s rroot) /\
  rget s rphase = SWEEP_PHASE /\ rget s rnumfree = set_card free /\
  rget s rsweepsum <= NUM_SWEEP /\
  ptrP ohpSt /\ ptrP (rget s rhpEnd) /\ 
  set_diff W swept = W /\
  objs = rangeMSObjs ohpSt (rget s rhpEnd) /\
  swept = rangeMSObjs ohpSt (11 + rget s rsweep) /\
  (exists k, rget s rsweep = ohpSt + 12 * k) /\
  set_eqv (set_union (set_union B W) free) objs /\  
  s |||- bobjHdrs B (4 + rget s rsweep) UNMARK MARKED ** 
    objHdrs (set_union W free) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, 
      (ohpSt::(rget s rhpEnd)::freePtr::stBot::stBuff::stBot::sweepAt::numfree::nil) **
      objHp objs W ** ifreeList (rget s rfree) free ** (A && objHp B B).

Definition isweepLoopPre := fun s =>
  exists B, exists W, exists objs, exists free, exists swept,
    exists ohpSt, exists freePtr, exists stBot, exists stBuff,
      exists sweepAt, exists numfree,
	isweepLoopPreP s B W objs free swept ohpSt freePtr stBot stBuff 
	sweepAt numfree	top.

Definition isweepLoopGuar := fun s s' =>
  (forall B W objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A,
    isweepLoopPreP s B W objs free swept ohpSt freePtr stBot stBuff 
    sweepAt numfree A ->
    exists W', exists free', exists swept', exists freePtr',
      exists sweepAt', exists numfree',
	isweepPost s' B W' objs free' swept' ohpSt (rget s rhpEnd) freePtr' stBot
	stBuff sweepAt' numfree' A /\
	(rget s' rsweep + 12 * (rget s rsweepsum) = rget s rsweep + 12 * (rget s' rsweepsum))) /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0.

Definition isweepLoopImpl :=
  addiu rtemp rsweep 12;
  sltu rtemp rhpEnd rtemp;
  bne rtemp zero ISWEEPFIN;;
  addiu rtemp zero NUM_SWEEP;
  subu rtemp rtemp rsweepsum;
  beq rtemp zero IDONESOMESWEEP;;
  lw v0 0 rsweep;
  beq v0 zero ISWEEPADD;; (* zero UNMARK *)
  sw zero 0 rsweep;
  j ISWEEPNEXT.

Definition isweepAddPreP
  s B W objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A :=
  okFieldVal B (rget s rroot) /\
  rget s rphase = SWEEP_PHASE /\ rget s rsweepsum < NUM_SWEEP /\ 
  rget s rnumfree = set_card free /\ 12 + rget s rsweep <= rget s rhpEnd /\
  ptrP ohpSt /\ ptrP (rget s rhpEnd) /\ 
  set_diff W swept = W /\
  objs = rangeMSObjs ohpSt (rget s rhpEnd) /\
  swept = rangeMSObjs ohpSt (11 + (rget s rsweep)) /\
  (exists k, rget s rsweep = ohpSt + 12 * k) /\
  set_eqv (set_union (set_union B W) free) objs /\
  set_In (4 + rget s rsweep) (set_union W free)/\
  s |||- bobjHdrs B (4 + rget s rsweep) UNMARK MARKED ** 
    objHdrs (set_union W free) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, 
      (ohpSt::(rget s rhpEnd)::freePtr::stBot::stBuff::stBot::
	sweepAt::numfree::nil) ** objHp objs W ** 
      ifreeList (rget s rfree) free ** (A && objHp B B).

Definition isweepAddPre := fun s =>
  exists B, exists W, exists objs, exists free, exists swept,
    exists ohpSt, exists freePtr, exists stBot, exists stBuff,
      exists sweepAt, exists numfree,
	isweepAddPreP s B W objs free swept ohpSt freePtr stBot stBuff 
	sweepAt numfree	top.

Definition isweepAddGuar := fun s s' =>
  (forall B W objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A,
    isweepAddPreP s B W objs free swept ohpSt freePtr stBot stBuff 
    sweepAt numfree A ->
    exists W', exists free', exists swept', exists freePtr',
      exists sweepAt', exists numfree',
	isweepPost s' B W' objs free' swept' ohpSt (rget s rhpEnd) freePtr' stBot
	stBuff sweepAt' numfree' A /\
	(rget s' rsweep + 12 * (rget s rsweepsum) = rget s rsweep + 12 * (rget s' rsweepsum))) /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0.

Definition isweepAddImpl :=
  lw v0 8 rsweep;
  addiu rtemp zero INFREELIST;
  beq v0 rtemp ISWEEPNEXT;;
  sw rfree 4 rsweep;
  sw rtemp 8 rsweep;
  addiu rfree rsweep HDSIZE;
  addiu rnumfree rnumfree 1;
  j ISWEEPNEXT.

Definition isweepNextPreP
  s B W objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A :=
  okFieldVal B (rget s rroot) /\
  rget s rphase = SWEEP_PHASE /\ rget s rsweepsum < NUM_SWEEP /\ 
  rget s rnumfree = set_card free /\
  ptrP ohpSt /\ ptrP (rget s rhpEnd) /\ 
  set_diff W swept = W /\
  objs = rangeMSObjs ohpSt (rget s rhpEnd) /\
  swept = rangeMSObjs ohpSt (23 + (rget s rsweep)) /\
  (exists k, rget s rsweep = ohpSt + 12 * k) /\
  set_eqv (set_union (set_union B W) free) objs /\
  s |||- bobjHdrs B (12 + 4 + rget s rsweep) UNMARK MARKED ** 
    objHdrs (set_union W free) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, 
      (ohpSt::(rget s rhpEnd)::freePtr::stBot::stBuff::stBot::
	sweepAt::numfree::nil) ** objHp objs W ** 
      ifreeList (rget s rfree) free ** (A && objHp B B).

Definition isweepNextPre := fun s =>
  exists B, exists W, exists objs, exists free, exists swept,
    exists ohpSt, exists freePtr, exists stBot, exists stBuff,
      exists sweepAt, exists numfree,
	isweepNextPreP s B W objs free swept ohpSt freePtr stBot stBuff 
	sweepAt numfree	top.

Definition isweepNextGuar := fun s s' =>
  (forall B W objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A,
    isweepNextPreP s B W objs free swept ohpSt freePtr stBot stBuff 
    sweepAt numfree A ->
    exists W', exists free', exists swept', exists freePtr',
      exists sweepAt', exists numfree',
	isweepPost s' B W' objs free' swept' ohpSt (rget s rhpEnd) freePtr' stBot
	stBuff sweepAt' numfree' A /\
	(rget s' rsweep + 12 * (rget s rsweepsum) = rget s rsweep + 12 * (rget s' rsweepsum))) /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0.

Definition isweepNextImpl :=
  addiu rsweep rsweep 12;
  addiu rsweepsum rsweepsum 1;
  j ISWEEPLOOP.

Definition isweepFinishedPreP
  s B objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A :=
  okFieldVal B (rget s rroot) /\
  rget s rphase = SWEEP_PHASE /\ rget s rsweepsum <= NUM_SWEEP /\ 
  rget s rnumfree = set_card free /\ 12 + rget s rsweep > rget s rhpEnd /\
  ptrP ohpSt /\ ptrP (rget s rhpEnd) /\ 
  objs = rangeMSObjs ohpSt (rget s rhpEnd) /\
  set_disj B free /\
  swept = rangeMSObjs ohpSt (11 + (rget s rsweep)) /\ (* swept should be objs *)
  (exists k, rget s rsweep = ohpSt + 12 * k) /\ (* rsweep should be rhpEnd *)
  set_eqv (set_union B free) objs /\
  s |||- objHdrs objs UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, 
      (ohpSt::(rget s rhpEnd)::freePtr::stBot::stBuff::stBot::
	sweepAt::numfree::nil) ** 
    ifreeList (rget s rfree) free ** (A && objHp B B).

Definition isweepFinishedPre := fun s =>
  exists B, exists objs, exists free, exists swept,
    exists ohpSt, exists freePtr, exists stBot, exists stBuff,
      exists sweepAt, exists numfree,
	isweepFinishedPreP s B  objs free swept ohpSt freePtr stBot stBuff 
	sweepAt numfree	top.

Definition isweepFinishedGuar := fun s s' =>
  (forall B objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A,
    isweepFinishedPreP s B objs free swept ohpSt freePtr stBot stBuff 
    sweepAt numfree A ->
    exists free', exists swept', exists freePtr',
      exists sweepAt', exists numfree',
	isweepPost s' B empty_set objs free' swept' ohpSt (rget s rhpEnd) freePtr' stBot
	stBuff sweepAt' numfree' A /\
	(sweepAt' = ohpSt /\ rget s' rsweep = rget s rsweep /\ 
	  rget s' rnumfree = rget s rnumfree /\rget s' rsweepsum = rget s rsweepsum)) /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0.

Definition isweepFinishedImpl :=
  addiu rphase zero IDLE_PHASE;
  lw rtemp 0 rgcInfo; (* rtemp = ohpSt *)
  sw rfree 8 rgcInfo;
  sw rtemp 24 rgcInfo;
  sw rnumfree 28 rgcInfo;
  jr ra.

Definition isweepDoneSomePreP
  s B W objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A :=
  okFieldVal B (rget s rroot) /\
  rget s rphase = SWEEP_PHASE /\ rget s rsweepsum = NUM_SWEEP /\ 
  rget s rnumfree = set_card free /\ 12 + rget s rsweep <= rget s rhpEnd /\
  ptrP ohpSt /\ ptrP (rget s rhpEnd) /\ 
  set_diff W swept = W /\
  objs = rangeMSObjs ohpSt (rget s rhpEnd) /\
  swept = rangeMSObjs ohpSt (11 + (rget s rsweep)) /\
  (exists k, rget s rsweep = ohpSt + 12 * k) /\
  set_eqv (set_union (set_union B W) free) objs /\
  s |||- bobjHdrs B (4 + rget s rsweep) UNMARK MARKED ** 
    objHdrs (set_union W free) UNMARK **
    buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->, 
      (ohpSt::(rget s rhpEnd)::freePtr::stBot::stBuff::stBot::
	sweepAt::numfree::nil) ** objHp objs W ** 
      ifreeList (rget s rfree) free ** (A && objHp B B).

Definition isweepDoneSomePre := fun s =>
  exists B, exists W, exists objs, exists free, exists swept,
    exists ohpSt, exists freePtr, exists stBot, exists stBuff,
      exists sweepAt, exists numfree,
	isweepDoneSomePreP s B W objs free swept ohpSt freePtr stBot stBuff 
	sweepAt numfree	top.

Definition isweepDoneSomeGuar := fun s s' =>
  (forall B W objs free swept ohpSt freePtr stBot stBuff sweepAt numfree A,
    isweepDoneSomePreP s B W objs free swept ohpSt freePtr stBot stBuff 
    sweepAt numfree A ->
    exists W', exists free', exists swept', exists freePtr',
      exists sweepAt', exists numfree',
	isweepPost s' B W' objs free' swept' ohpSt (rget s rhpEnd) freePtr' stBot
	stBuff sweepAt' numfree' A /\
	(rget s' rsweep = rget s rsweep /\ rget s' rnumfree = rget s rnumfree /\
	  rget s' rsweepsum = rget s rsweepsum)) /\
  calleeSavedOk s s' /\ rget s rcst1 = rget s' rcst1 /\ rget s rcst0 = rget s' rcst0.

Definition isweepDoneSomeSweepImpl :=
  sw rfree 8 rgcInfo;
  sw rsweep 24 rgcInfo;
  sw rnumfree 28 rgcInfo;
  jr ra.

(* do_gc() the sweepAt pointer should point to ohpSt initially *)
Definition HpUnchangeNotInF s s' objs free :=
  forall x, set_In x objs /\ ~ set_In x free -> hget s' x = hget s x /\
    hget s' (addrOffset + x) = hget s (addrOffset + x).

Definition HpUnchangeInB s s' B :=
  forall x, set_In x B -> hget s' x = hget s x /\
    hget s' (addrOffset + x) = hget s (addrOffset + x).

Definition imarkSweepGCInitPreP 
  s B free objs ohpSt ohpEnd freePtr stBot stBuff numfree A :=
  okFieldVal B (rget s rroot) /\  numfree = set_card free /\
  rget s rphase = IDLE_PHASE /\
  ptrP ohpSt /\ ptrP ohpEnd /\
  objs = rangeMSObjs ohpSt ohpEnd /\
  set_eqv (B |_| free) objs /\
  s |||- objHdrs (B |_| free) UNMARK ** buffStackSet stBot stBot stBuff nil **
    rget s rgcInfo |->,
      (ohpSt::ohpEnd::freePtr::stBot::stBuff::stBot::ohpSt::numfree::nil) ** 
      ifreeList freePtr free ** (A && objHp B B).

Definition imarkSweepGCInitPre := fun s =>
  exists B, exists free, exists objs, exists  ohpSt, 
    exists ohpEnd, exists freePtr, exists stBot, 
      exists stBuff, exists numfree,
	imarkSweepGCInitPreP s B free objs ohpSt ohpEnd freePtr stBot 
	stBuff numfree top.

Definition imarkSweepGCInitGuar := fun s s' =>
  (forall B free objs ohpSt ohpEnd freePtr stBot stBuff numfree A,
    imarkSweepGCInitPreP s B free objs ohpSt ohpEnd freePtr stBot stBuff
    numfree A ->
    (numfree < FREE_THRESHOLD -> HpUnchangeInB s s' B /\
      exists G, exists W, exists stTop,
	imarkPreP s' empty_set G W free objs ohpSt ohpEnd freePtr stBot
	stBuff stTop numfree A) /\
    (numfree >= FREE_THRESHOLD -> HpUnchangeInB s s' B /\
      imarkSweepGCInitPreP s' B free objs ohpSt ohpEnd freePtr stBot stBuff
      numfree A)) /\
  calleeSavedOk s s' /\ rget s rcst0 = rget s' rcst0.

Definition imarkSweepGCPre := fun s =>
  (rget s rphase = MARK_PHASE /\ imarkPre s) \/
  (rget s rphase = SWEEP_PHASE /\ isweepPre s) \/
  (rget s rphase = IDLE_PHASE /\ imarkSweepGCInitPre s).

Definition imarkSweepGCGuar s s' := 
  (rget s rphase = MARK_PHASE /\ 
    (forall B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree A,
      imarkPreP s B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree A ->
      exists sum, exists stTop', exists B', exists G', exists W',
	HpUnchangeNotInF s s' objs free /\
	imarkPost s' B' G' W' free objs ohpSt ohpEnd freePtr stBot stBuff stTop' numfree A
	/\  sum = set_card (set_diff B' B) /\ sum <= NUM_MARK /\ set_subset B B') /\
    calleeSavedOk s s' /\ rget s rcst0 = rget s' rcst0) \/
  (rget s rphase = SWEEP_PHASE /\ 
    (forall B W objs free swept ohpSt ohpEnd freePtr stBot stBuff sweepAt numfree A,
      isweepPreP s B W objs free swept ohpSt ohpEnd freePtr stBot stBuff 
      sweepAt numfree A ->
      exists num, exists W', exists free', exists swept', exists freePtr',
	exists sweepAt',exists numfree',
	  HpUnchangeInB s s' B /\
	  isweepPost s' B W' objs free' swept' ohpSt ohpEnd freePtr' stBot stBuff 
	  sweepAt' numfree' A /\ (rget s' rsweep = sweepAt + 12 * num /\ num <= NUM_SWEEP)) /\
    calleeSavedOk s s' /\ rget s rcst0 = rget s' rcst0) \/
  (rget s rphase = IDLE_PHASE /\ imarkSweepGCInitGuar s s').

(* we suppose sweepAt is initially set to ohpSt so
   there's no need to set it to ohpSt when num_free < free_threshold *)

Definition imarkSweepGCImpl :=
  addiu rcst1 ra 0;
  addiu rtemp zero MARK_PHASE;
  beq rphase rtemp IMARK;;
  addiu rtemp zero SWEEP_PHASE;
  beq rphase rtemp ISWEEP;;
  addiu rtemp zero IDLE_PHASE;
  bne rphase rtemp RETURN1;;
  lw rtemp 28 rgcInfo;
  addiu r31 zero FREE_THRESHOLD;
  sltu rtemp rtemp r31;
  beq rtemp zero RETURN1;;
  addiu rphase zero MARK_PHASE;
  lw rbot 12 rgcInfo;
  lw rbuff 16 rgcInfo;
  lw rtop 20 rgcInfo; (* stack reinstall*)
  addiu a0 rroot 0;
  jal MARKFIELD GCINITRETURN.

Definition igcInitRetPreP
  s G W free objs ohpSt ohpEnd freePtr stTop numfree A :=
  okFieldVal G (rget s rroot) /\ numfree = set_card free /\
  rget s rphase = MARK_PHASE /\ ptrP ohpSt /\ ptrP ohpEnd /\ 
  objs = rangeMSObjs ohpSt ohpEnd /\
  set_eqv ((G |_| W) |_| free) objs /\ 
  s |||- objHdrs G MARKED ** objHdrs (W |_| free) UNMARK **
    setHp G s **
    rget s rgcInfo |->,
      (ohpSt::ohpEnd::freePtr::(rget s rbot)::(rget s rbuff)::stTop::ohpSt::numfree::nil) **
      ifreeList freePtr free **
      (A && objHp (G |_| W) (G |_| W)).

Definition igcInitRetPre := fun s =>
  exists G, exists W, exists free, exists objs, exists ohpSt, 
    exists ohpEnd, exists freePtr, exists stTop, exists numfree,
	igcInitRetPreP s G W free objs ohpSt ohpEnd freePtr stTop numfree top.

Definition igcInitRetGuar := fun s s' =>
  (forall G W free objs ohpSt ohpEnd freePtr stTop numfree A,
    igcInitRetPreP s G W free objs ohpSt ohpEnd freePtr stTop numfree A ->
    imarkPreP s' empty_set G W free objs ohpSt ohpEnd freePtr
    (rget s rbot) (rget s rbuff) (rget s rtop) numfree A) /\
  (forall r, r <> a0 -> r <> ra -> rget s' r = rget s r) /\
  rget s' ra = rget s rcst1 /\ rget s rcst0 = rget s' rcst0.

Definition igcInitRetImpl :=
  sw rtop 20 rgcInfo;
  addiu ra rcst1 0;
  jr ra.

(* we add an additional condition (phase <> IDLE_PHASE) for we will
  mark all malloced obj black because we suppose sweepAt is initially 
  set to ohpSt so that every new >= sweepAt (= ohpSt) , then we will 
  collect no garbage because everything is marked.

  newObj 
  obj_t* newObj(obj_t* root) {
    obj_t* new;
  
    do_gc(root);
  
    if (freeList == NULL) {
      // error: out of space
      return NULL;
    }
  
    new = freeList;
    freeList = freeList->field1;
    num_free -= 1;
    if (phase <> IDLE_PHASE /\ new >= sweepAt) {
      new->color = BLACK;
    }
  
    return new;
  }
*)

Definition newObjPre := imarkSweepGCPre.

Definition newMarkPost 
  s B G W free free' objs ohpSt ohpEnd freePtr freePtr' stBot stBuff stTop numfree :=
  (free <> empty_set -> ~ set_In freePtr B /\ rget s v0 = freePtr /\
    imarkPreP s (set_add freePtr B) G W free' objs ohpSt ohpEnd freePtr' stBot stBuff
    stTop (numfree - 1) (objHp ((set_add freePtr B) |_| G) (set_add freePtr B) ** 
      objHp (((set_add freePtr B) |_| G) |_| W) (G |_| W))) /\
  (free = empty_set -> rget s v0 = NULL /\
    imarkPreP s B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree
    (objHp (B |_| G) B ** objHp ((B |_| G) |_| W) (G |_| W))).

Definition newSweepPost
  s B W objs free free' swept ohpSt ohpEnd freePtr freePtr' stBot stBuff sweepAt numfree :=
  (free <> empty_set -> ~ set_In freePtr B /\ rget s v0 = freePtr /\
    isweepPreP s (set_add freePtr B) W objs free' swept ohpSt ohpEnd freePtr' stBot 
    stBuff sweepAt (numfree - 1) (objHp (set_add freePtr B) (set_add freePtr B))) /\
  (free = empty_set -> rget s v0 = NULL /\
    isweepPreP s B W objs free swept ohpSt ohpEnd freePtr stBot stBuff
    sweepAt numfree (objHp B B)).

Definition newIdlePost
  s B free free' objs ohpSt ohpEnd freePtr freePtr' stBot stBuff numfree :=
  (free <> empty_set -> ~ set_In freePtr B /\ rget s v0 = freePtr /\
    imarkSweepGCInitPreP s (set_add freePtr B) free' objs ohpSt ohpEnd freePtr'
    stBot stBuff (numfree - 1) (objHp (set_add freePtr B) (set_add freePtr B))) /\
  (free = empty_set -> rget s v0 = NULL /\
    imarkSweepGCInitPreP s B free objs ohpSt ohpEnd freePtr stBot stBuff numfree
    (objHp B B)).

Definition newObjMarkGuar := fun s s' =>
  (forall B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree,
    imarkPreP s B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree 
    (objHp (B |_| G) B ** objHp ((B |_| G) |_| W) (G |_| W)) ->
    (exists stTop', exists B', exists G', exists W', exists free', exists freePtr',
      HpUnchangeNotInF s s' objs free /\ 
      newMarkPost s' B' G' W' free free' objs ohpSt ohpEnd freePtr freePtr' stBot
      stBuff stTop' numfree) \/ 
    (exists B', exists W', exists free', exists freePtr',
      HpUnchangeInB s s' B' /\ 
      newSweepPost s' B' W' objs free free' empty_set ohpSt ohpEnd freePtr freePtr'
      stBot stBuff ohpSt numfree)) /\
  calleeSavedOk s s'.

Definition newObjSweepGuar := fun s s' =>
  (forall B W objs free swept ohpSt ohpEnd freePtr stBot stBuff sweepAt numfree,
    isweepPreP s B W objs free swept ohpSt ohpEnd freePtr stBot stBuff 
    sweepAt numfree (objHp B B) ->
    (exists W', exists free', exists swept', exists freePtr',
      exists sweepAt', exists numfree', exists free'', exists freePtr'',
	HpUnchangeInB s s' B /\
	newSweepPost s' B W' objs free' free'' swept' ohpSt ohpEnd
	freePtr' freePtr'' stBot stBuff sweepAt' numfree') \/ 
    (exists free', exists freePtr', exists numfree', exists free'',
      exists freePtr'',
	HpUnchangeInB s s' B /\
	newIdlePost s' B free' free'' objs ohpSt ohpEnd freePtr' freePtr'' stBot stBuff numfree')) /\
  calleeSavedOk s s'.

Definition newObjIdleGuar := fun s s' =>
  (forall B free objs ohpSt ohpEnd freePtr stBot stBuff numfree,
    imarkSweepGCInitPreP s B free objs ohpSt ohpEnd freePtr stBot stBuff 
    numfree (objHp B B) ->
    (numfree < FREE_THRESHOLD -> HpUnchangeInB s s' B /\
      exists G, exists W, exists stTop, exists freePtr', exists free',
	newMarkPost s' empty_set G W free free' objs ohpSt ohpEnd freePtr freePtr'
	stBot stBuff stTop numfree) /\
    (numfree >= FREE_THRESHOLD -> HpUnchangeInB s s' B /\
      exists freePtr', exists free',
	~set_In freePtr B /\ rget s' v0 = freePtr /\
	imarkSweepGCInitPreP s' (set_add freePtr B) free' objs ohpSt ohpEnd freePtr'
	stBot stBuff (numfree - 1) (objHp (set_add freePtr B) (set_add freePtr B)))) /\
  calleeSavedOk s s'.

Definition newObjGuar := fun s s' =>
  (rget s rphase = MARK_PHASE /\ newObjMarkGuar s s') \/
  (rget s rphase = SWEEP_PHASE /\ newObjSweepGuar s s') \/
  (rget s rphase = IDLE_PHASE /\ newObjIdleGuar s s').

Definition newObjGCImpl :=
  addiu rcst0 ra 0;
  jal DOGC NEWOBJNEW.

Definition newObjNewPre := imarkSweepGCPre.

Definition newObjNewMarkGuar := fun s s' =>
  (forall B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree,
    imarkPreP s B G W free objs ohpSt ohpEnd freePtr stBot stBuff stTop numfree
    (objHp (B |_| G) B ** objHp ((B |_| G) |_| W) (G |_| W)) ->
    exists free', exists freePtr',
      HpUnchangeNotInF s s' objs free /\
      newMarkPost s' B G W free free' objs ohpSt ohpEnd freePtr freePtr' stBot
      stBuff stTop numfree) /\
  (forall r, r <> v0 -> r <> ra -> r <> rtemp -> r <> rnumfree 
    -> rget s' r = rget s r) /\ rget s rcst0 = rget s' ra.

Definition newObjNewSweepGuar := fun s s' =>
  (forall B W objs free swept ohpSt ohpEnd freePtr stBot stBuff sweepAt numfree,
    isweepPreP s B W objs free swept ohpSt ohpEnd freePtr stBot stBuff 
    sweepAt numfree (objHp B B) ->
    exists free', exists freePtr',
      HpUnchangeInB s s' B /\
      newSweepPost s' B W objs free free' swept ohpSt ohpEnd freePtr freePtr'
      stBot stBuff sweepAt numfree) /\
  (forall r, r <> v0 -> r <> ra -> r <> rtemp -> r <> rnumfree 
    -> rget s' r = rget s r) /\ rget s rcst0 = rget s' ra.

Definition newObjNewIdleGuar := fun s s' =>
  (forall B free objs ohpSt ohpEnd freePtr stBot stBuff numfree,
    imarkSweepGCInitPreP s B free objs ohpSt ohpEnd freePtr stBot stBuff
    numfree (objHp B B) ->
    exists free', exists freePtr',
      HpUnchangeInB s s' B /\
      newIdlePost s' B free free' objs ohpSt ohpEnd freePtr freePtr' stBot 
      stBuff numfree) /\
  (forall r, r <> v0 -> r <> ra -> r <> rtemp -> r <> rnumfree 
    -> rget s' r = rget s r) /\ rget s rcst0 = rget s' ra.

Definition newObjNewGuar := fun s s' =>
  (rget s rphase = MARK_PHASE /\ newObjNewMarkGuar s s') \/
  (rget s rphase = SWEEP_PHASE /\ newObjNewSweepGuar s s') \/
  (rget s rphase = IDLE_PHASE /\ newObjNewIdleGuar s s').
  
Definition newObjNewImpl :=
  lw v0 8 rgcInfo; (* freePtr => v0 // new *)
  addiu rtemp zero NULL;
  beq rtemp v0 RETURN0;; (* freePtr ?= NULL *)
  lw rtemp 0 v0;  (* rtemp = freePtr -> next *)
  sw rtemp 8 rgcInfo;
  addiu rtemp zero NULL; (* init fields to NULL *)
  sw rtemp 0 v0;
  sw rtemp 4 v0;
  lw rnumfree 28 rgcInfo; (* numfree => rtemp *)
  addiu rtemp zero 1;
  subu rnumfree rnumfree rtemp;
  sw rnumfree 28 rgcInfo;  (* numfree - 1 *)
  addiu rtemp zero IDLE_PHASE;
  beq rtemp rphase RETURN0;;
  lw rtemp 24 rgcInfo;  (* sweepAt => rtemp *)
  addiu rtemp rtemp HDSIZE;
  sltu rtemp v0 rtemp;  (* new - sweepAt *)
  bne rtemp zero RETURN0;;
  addiu rtemp zero HDSIZE;
  subu rtemp v0 rtemp;
  addiu r31 zero MARKED;
  sw r31 0 rtemp;
  addiu ra rcst0 0;
  jr ra.

(* Lemmas *)
Lemma uhsetHOS : forall s a v A,
  (Heap.uset (heapOf s) a v ||- A) ->
  heapOf (uhset s a v) ||- A.
Proof.
  intros s a v A.
  unfold uhset; unfold Heap.uset; unfold heapOf.
  trivial.
Qed.

Lemma empty_bfstk_eqv_b_t : forall bbot bbuff btop h A,
  (h ||- buffStackSet bbot btop bbuff empty_set ** A) ->
  bbot = btop.
Proof.
  unfold buffStackSet; unfold buffArrayStack.
  intros.
  linLogElim H.
  destruct H4 as [Xeq _].
  destruct H7 as [Heq _].
  clear H.
  subst btop.
  destruct x; simpl.
  trivial.
  elimtype False.
  destruct (Xeq a) as [_ IB]; simpl in IB; auto.
Qed.

Lemma notInInsert_LengthInc : forall v B,
  ~set_In v B -> length (set_insert v B)= S (length B).
Proof.
  intros.
  induction B.
  unfold length; unfold set_insert.
  reflexivity.
  unfold set_insert.
  destruct (POnat.POnat.compareDec v a).
  destruct s.
  simpl; auto.
  destruct H.
  apply set_add_intro1; assumption.
  simpl.
  assert (~ set_In v B).
  intro.
  apply H.
  apply set_add_intro2; assumption.
  assert (He := IHB H0).
  rewrite <- He.
  reflexivity.
Qed.
  
Lemma notInPlus : forall v B B',
  set_In v B' -> ~set_In v B ->
  set_card (set_diff B' B) = set_card (set_diff B' (v::B)) + 1.
Proof.
  intros.
  assert (Ht := removeIsDiffStep v B' B).
  unfold set_add in Ht.
  rewrite <- Ht; clear Ht.
  unfold set_card.
  assert (set_canon (set_diff B' B) = 
    set_insert v (set_remove v (set_canon (set_diff B' B)))).
  apply canonInsertRemoveEq.
  apply set_canonCanonP.
  apply set_canon_intro.
  apply set_diff_intro; auto.
  rewrite H1; clear H1.
  assert (Ht := set_canonRemoveSwap v (set_diff B' B)).
  rewrite Ht; clear Ht.
  replace (length (set_remove v (set_canon (set_diff B' B))) + 1) with
    ( S (length (set_remove v (set_canon (set_diff B' B))))).
  apply notInInsert_LengthInc.
  intro.
  assert (v <> v).
  eapply set_remove_elim2.
  apply H1.
  elim H2.
  reflexivity.
  omega.
Qed.

Lemma notInPlus' : forall B v,
  ~set_In v B -> 1 + (set_card B) = set_card (v::B).
Proof.
  unfold set_card.
  unfold set_canon.
  fold set_canon.
  intros.
  symmetry.
  simpl.
  apply notInInsert_LengthInc.
  intro.
  apply H.
  apply set_canon_elim; assumption.
Qed.

Lemma OkfieldValnotInfree :
  forall objs x,
    okFieldVal objs x -> ~ set_In INFREELIST objs
    -> x <> INFREELIST.
Proof.
  intros objs x Hok Hnotin.
  destruct Hok as [Hat | Hin].
  inversion_clear Hat.
  intro.
  eapply not_even_and_odd.
  apply H.
  assert (n = 1); auto.
  rewrite H1.
  auto with arith.
  intro.
  apply Hnotin.
  rewrite H in Hin; trivial.
Qed.

Lemma inFreeListnotinMSobjs :
  forall objs x y,
    objs = rangeMSObjs x y -> ~ set_In INFREELIST objs.
Proof.
  intros objs x y Hobjs Hin.
  assert (H := MSPtrGtFour _ _ _ _ Hin Hobjs).
  omega.
Qed.

Lemma inFreeListVal :
  forall free freePtr v h A,
    (h ||- ifreeList freePtr free ** A) -> set_In v free -> 
    h ||- lexists x,
      (v |-> x ** 4 + v |-> INFREELIST ** top).
Proof.
  induction free; intros.
  inversion H0.
  unfold ifreeList in H.
  linLogElim H.
  compare v a; intro He.
  subst v.  
  exists x.
  rewrite  <- hpredTrueEq.
  apply (sconjImplR H).
  clear H; intros.
  apply (sconjImplR H).
  clear H h; intros.
  apply topIntro.
  assert (set_In v free).
  eapply set_add_elim2.
  apply H0.
  apply He.
  hypReplace H (sconjAssocL (sconjAssocL (sconjComm 
    (sconjAssocL (sconjComm H))))).
  fold ifreeList in H.
  eapply IHfree.
  apply H.
  apply H1.
Qed.

Lemma set_diff_add_r :
  forall A B C v,
    set_diff A B = C -> ~ set_In v A ->
    set_diff A (v::B) = C.
Proof.
  intros.
  subst C.
  assert (He := removeIsDiffStep v A B).
  unfold set_add in He.
  rewrite <- He.
  apply set_removeEq.
  intro.
  apply H0.
  destruct (set_diff_elim H) as [Ho _].
  assumption.
Qed.

Lemma set_InUnion_notinL_inR :
  forall A B v,
    set_In v (set_union A B) -> ~ set_In v A ->
    set_In v B.
Proof.
  intros A B v HinU HnotinA.
  destruct (set_union_elim HinU) as [HinA | HinB].
  elimtype False; apply HnotinA; apply HinA.
  assumption.
Qed.

Lemma set_InUnion_notinR_inL :
  forall A B v,
    set_In v (set_union A B) -> ~ set_In v B ->
    set_In v A.
Proof.
  intros A B v HinU HnotinB.
  destruct (set_union_elim HinU) as [HinA | HinB].
  assumption.
  elimtype False; apply HnotinB; apply HinB.
Qed.

Lemma In_objHdrs_Val :
  forall H A a v h,
    (h ||- objHdrs H a ** A) -> set_In v H ->
    Heap.get h (v - 4) = Some a.
Proof.
  unfold objHdrs.
  intros H A a v h HobjH Hin.
  eapply contGetEqL.
  apply (sconjAssocL (fasElimL Hin HobjH)).
Qed.

Lemma bobjHdrsIntroLt : forall s h m n p,
  (forall x, set_In x s -> x < p) ->
  h ||- objHdrs s m ->
    h ||- bobjHdrs s p m n.
Proof.
  unfold bobjHdrs; unfold objHdrs; unfold fas; intros.
  assert (forall x, set_In x (set_canon s) -> x < p).
  intros.
  apply H.
  apply (set_canon_elim _ _ H1).
  rename H1 into Hx.
  clear H.
  generalize h m n p H0 Hx.
  clear h m n p H0 Hx.
  induction (set_canon s); intros.
  apply H0.
  assert (IN : set_In a (a::s0)).
  left; reflexivity.
  simpl in H0.
  simpl.
  apply (sconjImpl H0); clear h H0; intros h Hh.
  rewrite blt_true.
  apply Hh.
  apply Hx; assumption.
  apply IHs0.
  exact Hh.
  intros.
  apply Hx; right; exact H.
Qed.

Lemma heap_joinEmptyL : forall h,
  heap_join empty_heap h h.
Proof.
  constructor.
  intros.
  assert (forall k, HeapMap.MapDef.get empty_heap k = None).
  intros.
  simpl.
  reflexivity.
  rewrite (H k).
  destruct (HeapMap.MapDef.get h k);
    constructor.
Qed.

Lemma objHdrsSplit : forall A B h n,
  h ||- objHdrs (set_union A B) n ->
    set_disj A B ->
  h ||- objHdrs A n ** objHdrs B n.
Proof.
  assert (F : forall A B n, set_canonP A -> 
    forall h,
      h ||- objHdrs (set_union A B) n->
	set_disj A B ->
	h ||- objHdrs A n ** objHdrs B n).
  induction 1; intros h Hp Hdisj.
  hypReplace Hp (unitIL Hp).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold objHdrs.
  apply fasIntroE; assumption.
  assumption.
  
  assert (IA : set_In a (set_union (set_add a s) B)).
  apply set_union_intro1; apply set_add_intro1; reflexivity.
  hypReplace Hp (objHdrsElim _ _ _ _ IA Hp); clear IA.
  assert (Hp' : h ||- (a - 4 |-> n) ** objHdrs (set_union s B) n).
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  replace (set_union (set_add a s) B) with 
    (set_add a (set_union s B)) in Hp.
  simpl in Hp.
  destruct (POnat.POnat.eqDec a a).
  rewrite not_In_removeEq in Hp.
  assumption.
  intro I.
  destruct (set_union_elim I).
  assert (a < a).
  apply H0; assumption.
  omega.
  apply (set_disj_elim Hdisj a).
  apply set_add_intro1; reflexivity.
  assumption.
  elimtype False; apply n0; reflexivity.
  reflexivity.
  clear Hp.
  assert (Hp : h ||- (a - 4 |-> n) ** objHdrs s n ** objHdrs B n).
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp.
  apply IHset_canonP.
  assumption.
  apply set_disj_intro; intros k I1 I2.
  apply (set_disj_elim Hdisj k).
  apply set_add_intro2; assumption.
  assumption.

  clear Hp'.
  hypReplace Hp (sconjAssocR Hp).
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply objHdrsIntro; assumption.

  intros A B h n Hp Hdisj.
  assert (R := set_canonEqv A).
  assert (Q := F (set_canon A) B n (set_canonCanonP A) h); clear F.
  assert (Hp' : h ||- objHdrs (set_canon A) n ** objHdrs B n).
  apply Q.
  rewrite objHdrsCEq; setoid_rewrite <- R; assumption.
  setoid_rewrite <- R; assumption.
  apply (sconjImplL Hp'); clear Hp; intros h2 Hp.
  rewrite objHdrsCEq; setoid_rewrite R; assumption.
Qed.

Lemma objHp_ifreelist_disj : 
  forall B N free p h objs x y A,
    objs = rangeMSObjs x y -> set_subset N objs ->
    h ||- ifreeList p free ** objHp N B ** A ->
      set_disj B free.
Proof.
  intros B N free p h objs x y A Hobjs Hsubset Hp.
  apply set_disj_intro.
  intros a HinB HinF.
  assert (HgetF :=  inFreeListVal _ _ _ _ _ Hp HinF).
  assert (Hgeta : get h (4 + a) = Some INFREELIST).
  linLogElim HgetF.
  eapply contGetEqL.
  apply (sconjAssocL (sconjComm HgetF)).
  unfold objHp in Hp.
  hypReplace Hp (sconjAssocL (sconjComm Hp)).
  hypReplace Hp (sconjAssocL (fasElimL HinB Hp)).
  simpl in Hp.
  unfold okObjAt in Hp; unfold okFieldAt in Hp.
  linLogElim Hp.
  hypReplace Hp (sconjAssocL (sconjComm Hp)).
  assert (Hgeta' : Heap.get h (4 + a) = Some x1).
  eapply contGetEqL.
  apply Hp.
  rewrite Hgeta in Hgeta'; inversion Hgeta'.
  subst x1; clear Hp H3 x0 Hgeta'.
  destruct H6 as [Hatom | Hin].
  inversion_clear Hatom.
  inversion_clear H.
  inversion_clear H0.
  assert (HinObjs := Hsubset _ Hin).
  assert (Htt := inFreeListnotinMSobjs _ _ _ Hobjs).
  apply Htt; trivial.
Qed.

Lemma objHp_ifreelist_disj_union : 
  forall B C M N free p h objs x y A,
    objs = rangeMSObjs x y -> set_subset M objs ->
    set_subset N objs ->
    h ||- ifreeList p free ** objHp M B ** objHp N C ** A ->
      set_disj (B |_| C) free.
Proof.
  intros B C M N free p h objs x y A Hobjs HsubM HsubN Hp.
  assert (BFdisj := objHp_ifreelist_disj _ _ _ _ _ _ _ _ _ Hobjs HsubM Hp).
  sconjLift3 Hp; sconjLift2 Hp.
  assert (CFdisj := objHp_ifreelist_disj _ _ _ _ _ _ _ _ _ Hobjs HsubN Hp).
  apply set_unionDisjIntro; assumption.
Qed.

Lemma ifreeList_empty_NULL : forall h freePtr A,
  h ||- ifreeList freePtr empty_set ** A ->
    freePtr = NULL.
Proof.
  simpl.
  intros h freePtr A Hp.
  linLogElim Hp.
  assumption.
Qed.

Lemma ifreeList_NULL_empty : forall h free A,
  h ||- ifreeList NULL free ** A ->
    free = empty_set.
Proof.
  intros h free A Hp.
  compare free empty_set.
  trivial.
  intro He.
  assert (Hobj := notEmptyIsCons He).
  destruct Hobj as [x [objs Hobjs]].
  subst free.
  unfold ifreeList in Hp.
  linLogElim Hp.
  elim H6; reflexivity.
  auto with arith.
Qed.

Lemma set_unionEmptyL : forall s, s = set_union empty_set s.
Proof.
  induction s;auto.
Qed.

Lemma bobjHdrsCEq : forall  A v m n,
  bobjHdrs A v m n = (foralls x in A,
    x - 4 |-> if blt_nat x v then m else n).
Proof.
  reflexivity.
Qed.

Lemma beq_rnum_false_neq : forall r r',
  beq_rnum r r' = false -> r <> r'.
Proof.
  intros r r' Hne He.
  subst.
  rewrite beq_rnum_refl in Hne.
  discriminate.
Qed.

Lemma HeapgetNeq : forall l l' s v,
  ptrP l -> ptrP l' -> l <> l' ->
  Heap.get (Heap.uset s l' v) l = Heap.get s l.
Proof.
  intros l l' s v Pl Pl' NE.
  unfold Heap.get, Heap.uset; simpl.
  destruct (POnat.POnat.eqDec (div4 l) (div4 l')).

  elf.
  destruct (ptrP_elim Pl) as [k Pll]; clear Pl;
    destruct (ptrP_elim Pl') as [k' Pll']; clear Pl';
      subst l l'.
  rewrite div44eq in e; rewrite div44eq in e.
  auto.

  reflexivity.
Qed.

Lemma ifreeList_cont_notIn : forall h free freePtr x v A,
  h ||- ifreeList freePtr free ** x |-> v ** A ->
    ~ set_In x free.
Proof.
  induction free; intros freePtr x v A Hp Hxin.
  apply Hxin. 
  unfold ifreeList in Hp; fold ifreeList in Hp.
  linLogElim Hp.
  subst freePtr.
  compare x a; intro He.
  subst a.
  sconjLift4 Hp.
  apply (sconjDisjL Hp).
  sconjLift4 Hp.
  sconjLift4 Hp.
  apply (IHfree x0 x v _ Hp).
  elimSetIn.
  elim He; reflexivity.
Qed.
  
Lemma objHp_cont_neq : forall h objs objs' x y v A,
  set_In x objs ->
  h ||- objHp objs' objs ** y |-> v ** A ->
    x <> y /\ 4 + x <> y.
Proof.
  intros h objs objs' x y v A Hin Hp.
  hypReplace Hp (sconjAssocL (fasElimL Hin Hp)).
  unfold okObjAt, okFieldAt in Hp.
  linLogElim Hp.
  sconjLift4 Hp.
  split; eapply sconjNeq; linLogIntro.
  sconjLift2 Hp.
  apply H.
  sconjLift3 Hp.
  apply H.
Qed.
  
Lemma in_rangeMsobj_ptrP : forall x y p,
  ptrP x -> ptrP y -> set_In p (rangeMSObjs x y) ->
  ptrP p.
Proof.
  intros x y p Px Py Hobjs.
  destruct (rangeMSObjsElim p x y Hobjs) as [_ [k Hep]].
  destruct (ptrP_elim Px) as [k0 Hex].
  subst x.
  eapply ptrP_intro.
  replace (4 * k0 + 4 + 12 * k) with (4 * (k0 + (1 + 3 * k))) in Hep.
  apply Hep.
  rewrite (mult_plus_distr_l 4 k0 (1 + 3 * k)).
  replace (4 * (1 + 3 * k)) with (4 + 12 * k).
  auto with arith.
  rewrite (mult_plus_distr_l 4 1 (3 * k)).
  rewrite (mult_1_r 4).
  replace (12 * k) with (4 * (3 * k)).
  reflexivity.
  rewrite (mult_assoc 4 3 k).
  compute.
  reflexivity.
Qed.

Lemma bobjHdrsIntro2 : forall h s x p m n,
  x < p ->
  h ||- x - 4 |-> m ** bobjHdrs s p m n ->
    h ||- bobjHdrs (set_add x s) p m n.
Proof.
  intros h s x p m n X H.
  unfold bobjHdrs.
  apply fasIntro.
  rewrite blt_true; assumption.
  intro I.
  hypReplace H (sconjImplR H _ (fun h => fasElim (h := h) I)).
  assert (h ||- ((x - 4|-> m ** x - 4|-> m) ** top)).
  apply (sconjImpl (sconjAssocR H)); trivial.
  rewrite blt_true; trivial.
  intros; linLogIntro.
  assert (x - 4 <> x - 4).
  linLogElim H0.
  apply H1; reflexivity.
Qed.

Ltac regRewrite1 R :=
  regRewrite R; [idtac | 
    rewrite beq_rnum_neq; [reflexivity | discriminate]].

Ltac regRewrite2 R :=
  regRewrite R; [idtac | 
    rewrite beq_rnum_neq; [reflexivity | discriminate] |
      rewrite beq_rnum_neq; [reflexivity | discriminate]].

Definition imarkP f :=
  if EqNat.beq_nat MARKFIELD f then
    SomeT (bspeci markFieldPre markFieldGuar)
    else if EqNat.beq_nat IMARKLOOP f then
      SomeT (bspeci imarkLoopPre imarkLoopGuar)
      else if EqNat.beq_nat IMARKLOOPBODY f then
	SomeT (bspeci imarkLoopBodyPre imarkLoopBodyGuar)
	else if EqNat.beq_nat STACKEMPTY f then
	  SomeT (bspeci isEmptyPre isEmptyGuar)
	  else if EqNat.beq_nat STACKPOP f then
	    SomeT (bspeci popPre popGuar)
	    else if EqNat.beq_nat IMARKFIRST f then
	      SomeT (bspeci imarkFirstPre imarkFirstGuar)
	      else if EqNat.beq_nat IMARKSECOND f then
		SomeT (bspeci imarkSecondPre imarkSecondGuar)
		else if EqNat.beq_nat IMARKSUMPLUS f then
		  SomeT (bspeci imarkSumPlusPre imarkSumPlusGuar)
		  else if EqNat.beq_nat IDONESOMEMARK f then
		    SomeT (bspeci imarkDoneSomeMarkPre imarkDoneSomeMarkGuar)
		    else if EqNat.beq_nat IMARKFINISHED f then
		      SomeT (bspeci imarkMarkFinishedPre imarkMarkFinishedGuar)
		      else
			NoneT _.

Ltac PsimplIMark := 
  unfold imarkP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

Definition isweepP f :=
  if EqNat.beq_nat ISWEEPLOOP f then
    SomeT (bspeci isweepLoopPre isweepLoopGuar)
    else if EqNat.beq_nat ISWEEPADD f then
      SomeT (bspeci isweepAddPre isweepAddGuar)
      else if EqNat.beq_nat ISWEEPNEXT f then
	SomeT (bspeci isweepNextPre isweepNextGuar)
	else if EqNat.beq_nat ISWEEPFIN f then
	  SomeT (bspeci isweepFinishedPre isweepFinishedGuar)
	  else if EqNat.beq_nat IDONESOMESWEEP f then
	    SomeT (bspeci isweepDoneSomePre isweepDoneSomeGuar)
	    else NoneT _.

Ltac PsimplISweep := 
  unfold isweepP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

Ltac iseqWeakSolverlg Psolver := 
  match goal with
    | |- iseqOkS ?P _ _ _ => 
      unfold bne; unfold beq;
      eapply UseWeakIseq;
	[repeat (iseqWeakSolver' Psolver P) | idtac]
  end.

Definition imarkSweepGCP f :=
  if EqNat.beq_nat ISWEEP f then
    SomeT (bspeci isweepPre isweepGuar)
    else if EqNat.beq_nat IMARK f then
      SomeT (bspeci imarkPre imarkGuar)
      else if EqNat.beq_nat RETURN1 f then
	SomeT (bspeci return1Pre return1Guar)
	else if EqNat.beq_nat MARKFIELD f then
	  SomeT (bspeci markFieldPre markFieldGuar)
	  else if EqNat.beq_nat GCINITRETURN f then
	    SomeT (bspeci igcInitRetPre igcInitRetGuar)
	    else
	      NoneT _.

Ltac PsimplimarkSweepGC :=
  unfold imarkSweepGCP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

Definition imarkSweepNewObjP f :=
  if EqNat.beq_nat RETURN0 f then
    SomeT (bspeci return0Pre return0Guar)
    else if EqNat.beq_nat DOGC f then
      SomeT (bspeci imarkSweepGCPre imarkSweepGCGuar)
      else if EqNat.beq_nat NEWOBJNEW f then
	SomeT (bspeci newObjNewPre newObjNewGuar)
	else
	  NoneT _.

Ltac PsimplimarkSweepNewObj :=
  unfold imarkSweepNewObjP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.
