Require Import OptUtil.
Require Import StateTheory.
Require Import EqNat.
Require Import WeakSCAP.
Require Import SpecUtil.
Require Import VGC.
Require Import LinkedList.


Set Implicit Arguments.
Unset Strict Implicit.


Module ListRevGuar <: BASIC_MGI_ARG.

  Definition gcGuar vs vs' rr :=
    (forall r, In r rr -> rget vs r = NULL -> rget vs' r = NULL) /\
    (forall l1 l2 r1 r2,
      vs |||- NatListAt l1 (rget vs r1) ** NatListAt l2 (rget vs r2) ** top ->
	incl (r1::r2::nil) rr ->
	vs' |||- NatListAt l1 (rget vs' r1) ** NatListAt l2 (rget vs' r2) ** top).

End ListRevGuar.

Module ListRevMGI := BasicMGI (ListRevGuar).


(*
List rev mutator implementation

 list_rev(i, j) {
   j = null;
   while (i <> null){
     k = i->next;
     i->next = j;
     j = i;
     i = k;
   }
 }

*)


Definition REV := 0.
Definition LOOP1 := 1.
Definition LOOP2 := 2.
Definition LOOP3 := 3.
Definition FIN := 4.


Notation READ2  := ListRevMGI.read2Lbl.
Notation WRITE2 := ListRevMGI.write2Lbl.


Notation ri := t0.
Notation rj := s0.

Definition revEnter :=
  addiu rj zero NULL;
  j LOOP1.

Definition loop1 :=
  addiu ast zero NULL;
  beq ri ast FIN;;
  jal READ2 LOOP2.

Definition loop2 :=
  addu a0 rj zero;
  jal WRITE2 LOOP3.

Definition loop3 :=
  addiu rj ri 0;
  addiu ri v0 0;
  j LOOP1.

Definition fin :=
  jr ra.


Module MutatorVerif (V : VIEW_GC with Module I := ListRevMGI).

  Definition revPre s :=
    exists l, exists vs,
      s |||- V.mview vs s (ri::nil) ** top /\
      vs |||- NatListAt l (rget vs ri) ** top.

  Definition revGuar s s' :=
    forall vs A l,
      s |||- V.mview vs s (ri::nil) ** A ->
	vs |||- NatListAt l (rget vs ri) ** top ->
          exists vs',
            s' |||- V.mview vs' s' (rj::nil) ** A /\
	      vs' |||- NatListAt (rev l) (rget vs' rj) ** top.

  Definition loopPre s :=
    exists l1, exists l2, exists vs,
      s |||- V.mview vs s (rj::ri::nil) ** top /\
	vs |||- NatListAt l1 (rget vs ri) ** NatListAt l2 (rget vs rj) ** top.
  
  Definition loopGuar s s' :=
    forall vs A l1 l2, 
      s |||- V.mview vs s (rj::ri::nil) ** A ->
	vs |||- NatListAt l1 (rget vs ri) ** NatListAt l2 (rget vs rj) ** top ->
	  exists vs', 
            s' |||- V.mview vs' s' (rj::ri::nil) ** A /\
	      vs' |||- NatListAt ((rev l1)++l2) (rget vs' rj) ** top.

  Definition loop3Pre s :=
    loopPre (rset (rset s rj (rget s ri)) ri (rget s v0)).

  Definition loop3Guar s s' :=
    loopGuar (rset (rset s rj (rget s ri)) ri (rget s v0)) s'.

  Notation argSpecAdder := (fun sp =>
    sp V.mview ListRevMGI.rroot ListRevMGI.gcGuar ListRevMGI.presReg).

  Notation readPre := (readPre V.mview ListRevMGI.rroot).
  Notation readGuar := (argSpecAdder readGuar).
  Notation writePre := (writePre V.mview ListRevMGI.rroot).
  Notation writeGuar := (argSpecAdder writeGuar).
  Notation write2Guar := (writeGuar 4).
  Notation read2Spec := (argSpecAdder read2Spec).
  Notation write2Spec := (argSpecAdder write2Spec).


  Definition loop2Pre s :=
    writePre 4 (rset s a0 (rget s rj)) /\
    forall s', write2Guar (rset (rset s a0 (rget s rj)) ra LOOP3) s' -> loop3Pre s'.

  Definition loop2Guar s s'' :=
    exists s',
      write2Guar (rset (rset s a0 (rget s rj)) ra LOOP3) s' /\
      loop3Guar s' s''.

  Definition finPre (s:state) := True.

  Definition finGuar (s s':state) := s = s'.

  Definition revSpec := bspeci revPre revGuar.
  Definition loopSpec := bspeci loopPre loopGuar.
  Definition loop2Spec := bspeci loop2Pre loop2Guar.
  Definition loop3Spec := bspeci loop3Pre loop3Guar.
  Definition finSpec := bspeci finPre finGuar.
  
  Definition mPhi (f:nat) :=
    if beq_nat REV f then
      SomeT revSpec
    else if beq_nat LOOP1 f then
      SomeT loopSpec
    else if beq_nat LOOP2 f then
      SomeT loop2Spec
    else if beq_nat LOOP3 f then 
      SomeT loop3Spec
    else if beq_nat FIN f then
      SomeT finSpec
    else if beq_nat READ2 f then 
      SomeT read2Spec
    else if beq_nat WRITE2 f then
      SomeT write2Spec
    else NoneT _.

  Ltac specUnfolder := 
    unfold revSpec;
    unfold loopSpec;
    unfold loop2Spec;
    unfold loop3Spec;
    unfold finSpec;
    unfold VGC.read2Spec;
    unfold VGC.write2Spec.
  
  Ltac Psimpl :=
    unfold mPhi;
    unfold VGC.read2Spec;
    unfold VGC.write2Spec;
    rewrite beq_true;
    repeat (rewrite beq_false; [idtac | discriminate]);
    specUnfolder; reflexivity.
  
  Ltac nonRootSolver :=
    sconjMatch;
    apply V.setNonRootOk;
    [try assumption |
      let X := fresh in
        (simpl; intro X; destruct X as [X | X]; [discriminate X | intuition])].

  (* verifying the mutator *)

  Lemma revOk : iseqOkS mPhi revPre revGuar revEnter.
  Proof.
    unfold revEnter.
    iseqWeakSolver Psimpl.

    cbv beta.
    intros s [l [vs [V Hp]]].
    rewrite addiuStep.
    stSimpl; simpl.
    unfold revPre; unfold revGuar; unfold loopPre; unfold loopGuar.
    simpl; stSimpl.
  
    constructor.

    destruct s as [h rf].
    exists l; exists (nil (A:=nat)); 
      exists (rset vs rj (RFile.get (RFile.set rf rj 1) rj)); constructor.
    apply V.addAtomRoot with (r := rj) (rf := (RFile.set rf rj 1)).
    nonRootSolver.
    discriminate.
    simpl; intuition; discriminate.
    compute; constructor; constructor.
    unfold NatListAt; stSimpl; simpl.
    linLogIntro.
    exact Hp.

    clear l vs V Hp.
    intros s' F vs' A l V' Hp'.
    destruct s as [h rf].
    destruct (F (rset vs' rj (RFile.get (RFile.set rf rj 1) rj))
      A l (nil (A:=nat))) as [vs'' [V'' E]]; clear F.
    apply V.addAtomRoot with (r := rj) (rf := (RFile.set rf rj 1)).
    nonRootSolver.
    discriminate.
    simpl; intuition; discriminate.
    compute; constructor; constructor.
    unfold NatListAt; stSimpl; simpl.
    linLogIntro.
    exact Hp'.
    
    exists vs''; constructor.
    sconjMatch.
    assert (Ic : incl (rj::nil) (rj::ri::nil)).
    intros a [I | I]; try (subst a); intuition.
    apply (V.rootWeaken Hp Ic).
    rewrite <- app_nil_end in E.
    exact E.
  Qed.

  Lemma loop3Ok : iseqOkS mPhi loop3Pre loop3Guar loop3.
  Proof.
    unfold loop3.
    iseqWeakSolver Psimpl.
    
    cbv beta.
    intro s.
    rewrite addiuStep.
    rewrite addiuStep.
    unfold loop3Pre; unfold loop3Guar.
    auto.
  Qed.

  Lemma writeRApres : forall k s s', 
    writePre k s -> writeGuar k s s' ->
    rget s ra = rget s' ra.
  Proof.
    unfold VGC.writeGuar.
    intros k s s' _ [_ G].
    rewrite (regPresEq G).
    reflexivity.
    constructor; reflexivity.
  Qed.

  Lemma loop2Ok : iseqOkS mPhi loop2Pre loop2Guar loop2.
  Proof.
    unfold loop2.
    iseqWeakSolver Psimpl.
    apply writeRApres.

    cbv beta.
    intro s.
    rewrite adduStep.
    stSimpl.
    replace (rget s rj + 0) with (rget s rj); [idtac | auto].
    unfold loop2Pre; unfold loop2Guar.
    intuition.
    destruct H0 as [vs [Hp Vhp]]; exists vs; constructor.
    simpl; simpl in Hp; nonRootSolver.
    discriminate.
    discriminate.
    assumption.
  Qed.

  Lemma finOk : iseqOkS mPhi finPre finGuar fin.
  Proof.
    unfold fin.
    iseqWeakSolver Psimpl.
    intuition.
  Qed.

  Lemma readRApres : forall k s s', 
    readPre k s -> readGuar k s s' ->
    rget s ra = rget s' ra.
  Proof.
    unfold VGC.readGuar.
    intros k s s' _ [_ G].
    rewrite (regPresEq G).
    reflexivity.
    split.
    constructor; reflexivity.
    discriminate.
  Qed.

  Lemma app_cons_assoc : forall (A : Set) (l1 l2 : list A) (x : A),
    l1 ++ x :: l2 = (l1 ++ x :: nil) ++ l2.
  Proof.
    induction l1; intros.
    reflexivity.
    simpl.
    rewrite IHl1.
    reflexivity.
  Qed.

  Lemma loop1Ok : iseqOkS mPhi loopPre loopGuar loop1.
  Proof.
    unfold loop1.
    iseqWeakSolver Psimpl.
    apply readRApres.

    assert (Iri : In ri (rj::ri::nil)).
    right; left; reflexivity.

    assert (NIast : ~ In ast (V.I.rgcInfo::rj::ri::nil)).
    intros [I | [I | [I | I]]]; auto; discriminate.

    intros s.  
    unfold loopPre.
    rewrite addiuStepZero.
    stSimpl.
    simpl.

    compare (rget s ri) NULL.

    (* case: list is empty *)
    unfold loopGuar.
    intro R; rewrite R; rewrite beq_true.
    destruct (beq_nat (rget s rj) NULL).
    intros [l1 [l2 [vs [Hs Hv]]]].
    constructor.
    unfold finPre; tauto.
    clear Hs Hv vs l1 l2.
    unfold finGuar.
    intros s' E; subst s'.
    stSimpl.
    intros vs' A l1 l2 V' Hp'.
    exists vs'; constructor.
    sconjMatch.
    apply (V.setNonRootOk NULL Hp NIast).

    rewrite <- (V.atomRegEq _ V' Iri) in Hp';
      unfold rget in R.
    rewrite R in Hp'.
    destruct l1; unfold NatListAt in Hp'; simpl in Hp'; linLogElim Hp'; 
      intuition.
    left; rewrite R.
    unfold NULL; constructor; constructor.

    intros [l1 [l2 [vs [Hs Hv]]]].
    constructor.
    unfold finPre; tauto. 
    clear Hs Hv vs l1 l2.
    unfold finGuar.
    intros s' E; subst s'.
    stSimpl.
    intros vs' A l1 l2 V' Hp'.
    exists vs'; constructor.
    sconjMatch.
    apply (V.setNonRootOk NULL Hp NIast).
    rewrite <- (V.atomRegEq _ V' Iri) in Hp';
      unfold rget in R.
    rewrite R in Hp'.
    destruct l1; unfold NatListAt in Hp'; simpl in Hp'; linLogElim Hp'; 
      intuition.
    left; rewrite R.
    unfold NULL; constructor; constructor.
    
    (* case: list isn't empty *)
    intro Ne; rewrite (beq_false Ne).
    intros [l1 [l2 [vs [Hs Hv]]]]; apply callChain;
      unfold loop2Pre; unfold loop2Guar;
	unfold loop3Pre; unfold loop3Guar;
	  unfold loopPre; unfold loopGuar.
    
    assert (V := Hv).
    destruct l1; unfold NatListAt in Hv; simpl in Hv; linLogElim Hv;
      fold NatListAt in Hv.
    rewrite <- (V.atomRegEq _ Hs Iri) in H1.
    unfold rget in Ne; rewrite H1 in Ne.
    elimtype False; apply Ne; reflexivity.
    right; rewrite H1; unfold NULL; constructor; constructor.

    repeat split.
    exists vs; constructor; simpl.
    repeat nonRootSolver.
    assert (Ic : incl (ri::nil) (rj::ri::nil)).
    intros a [I | I]; try (subst a); intuition.
    apply (V.rootWeaken Hp Ic).
    discriminate.
    discriminate.
    apply lexIL; apply lexI with (x := x).
    sconjStep' 2.
    
    destruct H as [G1 G2].
    destruct (G1 vs top (rj::ri::nil)) as [vs' [G3 [V' Hp']]].
    simpl; repeat nonRootSolver.
    discriminate.
    discriminate.
    discriminate.
    discriminate.
    right; left; trivial.
    simpl; intros r [I | [I | I]]; 
      [right; right; left; auto | left; rewrite <- I; reflexivity | elf; assumption].

    exists (rset vs' a0 (rget vs' rj)); split.
    assert (Ic : incl (ri::a0::nil) (a0::v0::rj::ri::nil)).
    intros a [I | [I | I]]; (try subst a); intuition.
    stSimpl; sconjMatch.
    refine (V.rootWeaken _ Ic).
    replace (rfileOf (rset s' a0 (rget s' rj))) with 
      (RFile.set (rfileOf s') a0 (RFile.get (rfileOf s') rj)); [idtac | reflexivity].
    apply V.rootCopyOk.
    assumption.
    right; left; reflexivity.
    simpl; intros [E | [E | E]]; try (discriminate E); elf; assumption.
    linLogIntro; apply Hp'.

    intros s'' [G2 GR2].
    destruct H as [G1 GR1].
    destruct (G1 vs top (rj::ri::nil)) as [vs' [G3 [V' Hv']]]; clear G1.
    simpl; repeat (nonRootSolver; try discriminate).
    right; left; trivial.
    intros r [I | [I | I]].
    right; right; left; symmetry; assumption.
    left; symmetry; assumption.
    elf; auto.

    destruct (G2 (rset vs' a0 (rget vs' rj)) top (a0::v0::rj::ri::nil)) as [vs'' [G4 V'']]; clear G2.
    stSimpl; scm.
    simpl; nonRootSolver; try discriminate.
    replace (rget s' rj) with (RFile.get (rfileOf s') rj); [idtac | reflexivity].
    apply V.rootCopyOk.
    assumption.
    right; left; reflexivity.
    intros [E | [E | E]]; [discriminate E | discriminate E | elf; assumption].
    intros a [I | [I | I]]; try (subst a); intuition.
    exists l1; exists (n::l2).
    stSimpl.
    exists (rset (rset vs'' rj (rget vs'' ri)) ri (rget vs'' v0)); split.
    destruct s'' as [h rf]; sconjMatch.

    assert (Ic : In ri (a0::v0::rj::ri::nil)).
    simpl; intuition.
    assert (N : ~In rj (zero :: V.I.rgcInfo :: nil)).
    simpl; intuition; discriminate.
    hypReplace Hp (V.rootCopyOk _ Hp Ic N).
    clear Ic N; assert (Ic : In v0 (rj :: a0 :: v0 :: rj :: ri :: nil)).
    simpl; intuition.
    assert (N : ~ In ri (zero :: V.I.rgcInfo :: nil)).
    simpl; intuition; discriminate.
    hypReplace Hp (V.rootCopyOk _ Hp Ic N).
    clear Ic N; assert (Ic : incl (rj::ri::nil) (ri::rj::a0::v0::rj::ri::nil)). 
    intros a [I | [I | I]]; (try subst a); intuition.
    generalize (V.rootWeaken Hp Ic).
    stSimpl; auto.

    destruct G3 as [_ G3].
    destruct G4 as [_ G4].
    assert (Ic : incl (ri::rj::nil) (rj::ri::nil)). 
    intros a [I | [I | I]]; (try subst a); intuition.
    hypReplace Hv (G3 _ _ _ _ V Ic); clear Ic G3.
    generalize G4; stSimpl; clear G4; intro.
    unfold NatListAt in Hv; simpl in Hv; linLogElim Hv.
    apply G4; clear G4; stSimpl; simpl.
    sconjLift2 Hv.
    assert (x0 = rget vs' v0).
    apply (contContEq Hv Hv').
    subst x0.
    hypReplace Hv (contUpdateSt Hv (rget vs' rj)).
    unfold NatListAt; simpl; linLogIntro.
    sconjLift2 Hv; sconjLift3 H; apply H0.
    intros a [I | [I | I]]; (try subst a); intuition.

    rename s' into s1.
    intros s'' [s2 [G2 G3]].
    clear l1 l2 V Hs n H3 Hv vs x.
    rename H into G1.
    intros vs A l1 l2 Hs Hv.

    assert (V := Hv).
    destruct l1; unfold NatListAt in Hv; simpl in Hv; linLogElim Hv;
      fold NatListAt in Hv.
    assert (I : In ri (rj::ri::nil)).
    right; left; reflexivity.
    rewrite <- (V.atomRegEq _ Hs I) in H1.
    unfold rget in Ne; rewrite H1 in Ne.
    elimtype False; apply Ne; reflexivity.
    right; rewrite H1; unfold NULL; constructor; constructor.

    destruct G1 as [G1 GR1].
    rename G3 into G.
    destruct (G1 vs A (rj::ri::nil)) as [vs' [G3 [V' Hv']]]; clear G1.
    simpl; repeat (nonRootSolver; try discriminate).
    right; left; trivial.
    intros r [I | [I | I]]; 
      [right; right; left | left | elf]; auto.

    destruct G2 as [G2 GR2].
    destruct (G2 (rset vs' a0 (rget vs' rj)) A (a0::v0::rj::ri::nil)) as [vs'' [G4 V'']]; clear G2.
    stSimpl; scm.
    replace (rfileOf (rset (rset s1 a0 (rget s1 rj)) ra LOOP3)) with
      (RFile.set (RFile.set (rfileOf s1) a0 (RFile.get (rfileOf s1) rj)) ra LOOP3); 
      [idtac | reflexivity].
    nonRootSolver; try discriminate.
    apply V.rootCopyOk.
    assumption.
    right; left; reflexivity.
    intros [I | [I | I]]; try discriminate; intuition.
    intros a [I | [I | I]]; try (subst a); intuition.

    replace (rev (n::l1)++l2) with (rev l1 ++ n::l2).
    apply G with (vs := (rset (rset vs'' rj (rget vs'' ri)) ri (rget vs'' v0)))
      (A := A) (l1 := l1) (l2 := n::l2).
    destruct s2 as [h rf]; sconjMatch.
    assert (Ic : In ri (a0::v0::rj::ri::nil)).
    simpl; intuition.
    assert (N : ~ In rj (zero :: V.I.rgcInfo :: nil)).
    simpl; intuition; discriminate.
    hypReplace Hp (V.rootCopyOk _ Hp Ic N).
    clear Ic N; assert (Ic : In v0 (rj::a0::v0::rj::ri::nil)). 
    simpl; intuition.
    assert (N : ~ In ri (zero :: V.I.rgcInfo :: nil)).
    simpl; intuition; discriminate.
    hypReplace Hp (V.rootCopyOk _ Hp Ic N).
    clear Ic N; assert (Ic : incl (rj::ri::nil) (ri::rj::a0::v0::rj::ri::nil)). 
    intros a [I | [I | I]]; (try subst a); intuition.
    generalize (V.rootWeaken Hp Ic).
    stSimpl; auto.

    destruct G3 as [_ G3].
    destruct G4 as [_ G4].
    assert (Ic : incl (ri::rj::nil) (rj::ri::nil)). 
    intros a [I | [I | I]]; (try subst a); intuition.
    hypReplace Hv (G3 _ _ _ _ V Ic); clear Ic G3.
    generalize G4; stSimpl; clear G4; intro.
    unfold NatListAt in Hv; simpl in Hv; linLogElim Hv.
    apply G4; clear G4; stSimpl; simpl.
    sconjLift2 Hv.
    assert (x0 = rget vs' v0).
    apply (contContEq Hv Hv').
    subst x0.
    hypReplace Hv (contUpdateSt Hv (rget vs' rj)).
    unfold NatListAt; simpl; linLogIntro.
    instantiate (1 := rget vs' rj).
    sconjLift2 Hv; sconjLift3 Hv; assumption.
    intros a [I | [I | I]]; (try subst a); intuition.

    simpl; rewrite <- app_cons_assoc; reflexivity.
  Qed.

End MutatorVerif.



(** Showing the NullGC is compatible.  *)
Require Import NullGC.


Module NullGCCompat <: MG_SUB_INTERFACE.
  Module I := NullGCMGI.
  Module I' := ListRevMGI.


  Lemma gcGuarSub : forall s s' rr,
    I.gcGuar s s' rr -> I'.gcGuar s s' rr.
  Proof.
    unfold I.gcGuar; unfold NullGCGuar.gcGuar.
    intros s s' rr [G Req].
    unfold I'.gcGuar; unfold ListRevGuar.gcGuar.
    constructor.
    intros r I E1; rewrite <- Req; assumption.
    intros l1 l2 r1 r2 Hp II.
    rewrite <- Req; [idtac | apply II; simpl; auto].
    rewrite <- Req; [idtac | apply II; simpl; auto].
    apply sconjAssocL.
    hypReplace Hp (sconjAssocR Hp).
    destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]]; clear Hp.
    eapply sconjInsert.
    2:eauto.

    assert (Hp' : exists objs, exists objs',
      ha ||- minObjHp ((rget s r1::nil) |_| (rget s r2::nil)) (objs |_| objs')).
    assert (Hp' : 
      ha ||- (lexists objs, minObjHp (rget s r1::nil) objs) **
  	     (lexists objs, minObjHp (rget s r2::nil) objs)).
    sconjMatch; eapply natListIsMinObjHp; eauto.
    linLogElim Hp'; econstructor; econstructor; apply minObjHpMerge; eauto.

    destruct Hp' as [objs [objs' Hp']]; simpl in Hp'.
    apply (G (objs |_| objs') ha (r1::r2::nil) II); clear G.
(*    intros r Ir.
    assert (s |||- (!(okFieldVal (objs |_| objs') (rget s r)) ** top) ** top).
    sconjMatch.
    linLogIntro; eapply minObjHpRootOk; eauto.
    simpl; simpl in Ir.
    destruct Ir as [X | [X | X]]; subst; auto.
    linLogElim H; assumption.*)

    eapply sconjIntro.
    3:eauto.
    apply aconjIntro.
    assumption.
    intro; rewrite Heap.HeapMap.map_canonEqv; reflexivity.
    linLogIntro.

  Qed.

    
  Lemma rootEq : I.rroot = I'.rroot.
  Proof.
    reflexivity.
  Qed.

  Lemma gcInfoEq : I.rgcInfo = I'.rgcInfo.
  Proof.
    reflexivity.
  Qed.

  Lemma nullEq : I.NULL = I'.NULL.
  Proof.
    reflexivity.
  Qed.

  Lemma presRegSub : forall r,
    I'.presReg r -> I.presReg r.
  Proof.
    intros; assumption.
  Qed.

  Lemma atomPSub : forall n,
    I'.atomP n -> I.atomP n.
  Proof.
    intros; assumption.
  Qed.

  Lemma gcPhiDomSub : set_subset I.gcPhiDom I'.gcPhiDom.
  Proof.
    intro; auto.
  Qed.

  Lemma allocLblEq : I.allocLbl = I'.allocLbl.
  Proof.
    reflexivity.
  Qed.

  Lemma read1LblEq : I.read1Lbl = I'.read1Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma read2LblEq : I.read2Lbl = I'.read2Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma write1LblEq : I.write1Lbl = I'.write1Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma write2LblEq : I.write2Lbl = I'.write2Lbl.
  Proof.
    reflexivity.
  Qed.


End NullGCCompat.


Module ListRevNullGC := CoerceGC NullGCImpl NullGCCompat.


Module ListRevGCed := MutatorVerif ListRevNullGC.




(** Showing that the Cheney collector is compatible *)

Require Import CheneyGC.
Require Import CheneyLemma.


Module CheneyGCCompat <: MG_SUB_INTERFACE.
  Module I := CheneyGCMGI.
  Module I' := ListRevMGI.

  Require Import FnMap.


  Lemma natListAtFwd : forall l h h' x x' phi objs objs',
    h ||- NatListAt l x ->
    h ||- objHp objs objs ->
    h' ||- (foralls x in objs', objIsFwdedH phi h x) ->
    iso_set phi objs objs' ->
    mapX phi x = Some x' ->
    h' ||- NatListAt l x'.
  Proof.
    assert (NullOdd : odd NULL).
    unfold NULL; constructor; constructor.

    unfold NatListAt.
    induction l; intros h h' x x' phi objs objs' Hp HpO Hp' Iso Meq.
    
    (* base case *)
    simpl in Hp; lle Hp; subst x.
    unfold mapX in Meq; rewrite isOddTrue in Meq.
    injection Meq; clear Meq; intro; subst x'.
    destruct objs.
    destruct objs'.
    simpl; lli.
    apply (fasElimE Hp').
    reflexivity.
    elimtype False; clear Hp HpO Hp'.
    destruct Iso as [Iso [_ [_ Iso']]].
    destruct (Iso' a) as [x Peq]; clear Iso'.
    left; reflexivity.
    destruct (Iso x) as [_ Q]; apply Q; rewrite Peq; discriminate.
    elimtype False; clear Hp' Iso.

    assert (I : set_In a (a::objs)).
    left; reflexivity.

    hypReplace HpO (fasElim I HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO; lle HpO.
    eapply unitContElim; eauto.
    assumption.


    (* inductive case *)
    simpl in Hp; lle Hp; clear H1.
    unfold mapX in Meq; rewrite isOddFalse in Meq.
    assert (I' : set_In x' objs').
    destruct Iso as [_ [Iso _]].
    eapply Iso; apply Meq.

    hypReplace Hp' (fasElim I' Hp'); cbv beta in Hp'; unfold objIsFwdedH at 1 in Hp'; lle Hp'.
    destruct H3 as [y [v0 [v1 [Peq [HpY [Meq0 Meq1]]]]]].
    lle HpY.
    assert (x = y).
    destruct Iso as [_ [_ [X _]]]; eapply X; try (instantiate (1 := x')); assumption.
    subst y.
    assert (U := contContEq HpY Hp).
    assert (v1 = x0).
    sconjLift2 HpY; sconjLift2 Hp; exact (contContEq HpY Hp).
    subst v0 v1; clear HpY.
    unfold mapX in Meq0; rewrite isOddTrue in Meq0.
    injection Meq0; clear Meq0; intro; subst x1.
    simpl; lli.
    intro; subst x'.
    apply (not_even_and_odd NULL).
    apply ptrPeven; eapply contElim1L; apply Hp'.
    assumption.

    instantiate (1 := x2).
    clear Peq.
    rename x0 into y; rename x2 into y'.
    hypReplace Hp (sconjAssocR Hp).

    assert (I : set_In x objs).
    destruct Iso as [Iso _]; destruct (Iso x).
    apply H0; rewrite Meq; discriminate.
    
    hypReplace HpO (objHpElim I HpO); unfold okObjAt in HpO; unfold okFieldAt in HpO;
      lle HpO; hypReplace HpO (sconjAssocR HpO).
    clear H3 H6.
    sconjLift2of2 Hp; hypReplace Hp (sconjExtract Hp); destruct Hp as [ha [Hp HpA]].
    sconjLift2of2 HpO; hypReplace HpO (sconjExtract HpO); destruct HpO as [ha' [HpO HpA']].
    sconjLift2of2 Hp; sconjLift2of2 HpO.

    assert (Iso' : iso_set (map_remove _ _ eqDecB phi x) (set_remove x objs) (set_remove x' objs')).
    apply iso_setRemove; assumption.

    cut (Heap.heap_eqv ha ha').
    intro HeqA.
    
    cut (ha ||- objHp (set_remove x objs) (set_remove x objs)).
    intro HpA''.

    cut (h' ||- x' |-> natEnc a ** 4 + x' |-> y' ** 
      (foralls z in set_remove x' objs', objIsFwdedH (map_remove _ _ eqDecB phi x) ha z)).

    intro Hp'B; clear Hp'; rename Hp'B into Hp'.
    sconjMatch.
    eapply IHl.
    apply HpA.
    apply HpA''.
    apply Hp0.
    assumption.
    unfold mapX; unfold mapX in Meq1; destruct (isOdd y).
    assumption.
    unfold map_remove; unfold eqDecB; destruct (POnat.POnat.eqDec x y).
    elimtype False.
    2:assumption.
    subst y.
    assert (X : h ||- (x |-> (natEnc a) ** top) ** (x |->__ ** top)).
    sconjMatch.
    destruct l.
    simpl in HpA; lle HpA; subst x; lle HpO.
    elimtype False; apply (not_even_and_odd NULL).
    apply ptrPeven; eapply contElim1L; eauto.
    assumption.
    simpl in HpA; lle HpA.
    eapply heapEqvTrue.
    apply Heap.heap_eqvSymm; apply eqvPredEqv; eauto.
    lli; sconjMatch.
    lle X.

    (* forward on the smaller phi *)
    sconjMatch.
    apply (fasImplI Hp0); clear Hp0.
    clear IHl.

    unfold objIsFwdedH; intros x2 h1 Ix Hp0.
    lle Hp0.
    rename Meq1 into Meq1'.
    destruct H1 as [z [v0 [v1 [Peq' [HpZ [Meq0 Meq1]]]]]].
    lli.
    2:eauto.
    exists z; exists v0; exists v1.
    clear h1 Hp0.

    assert (Iz : set_In z (set_remove x objs)).
    destruct Iso' as [Iso' [_ [_ Iso'']]].
    destruct (Iso'' _ Ix) as [z' PeqZ].
    assert (z' = z).
    destruct Iso as [_ [_ [X _]]]; eapply X.
    2:eauto.
    unfold map_remove in PeqZ; unfold eqDecB in PeqZ;
      destruct (POnat.POnat.eqDec x z').
    discriminate PeqZ.
    assumption.
    subst z'.
    destruct (Iso' z) as [_ Q]; apply Q.
    rewrite PeqZ; discriminate.
    
    unfold map_remove; rewrite Peq'.

    assert (NE : z <> x).
    exact (set_remove_elim2 _ Iz).

    unfold eqDecB; destruct (POnat.POnat.eqDec x z).
    elimtype False; auto.
    constructor.
    reflexivity.
    lle HpZ.

    cut (ha ||- z |->, (v0 :: v1 :: nil) ** top).
    intro HA.
    constructor.
    assumption.
    lle HA.
    hypReplace HpA'' (objHpElim Iz HpA'').
    unfold okObjAt in HpA''; unfold okFieldAt in HpA''; lle HpA''.
    assert (U := contContEq HA HpA'').
    assert (v1 = x6).
    sconjLift2 HA; sconjLift2 HpA''; exact (contContEq HA HpA'').
    subst x5 x6.
    unfold mapX in Meq0, Meq1 |- * .
    constructor.
    assert (U := isOddOk v0).
    destruct (isOdd v0).
    assumption.
    destruct H3.
    elimtype False; apply (not_even_and_odd v0); assumption.
    destruct (POnat.POnat.eqDec x v0).
    2:assumption.
    assert (X := set_remove_elim2 _ H); elimtype False; auto.
    assert (U := isOddOk v1).
    destruct (isOdd v1).
    assumption.
    destruct H6.
    elimtype False; apply (not_even_and_odd v1); assumption.
    destruct (POnat.POnat.eqDec x v1).
    2:assumption.
    assert (X := set_remove_elim2 _ H); elimtype False; auto.

    (* show that this part of ha is the same as h *)
    lli.
    lle HpO.

    hypReplace HpA'' (objHpElim Iz HpA''); unfold okObjAt in HpA'';
      unfold okFieldAt in HpA''; lle HpA''.
    assert (Sub : Heap.subHeap ha h).
    sconjLift2of2 Hp; exact (sconjEqSubHeap Hp).
    assert (Q := subHeapContEq Sub HpZ HpA'').
    sconjLift2 HpZ; sconjLift2 HpA''.
    assert (Q' := subHeapContEq Sub HpZ HpA'').
    subst x5 x6.
    sconjLift2 HpA''; sconjMatch.
    hypReplace HpA (natListIsMinObjHp _ _ _ HpA).
    lle HpA.
    unfold minObjHp in HpA; destruct (aconjElim HpA) as [_ HpA''].
    unfold okObjHp in HpA''; lle HpA''.
    
    assert (H11 : ha ||- objHp objs (set_remove x objs)).
    eapply heapEqvTrue; eauto.

    assert (X := objHpEqObjs _ _ _ _ _ HpA'' H11).
    unfold hpredTrue; rewrite <- X; assumption.

    clear IHl Hp' HpA' HpA Iso' Meq1 Meq I I'.
    lle Hp; lle HpO.
    assert (X := contContEq Hp HpO).
    subst x0.
    assert (y = x1).
    sconjLift2 Hp; sconjLift2 HpO; exact (contContEq Hp HpO).
    subst x1.
    hypReplace Hp (sconjAssocR Hp).
    hypReplace HpO (sconjAssocR HpO).
    clear Iso y' objs objs' phi NullOdd l h' x'.
    rename HpO into Hp'; rename ha into h1; rename ha' into h2.
    destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]]; clear Hp.
    destruct (sconjElim Hp') as [ha' [hb' [J' [HA' HB']]]]; clear Hp'.

    assert (E : Heap.heap_eqv ha ha').
    clear hb hb' J J' HB HB'.
    hypReplace HA (unitIR HA).
    hypReplace HA' (unitIR HA').
    assert (HA2 : ha ||- (x |-> natEnc a ** S (S (S (S x))) |-> y) ** 
      (Heap.heap_eqv Heap.empty_heap)).
    sconjMatch.
    assert (HA2' : ha' ||- (x |-> natEnc a ** S (S (S (S x))) |-> y) ** 
      (Heap.heap_eqv Heap.empty_heap)).
    sconjMatch.
    clear HA HA'.
    lle HA2; lle HA2'.
    sconjLift3of3 HA2; sconjLift2 HA2.
    sconjLift3of3 HA2'; sconjLift2 HA2'.
    hypReplace HA2 (sconjComm (unitIR (heapEqvAdd HA2))).
    hypReplace HA2' (sconjComm (unitIR (heapEqvAdd HA2'))).
    sconjLift2 HA2; sconjLift3of3 HA2.
    sconjLift2 HA2'; sconjLift3of3 HA2'.
    hypReplace HA2 (heapEqvAdd HA2).
    hypReplace HA2' (heapEqvAdd HA2').
    lle HA2; lle HA2'.
    eapply Heap.heap_eqvTrans.
    apply Heap.heap_eqvSymm; apply eqvPredEqv; apply HA2.
    apply eqvPredEqv; assumption.
    eapply Heap.heap_eqvTrans.
    instantiate (1 := hb).
    apply eqvPredEqv; assumption.
    eapply Heap.heap_eqvTrans.
    instantiate (1 := hb').
    2:apply Heap.heap_eqvSymm; apply eqvPredEqv; assumption.
    eapply Heap.HeapMap.ABJoin.mapJoinJoinEqvR.
    apply J.
    eapply Heap.HeapMap.ABJoin.mapJoinEqv.
    apply J'.
    apply Heap.heap_eqvSymm; assumption.
    apply Heap.heap_eqvRefl.
    apply Heap.heap_eqvRefl.
    unfold natEnc.
    constructor.
    apply even_mult_l.
    constructor; constructor; constructor; constructor.
    
    apply ptrPeven.
    eapply contElim1L.
    apply Hp.

  Qed.


  Lemma doubleNatListAtFwd : forall h l1 l2 x1 x2,
    h ||- NatListAt l1 x1 ** NatListAt l2 x2 -> forall objs,
    h ||- objHp objs objs -> forall phi objs',
    iso_set phi objs objs' -> forall h',
    h' ||- (foralls x in objs', objIsFwdedH phi h x) -> forall x1' x2',
    mapX phi x1 = Some x1' ->
    mapX phi x2 = Some x2' ->
    h' ||- NatListAt l1 x1' ** NatListAt l2 x2'.
  Proof.
    intros h l1 l2 x1 x2 Hp objs Hp' phi objs' Iso h' Hp'' x1' x2' Meq1 Meq2.

    assert (Hp2x : 
      h ||- (lexists objs, NatListAt l1 x1 && minObjHp (x1::nil) objs) **
	(lexists objs, NatListAt l2 x2 && minObjHp (x2::nil) objs)).
    clear Hp'; sconjMatch;     
      assert (Hp' := natListIsMinObjHp _ _ _ Hp); lle Hp'; existsl x; apply aconjIntro; assumption.
    lle Hp2x.
    clear Hp.
    rename x into objs1; rename x0 into objs2.
    
    assert (OE : set_eqv objs (objs1 |_| objs2)).
    eapply objHpEqObjs.
    apply Hp'.
    instantiate (1 := objs1 |_| objs2).
    apply objHpMerge; sconjMatch.

    eapply objHpWeaken.
    instantiate (1 := objs1).
    apply (minObjHpElim1 (aconjER Hp)).
    intros x I; apply set_union_intro1; assumption.
    eapply objHpWeaken.
    instantiate (1 := objs2).
    apply (minObjHpElim1 (aconjER Hp)).
    intros x I; apply set_union_intro2; assumption.

    destruct (iso_split Iso (X' := objs1)) as [objs1' X1].
    rewrite OE; intros x I; apply set_union_intro1; assumption.
    destruct (iso_split Iso (X' := objs2)) as [objs2' X2].
    rewrite OE; intros x I; apply set_union_intro2; assumption.
    destruct X1 as [phi1 [Iso1 [Sub1 PSub1]]].
    destruct X2 as [phi2 [Iso2 [Sub2 PSub2]]].
    
    assert (OE' : set_eqv objs' (objs1' |_| objs2')).
    constructor; intro I.
    destruct Iso as [Iso0 [_ [_ Iso]]].
    destruct (Iso _ I) as [b Beq]; clear Iso I Hp''.
    assert (I : set_In b (objs1 |_| objs2)).
    rewrite <- OE; destruct (Iso0 b) as [_ IO]; clear Iso0 OE.
    apply IO; rewrite Beq; discriminate.
    destruct (set_union_elim I) as [I' | I']; clear I.

    apply set_union_intro1; clear phi2 Iso2 Sub2 PSub2 Meq1 Meq2.
    hypReplace PSub1 (PSub1 b).
    destruct Iso1 as [Iso1 [X _]].
    apply (X b); clear X.
    destruct (Iso1 b) as [Iso1' _]; clear Iso1.
    hypReplace Iso1' (Iso1' I').
    destruct (phi1 b).
    rewrite (PSub1 a0) in Beq; auto.
    elimtype False; apply Iso1'; reflexivity.

    apply set_union_intro2; clear phi1 Iso1 Sub1 PSub1 Meq1 Meq2.
    hypReplace PSub2 (PSub2 b).
    destruct Iso2 as [Iso2 [X _]].
    apply (X b); clear X.
    destruct (Iso2 b) as [Iso2' _]; clear Iso2.
    hypReplace Iso2' (Iso2' I').
    destruct (phi2 b).
    rewrite (PSub2 a0) in Beq; auto.
    elimtype False; apply Iso2'; reflexivity.

    destruct (set_union_elim I) as [I' | I']; clear I.
    
    clear objs2' phi2 Iso2 Sub2 PSub2.
    destruct Iso1 as [Iso1 [_ [_ X]]].
    destruct (X _ I') as [b Beq]; clear X I'.
    hypReplace PSub1 (PSub1 _ _ Beq); clear Beq.
    destruct Iso as [_ [X _]]; exact (X _ _ PSub1).

    destruct Iso2 as [Iso2 [_ [_ X]]].
    destruct (X _ I') as [b Beq]; clear X I'.
    hypReplace PSub2 (PSub2 _ _ Beq); clear Beq.
    destruct Iso as [_ [X _]]; exact (X _ _ PSub2).

    assert (Hp3 : h' ||- (foralls x in objs1', objIsFwdedH phi h x) **
      (foralls x in objs2', objIsFwdedH phi h x)).
    apply fasSplit.
    unfold hpredTrue; rewrite <- OE'; assumption.
    eapply iso_pres_disj; eauto.
    rewrite <- OE; rewrite <- OE'; assumption.
    assert (Hp2' : h ||- (foralls x in objs1, x |->__) ** (foralls x in objs2, x |->__) ** unit).
    lli; sconjMatch; exact (objHpIsObjs (minObjHpElim1 (aconjER Hp))).
    apply (fasFasDisj Hp2'); intros; sconjMatch.

    clear Hp' Hp''.
    destruct (sconjExtract Hp2x) as [ha [Hp2xA HpA]].
    sconjLift2of2 Hp2x.
    destruct (sconjExtract Hp2x) as [hb [Hp2xB HpB]]; clear Hp2x.
    sconjMatch.

    (* first list *)
    eapply natListAtFwd.
    apply (aconjEL HpA).
    apply (minObjHpElim1 (aconjER HpA)).
    instantiate (1 := objs1').
    2:eauto.
    clear Meq1 Meq2 HpB.
    apply (fasImplI Hp); clear h0 Hp.
    intros x h' I Hp'.
    eapply objIsFwdedNarrow.
    3:eauto.
    instantiate (1 := h).
    clear Hp2xB; sconjMatch.
    eapply heapEqvTrue.
    apply Heap.heap_eqvSymm; apply eqvPredEqv; eauto.
    exact  (minObjHpElim1 (aconjER HpA)).
    eauto.
    exact (sconjEqSubHeap Hp2xA).
    exact (minObjHpElim1 (aconjER HpA)).
    assumption.
    assumption.
    destruct Iso as [_ [_ [X _]]]; exact X.

    assert (IO := isOddOk x1); unfold mapX in * ; destruct (isOdd x1).
    assumption.
    destruct (minObjHpRootOk (aconjER HpA) x1).
    left; reflexivity.
    elimtype False; apply (not_even_and_odd x1); assumption.
    destruct Iso1 as [Iso1 _]; destruct (Iso1 x1) as [Q _]; hypReplace Q (Q H); clear Iso1.
    hypReplace PSub1 (PSub1 x1).
    destruct (phi1 x1).
    rewrite <- Meq1; rewrite (PSub1 a); reflexivity.
    elimtype False; apply Q; reflexivity.

    (* second list *)
    eapply natListAtFwd.
    apply (aconjEL HpB).
    apply (minObjHpElim1 (aconjER HpB)).
    instantiate (1 := objs2').
    2:eauto.
    clear Meq1 Meq2 HpA.
    apply (fasImplI Hp); clear h0 Hp.
    intros x h' I Hp'.
    eapply objIsFwdedNarrow.
    3:eauto.
    instantiate (1 := h).
    clear Hp2xA; sconjMatch.
    eapply heapEqvTrue.
    apply Heap.heap_eqvSymm; apply eqvPredEqv; eauto.
    exact  (minObjHpElim1 (aconjER HpB)).
    eauto.
    exact (sconjEqSubHeap Hp2xB).
    exact (minObjHpElim1 (aconjER HpB)).
    assumption.
    assumption.
    destruct Iso as [_ [_ [X _]]]; exact X.

    assert (IO := isOddOk x2); unfold mapX in * ; destruct (isOdd x2).
    assumption.
    destruct (minObjHpRootOk (aconjER HpB) x2).
    left; reflexivity.
    elimtype False; apply (not_even_and_odd x2); assumption.
    destruct Iso2 as [Iso2 _]; destruct (Iso2 x2) as [Q _]; hypReplace Q (Q H); clear Iso2.
    hypReplace PSub2 (PSub2 x2).
    destruct (phi2 x2).
    rewrite <- Meq2; rewrite (PSub2 a); reflexivity.
    elimtype False; apply Q; reflexivity.

  Qed.


  Lemma gcGuarSub : forall s s' rr,
    I.gcGuar s s' rr -> I'.gcGuar s s' rr.
  Proof.
    unfold I.gcGuar; unfold NullGCGuar.gcGuar.
    intros s s' rr G.
    unfold I'.gcGuar; unfold ListRevGuar.gcGuar.
    constructor.

    (* show result is still NULL when we have a NULL root *)
    intros r I E1.
    destruct (G empty_set Heap.empty_heap (r::nil)) as [h' [_ X]].
    intros x II; simpl in II; destruct II; [subst x | elimtype False]; assumption.
    simpl; rewrite E1; clear E1 I r.
    assert (Hp : s |||- unit ** top).
    lli.
    sconjMatch.
    apply aconjIntro.
    constructor.
    simpl; intros; elimtype False; assumption.
    rewrite <- hpredTrueEq; unfold okObjHp; lli.
    simpl; intros v [I | I].
    subst v; left; unfold NULL; constructor; constructor.
    elimtype False; assumption.
    apply objHpIntroE; assumption.
    apply unitElim2; assumption.
    
    destruct X as [objs [objs' [phi [_ [_ [_ [X _]]]]]]].
    hypReplace X (X r).
    replace (rget (Heap.empty_heap, rfileOf s) r) with (rget s r) in X; [idtac | reflexivity].
    replace (rget (h', rfileOf s') r) with (rget s' r) in X; [idtac | reflexivity].
    rewrite E1 in X; clear E1 G objs objs'.
    assert (Y : Some (rget s' r) = Some NULL).
    rewrite <- X; clear X.
    unfold mapX.
    rewrite isOddTrue.
    reflexivity.
    unfold NULL; constructor; constructor.
    left; reflexivity.
    injection Y; intro; assumption.


    (* show that the sum of a list in the root is preserved *)
    intros l1 l2 r1 r2 Hp II.
    hypReplace Hp (sconjAssocR Hp).
    destruct (sconjExtract Hp) as [h [Hp' OHp]]; clear Hp.

    assert (OHp' : h ||- (lexists objs, minObjHp (rget s r1::nil) objs) **
      (lexists objs, minObjHp (rget s r2::nil) objs)).
    sconjMatch; eapply natListIsMinObjHp; eauto.

    lle OHp'; rename x into objs; rename  x0 into objs'.
    assert (OHp'' := minObjHpMerge OHp'); clear OHp'; simpl in OHp''.
    assert (Hp : s |||- (minObjHp (rget s r1 :: rget s r2 :: nil) (objs |_| objs') &&
      Heap.heap_eqv h) ** top).
    sconjMatch.
    apply aconjIntro.
    eapply heapEqvTrue.
    2:eauto.
    eapply Heap.heap_eqvTrans.
    2:apply Heap.heap_eqvSymm; apply Hp.
    apply Heap.HeapMap.map_canonEqv; assumption.
    assumption.

    clear Hp'.
    destruct (G _ _ _ II Hp) as [h' [Hp' Iso]]; clear G Hp II rr OHp'' objs objs'.
    apply sconjAssocL; sconjMatch.
    eapply heapEqvTrue.
    eapply Heap.heap_eqvTrans.
    2:apply Heap.heap_eqvSymm; apply Hp.
    apply Heap.HeapMap.map_canonEqv; assumption.
    
    clear h0 Hp; rename OHp into Hp1.
    destruct Iso as [objs [objs' [phi X]]].
    generalize dependent X; stSimpl; intro.
    destruct X as [Hp1' [Hp2 [Iso [Req HpF]]]].

    assert (E1 : mapX phi (rget s r1) = Some (rget s' r1)).
    replace (rget s r1) with (rget (h, rfileOf s) r1); [idtac | reflexivity].
    replace (rget s' r1) with (rget (h', rfileOf s') r1); [idtac | reflexivity].
    apply Req.
    left; reflexivity.

    assert (E2 : mapX phi (rget s r2) = Some (rget s' r2)).
    replace (rget s r2) with (rget (h, rfileOf s) r2); [idtac | reflexivity].
    replace (rget s' r2) with (rget (h', rfileOf s') r2); [idtac | reflexivity].
    apply Req.
    right; left; reflexivity.

    clear Req Hp2.
    eapply doubleNatListAtFwd; eauto.

  Qed.

    
  Lemma rootEq : I.rroot = I'.rroot.
  Proof.
    reflexivity.
  Qed.

  Lemma gcInfoEq : I.rgcInfo = I'.rgcInfo.
  Proof.
    reflexivity.
  Qed.

  Lemma nullEq : I.NULL = I'.NULL.
  Proof.
    reflexivity.
  Qed.

  Lemma presRegSub : forall r,
    I'.presReg r -> I.presReg r.
  Proof.
    intros; assumption.
  Qed.

  Lemma atomPSub : forall n,
    I'.atomP n -> I.atomP n.
  Proof.
    intros; assumption.
  Qed.

  Lemma gcPhiDomSub : set_subset I.gcPhiDom I'.gcPhiDom.
  Proof.
    intro; auto.
  Qed.

  Lemma allocLblEq : I.allocLbl = I'.allocLbl.
  Proof.
    reflexivity.
  Qed.

  Lemma read1LblEq : I.read1Lbl = I'.read1Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma read2LblEq : I.read2Lbl = I'.read2Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma write1LblEq : I.write1Lbl = I'.write1Lbl.
  Proof.
    reflexivity.
  Qed.

  Lemma write2LblEq : I.write2Lbl = I'.write2Lbl.
  Proof.
    reflexivity.
  Qed.


End CheneyGCCompat.


Module ListRevCheneyGC := CoerceGC CheneyGCImpl CheneyGCCompat.


Module ListRevCheneyGCed := MutatorVerif ListRevCheneyGC.





