Require Import MarkSweepGC.

(** ** Local specification map *)

Definition markP f :=
  if EqNat.beq_nat MARKFIELD f then
    SomeT (bspeci markFieldPre markFieldGuar)
    else if EqNat.beq_nat MARKLOOP f then
      SomeT (bspeci markLoopPre markLoopGuar)
      else if EqNat.beq_nat MARKLOOPBODY f then
	SomeT (bspeci markLoopBodyPre markLoopBodyGuar)
	else if EqNat.beq_nat STACKEMPTY f then
	  SomeT (bspeci isEmptyPre isEmptyGuar)
	else if EqNat.beq_nat STACKPOP f then
	  SomeT (bspeci popPre popGuar)
	  else if EqNat.beq_nat MARKFIRST f then
	    SomeT (bspeci markFirstPre markFirstGuar)
	    else if EqNat.beq_nat MARKSECOND f then
	      SomeT (bspeci markSecondPre markSecondGuar)
	      else if EqNat.beq_nat RETURN2 f then
		SomeT (bspeci return2Pre return2Guar)
		else
		  NoneT _.

Ltac PsimplMark := 
  unfold markP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.


(** ** Local lemmas *)
Lemma xNotInB : forall s x y z B G W A, 
  s |||- (x - 4 |-> 1 ** objHdrs B 1) ** A ->
  set_eqv (set_add x (set_union (set_union B G) W)) (rangeMSObjs y z) ->
  set_In x  (rangeMSObjs y z) -> ptrP y -> 
  ~ set_In x B.
Proof.  
  intros.
  assert (forall p, set_In p B -> p <> x).
  intros.
  assert (Hx := MSPtrGtFour _ _ _ _ H1 
    (refl_equal (rangeMSObjs y z))).
  assert (set_In p (rangeMSObjs y z)).
  destruct (H0 p).
  apply H4;elimSetIn.
  assert (Hp := MSPtrGtFour _ _ _ _ H4 
    (refl_equal (rangeMSObjs y z))).
  cut (x - 4 <> p - 4).
  intros.
  omega.
  assert (Hh := sconjAssocL (sconjComm (
    sconjAssocL H))); clear H.
  assert (H := sconjComm (sconjAssocR (
    fasElimL H3 Hh))); clear Hh.
  simpl in H.
  eapply sconjNeq.
  linLogIntroOld.
  existsl 1. 
  eapply sconjImplR. 
  apply H.
  intros h Hh.
  linLogIntroOld.
  exists 1;apply (sconjAssocL Hh).
  intro X.
  apply (H3 x);trivial.
Qed.

Lemma xNotInG : forall s x y z n B G W A, 
  s |||- (x - 4 |-> n ** objHdrs B 1 ** objHdrs G 1) ** A ->
  set_eqv (set_add x (set_union (set_union B G) W)) (rangeMSObjs y z) ->
  set_In x  (rangeMSObjs y z) -> ptrP y -> ~ set_In x G.
Proof.  
  intros.
  assert (forall p, set_In p G -> p <> x).
  intros.
  assert (Hx := MSPtrGtFour _ _ _ _ H1 
    (refl_equal (rangeMSObjs y z))).
  assert (set_In p (rangeMSObjs y z)).
  destruct (H0 p).
  apply H4;elimSetIn.
  assert (Hp := MSPtrGtFour _ _ _ _ H4 
    (refl_equal (rangeMSObjs y z))).
  cut (x - 4 <> p - 4).
  intros.
  omega.
  destruct (sconjElim H) as [h1 [h2 [_ [Hh1 Hh2]]]].
  assert (Hh := sconjAssocL (sconjComm Hh1));clear Hh1.
  clear h2 Hh2.
  destruct (sconjElim Hh) as [h2 [h [_ [_ Hh2]]]].
  clear H.
  assert (H := (sconjComm 
    (fasElimL H3 Hh2)));clear Hh2.
  simpl in H.
  eapply sconjNeq.
  linLogIntroOld.
  existsl n. 
  eapply sconjImplR. 
  apply H.
  intros h' Hh'.
  linLogIntroOld.
  exists 1;apply Hh'.
  intro X.
  apply (H3 x);trivial.
Qed.
  
Lemma xInBNotInG : forall s x y z m n B G W A, 
  s |||- (m - 4 |-> n ** objHdrs B 1 ** objHdrs G 1) ** A ->
  set_eqv (set_add m (set_union (set_union B G) W)) (rangeMSObjs y z) ->
  set_In x  (rangeMSObjs y z) -> ptrP y -> set_In x B -> 
  ~ set_In x G.
Proof.  
  intros.
  assert (forall p, set_In p G -> p <> x).
  intros.
  assert (Hx := MSPtrGtFour _ _ _ _ H1 
    (refl_equal (rangeMSObjs y z))).
  assert (set_In p (rangeMSObjs y z)).
  destruct (H0 p).
  apply H5;elimSetIn.
  assert (Hp := MSPtrGtFour _ _ _ _ H5 
    (refl_equal (rangeMSObjs y z))).
  cut (x - 4 <> p - 4).
  intros.
  omega.
  destruct (sconjElim H) as 
    [h1 [h2 [_ [Hh _]]]].
  destruct (sconjElim Hh) as 
    [h1' [h2' [_  [_ Hh']]]]; 
    clear h1 h2 Hh H.
  assert (H := (sconjComm (fasElimL H3 Hh')));
    clear Hh'.
  destruct (sconjElim (sconjAssocR H)) as 
    [h1 [h2 [_  [Hh _]]]]; clear h1' h2' H.
  assert (H := sconjComm (
    fasElimL H4 Hh));clear Hh.
  simpl in H.
  eapply sconjNeq.
  linLogIntroOld.
  existsl 1. 
  eapply sconjImplR. 
  apply H.
  intros h' Hh'.
  linLogIntroOld.
  exists 1;apply Hh'.
  intro X.
  apply (H4 x);trivial.
Qed.


(** ** Well-formedness Theorems *)

Opaque contL.

Theorem markSecondImplOk : 
  iseqOkS markP markSecondPre markSecondGuar markSecondImpl.
Proof.
  unfold markSecondImpl.
  iseqWeakSolver PsimplMark.

  unfold markFieldGuar.
  intros s s' _ [_ [_ Hr]].
  regEqSolver.

  intros s [B [G [W [ohpSt [ohpEnd [
    freePtr [lchd Hp]]]]]]].
  destruct Hp as [Hst [Hed [Hroot [ Hv0 
    [Hobjs [Hlc [Hlc' Hhp]]]]]]].
  aconjE2 Hhp.
  clear Hhp; rename Hh into Hhp.
  assert (Hhp' := (sconjAssocL (sconjComm 
    (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL 
	(sconjComm (sconjAssocR (sconjAssocR 
	  (sconjAssocR (sconjAssocR (sconjAssocR 
	    (sconjAssocR Hhp)))))))))))))).
  assert (okFieldVal (set_union (rget s v0 :: B) G) lchd).
  clear Hhp.
  destruct Hlc' as [F1 | F1]; 
    [left | right];elimSetIn. 
  clear Hlc';assert (Hlc' := H);clear H.
  unfold objHp in Hhp'; unfold fas in Hhp'; 
    unfold fas' in Hhp'; unfold okObjAt in Hhp'; 
      unfold okFieldAt in Hhp'; simpl in Hhp'.
  destruct (lexEL (sconjAssocL (sconjAssocL Hhp'))) as [x XX].
  destruct (bangEL (sconjAssocL XX)) as [Hlchd XXX];clear XX Hhp'.
  assert (x = lchd).
  assert (hget s (rget s v0) = Some x).
  unfold hget;eapply contGetEqL;apply XXX.
  rewrite Hlc in H;inversion_clear H;reflexivity.
  subst x.
  destruct (lexEL (sconjAssocL (sconjComm XXX))) as 
    [rchd XX];clear XXX.
  destruct (bangEL (sconjAssocL XX)) as [Hrchd XXX];clear XX.
  assert (Hhp' := sconjAssocR (sconjComm (sconjAssocR XXX))).
  assert (hget s (4 + rget s v0) = Some rchd).
  unfold hget;eapply contGetEqL;apply XXX.
  assert (HB := H);clear H.

  rewrite (lwStep XXX);stSimpl;auto;clear XXX.
  destruct Hrchd as [Hatom | Hlin].

  (* first branch : atomP rchd *)
  split.
  split.
  left;apply Hatom.

  intros s' [Hg [_ Hr]].
  destruct (Hg (or_introl _ Hatom)) as [Hh Hrtop].
  simpl in Hh;clear Hg.
  exists ((rget s v0)::B); exists G; exists W;
    exists ohpSt; exists ohpEnd; exists freePtr.
  unfold markLoopInv.
  rewrite <- Hrtop.
  stSimpl.
  regRewrite rroot.
  regRewrite rgcInfo.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  assert (~ set_In (rget s v0) B).
  clear Hhp'.
  eapply xNotInB.
  apply (sconjAssocR Hhp).
  apply Hobjs.
  apply Hv0.
  apply Hst.
  rename H into Hna.

  clear Hhp.
  repeat (split; trivial).
  right; exact Hroot.
  intro N.
  elimtype False.
  assert (set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  apply (MSPtrNotAtom _ _ _ _ H Hst 
    (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).
  assert (hget s (rget s v0) = Some lchd).
  unfold hget.
  eapply contGetEqL.
  apply (sconjAssocL Hhp').
  assert (Hhp'' := (sconjAssocL (sconjAssocL (unitEL 
    (sconjAssocL (sconjComm Hhp'))))));clear Hhp'.
  apply sconjAssocL; apply sconjAssocL; apply sconjAssocL; 
    apply sconjAssocL; apply sconjAssocL.
  eapply sconjImpl.
  apply Hhp''.
  clear Hhp''.
  intros h Hp.
  apply (sconjImplL Hp);
    clear h Hp; intros h Hp.
  repeat (apply sconjAssocR).
  eapply sconjImplL.
  apply (sconjAssocL (sconjAssocL (sconjAssocL Hp))).
  clear h Hp;intros h Hp.
  unfold objHdrs.
  apply fasIntro;auto.
  intros h Hp.
  apply aconjIntro.
  intros; linLogIntroOld.
  eapply sconjImplL.
  apply (sconjAssocR (sconjComm (sconjAssocR Hp))).
  clear h Hp;intros h Hp.

  unfold objHp.
  apply fasIntro;auto.
  eapply sconjImpl.
  apply Hp.
  clear h Hp;intros h Hp.
  unfold okObjAt; unfold okFieldAt; linLogIntroOld.
  existsl lchd;linLogIntroOld.
  existsl rchd;linLogIntroOld.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  apply Hp.
  left;auto.
  clear h Hp;intros h Hp.
  simpl;apply Hp.

  clear Hhp'.
  assert (Hhp' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocR (sconjAssocR 
      (sconjAssocR (sconjAssocR Hhp))))))); clear Hhp .
  assert (exists tempSt, Heap.get (heapOf s) 
    (rget s rgcInfo) = Some tempSt /\ tempSt = ohpSt).
  exists ohpSt.
  split;trivial.
  eapply contGetEqL.
  apply Hhp'.
  assert (exists tempEnd, Heap.get (heapOf s) 
    (4 + rget s rgcInfo) = Some tempEnd /\ tempEnd = ohpEnd).
  exists ohpEnd.
  split;trivial.
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).
  destruct H as [tempSt [Htst Hest]].
  subst ohpSt.
  destruct H0 as [tempEnd [Htend Heend]].
  subst ohpEnd.
  
  clear B G W freePtr Hhp' Hobjs Hst Hed
    Hroot Hv0 Hlc' Hlchd lchd Hlc.
  intros s'' [s' [[Hg1 [Hg2 Hr]] [Hg' [
    Hcall [Hcst1 Hcst2]]]]].
  clear Hg2.
  split.
  intros.
  destruct H as [lchd [Hst [Hed [Hroot [ Hv0 [Hobjs
    [Hlc [Hlc' Hhp]]]]]]]].
  assert (Hhp' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocR (sconjAssocR 
      (sconjAssocR (sconjAssocR Hhp))))))).
  assert (tempSt = ohpSt).
  clear Hhp.
  assert ( Heap.get (heapOf s) (rget s rgcInfo) = Some ohpSt). 
  eapply contGetEqL.
  apply Hhp'.
  rewrite H in Htst;inversion_clear Htst;reflexivity.
  assert (tempEnd = ohpEnd).
  assert (Heap.get (heapOf s) (4 + (rget s rgcInfo)) = Some ohpEnd). 
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).  
  rewrite H0 in Htend;inversion_clear Htend;reflexivity.
  subst tempEnd; subst tempSt; clear Htst Htend.
  
  destruct (Hg1 (or_introl _ Hatom)) as [Hh Hrtop].
  simpl in Hh;clear Hg1 Hcall.
  generalize Hg';clear Hg'.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  intro Hg'.
  apply (Hg' (rget s v0::B) G W) ;clear Hg'.
  
  assert (okFieldVal (set_union (rget s v0 :: B) G) lchd).
  clear Hhp' Hhp.
  destruct Hlc' as [F1 | F1]; [left | right];
    elimSetIn.
  clear Hlc'; rename H into Hlc'. 

  unfold markLoopInv.
  rewrite <- Hrtop.
  assert (~ set_In (rget s v0) B).
  clear Hhp'.
  eapply xNotInB.
  apply (sconjAssocR Hhp).
  apply Hobjs.
  apply Hv0.
  apply Hst.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rroot.
  regRewrite rgcInfo.
  stSimpl.
  repeat (split; trivial).
  right; exact Hroot.
  intro N.
  elimtype False.
  assert (set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  apply (MSPtrNotAtom _ _ _ _ H0 Hst 
    (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).
  clear Hhp'.
    
  destruct (sconjElim (sconjComm (sconjAssocR 
    (sconjAssocR (sconjAssocR (sconjAssocR 
    (sconjAssocR (sconjAssocR Hhp)))))))) as
  [h1 [h2 [J [Hh1 Hh2]]]].
  clear Hhp.
  repeat (apply sconjAssocL).
  apply sconjComm.
  apply sconjIntro with h1 h2.
  apply aconjIntro.
  apply (aconjEL Hh1).
  clear Hh2.
  hypReplace Hh1 (aconjER Hh1).
  assert (set_In (rget s v0) (rget s v0::empty_set)).
  left; reflexivity.
  unfold objHp in Hh1; unfold fas in Hh1; 
    unfold fas' in Hh1; simpl in Hh1; fold fas' in Hh1.
  unfold okObjAt in Hh1; unfold okFieldAt in Hh1; 
    linLogElimOld Hh1.
  hypReplace Hh1 (sconjAssocL (sconjComm
    (sconjAssocR Hh1))).
  assert (Heap.get (heapOf s) (rget s v0) = Some x).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply Hh1.
  assert (Heap.get (heapOf s) (4 + rget s v0) = Some x0).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply (sconjAssocL (
    sconjComm Hh1)).
  assert (x = lchd).
  clear Hh1.
  unfold hget in Hlc.
  rewrite Hlc in H1.
  inversion_clear H1; reflexivity.
  subst x.
  assert (x0 = rchd).
  clear Hh1.
  unfold hget in HB.
  rewrite HB in H2.
  inversion_clear H2; reflexivity.
  subst x0.
  clear H10 H9 H1 H2.
  apply (sconjImplL (sconjAssocR (
    sconjAssocR Hh1))); clear Hh Hh1; intros h Hh.

  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImpl Hh);
    clear h Hh;intros h Hh.
  unfold okObjAt; unfold okFieldAt; linLogIntroOld.
  existsl lchd;linLogIntroOld.
  existsl rchd;linLogIntroOld.
  eapply sconjImplR.
  apply Hh.
  clear h Hh;intros h Hp.
  apply Hp.
  left;auto.
  simpl;apply Hh.
  clear Hh1.
  apply (sconjImplL Hh2);
    clear Hh2; intros h Hh2.
  repeat (apply sconjAssocR). 
  apply (sconjImplL (sconjAssocL 
    (sconjAssocL (sconjAssocL Hh2)))).
  clear h Hh2; intros h Hp.
  unfold objHdrs.
  apply fasIntro;auto.
  rewrite Hh; apply J.

  clear Hg1 Hg'.
  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst2 Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | 
    [H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  elimtype False;auto.
  clear Hcall.
  split.
  rewrite <- Hcst1;regEqSolver. 
  destruct Hcst2 as [Hcst2 Hrt].
  split.
  rewrite <- Hcst2;regEqSolver.
  destruct Hrt.
  split.
  rewrite <- H;regEqSolver.
  rewrite <- H0;regEqSolver. 

  (* second branch, rchd <> atom *)
  split.
  split.
  clear Hhp'.
  right.
  split.
  clear Hhp.
  eapply MSPtrGtFour.  
  apply Hlin.
  reflexivity.
  split.
  clear Hhp.
  eapply MSPtrNotAtom.
  apply Hlin.
  apply Hst.
  reflexivity.
  destruct (Hobjs rchd) as [_ XXX].
  assert (Hl := XXX Hlin).
  inversion_clear Hl.
  subst rchd.
  exists G.
  right.
  split.
  eapply xNotInG.
  2: apply Hobjs.
  2: apply Hv0.
  2: trivial.
  eapply sconjImplL.
  apply (sconjAssocR (sconjAssocR Hhp)).
  intros h Hh.
  apply sconjAssocL;apply Hh.
  linLogIntroOld.
  existsl 1.
  eapply sconjImplR.
  apply Hhp.
  clear Hhp.
  intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjComm Hh)).
  clear h Hh; intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocL Hh)))).
  intros; linLogIntro.

  destruct (set_union_elim H) as [Hl | Hl].
  destruct (set_union_elim Hl) as [Hl' | Hl'].
  exists G.
  right.
  split.
  eapply xInBNotInG.
  2: apply Hobjs.
  2: apply Hlin.
  2: trivial.
  2: apply Hl'.
  eapply sconjImplL.
  apply (sconjAssocR (sconjAssocR Hhp)).
  intros h Hh.
  apply sconjAssocL;apply Hh.
  linLogIntroOld.
  existsl 1.
  assert (Hh := (sconjAssocL (fasElimL Hl' 
    (sconjAssocL (sconjComm Hhp)))));clear Hhp.
  eapply sconjImplR.
  apply Hh.
  clear Hh;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (
    sconjComm Hh))).
  clear h Hh; intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocL Hh)))).
  intros; linLogIntro.

  exists G.
  left.
  split;trivial.
  assert (Hh := (sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocL (
      sconjComm Hhp))))));clear Hhp.
  eapply sconjImplR.
  apply Hh.
  clear Hh;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocL Hh)))).
  intros; linLogIntro.

  exists G.
  right.
  split.
  intro Hin'.
  assert (Heap.get (heapOf s) (rchd - 4) = Some 1).
  eapply contGetEqL.
  apply (sconjAssocL (fasElimL  
    Hin' (sconjAssocL (sconjAssocL (sconjComm 
      (sconjAssocL (sconjComm Hhp))))))).
  assert (Heap.get (heapOf s) (rchd - 4) = Some 0).
  eapply contGetEqL.
  apply (sconjAssocL (fasElimL  
    Hl (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL (sconjAssocL 
	(sconjComm (sconjAssocL 
	  (sconjComm Hhp)))))))))).
  rewrite H0 in H1;inversion_clear H1.
  assert (Hh' := (sconjAssocL (fasElimL  
    Hl (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL 
	(sconjAssocL (sconjComm (sconjAssocL 
	  (sconjComm Hhp))))))))))).
  clear Hhp.
  linLogIntroOld.
  existsl 0.
  eapply sconjImplR.
  apply Hh'.
  clear Hh';intros h Hh.
  eapply sconjImplR.
  apply (sconjComm (sconjAssocR 
    (sconjAssocR Hh))).
  clear Hh h;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL  (sconjAssocL 
    (sconjComm (sconjAssocL Hh)))).
  intros; linLogIntro.

  intros s' [Hg1 [Hg2 Hr]].
  destruct (Hobjs rchd) as [_ XXX].
  assert (Hl := XXX Hlin).
  inversion_clear Hl.

  (* rchd = rget s v0 *)
  clear Hg2.
  subst rchd.
  assert (s |||- rget s v0 - 4 |-> 1 ** top).
  eapply sconjImplR.
  apply Hhp.
  clear Hhp.
  intros; linLogIntro.
  clear XXX.
  destruct (Hg1 (or_intror _ H)) as [Hh Hrtop];
    clear H Hg1.
  assert (~ set_In (rget s v0) B).
  eapply xNotInB.
  apply (sconjAssocR Hhp).
  apply Hobjs.
  apply Hv0.
  apply Hst.
  clear Hhp.
  exists (rget s v0::B); exists G; exists W; 
    exists ohpSt; exists ohpEnd;
      exists freePtr.
  unfold markLoopInv.
  rewrite <- Hrtop.
  regRewrite rroot.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  right; exact Hroot.
  intro N.
  elimtype False.
  assert (set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  apply (MSPtrNotAtom _ _ _ _ H0 Hst 
    (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).
  assert (Hhp'' := (sconjAssocL (sconjAssocL 
    (unitEL (sconjAssocL (
      sconjComm Hhp'))))));clear Hhp'.
  apply sconjAssocL; apply sconjAssocL;  apply sconjAssocL; 
    apply sconjAssocL; apply sconjAssocL.
  eapply sconjImpl.
  apply Hhp''.
  clear Hhp''.
  intros h Hp.
  apply (sconjImplL Hp);
    clear h Hp; intros h Hp.
  repeat (apply sconjAssocR).
  eapply sconjImplL.
  apply (sconjAssocL (sconjAssocL (
    sconjAssocL Hp))).
  clear h Hp;intros h Hp.
  unfold objHdrs.
  apply fasIntro;auto.
  clear Hhp''.
  intros h Hp.
  apply aconjIntro.
  intros; linLogIntro.
  eapply sconjImplL.
  apply (sconjAssocR (sconjComm (
    sconjAssocR Hp))).
  clear h Hp;intros h Hp.
  unfold objHp.
  apply fasIntro;auto.
  eapply sconjImpl.
  apply Hp.
  clear h Hp;intros h Hp.
  unfold okObjAt; unfold okFieldAt; linLogIntroOld.
  existsl lchd;linLogIntroOld.
  existsl (rget s v0);linLogIntroOld.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  apply Hp.
  right;elimSetIn. 
  trivial.

  destruct (set_union_elim H) as [Hl | Hl].

  (*  rchd \in B \cup G *)
  clear Hg2.
  assert (s |||- rchd - 4 |-> 1 ** top).
  clear Hhp'.
  destruct (set_union_elim Hl) as [Hl' | Hl'].
  assert (Hh := (sconjAssocL (fasElimL  Hl'
    (sconjAssocL (sconjComm Hhp)))));clear Hhp.
  eapply sconjImplR.
  apply Hh.
  intros; linLogIntro.
  assert (Hh := (sconjAssocL (fasElimL  Hl' 
    (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL 
	(sconjComm Hhp)))))))); clear Hhp.
  eapply sconjImplR.
  apply Hh.
  intros; linLogIntro.
  clear H.
  destruct (Hg1 (or_intror _ H0)) as [Hh Hrtop];
    clear H0 Hg1.
  assert (~ set_In (rget s v0) B).
  clear Hhp'.
  eapply xNotInB.
  apply (sconjAssocR Hhp).
  apply Hobjs.
  apply Hv0.
  apply Hst.
  clear Hhp.
  exists (rget s v0 ::B); exists G; exists W; 
    exists ohpSt; exists ohpEnd; exists freePtr.
  unfold markLoopInv.
  rewrite <- Hrtop.
  regRewrite rroot.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  right; exact Hroot.
  intro N.
  elimtype False.
  assert (set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  apply (MSPtrNotAtom _ _ _ _ H0 Hst 
    (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).

  assert (Hhp'' := (sconjAssocL (sconjAssocL 
    (unitEL (sconjAssocL (
      sconjComm Hhp'))))));clear Hhp'.
  apply sconjAssocL; apply sconjAssocL; 
    apply sconjAssocL; apply sconjAssocL.
  apply sconjAssocL.
  eapply sconjImpl.
  apply Hhp''.
  clear Hhp''.
  intros h Hp.
  apply (sconjImplL Hp);
    clear h Hp; intros h Hp.
  repeat (apply sconjAssocR).
  eapply sconjImplL.
  apply (sconjAssocL (sconjAssocL (
    sconjAssocL Hp))).
  clear h Hp;intros h Hp.
  unfold objHdrs.
  apply fasIntro;auto.
  clear Hhp''.
  intros h Hp.
  apply aconjIntro.
  intros; linLogIntro.
  eapply sconjImplL.
  apply (sconjAssocR (sconjComm (
    sconjAssocR Hp))).
  clear h Hp;intros h Hp.
  unfold objHp.
  apply fasIntro;auto.
  eapply sconjImpl.
  apply Hp.
  clear h Hp;intros h Hp.
  unfold okObjAt; unfold okFieldAt; linLogIntroOld.
  existsl lchd;linLogIntroOld.
  existsl rchd;linLogIntroOld.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  apply Hp.
  right;elimSetIn. 
  trivial.

  (*  rchd \in W *)
  clear Hg1.
  assert (~ set_In (rget s v0) B).
  clear Hhp' Hg2.
  eapply xNotInB.
  apply (sconjAssocR Hhp).
  apply Hobjs.
  apply Hv0.
  apply Hst.
  rename H0 into Hv0n.
  assert (s |||- rchd - 4 |-> 0 ** objHdrs G 1 ** 
    buffStackSet (rget s rbot) (rget s rtop) 
    (rget s rbuff) G **
    rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr 
      :: rget s rbot :: rget s rbuff :: nil) ** top **
    (objHp (set_add (rget s v0) 
      (set_union B G)) B **
      objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W) **
      objHp (rangeMSObjs ohpSt ohpEnd) 
      (rget s v0 :: empty_set)) **
      objHdrs B 1 ** objHdrs (set_remove rchd W) 0 ** 
      rget s v0 - 4 |-> 1).
  clear Hg2 Hhp'.
  assert (Hh := (sconjAssocL (fasElimL  Hl
    (sconjAssocL (sconjAssocL (
      sconjComm (sconjAssocL 
      (sconjAssocL (sconjComm (
	sconjAssocL (sconjComm Hhp))))))))))).
  clear Hhp.
  eapply sconjImplR.
  apply Hh.
  clear Hh;intros h Hh.
  apply sconjComm;apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (
    sconjComm Hh))).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  apply (sconjImplR (sconjAssocL Hh)).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply Hh.
  clear h Hh;intros h Hh.
  apply sconjAssocR;apply sconjComm;
    repeat (apply sconjAssocR).
  eapply sconjImplR.
  apply (sconjComm Hh).
  clear h Hh;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjComm;apply Hh.
  clear Hhp;rename H0 into Hhp.
  assert (~ atomP rchd).
  clear Hg2 Hhp Hhp'.
  assert (set_In rchd (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  eapply MSPtrNotAtom.
  apply H0.
  apply Hst.
  reflexivity.
  rename H0 into Hna; clear H.

  clear Hhp.
  exists (rget s v0::B); exists (rchd::G); 
    exists (set_remove rchd W); 
	exists ohpSt; exists ohpEnd; exists freePtr.
  unfold markLoopInv.
  regRewrite rroot.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  stSimpl.
  Opaque set_eqv.
  repeat (split; trivial).
  right; elimSetIn.
  intro N.
  elimtype False.
  assert (set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  apply (MSPtrNotAtom _ _ _ _ H Hst 
    (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).
  clear Hg2 Hhp'.
  elimSetIn.
  rewrite <- Hobjs.
  Transparent set_eqv.
  split; intro; elimSetIn.
  assert (Hhp'' := fasElimL  Hl 
    (sconjAssocL (sconjComm (sconjAssocL 
    (sconjAssocL (sconjAssocL (sconjAssocL 
    (sconjAssocL (sconjAssocL (unitEL 
    (sconjAssocL (sconjComm Hhp'))))))))))));
  clear Hhp'.
  assert (s |||- rchd - 4 |-> 0 **
    objHdrs G 1 ** buffStackSet (rget s rbot) 
    (rget s rtop) (rget s rbuff) G ** 
    rget s rgcInfo |->, 
      (ohpSt :: ohpEnd :: freePtr :: rget s rbot :: 
	rget s rbuff :: nil) ** top **
      objHp (set_add (rget s v0) 
	(set_union B G)) B **
      objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W) **
      rget s v0 |-> lchd ** 4 + rget s v0 |-> rchd ** 
	(objHdrs (set_remove rchd W) 0) ** 
	(rget s v0 - 4 |-> 1 ** objHdrs B 1)).
  clear Hg2.
  eapply sconjImplR.
  apply (sconjAssocL Hhp'').
  clear Hhp'';intros h Hp.
  apply sconjComm; apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (
    sconjComm Hp))).
  clear h Hp;intros h Hp.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hp).
  clear h Hp;intros h Hp.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hp).
  clear h Hp;intros h Hp.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hp).
  clear h Hp;intros h Hp.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hp).
  clear h Hp;intros h Hp.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hp).
  clear h Hp;intros h Hp.
  apply sconjAssocR.
  apply (sconjImplR Hp);
    clear h Hp; intros h Hp.
  apply sconjAssocR; apply sconjComm.
  eapply sconjImplR.
  apply Hp.
  trivial.
  clear XXX Hhp''.
  assert (Hhp := Hg2 _ _ Hna H);clear H Hg2.
  apply sconjComm; apply sconjAssocR. 
  eapply sconjImplR.
  apply Hhp.
  clear Hhp.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  intros h Hp.
  apply sconjAssocR; apply sconjComm; 
    repeat (apply sconjAssocR).
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  apply sconjAssocR.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  apply sconjAssocR.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  eapply sconjImpl.
  apply (sconjAssocR (sconjAssocR (
    sconjAssocR Hp))).
  clear h Hp;intros h Hp.
  apply aconjIntro.
  intros; linLogIntro.
  apply sconjComm.
  eapply sconjImpl.
  apply (sconjAssocL(sconjComm (
    sconjAssocL (sconjAssocL Hp)))).
  clear h Hp;unfold objHp; intros h Hp.
  eapply fasEqvSet.
  apply Hp.
  split; intro; elimSetIn.
  clear h Hp;unfold objHp; intros h Hp.
  apply fasIntro;auto.
  eapply sconjImpl.
  apply Hp.
  clear h Hp;unfold okObjAt; 
    unfold okFieldAt; intros h Hp.
  linLogIntroOld; existsl lchd.
  linLogIntroOld; existsl rchd.
  linLogIntroOld.
  destruct Hlc';[left | right];trivial.
  elimSetIn.
  eapply sconjImplR.
  apply Hp.
  trivial.
  right; elimSetIn.
  apply fasImpl0;auto.
  clear h Hp; unfold okObjAt; 
    unfold okFieldAt; intros x h Hx.
  linLogElimOld Hx.
  linLogIntroOld; existsl x0.
  linLogIntroOld; existsl x1.
  linLogIntroOld.
  destruct H3;[left | right];trivial.
  elimSetIn.
  eapply sconjImplR.
  apply Hx.
  trivial.
  destruct H4;[left | right];trivial.
  elimSetIn.
  clear h Hp; intros h Hp.
  eapply sconjImplL.
  apply (sconjComm Hp).
  clear h Hp; unfold objHdrs; intros h Hp.
  apply fasIntro;trivial.
   
  clear Hhp'.
  assert (Hhp' := sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocR (sconjAssocR 
    (sconjAssocR (sconjAssocR Hhp))))))); clear Hhp .
  assert (exists tempSt, Heap.get (heapOf s) 
    (rget s rgcInfo) = Some tempSt /\ tempSt = ohpSt).
  exists ohpSt.
  split;trivial.
  eapply contGetEqL.
  apply Hhp'.
  assert (exists tempEnd, Heap.get (heapOf s) 
    (4 + rget s rgcInfo) = Some tempEnd /\ tempEnd = ohpEnd).
  exists ohpEnd.
  split;trivial.
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).
  destruct H as [tempSt [Htst Hest]].
  subst ohpSt.
  destruct H0 as [tempEnd [Htend Heend]].
  subst ohpEnd.
  
  
  clear B G W freePtr Hhp' Hobjs Hst 
    Hed Hroot Hv0 Hlc' Hlchd Hlc lchd.
  intros s'' [s' [[Hg1 [Hg2  Hr]] [Hg' [
    Hcall [Hcst1 Hcst2]]]]].
  split.
  intros.
  destruct H as [lchd [Hst [Hed [Hroot [Hv0 
    [Hobjs [Hlc [Hlc' Hhp]]]]]]]].
  assert (Hhp' := sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocR (sconjAssocR (
      sconjAssocR (sconjAssocR Hhp))))))).
  assert (tempSt = ohpSt).
  clear Hg1 Hg2 Hhp.
  assert ( Heap.get (heapOf s) (rget s rgcInfo) = Some ohpSt). 
  eapply contGetEqL.
  apply Hhp'.
  rewrite H in Htst;
    inversion_clear Htst;reflexivity.
  assert (tempEnd = ohpEnd).
  assert (Heap.get (heapOf s) (4 + (rget s rgcInfo)) = Some ohpEnd). 
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).  
  rewrite H0 in Htend;inversion_clear Htend;reflexivity.
  subst tempEnd; subst tempSt; clear Htst Htend.

  generalize Hg';clear Hg'.
  clear Hcall Hhp'.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  intro Hg'.
  

  assert (okFieldVal (set_union (rget s v0 :: B) G) lchd).
  clear Hg1 Hg2 Hhp Hg'.
  destruct Hlc' as [F1 | F1]; [left | right];elimSetIn. 
  clear Hlc';assert (Hlc' := H);clear H.

  assert (~ set_In (rget s v0) B).
  clear Hg1 Hg2 Hg'.
  eapply xNotInB.
  apply (sconjAssocR Hhp).
  apply Hobjs.
  apply Hv0.
  apply Hst.
  assert (Hv0n := H);clear H.

  destruct (Hobjs rchd) as [_ XXX].
  assert (Hl := XXX Hlin).
  inversion_clear Hl.
  subst rchd.

  (* rchd = rget s v0 *)
  apply (Hg' (rget s v0::B) G W);auto;clear Hg'.

  clear Hg2 XXX.
  assert (s |||- rget s v0 - 4 |-> 1 ** top).
  clear Hg1.
  eapply sconjImplR.
  apply Hhp.
  intros; linLogIntro.
  destruct (Hg1 (or_intror _ H)) as 
    [Hh Hrtop];clear H Hg1.
  assert (~ set_In (rget s v0) B).
  eapply xNotInB.
  apply (sconjAssocR Hhp).
  apply Hobjs.
  apply Hv0.
  apply Hst.
  unfold markLoopInv.
  rewrite <- Hrtop.
  regRewrite rroot.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  right; elimSetIn.
  intro N.
  elimtype False.
  assert (set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  apply (MSPtrNotAtom _ _ _ _ H0 Hst 
    (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).
  destruct (sconjElim (sconjComm (sconjAssocR 
    (sconjAssocR (sconjAssocR (sconjAssocR 
    (sconjAssocR (sconjAssocR Hhp)))))))) as
  [h1 [h2 [J [Hh1 Hh2]]]]; clear Hhp.
  repeat (apply sconjAssocL).
  apply sconjComm.
  apply sconjIntro with h1 h2.
  clear Hh2.
  apply aconjIntro.
  apply (aconjEL Hh1).
  hypReplace Hh1 (aconjER Hh1).
  assert (set_In (rget s v0) (rget s v0::empty_set)).
  clear Hh1.
  left; reflexivity.
  unfold objHp in Hh1; unfold fas in Hh1; 
    unfold fas' in Hh1; simpl in Hh1; fold fas' in Hh1.
  unfold okObjAt in Hh1; unfold okFieldAt in Hh1; 
    linLogElimOld Hh1.
  hypReplace Hh1 (sconjAssocL (sconjComm
    (sconjAssocR Hh1))).
  assert (Heap.get (heapOf s) (rget s v0) = Some x).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply Hh1.
  assert (Heap.get (heapOf s) (4 + rget s v0) = Some x0).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply (sconjAssocL (
    sconjComm Hh1)).
  assert (x = lchd).
  clear Hh1.
  unfold hget in Hlc.
  rewrite Hlc in H1.
  inversion_clear H1; reflexivity.
  subst x.
  assert (x0 = (rget s v0)).
  clear Hh1.
  unfold hget in HB.
  rewrite HB in H2.
  inversion_clear H2; reflexivity.
  subst x0.
  clear H10 H9 H1 H2.
  apply (sconjImplL (sconjAssocR (
    sconjAssocR Hh1))); clear Hh1 Hh; intros h Hh.

  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImpl Hh);
    clear h Hh;intros h Hh.
  unfold okObjAt; unfold okFieldAt; linLogIntroOld.
  existsl lchd;linLogIntroOld.
  existsl (rget s v0);linLogIntroOld.
  eapply sconjImplR.
  apply Hh.
  clear h Hh;intros h Hp.
  apply Hp.
  right; elimSetIn. 
  simpl;apply Hh.
  repeat (apply sconjAssocR). 
  apply (sconjImplL (sconjAssocL (sconjAssocL 
    (sconjAssocL (sconjAssocL Hh2))))).
  intros h Hp.
  unfold objHdrs.
  apply fasIntro;auto.
  apply J.

  destruct (set_union_elim H) as [Hl | Hl].

  (*  rchd \in B \cup G *)
  clear Hg2.
  assert (s |||- rchd - 4 |-> 1 ** top).
  clear Hg1 Hg'.
  destruct (set_union_elim Hl) as [Hl' | Hl'].
  assert (Hh := (sconjAssocL (fasElimL  Hl'
    (sconjAssocL (sconjComm Hhp)))));clear Hhp.
  eapply sconjImplR.
  apply Hh.
  intros; linLogIntro.
  assert (Hh := (sconjAssocL (fasElimL  Hl' 
    (sconjAssocL (sconjAssocL (
      sconjComm (sconjAssocL 
	(sconjComm Hhp)))))))).
  eapply sconjImplR.
  apply Hh.
  intros; linLogIntro.
  clear H.
  destruct (Hg1 (or_intror _ H0)) as 
    [Hh Hrtop];clear H0 Hg1.
  apply (Hg' (rget s v0::B) G W);
    auto;clear Hg'.
  unfold markLoopInv.
  rewrite <- Hrtop.
  regRewrite rroot.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  assert (~ set_In (rget s v0) B).
  eapply xNotInB.
  apply (sconjAssocR Hhp).
  apply Hobjs.
  apply Hv0.
  apply Hst.
  repeat (split; trivial).
  right; elimSetIn.
  intro N.
  elimtype False.
  assert (set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  apply (MSPtrNotAtom _ _ _ _ H0 Hst 
    (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).
  destruct (sconjElim (sconjComm (sconjAssocR 
    (sconjAssocR (sconjAssocR  (sconjAssocR 
    (sconjAssocR (sconjAssocR Hhp)))))))) as
  [h1 [h2 [J [Hh1 Hh2]]]]; clear Hhp.
  repeat (apply sconjAssocL).
  apply sconjComm.
  apply sconjIntro with h1 h2.
  clear Hh2.
  apply aconjIntro.
  apply (aconjEL Hh1).
  hypReplace Hh1 (aconjER Hh1).
  assert (set_In (rget s v0) (rget s v0::empty_set)).
  left; reflexivity.
  unfold objHp in Hh1; unfold fas in Hh1; 
    unfold fas' in Hh1; simpl in Hh1; fold fas' in Hh1.
  unfold okObjAt in Hh1; unfold okFieldAt in Hh1; 
    linLogElimOld Hh1.
  hypReplace Hh1 (sconjAssocL (sconjComm
    (sconjAssocR Hh1))).
  assert (Heap.get (heapOf s) (rget s v0) = Some x).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply Hh1.
  assert (Heap.get (heapOf s) (4 + rget s v0) = Some x0).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply (sconjAssocL (
    sconjComm Hh1)).
  assert (x = lchd).
  clear Hh1.
  unfold hget in Hlc.
  rewrite Hlc in H1.
  inversion_clear H1; reflexivity.
  subst x.
  assert (x0 = rchd).
  clear Hh1.
  unfold hget in HB.
  rewrite HB in H2.
  inversion_clear H2; reflexivity.
  subst x0.
  clear H10 H9 H1 H2.
  apply (sconjImplL (sconjAssocR (
    sconjAssocR Hh1))); clear Hh Hh1; intros h Hh.

  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImpl Hh);
    clear h Hh;intros h Hh.
  unfold okObjAt; unfold okFieldAt; linLogIntroOld.
  existsl lchd;linLogIntroOld.
  existsl rchd;linLogIntroOld.
  eapply sconjImplR.
  apply Hh.
  clear h Hh;intros h Hp.
  apply Hp.
  right.
  elimSetIn.
  simpl;apply Hh.
  clear Hh1.
  repeat (apply sconjAssocR). 
  apply (sconjImplL (sconjAssocL (sconjAssocL 
    (sconjAssocL (sconjAssocL Hh2))))).
  intros h Hp.
  unfold objHdrs.
  apply fasIntro;auto.
  apply J.

  (*  rchd \in W *)
  clear Hg1.
  assert (s |||- rchd - 4 |-> 0 ** objHdrs G 1 ** 
    buffStackSet (rget s rbot) (rget s rtop) 
    (rget s rbuff) G **
    rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr 
      :: rget s rbot :: rget s rbuff :: nil) ** A' ** (A &&
        objHp (set_union (rget s v0 :: empty_set) 
	  (set_union B G)) B **
         objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W) **
         objHp (rangeMSObjs ohpSt ohpEnd) 
	 (rget s v0 :: empty_set)) **
      objHdrs B 1 ** objHdrs (set_remove rchd W) 0 ** 
      rget s v0 - 4 |-> 1).
  assert (Hh := (sconjAssocL (fasElimL  Hl
    (sconjAssocL (sconjAssocL (sconjComm 
      (sconjAssocL (sconjAssocL (sconjComm (
	sconjAssocL (sconjComm Hhp))))))))))).
  clear Hhp Hg2 Hg'.
  eapply sconjImplR.
  apply Hh.
  clear Hh;intros h Hh.
  apply sconjComm;apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (sconjComm Hh))).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply Hh.
  clear h Hh;intros h Hh.
  apply sconjAssocR;apply sconjComm;
    repeat (apply sconjAssocR).
  eapply sconjImplR.
  apply (sconjComm Hh).
  clear h Hh;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjComm;apply Hh.
  rename H0 into HXXX.
  clear Hhp.
  assert (~ atomP rchd).
  clear Hg2 Hg'.
  assert (set_In rchd (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  eapply MSPtrNotAtom.
  apply H0.
  apply Hst.
  reflexivity.
  assert (Hna := H0); clear H H0.

  apply (Hg' (rget s v0::B) (rchd::G) 
    (set_remove rchd W));auto; clear Hg'.
  unfold markLoopInv.
  regRewrite rroot.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  stSimpl.
  Opaque set_eqv.
  repeat (split; trivial).
  right; elimSetIn.
  intro N.
  elimtype False.
  assert (set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  apply (MSPtrNotAtom _ _ _ _ H Hst 
    (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).
  clear Hg2 HXXX.
  elimSetIn.
  rewrite <- Hobjs.
  Transparent set_eqv.
  split; intro; elimSetIn.

  assert (Hhp' := Hg2 _ _ Hna HXXX).
  hypReplace HXXX (aconjERR (sconjAssocR 
    (sconjComm (sconjAssocR (sconjAssocR 
       (sconjAssocR (sconjAssocR (sconjAssocR HXXX)))))))).
  
  hypReplace HXXX  (sconjAssocL (sconjComm (sconjAssocL 
    (sconjAssocL (sconjComm (sconjAssocL 
    (sconjAssocL (sconjComm (sconjAssocL 
      (sconjComm (sconjAssocL HXXX))))))))))).
  assert (In : set_In (rget s v0) (rget s v0::empty_set)).
  clear HXXX Hhp' Hg2.
  left; reflexivity.
  unfold objHp in HXXX; unfold fas in HXXX; 
    unfold fas' in HXXX; simpl in HXXX.
  fold fas' in HXXX.
  hypReplace HXXX (sconjComm 
    (unitEL (sconjAssocL (sconjComm (
    sconjAssocL HXXX))))).
  unfold okObjAt in HXXX; unfold okFieldAt in HXXX;
    linLogElimOld HXXX.
  assert (Heap.get (heapOf s) (rget s v0) = Some x).
  eapply contGetEqL; apply HXXX.
  assert (Heap.get (heapOf s) (4 + rget s v0) = Some x0).
  eapply contGetEqL; apply (sconjAssocL (
    sconjComm HXXX)).
  assert (x = lchd).
  clear HXXX Hhp' Hg2.
  unfold hget in Hlc.
  rewrite Hlc in H.
  inversion_clear H; reflexivity.
  subst x.
  assert (x0 = rchd).
  clear HXXX Hhp' Hg2.
  unfold hget in HB.
  rewrite HB in H0.
  inversion_clear H0; reflexivity.
  subst x0.
  clear H3 H6 H H0.
  
  hypReplace HXXX (sconjAssocL (sconjComm (
    sconjAssocR (sconjAssocR
    (sconjAssocR (sconjAssocR  HXXX)))))).
  fold fas' in HXXX.
  assert (s |||- rchd - 4 |-> 0 ** objHdrs G 1 ** 
    buffStackSet (rget s rbot) (rget s rtop) 
    (rget s rbuff) G **
    rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr 
      :: rget s rbot :: rget s rbuff :: nil) ** A' ** (
	objHp (rget s v0 :: set_union B G) B **
	objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W) **
	(((rget s v0 |-> lchd ** 4 + rget s v0 |-> rchd) ** 
	  objHdrs B 1) **
          objHdrs (set_remove rchd W) 0) ** rget s v0 - 4 |-> 1)).
  clear Hhp' Hg2.
  apply (sconjImplR HXXX); 
    clear HXXX; intros h Hh.
  apply (sconjImplR (sconjAssocL Hh)); 
    clear h Hh; intros h Hh.
  apply (sconjImplR (sconjAssocL Hh)); 
    clear h Hh; intros h Hh.
  apply sconjAssocL; apply sconjAssocL.
  apply sconjAssocL.
  apply (sconjImplL Hh); 
    clear h Hh; intros h Hh.
  apply sconjAssocR.
  apply sconjAssocR.
  simpl.
  exact Hh.
  assert (Hhp'' := Hg2 _ _ Hna H);
    clear HXXX H In Hg2.

  apply sconjAssocL; apply sconjAssocL;
    apply sconjComm; apply sconjAssocR.
  generalize Hhp';clear Hhp'.
  intro Hhp'.
  apply sconjComm.
  repeat (apply sconjAssocR).
  apply sconjComm.
  repeat (apply sconjAssocR).
  hypReplace Hhp' (sconjAssocL (
    sconjComm (sconjAssocR (
      sconjAssocR (sconjAssocR Hhp'))))).
  destruct (sconjElim Hhp') as [h1 [h2 [J [Hh1 Hh2]]]]; 
    clear Hhp'.
  apply sconjComm.
  repeat (apply sconjAssocL).
  apply sconjAssocR.
  apply sconjIntro with h1 h2.
  clear Hh2.
  apply aconjIntro.
  apply (aconjEL Hh1).
  hypReplace Hh1 (aconjER Hh1).
  assert (set_In (rget s v0) (rget s v0::empty_set)).
  clear Hh1 Hhp''.
  left; reflexivity.
  unfold objHp in Hh1; unfold fas in Hh1; 
    unfold fas' in Hh1; simpl in Hh1; fold fas' in Hh1.
  unfold okObjAt in Hh1; unfold okFieldAt in Hh1; 
    linLogElimOld Hh1.
  hypReplace Hh1 (sconjAssocL (sconjComm
    (sconjAssocR Hh1))).
  assert (Heap.get (heapOf s') (rget s v0) = Some x).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply Hh1.
  assert (Heap.get (heapOf s') (4 + rget s v0) = Some x0).
  apply heapGetSoundL with h1 h2; trivial.
  eapply contGetEqL; apply (sconjAssocL (
    sconjComm Hh1)).
  hypReplace Hhp'' (sconjAssocL  (sconjAssocL 
    (sconjAssocL  (sconjAssocL (sconjComm
      (sconjAssocR  (sconjAssocR  (sconjAssocR 
	(sconjAssocR (sconjAssocR Hhp'')))))))))).
  assert (Heap.get (heapOf s') (rget s v0) = Some lchd).
  eapply contGetEqL; apply Hhp''.
  assert (Heap.get (heapOf s') (4 + rget s v0) = Some rchd).
  eapply contGetEqL; apply (sconjAssocL (
    sconjComm Hhp'')).
  clear Hhp''.
  assert (x = lchd).
  rewrite H2 in H0.
  inversion_clear H0; reflexivity.
  subst x.
  assert (x0 = rchd).
  rewrite H3 in H1.
  inversion_clear H1; reflexivity.
  subst x0.
  clear H3 H0 H1 H2.
  apply (sconjImpl (sconjAssocR (
    sconjAssocR Hh1))); 
  clear Hh1; intros h Hh.

  unfold objHp.
  apply fasIntro; trivial.
  apply (sconjImpl Hh);
    clear h Hh;intros h Hh.
  unfold okObjAt; unfold okFieldAt; linLogIntroOld.
  existsl lchd;linLogIntroOld.
  existsl rchd;linLogIntroOld.
  destruct Hlc';[left | right];trivial.
  elimSetIn.
  apply (sconjImplR Hh).
  trivial.
  right; elimSetIn.
  eapply fasImpl0.
  2: unfold fas.
  2: apply Hh.
  clear h Hh; unfold okObjAt; 
    unfold okFieldAt; intros x h Hx.
  linLogElimOld Hx.
  linLogIntroOld; existsl x0.
  linLogIntroOld; existsl x1.
  linLogIntroOld.
  destruct H4;[left | right];trivial.
  elimSetIn.
  eapply sconjImplR.
  apply Hx.
  trivial.
  destruct H5;[left | right];trivial.
  elimSetIn.
  unfold objHp.
  unfold okObjAt.
  unfold okFieldAt.
  eapply fasEqvSet.
  unfold fas.
  apply Hh.
  split; intro; elimSetIn.
  clear Hh1 Hhp''.
  apply (sconjImplL (sconjAssocR Hh2)).
  clear Hh2.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  intros h Hh.
  apply (sconjImplL (sconjAssocR Hh)).
  clear h Hh; intros h Hh.
  apply (sconjImplL (sconjAssocR Hh)).
  clear h Hh; intros h Hh.
  apply sconjAssocR; apply sconjComm; apply sconjAssocR.
  apply (sconjImplR (sconjComm Hh)).
  clear h Hh; intros h Hh.
  apply (sconjImplR (sconjAssocL (sconjComm Hh))).
  clear h Hh; intros h Hh.
  unfold objHdrs; apply fasIntro; trivial.
  apply J.

  clear Hg1 Hg2 Hg'.
  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst2 Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | 
    [H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  elimtype False;auto.
  clear Hcall.
  split.
  rewrite <- Hcst1;regEqSolver. 
  destruct Hcst2 as [Hcst2 Hrt].
  split.
  rewrite <- Hcst2;regEqSolver.
  destruct Hrt.
  split.
  rewrite <- H;regEqSolver.
  rewrite <- H0;regEqSolver.
Qed.

Transparent contL.

Theorem markImplOk : iseqOkS markP markPre markGuar markImpl.
Proof.
  unfold markImpl.
  iseqWeakSolver PsimplMark.
  
  unfold markFieldGuar.
  intros s s' _ [_ [_ Hr]].
  regEqSolver.
  
  unfold markPre; unfold markPreP.
  intros s [objs [ohpSt [ohpEnd [freePtr [stBot 
    [stBuff [Hst [Hed [Hroot [Hobjs Hhp]]]]]]]]]].
  rewrite addiuStep;stSimpl.
  assert (Hhp' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocL (sconjComm Hhp))))).
  unfold contL in Hhp'.
  assert (Hhp'' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL (sconjAssocL 
	(sconjComm (sconjAssocL Hhp'))))))))));clear Hhp'.  
  rewrite (lwStep Hhp'');stSimpl.
  assert (Hhp' := (sconjAssocL (sconjComm Hhp'')));
    clear Hhp''.  
  rewrite (lwStep Hhp');stSimpl.
  rewrite addiuStep;stSimpl.
  rewrite addiuStep;stSimpl.
  2: simpl;reflexivity.
  2: auto with arith.
  2: simpl;reflexivity.
  2: auto with arith.
  split.
  split.
  destruct Hroot as [Hrt1 | Hroot]; [left | right].
  exact Hrt1.
  assert (Hroot2 : set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  rewrite <- Hobjs.
  apply Hroot.
  split.
  apply (MSPtrGtFour _ _ _ _ Hroot2 (refl_equal (rangeMSObjs ohpSt ohpEnd))). 
  split.
  apply (MSPtrNotAtom _ _ _ _ Hroot2 Hst (refl_equal (rangeMSObjs ohpSt ohpEnd))).
  exists empty_set.
  clear Hhp'.
  right.
  split.
  unfold empty_set;simpl;auto.
  unfold objHdrs in Hhp.
  assert (X := fasElimL  Hroot Hhp).
  simpl.
  eapply sconjImplR.
  linLogIntroOld;exists 0.
  apply (sconjAssocL X).
  clear Hhp X;intros h Hhp.
  unfold objHdrs;unfold fas;unfold fas';simpl.
  apply unitIL.
  eapply sconjImplR.
  apply (sconjAssocL (sconjComm Hhp)).
  intros; linLogIntro.

  clear Hhp'.
  intros s' [Hn [Hc Hr]].
  simpl in Hn; simpl in Hc.
  generalize Hn Hc Hr; clear Hn Hc Hr.
  stSimpl.
  intros.
  destruct Hroot as [Hrt1 | Hroot].
  exists empty_set; 
    exists (empty_set); 
      exists objs;
	exists ohpSt; exists ohpEnd; exists freePtr.
  repeat (split;auto).
  regRewrite rroot.
  left; exact Hrt1.
  regRewrite rbuff.
  regRewrite rbot. 
  regRewrite rgcInfo.
  destruct (Hn (or_introl _ Hrt1)) as [Hh Htop ].
  rewrite <- Htop.
  apply unitIL.
  apply unitIL.
  rewrite Hh.
  stSimpl.
  sconjStep' 5.
  apply aconjIntro.
  apply (aconjEL Hhp).
  hypReplace Hhp (aconjER Hhp).
  apply unitIL.
  unfold objHp in * .
  eapply fasImpl0.
  2: apply Hhp.
  unfold okObjAt; unfold okFieldAt; unfold okFieldVal.
  intros.
  linLogElim H.
  linLogIntro.
  instantiate (1:= x0).
  intuition.
  right; rewrite <- Hobjs; elimSetIn.
  right; rewrite <- Hobjs; elimSetIn.
  instantiate (1:= x1).
  trivial.
  intuition.
  right; rewrite <- Hobjs; elimSetIn.
  right; rewrite <- Hobjs; elimSetIn.

  exists empty_set; 
    exists ((rget s rroot)::empty_set); 
      exists (set_remove (rget s rroot) objs);
	exists ohpSt; exists ohpEnd; exists freePtr.
  repeat (split;auto).
  regRewrite rroot;stSimpl.
  right; elimSetIn.
  regRewrite rroot;stSimpl.
  intro N.
  elimtype False.
  assert (Hroot2 : set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  rewrite <- Hobjs.
  apply Hroot.
  apply (MSPtrNotAtom _ _ _ _ Hroot2 Hst (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).
  intro; rewrite <- Hobjs;  elimSetIn.
  intro; rewrite Hobjs;   elimSetIn.
  assert (s' |||- objHdrs (rget s rroot :: empty_set) 1 **  
    buffStackSet (rget s' rbot) (rget s' rtop) 
    (rget s' rbuff) (set_add(rget s rroot)  empty_set) ** 
    (objHdrs (set_remove (rget s rroot) objs) 0 **
      rget s' rgcInfo |->, (ohpSt :: ohpEnd :: freePtr :: 
	rget s' rbot :: rget s' rbuff :: nil) ** top **
      objHp objs objs)).
  apply Hc.
  assert (Hroot2 : set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  rewrite <- Hobjs.
  apply Hroot.
  apply (MSPtrNotAtom _ _ _ _ Hroot2 Hst (refl_equal (rangeMSObjs ohpSt ohpEnd))).
  assert (X := fasElimL  Hroot Hhp).
  eapply sconjImplR.
  apply (sconjAssocL X).
  clear Hhp X;intros h Hhp.
  unfold objHdrs;unfold fas;unfold fas';simpl.
  apply unitIL.
  eapply sconjImplR.
  apply (sconjAssocL (sconjComm Hhp)).
  clear Hhp h;intros h Hhp.
  eapply sconjImplR.
  apply (sconjComm Hhp).
  clear Hhp h; intros h Hhp.
  
  regRewrite rbot; regRewrite rbuff; regRewrite rgcInfo.
  stSimpl.
  apply (sconjAssocL (aconjERR (sconjAssocR Hhp))).
  apply sconjComm; apply sconjAssocR.
  eapply sconjImplR.
  apply H.
  clear Hhp H ;intros h Hhp.
  unfold objHdrs;unfold fas;unfold fas';simpl.
  apply unitIR.
  apply sconjComm; apply sconjAssocR.
  eapply sconjImplR.
  apply Hhp.
  clear Hhp h;intros h Hhp.
  apply sconjComm.
  eapply sconjImplR.
  apply Hhp.
  clear Hhp h;intros h Hhp.
  
  eapply sconjImplR.
  apply Hhp.
  clear Hhp h; unfold objHp;intros h Hp.
  eapply sconjImplR.
  apply Hp.
  clear Hp h; intros h Hp.
  apply aconjIntro.
  intros; linLogIntro.
  apply unitIL.
  eapply fasEqvSet.
  eapply fasImpl0.
  2: apply Hp.
  unfold okObjAt; unfold okFieldAt; unfold okFieldVal.
  intros.
  linLogElim H.
  linLogIntro.
  instantiate (1:= x0).
  intuition.
  right; rewrite <- Hobjs; elimSetIn.
  right; rewrite <- Hobjs; elimSetIn.
  instantiate (1:= x1).
  trivial.
  intuition.
  right; rewrite <- Hobjs; elimSetIn.
  right; rewrite <- Hobjs; elimSetIn.
  split;intro Ha; elimSetIn.

  intros s'' [s' [[Hatom [Hnatom Hr]] [Hg [Hcall [Hcst1 Hcst2]]]]].
  assert (exists tempBuff, Heap.get (heapOf s) 
    (16 + (rget s rgcInfo)) = Some tempBuff /\ tempBuff = stBuff).
  exists stBuff.
  split;auto.
  eapply contGetEqL.
  apply Hhp'.
  destruct H as [tempBuff [Ht1 He]];subst stBuff.
  assert (exists tempBot, Heap.get (heapOf s) 
    (12 + (rget s rgcInfo)) = Some tempBot /\ tempBot = stBot).
  exists stBot.
  split;auto.
  eapply contGetEqL.
  apply (sconjComm (sconjAssocR Hhp')).
  destruct H as [tempBot [Ht2 He]];subst stBot.
  clear Hhp'.
  assert (Hhp' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocR Hhp)))).
  assert (exists tempSt, Heap.get (heapOf s) 
    (rget s rgcInfo) = Some tempSt /\ tempSt = ohpSt).
  exists ohpSt.
  split;trivial.
  eapply contGetEqL.
  apply Hhp'.
  assert (exists tempEnd, Heap.get (heapOf s) 
    (4 + rget s rgcInfo) = Some tempEnd /\ tempEnd = ohpEnd).
  exists ohpEnd.
  split;trivial.
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).
  destruct H as [tempSt [Htst Hest]].
  subst ohpSt.
  destruct H0 as [tempEnd [Htend Heend]].
  subst ohpEnd.

  split.
  clear objs freePtr Hst Hed Hroot Hobjs Hhp' Hhp.
  intros.
  destruct H as [Hst [Hed [Hroot [Hobjs Hhp]]]].
  simpl in Hatom; simpl in Hnatom.
  assert (Hhp' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocR Hhp)))).
  assert (tempSt = ohpSt).
  assert ( Heap.get (heapOf s) (rget s rgcInfo) = Some ohpSt). 
  eapply contGetEqL.
  apply Hhp'.
  rewrite H in Htst;inversion_clear Htst;reflexivity.
  assert (tempEnd = ohpEnd).
  assert (Heap.get (heapOf s) 
    (4 + (rget s rgcInfo)) = Some ohpEnd). 
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).  
  rewrite H0 in Htend;inversion_clear Htend;reflexivity.
  subst tempEnd; subst tempSt; clear Htst Htend.
  clear Hhp'.
  assert (Hhp' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocL (sconjComm Hhp))))).
  unfold contL in Hhp'.
  assert (Hhp'' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL (sconjAssocL 
	(sconjComm (sconjAssocL Hhp'))))))))));clear Hhp'. 
  assert (tempBot = stBot).
  assert ( Heap.get (heapOf s) (12 + (rget s rgcInfo)) = Some stBot). 
  eapply contGetEqL.
  apply Hhp''.
  rewrite H in Ht2;inversion_clear Ht2;reflexivity.
  assert (tempBuff = stBuff).
  assert (Heap.get (heapOf s) (16 + (rget s rgcInfo)) = Some stBuff). 
  eapply contGetEqL.
  apply (sconjAssocL (sconjComm Hhp'')).  
  rewrite H0 in Ht1;inversion_clear Ht1;reflexivity.
  subst tempBot; subst tempBuff; clear Ht1 Ht2.
  rewrite (Hr rbuff) in Hg;
    autorewrite with RNum8;try discriminate;auto.
  rewrite (Hr rbot) in Hg;
    autorewrite with RNum8;try discriminate;auto.
  generalize Hg Hatom Hnatom Hr;stSimpl;clear Hg Hatom Hnatom Hr;intros.
  destruct Hroot as [Hrt1 | Hroot].
  apply (Hg empty_set empty_set objs
    ohpSt ohpEnd  freePtr A);clear Hg.
  repeat (split;auto).
  regRewrite rroot.
  left; exact Hrt1.
  regRewrite rbuff.
  regRewrite rbot. 
  regRewrite rgcInfo.
  destruct (Hatom (or_introl _ Hrt1)) as [Hh Htop ].
  rewrite <- Htop.
  apply unitIL.
  apply unitIL.
  rewrite Hh.
  stSimpl.
  sconjStep' 5.
  apply aconjIntro.
  apply (aconjEL Hhp).
  hypReplace Hhp (aconjER Hhp).
  apply unitIL.
  unfold objHp in * .
  eapply fasImpl0.
  2: apply Hhp.
  unfold okObjAt; unfold okFieldAt; unfold okFieldVal.
  intros.
  linLogElim H.
  linLogIntro.
  instantiate (1:= x0).
  intuition.
  right; rewrite <- Hobjs; elimSetIn.
  right; rewrite <- Hobjs; elimSetIn.
  instantiate (1:= x1).
  trivial.
  intuition.
  right; rewrite <- Hobjs; elimSetIn.
  right; rewrite <- Hobjs; elimSetIn.

  apply (Hg empty_set ((rget s rroot)::empty_set) 
    (set_remove (rget s rroot) objs) 
    ohpSt ohpEnd  freePtr A);clear Hg.
  stSimpl.
  repeat (split;auto).
  regRewrite rroot;stSimpl.
  right; elimSetIn.
  regRewrite rroot;stSimpl.
  intro N.
  elimtype False.
  assert (Hroot2 : set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  rewrite <- Hobjs.
  apply Hroot.
  apply (MSPtrNotAtom _ _ _ _ Hroot2 Hst (refl_equal (rangeMSObjs ohpSt ohpEnd)) N).
  intro; rewrite <- Hobjs;  elimSetIn.
  intro; rewrite Hobjs;   elimSetIn.
  assert (s' |||- objHdrs (rget s rroot :: empty_set) 1 **  
    buffStackSet (rget s' rbot) (rget s' rtop) 
    (rget s' rbuff) (set_add (rget s rroot) empty_set) ** 
    (objHdrs (set_remove (rget s rroot) objs) 0 **
      rget s' rgcInfo |->, (ohpSt :: ohpEnd :: freePtr :: 
	rget s' rbot :: rget s' rbuff :: nil) ** A' **
      (A && objHp objs objs))).

  apply Hnatom.
  assert (Hroot2 : set_In (rget s rroot) (rangeMSObjs ohpSt ohpEnd)).
  rewrite <- Hobjs.
  apply Hroot.
  apply (MSPtrNotAtom _ _ _ _ Hroot2 Hst (refl_equal (rangeMSObjs ohpSt ohpEnd))).
  assert (X := fasElimL  Hroot Hhp).
  eapply sconjImplR.
  apply (sconjAssocL X).
  clear Hhp X;intros h Hhp.
  unfold objHdrs;unfold fas;unfold fas';simpl.
  apply unitIL.
  eapply sconjImplR.
  apply (sconjAssocL (sconjComm Hhp)).
  clear Hhp h;intros h Hhp.
  eapply sconjImplR.
  apply (sconjComm Hhp).
  regRewrite rbot; regRewrite rbuff; regRewrite rgcInfo. 
  trivial.
  apply sconjComm; apply sconjAssocR.
  eapply sconjImplR.
  apply H.
  clear Hhp H ;intros h Hhp.
  unfold objHdrs;unfold fas;unfold fas';simpl.
  apply unitIR.
  apply sconjComm; apply sconjAssocR.
  assert (rget s' rbot = stBot).
  regRewrite rbot.
  trivial.
  subst stBot.
  assert (rget s' rbuff = stBuff).
  regRewrite rbuff.
  trivial.
  subst stBuff.
  eapply sconjImplR.
  apply Hhp.
  clear Hhp h;intros h Hhp.
  apply sconjComm.
  eapply sconjImplR.
  apply Hhp.
  clear Hhp h;intros h Hhp.
  eapply sconjImplR.
  apply Hhp.
  clear Hhp h;intros h Hhp.
  eapply sconjImplR.
  apply Hhp.
  clear Hhp h;intros h Hhp.
  apply aconjIntro.
  apply (aconjEL Hhp).
  apply unitIL.
  unfold objHp.
  unfold objHp in Hhp.
  eapply fasEqvSet.
  assert (XXX := aconjER Hhp).
  eapply fasImpl0.
  2: apply XXX.
  unfold okObjAt; unfold okFieldAt; unfold okFieldVal.
  intros.
  linLogElim H.
  linLogIntro.
  instantiate (1:= x0).
  intuition.
  right; rewrite <- Hobjs; elimSetIn.
  right; rewrite <- Hobjs; elimSetIn.
  instantiate (1:= x1).
  trivial.
  intuition.
  right; rewrite <- Hobjs; elimSetIn.
  right; rewrite <- Hobjs; elimSetIn.
  split;intro Ha; elimSetIn.

  clear Hg Hatom Hnatom Hhp';split.
  apply raAndOthers.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst2 Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | [
    H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  elimtype False;auto.
  clear Hcall.
  destruct Hcst2 as [Hcst2 [Hrt Hrt']].
  transitivity (rget s' rcst2);
    [regRewrite rcst2; trivial | apply Hrt].
  destruct Hcst2 as [Hcst2 [Hrt Hrt']].
  split.
  transitivity (rget s' rcst1);
    [regEqSolver | apply Hcst1]. 
  split.
  rewrite <- Hcst2.
  regEqSolver.
  transitivity (rget s' rroot).
  rewrite Hrt';reflexivity.
  regRewrite rroot;stSimpl.
  trivial.
Qed.

Theorem markLoopImplOk : 
  iseqOkS markP markLoopPre markLoopGuar markLoopImpl.
Proof.
  unfold markLoopImpl.
  iseqWeakSolver PsimplMark.

  unfold isEmptyGuar.
  intros s s' _ [_ [_ Hr]].
  regEqSolver.

  destruct emptyStackIsEmptySet as [G1 G2].
  intros s [B [G [W [ohpSt [ohpEnd [
    freePtr Hp]]]]]]; stSimpl.
  assert (s |||- (objHp (set_union B G) B ** 
    objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W)) ** top). 
  destruct Hp as [Hst [Hed [Hroot [Hroot' [Hobjs  Hhp]]]]].
  
  aconjE2 Hhp.
  clear Hhp; rename Hh into Hhp. 
  eapply sconjImplR.
  apply (sconjComm (sconjAssocR (
    sconjAssocR (sconjAssocR (sconjAssocR (
      sconjAssocR Hhp)))))).
  intros; linLogIntro.
  rename H into HXXX. 
  split.
  destruct Hp as [Hst [Hed [Hroot [Hroot' [Hobjs  Hhp]]]]].
  assert (Hhp' := sconjAssocL (sconjComm (
    sconjAssocR (sconjAssocR Hhp)))).
  split.
  apply G1.
  exists G.
  eapply sconjImplR.
  apply Hhp'.
  intros; linLogIntro.

  intros s' G'. 
  generalize (G2 _ _ G'); intros [Hg [Hh Hr]]; clear G'.
  exists B; exists G; exists W; exists ohpSt; exists ohpEnd; exists freePtr.
  split.
  eapply Hg.
  apply Hhp'.
  clear Hhp'.
  assert (forall r, r <> v0 -> 
    rget s' r = rget (rset s ra MARKLOOPBODY) r).
  intros;rewrite (Hr r);auto.
  clear Hr; rename H into Hr.
  unfold markLoopInv; stSimpl.
  regRewrite rroot; regRewrite rtop; 
  regRewrite rbot; regRewrite rbuff; 
  regRewrite rgcInfo; stSimpl.
  repeat (split;trivial).
  simpl in Hh;rewrite <- Hh;apply Hhp.
  
  destruct Hp as [Hst [Hed [Hroot [Hroot' [Hobjs  Hhp]]]]].
  assert (Hhp' := sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocR 
      (sconjAssocR (sconjAssocR Hhp)))))); 
  clear Hhp .
  assert (exists tempSt, Heap.get (heapOf s) 
    (rget s rgcInfo) = Some tempSt /\ tempSt = ohpSt).
  exists ohpSt.
  split;trivial.
  eapply contGetEqL.
  apply Hhp'.
  assert (exists tempEnd, Heap.get (heapOf s) 
    (4 + rget s rgcInfo) = Some tempEnd /\ tempEnd = ohpEnd).
  exists ohpEnd.
  split;trivial.
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).
  destruct H as [tempSt [Htst Hest]].
  subst ohpSt.
  destruct H0 as [tempEnd [Htend Heend]].
  subst ohpEnd.

  clear B G W freePtr Hhp' Hobjs
    Hst Hed Hroot HXXX Hroot'.
  intros s'' [s' [G' [Hg' [Hcall [
    Hcst1 Hcst2]]]]].
  destruct (G2 _ _ G') as [Hg [Hh Hr]];
    clear G1 G2 G'.
  split.
  intros.
  assert (XXX := H).
  destruct H as [Hst [Hed [Hroot [Hroot' [Hobjs  Hhp]]]]].
  assert (Hhp' := sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocR 
      (sconjAssocR (sconjAssocR Hhp)))))).
  assert (tempSt = ohpSt).
  assert ( Heap.get (heapOf s) (rget s rgcInfo) = Some ohpSt). 
  eapply contGetEqL.
  apply Hhp'.
  rewrite H in Htst;inversion_clear Htst;reflexivity.
  assert (tempEnd = ohpEnd).
  assert (Heap.get (heapOf s) (4 + (rget s rgcInfo)) = Some ohpEnd). 
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).  
  rewrite H0 in Htend;inversion_clear Htend;reflexivity.
  subst tempEnd; subst tempSt; clear Htst Htend.
  assert (H := XXX);clear XXX.

  assert (forall r, r <> v0  -> 
    rget s' r = rget (rset s ra MARKLOOPBODY) r).
  intros;rewrite (Hr r);auto.
  clear Hr; assert (Hr := H0);clear H0.
  generalize Hg';clear Hg'.
  regRewrite rbot; regRewrite rbuff; stSimpl.
  intro Hg'.
  apply (Hg' B G W).
  eapply Hg.
  apply (sconjAssocL (sconjComm (
    sconjAssocR (sconjAssocR Hhp)))).
  assert (forall r, r <> v0  -> 
    rget s' r = rget (rset s ra MARKLOOPBODY) r).
  intros;rewrite (Hr r);auto.
  clear Hr; assert (Hr := H0);clear H0.
  unfold markLoopInv; stSimpl.
  regRewrite rroot; regRewrite rtop; 
  regRewrite rbot; regRewrite rbuff; 
  regRewrite rgcInfo;stSimpl.
  repeat (split;trivial).
  simpl in Hh;rewrite <- Hh;apply Hhp.

  assert (forall r, r <> v0  -> 
    rget s' r = rget (rset s ra MARKLOOPBODY) r).
  intros;rewrite (Hr r);auto.
  clear Hr; assert (Hr := H);clear H.
  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst2 Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | [
    H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  elimtype False;auto.
  clear Hcall.
  split.
  rewrite <- Hcst1;regEqSolver. 
  destruct Hcst2 as [Hcst2 [Hrt Hrt']].
  repeat split.
  rewrite <- Hcst2;regEqSolver.
  rewrite <- Hrt;regEqSolver.
  rewrite <- Hrt';regEqSolver.
Qed.

Theorem markLoopBodyImplOk : 
  iseqOkS markP markLoopBodyPre markLoopBodyGuar markLoopBodyImpl.
Proof.
  unfold markLoopBodyImpl.
  iseqWeakSolver PsimplMark.

  unfold popGuar.
  intros s s' _ [_ Hr].
  regEqSolver.
  
  destruct popStackIsPopSet as [G1 G2].
  intros s [B [G [W [ohpSt [ohpEnd 
    [freePtr [Hsb Hp]]]]]]].
  destruct Hp as [Hst [Hed [Hroot [Hroot' [Hobjs Hhp]]]]].

  unfold doBtest;stSimpl.
  compare (rget s v0) 0;intro e;
    [rewrite e;rewrite beq_true | rewrite beq_false];simpl;auto.
  rewrite e in Hsb;simpl in Hsb.
  split.
  split.
  apply G1.
  exists G.
  split;auto.
  eapply sconjImplR.
  apply (sconjAssocL (sconjComm 
    (sconjAssocR (sconjAssocR Hhp)))).
  intros; linLogIntro.

  intros s' G'. 
  destruct (G2 _ _ G') as [Hg Hr];
    clear G1 G2 G'.
  assert (forall r, r <> v0 -> r <> rtop  -> 
    rget s' r = rget (rset s ra MARKFIRST) r).
  intros;rewrite (Hr r);auto.
  clear Hr; rename H into Hr. 
  
  destruct Hroot as [Hrt1 | Hroot].
  rewrite (Hroot' Hrt1) in Hsb.
  elimtype False; apply (Hsb (refl_equal nil)).

  exists B; exists (set_remove (rget s' v0) G); exists W;
    exists ohpSt; exists ohpEnd; exists freePtr.
  destruct (Hg G _ (sconjAssocL (sconjComm 
    (sconjAssocR (sconjAssocR Hhp)))) Hsb) as 
  [Hv0' Hhp'];clear Hg.
  
  split;trivial.
  split;trivial.
  split.
  regRewrite rroot;stSimpl.
  elimSetIn.
  split.
  elimSetIn.
  split.
  setoid_rewrite <- Hobjs.
  split; intro; elimSetIn.

  apply sconjAssocL;apply sconjAssocL;
    apply sconjAssocL;apply sconjComm.
  apply sconjAssocR.
  generalize Hhp';clear Hhp'.
  unfold setHp.
  regRewrite rbot;regRewrite rbuff;
  regRewrite rgcInfo;stSimpl;intro Hhp'.
  eapply sconjImplR.
  apply Hhp'.
  clear Hhp';intros h Hp.
  eapply sconjImpl.
  apply Hp.
  clear Hp h;intros h Hp.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.  
  apply aconjIntro.
  intros; linLogIntro.
  hypReplace Hp (aconjER Hp).
  eapply sconjImpl.
  apply Hp.
  clear Hp h; unfold objHp; intros h Hp.
  eapply fasImpl0.
  2: apply Hp.
  clear h Hp;  unfold okObjAt; 
    unfold okFieldAt;intros x h Hp.
  linLogElimOld Hp.
  linLogIntroOld.
  existsl x0;linLogIntroOld.
  existsl x1;linLogIntroOld.
  destruct H3;[left | right];trivial.
  elimSetIn.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  apply Hp.
  destruct H4;[left | right];trivial.
  elimSetIn.
  clear h Hp; unfold objHp; intros h Hx.
  simpl in Hx.
  assert (set_In (rget s' v0) (set_union G W)).
  elimSetIn.
  assert (Hx' := fasElim H Hx).
  eapply sconjImpl.
  apply (sconjComm Hx').
  clear Hx' Hx h; intros h Hh.
  eapply fasEqvSet.
  apply Hh.
  split; intro.
  elimSetIn.
  assert (~ set_In (rget s' v0) W).
  intro X.
  assert (Hx := fasElimL  X 
    (sconjAssocL (sconjComm (sconjAssocR Hhp)))).
  assert (hget s (rget s' v0 - 4) = Some 0).
  unfold hget;eapply contGetEqL;
    apply (sconjAssocL Hx);clear Hx.
  assert (Hx' := fasElimL  Hv0' 
    (sconjAssocL (sconjComm  Hhp))).
  assert (hget s (rget s' v0 - 4) = Some 1).
  unfold hget;eapply contGetEqL;
    apply (sconjAssocL Hx');clear Hx'.
  rewrite H2 in H1;inversion_clear H1.
  elimSetIn.

  clear h Hx H Hx';intros h Hh.
  apply fasIntro;auto.
  apply unitIR.
  apply Hh.
  
  clear h Hp; intros h Hp.
  apply (sconjImplL Hp).
  clear h Hp; intros h Hp.
  apply sconjAssocR; apply sconjComm; 
    apply sconjAssocR;
      apply (sconjImplR Hp).
  clear h Hp; intros h Hp.
  assert (Hx' := fasElim Hv0' Hp).
  apply sconjComm; exact Hx'.

  assert (Hhp' := sconjAssocL 
    (sconjAssocL (sconjComm (sconjAssocR 
    (sconjAssocR (sconjAssocR Hhp)))))); clear Hhp .
  assert (exists tempSt, Heap.get (heapOf s) 
    (rget s rgcInfo) = Some tempSt /\ tempSt = ohpSt).
  exists ohpSt.
  split;trivial.
  eapply contGetEqL.
  apply Hhp'.
  assert (exists tempEnd, Heap.get (heapOf s) 
    (4 + rget s rgcInfo) = Some tempEnd /\ tempEnd = ohpEnd).
  exists ohpEnd.
  split;trivial.
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).
  destruct H as [tempSt [Htst Hest]].
  subst ohpSt.
  destruct H0 as [tempEnd [Htend Heend]].
  subst ohpEnd.
  
  clear B G W freePtr Hhp' Hobjs 
    Hst Hed  Hroot Hsb Hroot'.
  intros s'' [s' [G' [Hg' [Hcall 
    [Hcst1 Hcst2]]]]].
  destruct (G2 _ _ G') as [Hg  Hr];
    clear G' G1 G2.
  assert (forall r, r <> v0  -> 
    r <> rtop  -> 
    rget s' r = rget (rset s ra MARKFIRST) r).
  intros;rewrite (Hr r);auto.
  clear Hr; rename H into Hr.
  split.
  intros.
  assert (Hsb := H);clear H.
  assert (XXX := H0);clear H0.
  destruct XXX as [Hst [Hed [Hroot [Hroot' [Hobjs  Hhp]]]]].
  assert (Hhp' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocR 
    (sconjAssocR (sconjAssocR Hhp)))))).
  assert (tempSt = ohpSt).
  assert ( Heap.get (heapOf s) (rget s rgcInfo) = Some ohpSt). 
  eapply contGetEqL.
  apply Hhp'.
  rewrite H in Htst;inversion_clear Htst;reflexivity.
  assert (tempEnd = ohpEnd).
  assert (Heap.get (heapOf s) (4 + (rget s rgcInfo)) = Some ohpEnd). 
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).  
  rewrite H0 in Htend;inversion_clear Htend;reflexivity.
  subst tempEnd; subst tempSt; clear Htst Htend.

  generalize Hg';clear Hg' Hhp'.
  regRewrite rbot; regRewrite rbuff; stSimpl.
  intro Hg'.

  rewrite e in Hsb; simpl in Hsb.

  destruct Hroot as [Hrt1 | Hroot].
  rewrite (Hroot' Hrt1) in Hsb.
  elimtype False; apply (Hsb (refl_equal nil)).

  apply (Hg' B (set_remove (rget s' v0) G) W);clear Hg'.
  destruct (Hg G _ (sconjAssocL (sconjComm 
    (sconjAssocR (sconjAssocR Hhp)))) Hsb) as 
  [Hv0' Hhp'];clear Hg.

  split;trivial.
  split;trivial.
  split.
  regRewrite rroot;stSimpl.
  elimSetIn.
  split.
  elimSetIn.
  split.
  rewrite <- Hobjs.
  split; intro; elimSetIn. 

  apply sconjAssocL;apply sconjAssocL;
    apply sconjAssocL;apply sconjComm.
  apply sconjAssocR.
  generalize Hhp';clear Hhp'.
  unfold setHp.
  regRewrite rbot;regRewrite rbuff;
  regRewrite rgcInfo;stSimpl;intro Hhp'.
  eapply sconjImplR.
  apply Hhp'.
  clear Hhp';intros h Hp.
  eapply sconjImpl.
  apply Hp.
  clear Hp h;intros h Hp.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  apply aconjIntro.
  apply (aconjEL Hp).
  hypReplace Hp (aconjER Hp).
  eapply sconjImpl.
  apply Hp.
  clear Hp h; unfold objHp; intros h Hp.
  eapply fasImpl0.
  2: apply Hp.
  clear h Hp;  unfold okObjAt; 
    unfold okFieldAt;intros x h Hp.
  linLogElimOld Hp.
  linLogIntroOld.
  existsl x0;linLogIntroOld.
  existsl x1;linLogIntroOld.
  destruct H3;[left | right];trivial.
  elimSetIn.
  eapply sconjImplR.
  apply Hp.
  clear h Hp;intros h Hp.
  apply Hp.
  destruct H4;[left | right];trivial.
  elimSetIn.
  clear h Hp; unfold objHp; intros h Hx.
  simpl in Hx.
  assert (set_In (rget s' v0) (set_union G W)).
  elimSetIn.
  assert (Hx' := fasElim H Hx).
  eapply sconjImpl.
  apply (sconjComm Hx').
  clear Hx' Hx h; intros h Hh.
  eapply fasEqvSet.
  apply Hh.
  split; intro. 
  elimSetIn.
  assert (~ set_In (rget s' v0) W).
  intro X.
  assert (Hx := fasElimL  X 
    (sconjAssocL (sconjComm (sconjAssocR Hhp)))).
  assert (hget s (rget s' v0 - 4) = Some 0).
  unfold hget;eapply contGetEqL;
    apply (sconjAssocL Hx);clear Hx.
  assert (Hx' := fasElimL  Hv0' 
    (sconjAssocL (sconjComm  Hhp))).
  assert (hget s (rget s' v0 - 4) = Some 1).
  unfold hget;eapply contGetEqL;
    apply (sconjAssocL Hx');clear Hx'.
  rewrite H2 in H1;inversion_clear H1.
  elimSetIn.

  clear h Hx H Hx';intros h Hh.
  apply fasIntro;auto.
  apply unitIR.
  apply Hh.
  
  clear h Hp; intros h Hp.
  apply (sconjImplL Hp).
  clear h Hp; intros h Hp.
  apply sconjAssocR; apply sconjComm; 
    apply sconjAssocR;
      apply (sconjImplR Hp).
  clear h Hp; intros h Hp.
  assert (Hx' := fasElim Hv0' Hp).
  apply sconjComm; exact Hx'.

  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst2 Hg Hg' Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | 
    [H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  elimtype False;auto.
  clear Hcall.
  split.
  rewrite <- Hcst1;regEqSolver. 
  destruct Hcst2 as [Hcst2 [Hrt Hrt']].
  repeat split.
  rewrite <- Hcst2;regEqSolver.
  rewrite <- Hrt;regEqSolver.
  rewrite <- Hrt';regEqSolver.

  split.
  unfold return2Pre.
  tauto.

  clear B G W  Hst Hed ohpSt ohpEnd 
    freePtr Hroot Hobjs Hhp Hsb Hroot'.
  unfold return2Guar.
  intros s' Hg.
  unfold markLoopBodyGuar.
  subst s';stSimpl.
  split.
  intros.
  assert (Hsb := H);clear H.
  assert (Hp := H0);clear H0.
  destruct (rget s v0).
  elimtype False;auto.
  simpl in Hsb;clear e.
  subst G.
  exists B;exists W.
  destruct Hp as [Hst [Hed [Hroot [Hroot' [Hobjs  Hhp]]]]].
  split;auto.
  split;auto.
  split;auto.
  stSimpl.
  rewrite <- set_unionEmpty in Hroot;apply Hroot.
  split.
  rewrite <- Hobjs.
  rewrite <- set_unionEmpty.
  apply set_eqvRefl.

  apply sconjAssocL.
  unfold empty_set in Hhp.
  apply sconjImplL with 
    ((objHdrs B 1 ** objHdrs nil 1) ** objHdrs W 0).
  2: intros h Hhp'.
  2: eapply sconjImplR.
  2: apply (sconjComm Hhp').
  2: clear Hhp' h;intros h Hhp'.
  2: unfold objHdrs in Hhp';unfold fas in Hhp'; 
    unfold fas' in Hhp';simpl in Hhp'.
  2: apply (unitER Hhp').
  assert (XXX := Hhp); unfold buffStackSet in XXX;
    unfold buffArrayStack in XXX; linLogElimOld XXX.
  destruct H12 as [XX _].
  replace (rget s rtop) with (rget s rbot) in Hhp.
  stSimpl.
  simpl in Hhp.
  rewrite <- set_unionEmpty in Hhp.
  apply (sconjAssocR (sconjAssocR Hhp)).
  destruct H9.
  rewrite (eqvUnitIsUnit H) in XX.
  simpl in XX;omega.

  split.
  calleeSavedOkSolver'.
  split.
  reflexivity.
  tauto.
Qed.

Theorem markFirstImplOk : 
  iseqOkS markP markFirstPre markFirstGuar markFirstImpl.
Proof.
  unfold markFirstImpl.
  iseqWeakSolver PsimplMark.

  unfold markFieldGuar.
  intros s s' _ [_ [_ Hr]].
  regEqSolver.

  intros s [B [G [W  [ohpSt [ohpEnd [freePtr Hp]]]]]].
  destruct Hp as [Hst [Hed [Hroot [ Hv0 
    [Hobjs  Hhp]]]]].
  assert (XXX := Hhp).
  aconjE2 Hhp.
  clear Hhp; rename Hh into Hhp.
  assert (Hhp' := (sconjAssocL (sconjComm 
    (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL 
	(sconjComm (sconjAssocR (sconjAssocR 
	  (sconjAssocR (sconjAssocR 
	    (sconjAssocR (sconjAssocR Hhp)))))))))))))).
  destruct (inObjHpGetSome' (x:= rget s v0) (s := s) (objs := (rget s v0)::empty_set) (objs' := (rangeMSObjs ohpSt ohpEnd)))
    as [[lchd [XA [HA Hlchd]]] [rchd [XB [HB Hrchd]]]]; elimSetIn.
  eapply sconjImplR.
  apply Hhp'.
  intros; linLogIntro.
  clear Hhp'.

  rewrite (lwStep HA);stSimpl;auto.
  destruct Hlchd as [Hatom | Hlin].

  (* first branch : atomP(rget s a0) *)
  split.
  split.
  left; apply Hatom.

  intros s' [Hg [_ Hr]].
  destruct (Hg (or_introl _ Hatom)) as [Hh Hrtop].
  simpl in Hh;clear Hg.
  exists B; exists G; exists W; 
    exists ohpSt; exists ohpEnd; exists freePtr; exists lchd.
  rewrite <- Hrtop. 
  regRewrite rroot. 
  regRewrite rbot.
  regRewrite v0.
  regRewrite rbuff. 
  regRewrite rgcInfo.  
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  assert (hget s (rget s v0) = Some lchd).
  unfold hget.
  eapply contGetEqL.
  apply HA.
  unfold hget in *;rewrite Hh;auto.
  left; trivial.

  assert (Hhp' := sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocR (sconjAssocR (
      sconjAssocR (sconjAssocR Hhp))))))); 
  clear Hhp .
  assert (exists tempSt, Heap.get (heapOf s) 
    (rget s rgcInfo) = Some tempSt /\ tempSt = ohpSt).
  exists ohpSt.
  split;trivial.
  eapply contGetEqL.
  apply Hhp'.
  assert (exists tempEnd, Heap.get (heapOf s) 
    (4 + rget s rgcInfo) = Some tempEnd /\ tempEnd = ohpEnd).
  exists ohpEnd.
  split;trivial.
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).
  destruct H as [tempSt [Htst Hest]].
  subst ohpSt.
  destruct H0 as [tempEnd [Htend Heend]].
  subst ohpEnd.
  
  clear B G W freePtr Hhp' Hobjs Hst XXX
    Hed Hroot Hv0 Hrchd.
  intros s'' [s' [[Hg1 [Hg2 Hr]] [Hg' [Hcall [Hcst1 Hcst2]]]]].
  split.
  intros.
  destruct H as [Hst [Hed [Hroot [ Hv0 [Hobjs
    Hhp]]]]].
  assert (Hhp' := sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocR (sconjAssocR (
      sconjAssocR (sconjAssocR Hhp))))))).
  assert (tempSt = ohpSt).
  assert ( Heap.get (heapOf s) (rget s rgcInfo) = Some ohpSt). 
  eapply contGetEqL.
  apply Hhp'.
  rewrite H in Htst;inversion_clear Htst;reflexivity.
  assert (tempEnd = ohpEnd).
  assert (Heap.get (heapOf s) (4 + (rget s rgcInfo)) = Some ohpEnd). 
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).  
  rewrite H0 in Htend;inversion_clear Htend;reflexivity.
  subst tempEnd; subst tempSt; clear Htst Htend.
  clear Hhp' Hcall.
  generalize Hg'; clear Hg'.
  regRewrite rbot.
  regRewrite rbuff.
  intro Hg'.
  apply (Hg' B G W) ;clear Hg'.
  exists lchd.
  
  destruct (Hg1 (or_introl _ Hatom)) as [Hh Hrtop].
  simpl in Hh;clear Hg1 Hg2.
  rewrite <- Hrtop.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite v0.
  regRewrite rgcInfo.
  regRewrite rroot.
  stSimpl.
  repeat (split; trivial).
  assert (hget s (rget s v0) = Some lchd).
  unfold hget.
  eapply contGetEqL.
  apply HA.
  unfold hget in *;rewrite Hh;auto.
  left;trivial.
  rewrite Hh; trivial.
  
  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst2 Hg1 Hg2 Hg' Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | [
    H | [H | [H | [H | H]]]]]]]];
    try ((subst r; regEqSolver) ||
      (elimtype False; apply H)).
  clear Hcall.
  split.
  rewrite <- Hcst1; regEqSolver. 
  destruct Hcst2 as [Hcst2 [Hrt Hrt']].
  repeat split.
  rewrite <- Hcst2; regEqSolver.
  rewrite <- Hrt; regEqSolver.
  rewrite <- Hrt'; regEqSolver.
  
(* second branch *)
  split.
  split.
  clear XXX.
  right.
  split.
  eapply MSPtrGtFour.  
  apply Hlin.
  apply (refl_equal (rangeMSObjs ohpSt ohpEnd)).
  split.
  eapply MSPtrNotAtom.
  apply Hlin.
  apply Hst.
  apply (refl_equal (rangeMSObjs ohpSt ohpEnd)).
  destruct (Hobjs lchd) as [_ XXX].
  assert (Hl := XXX Hlin).
  inversion_clear Hl.
  subst lchd.
  exists G.
  right.
  split.
  
  eapply xNotInG.
  2: apply Hobjs.
  2: apply Hv0.
  2: trivial.
  apply (sconjImplL
    (sconjAssocR (sconjAssocR Hhp)));
  intros h Hh.
  apply sconjAssocL;apply Hh.
  linLogIntroOld.
  existsl 1.
  eapply sconjImplR.
  apply Hhp.
  intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjComm Hh)).
  clear h Hh; intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocL Hh)))).
  intros; linLogIntro.

  destruct (set_union_elim H) as [Hl | Hl].
  destruct (set_union_elim Hl) as [Hl' | Hl'].
  exists G.
  right.
  split.
  
  eapply xInBNotInG.
  2: apply Hobjs.
  2: apply Hlin.
  2: trivial.
  2: apply Hl'.
  eapply sconjImplL.
  apply (sconjAssocR (sconjAssocR Hhp)).
  intros h Hh.
  apply sconjAssocL;apply Hh.
  linLogIntroOld.
  existsl 1.
  assert (Hh := (sconjAssocL (fasElimL  Hl' 
    (sconjAssocL (sconjComm Hhp)))));clear Hhp.
  eapply sconjImplR.
  apply Hh.
  clear Hh;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (sconjComm Hh))).
  clear h Hh; intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocL Hh)))).
  intros; linLogIntro.

  exists G.
  left.
  split;trivial.
  assert (Hh := (sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocL (sconjComm Hhp))))));
  clear Hhp.
  eapply sconjImplR.
  apply Hh.
  clear Hh;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocL Hh)))).
  intros; linLogIntro.

  exists G.
  right.
  split.
  intro Hin'.
  assert (Heap.get (heapOf s) (lchd - 4) = Some 1).
  eapply contGetEqL.
  apply (sconjAssocL (
    fasElimL  Hin' (sconjAssocL 
    (sconjAssocL (sconjComm (
      sconjAssocL (sconjComm Hhp))))))).
  assert (Heap.get (heapOf s) (lchd - 4) = Some 0).
  eapply contGetEqL.
  apply (sconjAssocL (fasElimL  
    Hl (sconjAssocL (sconjAssocL (
      sconjComm (sconjAssocL 
	(sconjAssocL (sconjComm (
	  sconjAssocL (sconjComm Hhp)))))))))).
  rewrite H0 in H1;inversion_clear H1.
  assert (Hh' := (sconjAssocL (fasElimL  Hl 
    (sconjAssocL (sconjAssocL (sconjComm (
      sconjAssocL (sconjAssocL (sconjComm (
	sconjAssocL (sconjComm Hhp))))))))))); clear Hhp.
  linLogIntroOld.
  existsl 0.
  eapply sconjImplR.
  apply Hh'.
  clear Hh';intros h Hh.
  eapply sconjImplR.
  apply (sconjComm (sconjAssocR (sconjAssocR Hh))).
  clear Hh h;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL  (sconjAssocL (
    sconjComm (sconjAssocL Hh)))).
  intros; linLogIntro.

  intros s' [Hg1 [Hg2 Hr]].
  assert (HhpO := XXX); clear XXX.
  destruct (Hobjs lchd) as [_ XXX].
  assert (Hl := XXX Hlin).
  inversion_clear Hl.

  (* lchd = rget s v0 *)
  clear Hg2.
  subst lchd.
  assert (s |||- rget s v0 - 4 |-> 1 ** top).
  eapply sconjImplR.
  apply Hhp.
  intros; linLogIntro.
  clear XXX.
  destruct (Hg1 (or_intror _ H)) as [Hh Hrtop];
    clear H Hg1.
  exists B; exists G; exists W; 
    exists ohpSt; exists ohpEnd;
    exists freePtr; exists (rget s v0).
  rewrite <- Hrtop.
  regRewrite rroot.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  regRewrite v0.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  
  repeat (split; trivial).
  assert (hget s (rget s v0) = Some (rget s v0)).
  unfold hget.
  eapply contGetEqL.
  apply HA.
  unfold hget in *;rewrite Hh; auto.
  right; elimSetIn.

  destruct (set_union_elim H) as [Hl | Hl].

  (*  lchd \in B \cup G *)
  clear Hg2.
  assert (s |||- lchd - 4 |-> 1 ** top).
  destruct (set_union_elim Hl) as [Hl' | Hl'].
  assert (Hh := (sconjAssocL (fasElimL  Hl'
    (sconjAssocL (sconjComm Hhp)))));clear Hhp.
  eapply sconjImplR.
  apply Hh.
  intros; linLogIntro.
  assert (Hh := (sconjAssocL (fasElimL  
    Hl' (sconjAssocL (sconjAssocL (
      sconjComm (sconjAssocL 
	(sconjComm Hhp)))))))).
  eapply sconjImplR.
  apply Hh.
  intros; linLogIntro.
  clear H.
  destruct (Hg1 (or_intror _ H0)) as [Hh Hrtop];clear H0.
  exists B; exists G; exists W; 
    exists ohpSt; exists ohpEnd; exists freePtr; exists lchd.
  rewrite <- Hrtop. 
  regRewrite rroot. 
  regRewrite rbot.
  regRewrite rbuff. 
  regRewrite rgcInfo. 
  regRewrite v0.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  assert (hget s (rget s v0) = Some lchd).
  unfold hget.
  eapply contGetEqL.
  apply HA.
  unfold hget in *;rewrite Hh; auto.
  right; elimSetIn.

  (*  lchd \in W *)
  assert (s |||- lchd - 4 |-> 0 ** objHdrs G 1 ** 
    buffStackSet (rget s rbot) (rget s rtop) (rget s rbuff) G **
    rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr 
      :: rget s rbot :: rget s rbuff :: nil) ** top **
    (objHp (set_add (rget s v0) 
      (set_union B G)) B ** 
      objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W) ** 
      objHp (rangeMSObjs ohpSt ohpEnd) ((rget s v0)::empty_set)) **
      objHdrs B 1 ** objHdrs (set_remove lchd W) 0 ** 
      rget s v0 - 4 |-> 1).
  clear HhpO Hg1 Hg2.
  assert (Hh := (sconjAssocL (fasElimL  
    Hl (sconjAssocL (sconjAssocL (
      sconjComm (sconjAssocL 
	(sconjAssocL (sconjComm (
	  sconjAssocL (sconjComm Hhp))))))))))).
  apply (sconjImplR Hh).
  clear Hh Hhp;intros h Hh.
  apply sconjComm;apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (sconjComm Hh))).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply (sconjImplR Hh);
    clear h Hh; intros h Hh.
  apply sconjAssocR;apply sconjComm; 
    repeat (apply sconjAssocR).
  eapply sconjImplR.
  apply (sconjComm Hh).
  clear h Hh;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjComm; apply Hh.
  clear Hhp; rename H0 into Hhp.
  assert (~ atomP lchd).
  clear Hg1 Hg2 HhpO.
  assert (set_In lchd (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  eapply MSPtrNotAtom.
  apply H0.
  apply Hst.
  apply (refl_equal (rangeMSObjs ohpSt ohpEnd)).
  rename H0 into Hna; clear H.
  clear Hg1.

  exists B; exists (set_add lchd G); exists (set_remove lchd W); 
    exists ohpSt; exists ohpEnd;
    exists freePtr; exists lchd.
  regRewrite rroot.
  regRewrite rbot.
  regRewrite rbuff.
  regRewrite rgcInfo.
  regRewrite v0.
  stSimpl.
  clear HhpO.
  Opaque set_eqv.
  repeat (split; trivial).
  clear Hhp Hg2.
  elimSetIn.
  rewrite <- Hobjs.
  Transparent set_eqv.
  split; intro; elimSetIn.
  unfold objHp in Hhp;unfold fas in Hhp; 
    simpl in Hhp.
  unfold okObjAt in Hhp; unfold okFieldAt in Hhp; 
    linLogElimOld Hhp.
  assert (hget s (rget s v0) = Some x).
  assert (Hhp' := (sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocL (
      sconjAssocL (sconjComm (sconjAssocL (
	sconjAssocL (sconjComm (sconjAssocL ( 
	  sconjAssocL (sconjComm (sconjAssocL (
	    sconjComm (sconjAssocR (
	      sconjAssocR (sconjAssocR (
		sconjAssocR (sconjAssocR (
		  sconjAssocR Hhp))))))))))))))))))))).
  unfold hget.
  eapply contGetEqL.
  apply Hhp'.
  assert (Hhp'' := Hg2 _ _ Hna Hhp);clear Hhp.
  assert (hget s' (rget s v0) = Some x).
  assert (Hhp' :=   (sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocL (sconjAssocL (
      sconjComm (sconjAssocL (sconjAssocL (
	sconjComm (sconjAssocL (
	  sconjComm (sconjAssocR (
	    sconjAssocR (sconjAssocR (
	      sconjAssocR (sconjAssocR (
		sconjAssocR Hhp'')))))))))))))))))).
  clear Hhp''.
  unfold hget.
  eapply contGetEqL.
  apply Hhp'.
  assert (hget s (rget s v0) = Some lchd).
  unfold hget.
  eapply contGetEqL.
  apply HA.
  rewrite H in H1;rewrite H0;apply H1.
  clear Hg2 Hhp.
  right;elimSetIn.
  assert (Hhp' := Hg2 _ _ Hna Hhp);clear Hhp.
  apply sconjComm; apply sconjAssocR; apply sconjComm;
    apply sconjAssocR; apply sconjAssocR.
  apply (sconjImplR Hhp');
    clear Hg2 Hhp'; intros h Hh.
  apply sconjAssocR;apply sconjComm;
    repeat (apply sconjAssocR).
  generalize Hh.
  regRewrite rbot.
  regRewrite rbuff.
  stSimpl.
  clear Hh; intro Hh.
  eapply sconjImplR.
  apply Hh.
  clear Hh h; intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply Hh.
  clear Hh h; intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply Hh.
  clear Hh h; intros h Hh.
  eapply sconjImpl.
  apply Hh.
  clear Hh h; intros h Hh.
  apply aconjIntro.
  intros; linLogIntro.
  eapply sconjImpl.
  apply Hh.
  clear Hh h;unfold objHp;intros h Hn.
  eapply fasImpl0.
  2: apply Hn.
  clear h Hn;unfold okObjAt; unfold okFieldAt; 
    intros x h Hx.
  eapply sconjImpl.
  apply Hx.
  clear h Hx;intros h Hh;linLogElimOld Hh.
  existsl x0;linLogIntroOld;auto.
  destruct H1 as [Hx | Hx];
    [left | right]; trivial.
  elimSetIn.
  
  clear h Hx ;intros h Hh;linLogElimOld Hh.
  existsl x0;linLogIntroOld;auto.
  destruct H1 as [Hx | Hx];
    [left | right]; trivial.
  elimSetIn.
  clear h Hh;intros h Hh.
  eapply sconjImplL.
  apply Hh.
  clear h Hh;unfold objHp; intros h Hh.
  eapply fasEqvSet.
  apply Hh.
  split;intro XX; elimSetIn.
  clear h Hh; intros h Hh.
  apply sconjAssocR.
  apply (sconjComm (sconjAssocR Hh)).
   
  clear XXX.
  assert (Hhp' := sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocR (sconjAssocR 
    (sconjAssocR (sconjAssocR Hhp))))))); clear Hhp .
  assert (exists tempSt, Heap.get (heapOf s) 
    (rget s rgcInfo) = Some tempSt /\ tempSt = ohpSt).
  exists ohpSt.
  split;trivial.
  eapply contGetEqL.
  apply Hhp'.
  assert (exists tempEnd, Heap.get (heapOf s) 
    (4 + rget s rgcInfo) = Some tempEnd /\ tempEnd = ohpEnd).
  exists ohpEnd.
  split;trivial.
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).
  destruct H as [tempSt [Htst Hest]].
  subst ohpSt.
  destruct H0 as [tempEnd [Htend Heend]].
  subst ohpEnd.
  
  
  clear B G W freePtr Hhp' Hobjs Hst
    Hed Hroot Hv0.
  intros s'' [s' [[Hg1 [Hg2  Hr]] [Hg' [
    Hcall [Hcst1 Hcst2]]]]].
  split.
  intros.
  destruct H as [Hst [Hed [Hroot [ Hv0 
    [Hobjs  Hhp]]]]].
  assert (Hhp' := sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocR 
      (sconjAssocR (sconjAssocR (
	sconjAssocR Hhp))))))).
  assert (tempSt = ohpSt).
  assert ( Heap.get (heapOf s) (rget s rgcInfo) = Some ohpSt). 
  eapply contGetEqL.
  apply Hhp'.
  rewrite H in Htst;inversion_clear Htst;reflexivity.
  assert (tempEnd = ohpEnd).
  assert (Heap.get (heapOf s) (4 + (rget s rgcInfo)) = Some ohpEnd). 
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hhp'))).  
  rewrite H0 in Htend;inversion_clear Htend;reflexivity.
  subst tempEnd; subst tempSt; clear Htst Htend.
  clear Hhp'.

  generalize Hg';clear Hg'.
  clear Hcall.
  regRewrite rbot. 
  regRewrite rbuff. 
  stSimpl.
  intro Hg'.
  

  destruct (Hobjs lchd) as [_ XXX].
  assert (Hl := XXX Hlin).
  inversion_clear Hl.
  subst lchd.
  clear Hg2.

  (* lchd = rget s v0 *)
  clear XXX.
  apply (Hg' B G W);auto;clear Hg'.
  exists (rget s v0).
  assert (s |||- rget s v0 - 4 |-> 1 ** top).
  eapply sconjImplR.
  apply Hhp.
  intros; linLogIntro.
  destruct (Hg1 (or_intror _ H)) as 
    [Hh Hrtop];clear H Hg1.
  rewrite <- Hrtop. 
  regRewrite rroot.
  regRewrite rbot. 
  regRewrite rbuff.
  regRewrite rgcInfo.
  regRewrite v0.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  clear Hhp.
  assert (hget s (rget s v0) = Some (rget s v0)).
  unfold hget.
  eapply contGetEqL.
  apply HA.
  unfold hget in *;rewrite Hh;auto.
  clear Hhp.
  right;elimSetIn.

  destruct (set_union_elim H) as [Hl | Hl].

  (*  lchd \in B \cup G *)
  clear Hg2.
  apply (Hg' B G W);auto;clear Hg'.
  exists lchd.
  assert (s |||- lchd - 4 |-> 1 ** top).
  destruct (set_union_elim Hl) as 
    [Hl' | Hl'].
  assert (Hh := (sconjAssocL (fasElimL  
    Hl'(sconjAssocL (sconjComm Hhp)))));clear Hhp.
  eapply sconjImplR.
  apply Hh.
  intros; linLogIntro.
  assert (Hh := (sconjAssocL (fasElimL  Hl' 
    (sconjAssocL (sconjAssocL (sconjComm (
      sconjAssocL (sconjComm Hhp)))))))).
  clear Hhp.
  eapply sconjImplR.
  apply Hh.
  intros; linLogIntro.
  clear H.
  destruct (Hg1 (or_intror _ H0)) as 
    [Hh Hrtop];clear H0.
  rewrite <- Hrtop.
  regRewrite rroot.
  regRewrite rbot. 
  regRewrite rbuff.
  regRewrite rgcInfo.
  regRewrite v0.
  stSimpl.
  rewrite Hh;clear Hr Hrtop.
  repeat (split; trivial).
  clear Hg1 Hhp.
  assert (hget s (rget s v0) = Some lchd).
  unfold hget.
  eapply contGetEqL.
  apply HA.
  unfold hget in *;rewrite Hh;auto.
  clear Hg1 Hhp.
  right;elimSetIn.

  (*  lchd \in W *)
  assert (s |||- lchd - 4 |-> 0 ** objHdrs G 1 ** 
    buffStackSet (rget s rbot) (rget s rtop) 
    (rget s rbuff) G **
    rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr 
      :: rget s rbot :: rget s rbuff :: nil) ** A' ** (A &&
        objHp (set_union (rget s v0 :: empty_set) 
	  (set_union B G)) B **
        objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W) **
        objHp (rangeMSObjs ohpSt ohpEnd) 
	(rget s v0 :: empty_set)) **
    objHdrs B 1 ** objHdrs (set_remove lchd W) 0 ** 
    rget s v0 - 4 |-> 1).
  clear Hg1 Hg2.
  assert (Hh := (sconjAssocL (fasElimL  Hl
    (sconjAssocL (sconjAssocL (sconjComm (
      sconjAssocL (sconjAssocL (sconjComm (
	sconjAssocL (sconjComm Hhp))))))))))).
  clear Hhp.
  eapply sconjImplR.
  apply Hh.
  clear Hh;intros h Hh.
  apply sconjComm;apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL (sconjAssocL (sconjComm Hh))).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjAssocR.
  apply (sconjImplR (sconjAssocL Hh));
    clear h Hh;intros h Hh.
  apply sconjAssocR.
  apply (sconjImplR Hh);
    clear h Hh;intros h Hh.
  apply sconjAssocR;apply sconjComm;
    repeat (apply sconjAssocR).
  eapply sconjImplR.
  apply (sconjComm Hh).
  clear h Hh;intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL Hh).
  clear h Hh;intros h Hh.
  apply sconjComm;apply Hh.
  rename H0 into HXXX.
  assert (~ atomP lchd).
  clear Hg1 Hg2 Hhp HXXX.
  assert (set_In lchd (rangeMSObjs ohpSt ohpEnd)).
  destruct (Hobjs lchd).
  apply H0;elimSetIn.
  eapply MSPtrNotAtom.
  apply H0.
  apply Hst.
  reflexivity.
  assert (Hna := H0); clear H H0.

  apply (Hg' B (lchd::G) (set_remove lchd W));
    clear Hg'; auto.
  exists lchd.

  assert (Hhp' := Hg2 _ _ Hna HXXX).
  hypReplace HXXX (aconjERR (sconjAssocR 
    (sconjComm (sconjAssocR 
      (sconjAssocR (sconjAssocR 
	(sconjAssocR (sconjAssocR HXXX)))))))).
  hypReplace HXXX  (sconjAssocL (sconjComm (sconjAssocL 
    (sconjAssocL (sconjComm (sconjAssocL 
    (sconjAssocL (sconjComm (sconjAssocL 
      (sconjComm (sconjAssocL HXXX))))))))))).
  clear Hhp.

  clear Hg1.
  assert (In : set_In (rget s v0) (rget s v0::empty_set)).
  left; reflexivity.
  unfold objHp in HXXX; unfold fas in HXXX; 
    unfold fas' in HXXX; simpl in HXXX.
  fold fas' in HXXX.
  hypReplace HXXX (sconjComm 
    (unitEL (sconjAssocL (sconjComm (
    sconjAssocL HXXX))))).
  unfold okObjAt in HXXX; unfold okFieldAt in HXXX;
    linLogElimOld HXXX.
  assert (Heap.get (heapOf s) (rget s v0) = Some x).
  eapply contGetEqL; apply HXXX.
  assert (Heap.get (heapOf s) (4 + rget s v0) = Some x0).
  eapply contGetEqL; apply (sconjAssocL (
    sconjComm HXXX)).
  assert (x = lchd).
  assert (Heap.get (heapOf s ) (rget s v0) = Some lchd).
  eapply contGetEqL; apply HA.
  rewrite H in H1.
  inversion_clear H1; reflexivity.
  subst x.
  assert (x0 = rchd).
  assert (Heap.get (heapOf s) (4 + rget s v0) = Some rchd).
  eapply contGetEqL; apply HB.
  rewrite H1 in H0.
  inversion_clear H0; reflexivity.
  subst x0.
  clear H3 H6 H H0.
  
  hypReplace HXXX (sconjAssocL (sconjComm (
    sconjAssocR (sconjAssocR 
    (sconjAssocR (sconjAssocR  HXXX)))))).
  fold fas' in HXXX.
  assert (s |||- lchd - 4 |-> 0 ** objHdrs G 1 ** 
    buffStackSet (rget s rbot) (rget s rtop) 
    (rget s rbuff) G **
    rget s rgcInfo |->, (ohpSt :: ohpEnd :: freePtr 
      :: rget s rbot :: rget s rbuff :: nil) **  A' ** (
	objHp (rget s v0 :: set_union B G) B **
	objHp (rangeMSObjs ohpSt ohpEnd) (set_union G W) **
	(((rget s v0 |-> lchd ** 4 + rget s v0 |-> rchd) ** 
	  objHdrs B 1) **
          objHdrs (set_remove lchd W) 0) ** rget s v0 - 4 |-> 1)).
  clear Hhp' Hg2.
  apply (sconjImplR HXXX); intros h Hh.
  apply (sconjImplR (sconjAssocL Hh)); 
    clear h Hh; intros h Hh.
  apply (sconjImplR (sconjAssocL Hh)); 
    clear h Hh; intros h Hh.
  apply sconjAssocL;   
    apply sconjAssocL; apply sconjAssocL.
  apply (sconjImplL Hh); 
    clear h Hh; intros h Hh.
  apply sconjAssocR.
  simpl.
  apply sconjAssocR.
  apply sconjAssocR.
  apply (sconjImplR  Hh); 
    clear h Hh; intros h Hh.
  apply sconjAssocR.
  apply (sconjImplR  Hh); 
    clear h Hh; intros h Hh.
  apply sconjAssocR.
  apply (sconjImplR  Hh); 
    clear h Hh; intros h Hh.
  apply sconjAssocR.
  apply (sconjImplR  Hh); 
    clear h Hh; intros h Hh.
  apply Hh.
  assert (Hhp'' := Hg2 _ _ Hna H);
    clear HXXX H In Hg2.

  regRewrite rroot.
  regRewrite rbot.
  regRewrite rbuff. 
  regRewrite rgcInfo.
  regRewrite v0.
  stSimpl.
  Opaque set_eqv.
  repeat (split; trivial).
  clear Hhp'' Hhp'.
  elimSetIn.
  rewrite <- Hobjs.
  Transparent set_eqv.
  split; intro; elimSetIn.
  hypReplace Hhp'' (sconjAssocL (sconjAssocL (
    sconjAssocL (sconjAssocL (sconjComm (
      sconjAssocR (sconjAssocR (sconjAssocR 
	(sconjAssocR (sconjAssocR Hhp'')))))))))).
  unfold hget.
  eapply contGetEqL.
  apply Hhp''.
  clear Hhp''  Hhp'.
  right;elimSetIn.

  clear Hhp''.
  repeat (apply sconjAssocL). 
  apply sconjComm.
  hypReplace Hhp' (sconjAssocL (
    sconjComm (sconjAssocR 
      (sconjAssocR (sconjAssocR Hhp'))))).
  destruct (sconjElim Hhp') as [h1 [h2 [J [Hh1 Hh2]]]];
    clear Hhp'.
  apply sconjIntro with h1 h2.
  clear Hh2.
  apply aconjIntro.
  apply (aconjEL Hh1).
  hypReplace Hh1 (aconjER Hh1).
  apply sconjAssocL.
  apply (sconjImplL (sconjAssocR Hh1)); intros h Hh.
  apply (sconjImpl Hh); clear h Hh; intros h Hh.
  unfold objHp.
  eapply fasImpl0.
  2:unfold objHp in Hh; apply Hh.
  clear h Hh  Hh1;unfold okObjAt; unfold okFieldAt; 
    intros x h Hx.
  apply (sconjImpl Hx);
    clear h Hx;intros h Hh.
  linLogElimOld Hh.
  existsl x0;linLogIntroOld;auto.
  destruct H1 as [Hx | Hx];
    [left; trivial | right; elimSetIn].
  linLogElimOld Hh.
  existsl x0;linLogIntroOld;auto.
  destruct H1 as [Hx | Hx];
    [left; trivial | right; elimSetIn].
  unfold objHp.
  eapply fasEqvSet.
  unfold objHp in Hh; apply Hh.
  split;intro XX; elimSetIn.
  apply (sconjImplL (sconjAssocR Hh2)).
  clear Hh1 Hh2.
  regRewrite rbot. regRewrite rbuff.
  stSimpl.
  intros h Hh.
  apply (sconjImplL (sconjAssocR Hh)).
  clear h Hh; intros h Hh.
  apply (sconjImplL (sconjAssocR Hh)).
  clear h Hh; intros h Hh.
  apply sconjAssocR.
  apply sconjComm.
  apply sconjAssocR.
  apply (sconjImplR (sconjComm Hh)).
  clear h Hh; intros h Hh.
  apply sconjAssocL.
  apply (sconjComm Hh).
  apply J.

  split.
  apply calleeSavedOkTrans' with s';
    [idtac | apply Hcall].
  clear Hcst1 Hcst2 Hg1 Hg2 Hg' Hcall.
  intros r XXX.
  destruct XXX as [H | [H | [H | [H | [
    H | [H | [H | [H | H]]]]]]]];
    try (subst r;regEqSolver).
  elimtype False;auto.
  clear Hcall.
  split.
  rewrite <- Hcst1;regEqSolver. 
  destruct Hcst2 as [Hcst2 [Hrt Hrt']].
  repeat split.
  rewrite <- Hcst2;regEqSolver.
  rewrite <- Hrt;regEqSolver.
  rewrite <- Hrt';regEqSolver.
Qed.
