Require Import MarkSweepGC.
(** ** Local specification map *)

Definition markFieldP f :=
  if EqNat.beq_nat STACKPUSH f then
    SomeT (bspeci pushPre pushGuar)
    else if EqNat.beq_nat RETURN3 f then
      SomeT (bspeci return3Pre return3Guar)
      else
	NoneT _.

Ltac PsimplMarkField := 
  unfold markFieldP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

(** ** markField () *)

Theorem markFieldImplOk : 
  iseqOkS markFieldP markFieldPre markFieldGuar markFieldImpl.
Proof.
  unfold markFieldImpl.
  iseqWeakSolver PsimplMarkField.
  
  unfold pushGuar; intros s s' _ [_ Hr].
  apply Hr; discriminate.

  unfold markFieldPre; unfold markFieldGuar.
  intros s Hp.
  rewrite addiuStep;stSimpl.
  rewrite andi1Step;stSimpl.
  destruct Hp as [Hp | Hp].
  rewrite (atomIsOdd Hp).
  unfold doBtest.
  rewrite beq_false;auto.
  simpl.
  unfold return3Pre;unfold return3Guar;stSimpl.
  split;auto.
  intros s' Hs';rewrite <- Hs';stSimpl.
  split;auto.
  split.
  intuition.
  regSolver.
  
  destruct Hp as [Ha0 [He [G Hp']]].
  rewrite (notAtomNotOdd He).
  unfold doBtest;unfold negb.
  rewrite beq_true;auto.
  rewrite addiuStep;stSimpl.
  rewrite subuStep;stSimpl.
  assert (s |||- rget s a0 - 4 |->_ ** 
    buffStackSet (rget s rbot) 
    (rget s rtop) (rget s rbuff) G ** top).
  destruct (set_In_dec (rget s a0) G) as [X1 | X2].
  linLogIntroOld.
  existsl 1.
  destruct Hp' as [[Hin Hp] | [Hin Hp]].
  assert (X := fasElimL X1 Hp).
  eapply sconjImplR.
  apply (sconjAssocL X).
  clear X; intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjComm Hh)).
  intros; linLogIntroOld.
  elimtype False;auto.
  destruct Hp' as [[Hin Hp] | [Hin Hp]].
  elimtype False;auto.
  eapply sconjImplR.
  apply Hp.
  intros h Hh.
  eapply sconjImplR.
  apply (sconjAssocL (sconjComm Hh)).
  intros; linLogIntroOld.

  replace (Bool.negb true) with false.
  2:reflexivity.
  assert (Hp := H);clear H.
  linLogElimOld Hp.
  rewrite (lwStep Hp);stSimpl;auto.
  rewrite addiuStep;stSimpl.
  compare x 1;intro e.
  split.
  unfold return3Pre.
  rewrite e;rewrite beq_true;auto.
  rewrite e;rewrite beq_true;auto.
  unfold return3Guar;stSimpl.
  intros s' Hs';rewrite <- Hs';stSimpl.
  split.
  intro H;destruct H;[intuition|simpl;auto].
  split.
  simpl.
  intros A G' _ Hhp.
  assert (X := contContEq Hp Hhp).
  rewrite e in X;discriminate X.
  regSolver.

  cut (~ set_In (rget s a0) G).
  intro XXX.
  rewrite beq_false;auto.
  rewrite (swStep Hp);stSimpl;auto.
  rewrite addiuStep;stSimpl.
  destruct pushStackIsPushSet as [G1 G2].
  split.
  split.
  apply G1.
  unfold stackPushPre; stSimpl.
  exists G.
  split;auto.
  replace (4 + (rget s a0 - (4 + 0))) with 
    (rget s a0);auto with arith.
  unfold setHp;stSimpl.
  simpl.
  assert (H := contUpdateSt Hp 1);clear Hp.
  eapply sconjImplR.
  apply (sconjAssocL (sconjComm H)).
  intros;intros; linLogIntroOld.
  
  intros s' [_ Hr].
  unfold return3Pre;trivial.

  unfold return3Guar; stSimpl;simpl.
  intros s' [s'' [G' Hs'']]. 
  destruct (G2 _ _ G') as [Hhp Hr];
    clear G1 G2 G'.
  generalize Hhp.
  unfold setHp.
  subst s';stSimpl.
  clear Hhp; intro Hhp.
  split.
  intro Hx;destruct Hx;elimtype False;auto.
  assert (hget s (rget s a0 - 4) = Some x).
  unfold hget.
  eapply contGetEqL; apply Hp.
  assert (hget s (rget s a0 - 4) = Some 1).
  unfold hget.
  eapply contGetEqL; apply H.
  rewrite H1 in H0.
  apply e.
  inversion_clear H0;reflexivity.

  split.
  simpl.
  clear Hr Hp Hp' G XXX;intros A G _ Hp.
  replace (S (S (S (S (rget s a0 - 4))))) with 
    (rget s a0) in Hhp.
  apply sconjComm; apply sconjAssocR.
  cut (~ set_In (rget s a0) G).
  intro XXX.
  apply Hhp;auto.
  apply sconjAssocL; apply sconjComm.
  unfold objHdrs; apply fasIntroL;auto.
  apply sconjAssocR.
  simpl.
  apply contUpdateSt with 0;apply Hp.
  assert (hget s (rget s a0 - 4) = Some 0).
  unfold hget.
  eapply contGetEqL; apply Hp.
  apply neqNotIn.  intros y Hy Heq.
  assert (Hp' := sconjAssocL (sconjComm Hp));
    clear Hp.
  assert (Hp := sconjAssocL 
    (fasElimL Hy Hp')).
  simpl in Hp.
  assert (hget s (y - 4) = Some 1).
  unfold hget.
  eapply contGetEqL; apply Hp.
  subst y;rewrite H in H0.
  apply e.
  inversion_clear H0;reflexivity.
  omega.
  intros.
  generalize (Hr r H1 H).
  induction r; stSimpl;auto.
  generalize (Hr rcst3).
  stSimpl; intros.
  symmetry; apply H3; discriminate.

  intro Hin.
  destruct Hp' as [[Hin' Hp'] | [Hin' Hp']].
  assert (Hp'' := sconjAssocL 
    (fasElimL Hin Hp')).
  simpl in Hp''.
  assert (hget s ((rget s a0) - 4) = Some 1).
  unfold hget.
  eapply contGetEqL; apply Hp''.
  assert (hget s ((rget s a0) - 4) = Some x).
  unfold hget.
  eapply contGetEqL; apply Hp.
  rewrite H in H0.
  apply e.
  inversion_clear H0;reflexivity.
  auto.
Qed.
