Require Import SCAP.
Require Import CAP.

Require Import MarkSweepGC.
Require Import MarkField.
Require Import Mark.
Require Import Sweep.

(** ** Local specification map *)

Definition markSweepP f :=
  if EqNat.beq_nat SWEEP f then
    SomeT (bspeci sweepPre sweepGuar)
    else if EqNat.beq_nat MARK f then
      SomeT (bspeci markPre markGuar)
      else if EqNat.beq_nat MARKSWEEPSWEEP f then
	SomeT (bspeci sweepPre markSweepSweepGuar)
	else if EqNat.beq_nat RETURN1 f then
	  SomeT (bspeci return1Pre return1Guar)
	  else
	    NoneT _.

Ltac PsimplmarkSweep := 
  unfold markSweepP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

(** ** Well-formedness Theorems *)

Theorem markSweepSweepImplOk : 
  iseqOkS markSweepP markSweepSweepPre 
  markSweepSweepGuar markSweepSweepImpl.
Proof.
  unfold markSweepSweepImpl.
  iseqWeakSolver PsimplmarkSweep.

  intros.
  destruct H0 as [_ [G _]].
  apply (G ra).
  unfold calleeSaved.
  simpl.
  intuition.

  unfold sweepPre; unfold sweepGuar; 
    unfold return1Pre; unfold sweepPreP; 
      unfold return1Guar; unfold markSweepSweepGuar.
  split.
  split; trivial.

  clear H.
  stSimpl.
  intros s'' [s' [[G [Gcall [Gcst Groot]]]]]. 
  intros Hs''; subst s''.
  stSimpl.
  split.
  intros.

  assert ((forall x : POnat.POnat.A, set_In x B ->
    hget s' x = hget s x /\ 
    hget s' (4 + x) = hget s (4 + x))).
  intros.
  destruct H as [Hst [Hed [Hroot [He Hh]]]]. 
  aconjE1 Hh.
  sconjLR Hh0.
  assert (Hh' := (fasElimL  H0 
    (sconjComm Hh0))); clear Hh0.
  unfold okObjAt in Hh'; 
    unfold okFieldAt in Hh'; linLogElimOld Hh'.
  assert (hget s x = Some x0).
  unfold hget; eapply contGetEqL; apply Hh'.
  assert (hget s (4 + x) = Some x1).
  unfold hget; eapply contGetEqL; 
    apply (sconjAssocL (sconjComm Hh')).
  hypReplace Hh' (sconjComm (sconjAssocR 
    (sconjAssocR Hh'))).
  
  clear Hh.
  destruct (G B W ohpSt ohpEnd freePtr stBot stBuff 
    ((x |-> x0 ** 4 + x |-> x1) **
      (foralls x in set_remove x B,
        (lexists y, !okFieldVal B y ** x |-> y) **
        (lexists y, !okFieldVal B y ** 4 + x |-> y))) A
  ) as [free' [freePtr' P']].
  repeat (split; trivial).
  hypReplace Hh' (aconjGR Hh').
  repeat (apply sconjAssocL).
  apply (sconjImpl Hh'); clear Hh'; intros h Hh.
  sconjLR Hh.
  apply Hh.
  hypReplace Hh (aconjEL Hh).
  apply aconjIntro; trivial.
  unfold objHp. 
  apply fasEqvSet with (x::set_remove x B).
  eapply fasIntro.
  apply (sconjImplL Hh).
  clear h Hh; intros h Hh.
  unfold okObjAt; unfold okFieldAt; linLogIntroOld.
  existsl x0; linLogIntroOld.
  existsl x1; linLogIntroOld.
  apply Hh.
  intro X.
  apply (set_remove_elim2 _ X (refl_equal x)).
  split; intros; elimSetIn.
  clear Hst Hed Hroot Hh' He.
  destruct P' as [Hst [Hed [Hroot [He Hh]]]].
  aconjE1 Hh; clear Hh.
  assert (Hh := sconjAssocL (sconjAssocL (sconjComm 
    (sconjAssocR (sconjAssocR (sconjAssocR (
      (sconjAssocR Hh0)))))))).
  assert (hget s' x = Some x0).
  unfold hget; eapply contGetEqL; apply Hh.
  assert (hget s' (4 + x) = Some x1).
  unfold hget; eapply contGetEqL; 
    apply (sconjAssocL (sconjComm Hh)).
  rewrite H; rewrite H1; rewrite H2; rewrite H3; 
    split; reflexivity.

  destruct (G B W ohpSt ohpEnd freePtr stBot stBuff 
  (objHp B B) A H) as [free' [freePtr' P']].
  exists free'; exists freePtr'.
  split; trivial.
  
  split.
  intuition.
  split.
  assert (Gcall' : calleeSavedOk' s'
    (rset s' ra (rget s' rcst1))).
  calleeSavedOkSolver'.
  eapply calleeSavedOkTrans'.
  2: apply Gcall'.
  eapply calleeSavedOkTrans'.
  2: apply othersNotRa; trivial.
  2: apply Gcall.
  calleeSavedOkSolver'. 
  split; intuition. 
Qed.


Theorem markSweepGCImplOk : 
  iseqOkS markSweepP markSweepGCPre markSweepGCGuar markSweepGCImpl.
Proof.
  unfold markSweepGCImpl.
  iseqWeakSolver PsimplmarkSweep.

  intros.
  destruct H0 as [_ [G _]].
  apply (G ra).
  unfold calleeSaved.
  simpl.
  intuition.

  unfold markSweepGCGuar; unfold markSweepSweepGuar.
  unfold sweepPre; unfold sweepGuar; unfold markPre;
    unfold markGuar; unfold sweepPreP; unfold markPreP;
      unfold markPost.
  split.
  split; trivial.

  stSimpl.
  intros s' [G [Gcall [Gcst Groot]]].
  destruct H as [objs [ohpSt [ohpEnd [freePtr [
    stBot [stBuff H]]]]]].
  destruct (G objs ohpSt ohpEnd freePtr stBot stBuff 
    top top H) as [B [W P']]; clear G.

  exists B; exists W; exists ohpSt;
    exists ohpEnd; exists freePtr; exists stBot;
      exists stBuff.
  destruct P' as [Hst [Hed [Hroot [ He Hh]]]].
  repeat (split; trivial).
  aconjE2 Hh.
  apply sconjAssocL; apply sconjAssocL; 
    apply sconjAssocL.
  apply (sconjImpl (sconjAssocR (sconjAssocR (
    sconjAssocR Hh0)))); clear Hh Hh0; intros h Hh.
  apply sconjAssocR.
  apply (sconjImplL (sconjAssocL Hh)).
  clear h Hh; intros h Hh.
  apply (sconjComm Hh).
  apply (sconjImplR Hh);
    clear h Hh; intros h Hh.
  apply sconjComm.
  apply (sconjImplL Hh);
    clear h Hh; intros h Hh.
  apply aconjIntro.
  intros; linLogIntroOld.
  exact Hh.
  
  clear H.
  stSimpl.
  intros s'' [s' [[G' [Gcall' [Gcst' Groot']]] 
    [G'' [Gcall'' [Gra'' Groot'']]]]].
  split.
  intros.

  assert ((forall x : POnat.POnat.A, set_In x objs ->
    hget s' x = hget s x /\ 
    hget s' (4 + x) = hget s (4 + x))).
  intros.
  destruct H as [Hst [Hed [Hroot [Hobjs Hh]]]].
  aconjE1 Hh.
  sconjLR Hh0.
  assert (Hh' := (fasElimL  H0 
    (sconjComm Hh0))); clear Hh0.
  unfold okObjAt in Hh'; 
    unfold okFieldAt in Hh'; linLogElimOld Hh'.
  assert (hget s x = Some x0).
  unfold hget; eapply contGetEqL; apply Hh'.
  assert (hget s (4 + x) = Some x1).
  unfold hget; eapply contGetEqL; 
    apply (sconjAssocL (sconjComm Hh')).
  hypReplace Hh' (sconjComm (sconjAssocR 
    (sconjAssocR Hh'))).
  
  clear Hh.
  destruct (G' objs ohpSt ohpEnd freePtr stBot stBuff 
    ((x |-> x0 ** 4 + x |-> x1) **
      (foralls x in set_remove x objs,
        (lexists y, !okFieldVal objs y ** x |-> y) **
        (lexists y, !okFieldVal objs y ** 4 + x |-> y))) A
  ) as [B [W P']].
  repeat (split; trivial).
  hypReplace Hh' (aconjGR Hh').
  apply sconjAssocL.
  apply sconjAssocL.
  apply sconjAssocL.
  apply (sconjImpl Hh'); clear Hh'; intros h Hh.
  apply sconjAssocR.
  apply sconjAssocR.
  apply Hh.
  hypReplace Hh (aconjEL Hh).
  apply aconjIntro; trivial.
  unfold objHp. 
  apply fasEqvSet with (x::set_remove x objs).
  eapply fasIntro.
  apply (sconjImplL Hh).
  clear h Hh; intros h Hh.
  unfold okObjAt; unfold okFieldAt; linLogIntroOld.
  existsl x0; linLogIntroOld.
  existsl x1; linLogIntroOld.
  apply Hh.
  intro X.
  apply (set_remove_elim2 _ X (refl_equal x)).
  split; intros; elimSetIn.
  clear Hst Hed Hroot Hobjs Hh'.
  destruct P' as [Hst [Hed [Hroot [He Hh]]]].
  aconjE1 Hh; clear Hh.
  assert (Hh := sconjAssocL (sconjAssocL (sconjComm 
    (sconjAssocR (sconjAssocR 
      (sconjAssocR (sconjAssocR Hh0))))))).
  assert (hget s' x = Some x0).
  unfold hget; eapply contGetEqL; apply Hh.
  assert (hget s' (4 + x) = Some x1).
  unfold hget; eapply contGetEqL; 
    apply (sconjAssocL (sconjComm Hh)).
  rewrite H; rewrite H1; rewrite H2; rewrite H3; 
    split; reflexivity.

  assert (Hobjs : set_eqv objs (rangeMSObjs ohpSt ohpEnd)).
  intuition.
  destruct (G' objs ohpSt ohpEnd freePtr stBot stBuff 
    (objHp objs objs) A H) as [B [W P']]; clear G' H.

  destruct (G'' B W ohpSt ohpEnd freePtr stBot
    stBuff A) as [free' [freePtr' [X'' P'']]]; 
  clear G''.
  destruct P' as [Hst [Hed [Hroot [He Hh]]]].
  repeat (split; trivial).
  apply sconjAssocL; apply sconjAssocL; 
    apply sconjAssocL; apply sconjAssocL.
  apply (sconjImpl (sconjAssocR
     (sconjAssocR (sconjAssocR (
    sconjAssocR Hh))))); clear Hh; intros h Hh.
  apply sconjAssocR.
  apply (sconjImplL (sconjAssocL Hh)).
  clear h Hh; intros h Hh.
  apply (sconjImplL Hh).
  clear h Hh; intros h Hh.
  apply (sconjComm Hh).
  assert (Hh1 := aconjER Hh); clear Hh.
  apply (sconjImplR (sconjComm Hh1)).
  intros.
  apply aconjIntro; assumption.
  exists B; exists free'; exists freePtr'.
  split; trivial.
  destruct P'' as [_ [_ [_ [In _]]]].
  intros.
  assert (In' : set_In x objs).
  rewrite Hobjs.
  rewrite <- In.
  elimSetIn.
  destruct (X'' x H).
  destruct (H0 x In').
  rewrite H1; rewrite H2; rewrite H3; rewrite H4.
  split; reflexivity.
  
  clear G' G''.
  destruct Groot'.
  split.
  rewrite Gcall''.
  rewrite <- H.
  reflexivity.
  generalize Gcst' H H0; clear Gcst' H H0.
  stSimpl; intros.
  destruct Groot''.
  split.
  rewrite <- Gcst' in H1.
  apply raAndOthers; trivial.
  eapply calleeSavedOkTrans'.
  2: apply Gra''.
  assert (Gcall : calleeSavedOk' s
    (rset (rset s rcst1 (0 + rget s ra)) ra MARKSWEEPSWEEP)).
  clear H.
  calleeSavedOkSolver'.
  eapply calleeSavedOkTrans'.
  apply Gcall.
  apply othersNotRa; trivial.
  rewrite <- H0; apply H2.
Qed.


Require Import VGC.
Require Import List.

(*
Module MarkSweepGCGuar <: BASIC_MGI_ARG.
  Definition gcGuar (s s':state) (rr:list rnum) := 
    (forall objs h rr',
      incl rr' rr ->
      s |||- (minObjHp (List.map (rget s) rr') objs && Heap.heap_eqv h) ** top ->
      s' |||- Heap.heap_eqv h ** top) /\
    forall r, In r rr -> rget s r = rget s' r.

End MarkSweepGCGuar.

Module MarkSweepGCMGI := BasicMGI (MarkSweepGCGuar).
*)

Module MarkSweepGC <: VIEW_GC with Module I := BasicGCMGI.

  Module I := BasicGCMGI.

  Notation rgcInfo := s7.

  Definition mview vs rf rr :=
    lexists ohpSt, lexists ohpEnd, lexists freePtr, lexists stBot, lexists stBuff,
    lexists objs, lexists free,
    !(~In rgcInfo rr /\ 
      ptrP ohpSt /\ ptrP ohpEnd /\ 
      set_eqv (set_union objs free) (rangeMSObjs ohpSt ohpEnd) /\
      (forall r, In r rr -> rget vs r = RFile.get rf r /\ okFieldVal objs (RFile.get rf r)) /\
      vs |||- objHp objs objs) **
    Heap.heap_eqv (heapOf vs) ** freeList freePtr free **
    objHdrs (rangeMSObjs ohpSt ohpEnd) UNMARK ** buffStackSet stBot stBot stBuff nil **
    RFile.get rf rgcInfo |->, (ohpSt::ohpEnd::freePtr::stBot::stBuff::nil).

  Lemma atomRegEq : forall h vs rf rr A r,
    h ||- mview vs rf rr ** A ->
    In r rr ->
    I.atomP (RFile.get rf r) \/ I.atomP (rget vs r) ->
    RFile.get rf r = rget vs r.
  Proof.
    intros h vs rf rr A r Hp I D.
    unfold mview in Hp; linLogElim Hp.
    destruct H3 as [_ [_ [_ [_ [Req Vhp]]]]].
    destruct (Req _ I) as [X _].
    auto.
  Qed.

  Lemma setNonRootOk : forall h vs rf rr r v,
    h ||- mview vs rf rr ->
    ~In r (rgcInfo::rr) ->
    h ||- mview vs (RFile.set rf r v) rr.
  Proof.
    unfold mview.
    intros h vs rf rr r v Hp I.
    linLogElim Hp.
    destruct H1 as [rrOk [stOk [edOk [objOk [Req Vhp]]]]].
    linLogIntro.
    split.
    assumption.
    split.
    apply stOk.
    split.
    apply edOk.
    split.
    apply objOk.
    split; auto.
    intros r0 H; destruct (Req _ H) as [X Y].
    destruct (regCompare r r0) as [R | R].
    elimtype False; subst r; intuition.
    replace (RFile.get (RFile.set rf r v) r0) with (RFile.get rf r0).
    auto.
    unfold RFile.set; unfold RFile.get.
    destruct (r0compare r0).
    subst r0; rewrite beq_rnum_refl; reflexivity.
    rewrite beq_rnum_neq.
    2:auto.
    rewrite (beq_rnum_neq R).
    reflexivity.
    replace (RFile.get (RFile.set rf r v) rgcInfo) with (RFile.get rf rgcInfo).
    apply Hp.
    unfold RFile.set; unfold RFile.get.
    assert (r <> rgcInfo).
    intro E; subst r; intuition.
    rewrite (beq_rnum_neq H); reflexivity.
  Qed.

  
  Lemma rootCopyOk : forall h vs rf rr rs rd,
    h ||- mview vs rf rr ->
    In rs rr -> 
    ~In rd (zero::rgcInfo::nil) ->
    h ||- mview (rset vs rd (rget vs rs)) (RFile.set rf rd (RFile.get rf rs)) (rd::rr).
  Proof.
    unfold mview.
    intros h vs rf rr rs rd Hp I0 Nz.
    assert (NE1 : rd <> zero).
    intro E; subst rd; intuition.
    assert (NE2 : rd <> rgcInfo).
    intro E; subst rd; intuition.
    linLogElim Hp.
    destruct H1 as [rrOk [stOk [edOk [objOk [Req Vhp]]]]].
    linLogIntro.

    constructor.
    simpl; intro X; destruct X.
    subst rd; intuition.
    auto.
    split.
    apply stOk.
    split.
    apply edOk.
    split.
    apply objOk.
    split; trivial.
    intros r I.
    simpl in I; destruct I as [R | R].
    subst r; stSimpl.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) rd) with
      (RFile.get rf rs).
    apply Req; apply I0; simpl; auto.
    unfold RFile.set; unfold RFile.get.
    rewrite beq_rnum_refl.
    rewrite (beq_rnum_neq (r:=zero) (r':=rd)).
    reflexivity.
    auto.

    destruct (regCompare rd r) as [R' | R'].
    
    subst r; stSimpl.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) rd) with
      (RFile.get rf rs).
    auto.
    unfold RFile.set; unfold RFile.get.
    rewrite (beq_rnum_neq (r:=zero) (r':=rd)).
    rewrite beq_rnum_refl.
    reflexivity.
    auto.

    stSimpl.
    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) r) with
      (RFile.get rf r).
    auto.
    unfold RFile.get; unfold RFile.set.
    rewrite (beq_rnum_neq R'); reflexivity.

    replace (RFile.get (RFile.set rf rd (RFile.get rf rs)) rgcInfo)
      with (RFile.get rf rgcInfo).
    apply Hp.
    unfold RFile.get; unfold RFile.set.
    rewrite (beq_rnum_neq NE2); reflexivity.

  Qed.


  Lemma rootWeaken : forall h vs rf rr rr',
    h ||- mview vs rf rr ->
    incl rr' rr ->
    h ||- mview vs rf rr'.
  Proof.
    unfold mview.
    intros h vs rf rr rr' Hp IR.
    linLogElim Hp.
    destruct H1 as [rrOk [stOk [edOk [objOk [Req Vhp]]]]].
    linLogIntro.
    constructor.
    auto.
    split.
    apply stOk.
    split.
    apply edOk.
    split.
    apply objOk.
    split; trivial.

    intros r I; apply Req.
    auto.
    apply Hp.
  Qed.

  Lemma addAtomRoot : forall h vs rf rr A r,
    h ||- mview vs rf rr ** A ->
    ~In r (zero::I.rgcInfo::rr) ->
    I.atomP (RFile.get rf r) ->
    h ||- mview (rset vs r (RFile.get rf r)) rf (r::rr) ** A.
  Proof.
    unfold mview.
    intros h vs rf rr A r Hp NI Atom.
    assert (NE1 : r <> zero).
    intro E; subst r; intuition.
    assert (NE2 : r <> rgcInfo).
    intro E; subst r; apply NI; right; left; reflexivity.
    linLogElim Hp.
    destruct H3 as [rrOk [stOk [edOk [objOk [Req Vhp]]]]].
    linLogIntro.

    constructor.
    simpl; intros [X | X]; auto.
    split.
    apply stOk.
    split.
    apply edOk.
    split.
    apply objOk.
    split; trivial.

    intros r' I.
    simpl in I; destruct I as [R | R].
    subst r'; stSimpl.
    constructor.
    reflexivity.
    left; assumption.

    assert (r' <> r).
    intro X; subst r'; simpl in NI; auto.
    stSimpl; apply Req; assumption.
    apply Hp.
  Qed.

  Notation ALLOCLOOP := 15.
  Notation ALLOCRETPTR := 16.
  
  Notation allocImpl :=
    (lw v0 8 rgcInfo; (*  fp  *)
      bne v0 zero ALLOCRETPTR;; (* if fp = 0 then goto L2 *)
      addiu rcst0 ra 0;
      jal MARKSWEEPGC ALLOCLOOP).
  
  Definition allocRetPtrPre s :=
    exists fpn, exists fpv, exists fp,
      s |||- fp |->, (fpv :: fpn :: nil) ** (8 + rget s rgcInfo) |-> fp ** top.
  
  Definition allocRetPtrGuar s s' :=
    forall fp fpn fpv,
      s |||- fp |->, (fpv :: fpn :: nil) ** (8 + rget s rgcInfo) |-> fp ** top ->
      s' = (uhset (uhset (rset (uhset (rset (rset s v0 fp) ast fpn) (8 + rget s rgcInfo) fpn) ast I.NULL)
	fp I.NULL) (4 + fp) I.NULL).
      
  Notation allocRetPtrImpl :=
    (lw v0 8 rgcInfo;
      lw ast 4 v0; (* ast = fp -> next *)
      sw ast 8 rgcInfo;
      addiu ast zero I.NULL;
      sw ast 0 v0; 
      sw ast 4 v0;
      jr ra).

  Definition allocLoopPre s := 
    exists fpn, exists fpv, exists fp,
      s |||- (if beq_nat fp 0 then unit else (fp |->, (fpv :: fpn :: nil))) ** (8 + rget s rgcInfo) |-> fp ** top.


  Definition allocLoopGuar s s' :=
    forall fpn fpv fp,
      s |||- (if beq_nat fp 0 then unit else (fp |->, (fpv :: fpn :: nil))) ** 
	(8 + rget s rgcInfo) |-> fp ** top ->
	  if beq_nat fp 0 then False else
	    s' = (uhset (uhset (rset (uhset (rset (rset (rset (rset s ra (rget s rcst0)) v0 fp) v0 fp) ast fpn) 
	      (8 + rget s rgcInfo) fpn) ast I.NULL) fp I.NULL) (4 + fp) I.NULL).

  Notation allocLoopImpl :=
    (addiu ra rcst0 0;
      lw v0 8 rgcInfo;
      bne v0 zero ALLOCRETPTR;;
      j ALLOCLOOP).

  Definition phi1 n :=
    (* return *)
    if beq_nat n RETURN then
      SomeT (bspeci returnPre returnGuar)
    else if beq_nat n RETURN1 then
      SomeT (bspeci return1Pre return1Guar)
    else if beq_nat n RETURN2 then
      SomeT (bspeci return2Pre return2Guar)
    else if beq_nat n RETURN3 then
      SomeT (bspeci return3Pre return3Guar)
    else 
      NoneT _.

  Definition phi2 n :=
    (* stack *)
    if beq_nat n STACKEMPTY then
      SomeT (bspeci isEmptyPre isEmptyGuar)
    else if beq_nat n STACKEMPTYT then
      SomeT (bspeci isEmptyTPre isEmptyTGuar)
    else if beq_nat n STACKPUSH then
      SomeT (bspeci pushPre pushGuar)
    else if beq_nat n STACKPOP then
      SomeT (bspeci popPre popGuar)
    else if beq_nat n STACKLOOP then
      SomeT (bspeci loopPre loopGuar)
    (* markField *)
    else if beq_nat n MARKFIELD then
     SomeT (bspeci markFieldPre markFieldGuar)
    else 
      phi1 n.

  Definition phi3 n :=
    (* mark *)
    if beq_nat n MARK then
      SomeT (bspeci markPre markGuar)
    else if beq_nat n MARKLOOP then
     SomeT (bspeci markLoopPre markLoopGuar)
    else if beq_nat n MARKLOOPBODY then
     SomeT (bspeci markLoopBodyPre markLoopBodyGuar)
    else if beq_nat n MARKFIRST then
     SomeT (bspeci markFirstPre markFirstGuar)
    else if beq_nat n MARKSECOND then
     SomeT (bspeci markSecondPre markSecondGuar)
    else 
      phi2 n.

  Definition phi4 n :=
    (* sweep *)
    if beq_nat n SWEEP then
      SomeT (bspeci sweepPre sweepGuar)
    else if beq_nat n SWEEPLOOP then
      SomeT (bspeci sweepLoopPre sweepLoopGuar)
    else if beq_nat n SWEEPADD then
      SomeT (bspeci sweepAddPre sweepAddGuar)
    else if beq_nat n SWEEPNEXT then
      SomeT (bspeci sweepNextPre sweepNextGuar)
    else if beq_nat n SWEEPFIN then
      SomeT (bspeci sweepFinPre sweepFinGuar)
    (* markSweepGC *)
    else if beq_nat n MARKSWEEPGC then
      SomeT (bspeci markSweepGCPre markSweepGCGuar)
    else if beq_nat n MARKSWEEPSWEEP then
      SomeT (bspeci markSweepSweepPre markSweepSweepGuar)
    else
      phi3 n.

  Definition phi n :=
    (* allocator & barrier *)
    if beq_nat n I.allocLbl then
      SomeT (allocSpec mview I.rroot I.gcGuar I.presReg I.NULL)
    else if beq_nat n I.read1Lbl then
      SomeT (read1Spec mview I.rroot I.gcGuar I.presReg)
    else if beq_nat n I.read2Lbl then
      SomeT (read2Spec mview I.rroot I.gcGuar I.presReg)
    else if beq_nat n I.write1Lbl then
      SomeT (write1Spec mview I.rroot I.gcGuar I.presReg)
    else if beq_nat n I.write2Lbl then
      SomeT (write2Spec mview I.rroot I.gcGuar I.presReg)
    else if beq_nat n ALLOCLOOP then
      SomeT (bspeci allocLoopPre allocLoopGuar)
    else if beq_nat n ALLOCRETPTR then
      SomeT (bspeci allocRetPtrPre allocRetPtrGuar)
    else phi4 n.

  Ltac PhiOkSolver :=
    unfold phi; rewrite beq_true;
    repeat (rewrite beq_false; [idtac | discriminate]);
    simpl; intuition.

  Lemma gcGuarRefl : forall vs rr,
    I.gcGuar vs vs rr.
  Proof.
    apply basicGcGuarRefl.
  Qed.

  Lemma gcGuarSetNonRoot : forall vs rr r v,
    ~In r rr ->
    I.gcGuar vs (rset vs r v) rr.
  Proof.
    intros vs rr r v NI.
    constructor.
    intros objs h rr' _ HpOk.
    destruct (aconjElim (sconjDistrib2 HpOk)) as [_ X].
    assumption.
    intros r' IR.
    assert (r <> r').
    intro E; subst r'; auto.
    stSimpl; reflexivity.
  Qed.

  Notation allocPre' := (allocPre mview rroot).
  Notation allocGuar' := (allocGuar mview rroot I.gcGuar I.presReg).

  Lemma natEqDec : forall m n : nat, m = n \/ m <> n.
  Proof.
    intros m n; compare m n; auto.
  Qed.
  
  Ltac gcPhiDomOkSolver :=
    match goal with
      | Ne : (if beq_nat ?f ?x then _ else _) <> NoneT bspec |- _ => 
	destruct (natEqDec f x) as [R | R];
	  [subst f; rewrite beq_true in Ne;
	    apply setRangeIntro; omega |
	      rewrite (beq_false R) in Ne; clear R]
    end.

  Lemma gcPhiDomOk : forall f,
    phi f <> NoneT _ -> set_In f I.gcPhiDom.
  Proof.
    unfold I.gcPhiDom.
    unfold phi.
    intros f Ne.
    unfold I.allocLbl in Ne.
    unfold I.read1Lbl in Ne.
    unfold I.read2Lbl in Ne.
    unfold I.write1Lbl in Ne.
    unfold I.write2Lbl in Ne.

    repeat gcPhiDomOkSolver.

    unfold phi4 in Ne.

    repeat gcPhiDomOkSolver.

    unfold phi3 in Ne.

    repeat gcPhiDomOkSolver.

    unfold phi2 in Ne.

    repeat gcPhiDomOkSolver.

    unfold phi1 in Ne.

    repeat gcPhiDomOkSolver.

    tauto.
  Qed.

  Lemma read1PhiOk :
    optPredT (phi I.read1Lbl)
      (specWeaker (read1Spec mview I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma read2PhiOk :
    optPredT (phi I.read2Lbl)
      (specWeaker (read2Spec mview I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write1PhiOk :
    optPredT (phi I.write1Lbl)
      (specWeaker (write1Spec mview I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma write2PhiOk :
    optPredT (phi I.write2Lbl)
      (specWeaker (write2Spec mview I.rroot I.gcGuar I.presReg)).
  Proof.
    PhiOkSolver.
  Qed.

  Lemma allocPhiOk :
    optPredT (phi I.allocLbl)
      (specWeaker (allocSpec mview I.rroot I.gcGuar I.presReg I.NULL )).
  Proof.
    PhiOkSolver.
  Qed.

  Notation readkImpl := (fun k => lw v0 k rroot; jr ra).
  Notation writekImpl := (fun k => sw a0 k rroot; jr ra).

  Notation readPre' := (readPre mview rroot).
  Notation readGuar' := (readGuar mview rroot I.gcGuar I.presReg).

  Lemma readkImplOk : 
    forall k, k = 0 \/ k = 4 ->
      iseqOkS phi (readPre' k) (readGuar' k) (readkImpl k).
  Proof.
    intros k D.
    iseqWeakSolver idtac.
    unfold readPre; unfold mview.
    intros s [vs [Hp Vhp']].
    linLogElim Hp.
    destruct H3 as [rrOk [stOk [edOk [objOk [Req0 Vhp]]]]].
    linLogElim Vhp'.
    assert (Hp2 := sconjInsert Hp Vhp'); linLogElim Hp2.
    destruct (Req0 rroot) as [Req rootOk]; clear Req0.
    simpl; auto.
    assert (RX : rgcInfo <> rroot).
    simpl in rrOk; auto.
    clear rrOk.

    assert (HH : s |||- (!even(rget s rroot) ** top) ** top).
    clear Hp; sconjMatch.
    linLogIntro; apply ptrPeven; assert (P : ptrP (k + rget s rroot)).
    rewrite Req in Hp; eapply contElim1; eauto.
    destruct (ptrP_elim P) as [y Y]; destruct D; subst k;
      [apply (ptrP_intro y); auto | apply (ptrP_intro (pred y)); omega].

    linLogElim HH; clear HH; rename H1 into Ev.
    destruct rootOk.

    elimtype False; eapply not_even_and_odd; eauto.

    hypReplace Vhp (objHpElim H Vhp).
    unfold okObjAt in Vhp; unfold okFieldAt in Vhp; linLogElim Vhp.
    assert (Hp3 := sconjInsert Hp Vhp); linLogElim Hp3; clear Vhp.
    
    assert (Dstep : exists v,
      cstep (lw v0 k rroot) s = Some (rset s v0 v) /\ 
      ((k = 0 /\ v = x7) \/ (k = 4 /\ v = x8))).
    destruct D; [exists x7 | exists x8]; subst k; constructor; auto.
    eapply lwStep. eauto.
    auto.
    auto.
    sconjLift2 Hp3; eapply lwStep; eauto.

    destruct Dstep as [v [Ceq Veq]].
    rewrite Ceq.
    constructor.
    trivial.

    clear Hp Vhp' Hp2 stOk edOk.
    assert (Hp0 : s |||- k + rget s rroot |-> v ** top).
    destruct Veq as [[X Y] | [X Y]].
    subst k v; simpl; sconjMatch.
    subst k v; simpl; sconjLift2 Hp3; sconjMatch.

    clear vs Req H H4 H7 Hp3 x x0 x1 x2 x3 Veq Ceq objOk.
    intros s' X; subst s'.
    unfold readGuar.

    constructor.

    intros vs A rr Hp Irr NIrr.
    linLogElim Hp.
    destruct H3 as [rrOk [stOk [edOk [objOk [Req0 Vhp]]]]].
    destruct (Req0 _ Irr) as [Req rootOk].
    destruct rootOk as [H | I].
    
    elimtype False; eapply not_even_and_odd; eauto.

    assert (NIv0 : ~In v0 rr).
    intro II; destruct (NIrr v0 II) as [I' | I']; clear II;
      [discriminate I' | calleeSavedSolver I'].

    exists (rset vs v0 v).
    assert (Vhp0 := Vhp).
    hypReplace Vhp (objHpElim I Vhp).
    unfold okObjAt in Vhp; unfold okFieldAt in Vhp; linLogElim Vhp.
    assert (Hp2 := sconjInsert Hp Vhp); linLogElim Hp2.
    constructor.
    apply gcGuarSetNonRoot; assumption.
    
    stSimpl.
    constructor.
    linLogIntro.
    constructor.
    simpl; intro DZ; destruct DZ.
    discriminate H.
    auto.
    split.
    apply stOk.
    split.
    apply edOk.
    split.
    apply objOk.
    split; trivial.
    simpl; intros r [E | II].
    subst r; stSimpl.
    unfold RFile.get; unfold RFile.set.
    rewrite beq_rnum_neq.
(*     rewrite beq_rnum_refl. *)
    constructor.
    reflexivity.
    destruct D; subst k.
    rewrite (contContEq Hp0 Hp2); assumption.
    sconjLift2 Vhp; sconjLift2 Hp2; 
      rewrite (contContEq Hp0 Hp2); assumption.
    discriminate.
    destruct (regCompare r v0) as [X | X].
    subst r; elimtype False; auto.
    stSimpl.
    assert (RFile.get (RFile.set (rfileOf s) v0 v) r  = RFile.get (rfileOf s) r).
    simpl.
    unfold RFile.get; unfold RFile.set.
    destruct (r0compare r).
    subst r; stSimpl; rewrite beq_rnum_refl; auto.
    rewrite beq_rnum_neq.
    rewrite beq_rnum_neq.
    reflexivity.
    auto.
    auto.
    rewrite H.
    apply Req0; trivial.
    apply Hp.
    
    rewrite Req.
    destruct D; subst k.
    rewrite (contContEq Hp0 Hp2).
    sconjMatch.
    sconjLift2 Hp2; rewrite (contContEq Hp0 Hp2).
    sconjLift2 Vhp; sconjMatch.

    intros r [A B]; stSimpl; reflexivity.

  Qed.


  Lemma read1ImplOk : iseqOkS phi (readPre' 0) (readGuar' 0) (readkImpl 0).
  Proof.
    apply (readkImplOk 0); auto.
  Qed.

  Lemma read2ImplOk : iseqOkS phi (readPre' 4) (readGuar' 4) (readkImpl 4).
  Proof.
    apply (readkImplOk 4); auto.
  Qed.


  Notation writePre' := (writePre mview rroot).
  Notation writeGuar' := (writeGuar mview rroot I.gcGuar I.presReg).


  Lemma writekImplOk : 
    forall k, k = 0 \/ k = 4 ->
      iseqOkS phi (writePre' k) (writeGuar' k) (writekImpl k).  
  Proof.
    intros k D.
    iseqWeakSolver idtac.

    intros s [vs [Hp Vhp]].
    unfold mview in Hp; linLogElim Hp.
    destruct H3 as [rrOk [stOk [edOk [objOk [Req0 Vhp']]]]].
    clear rrOk.

    linLogElim Vhp.
    assert (Hp2 := sconjInsert Hp Vhp); linLogElim Hp2.

    destruct (Req0 rroot) as [Req rootOk].
    simpl; auto.
    rewrite (swStep Hp2); auto.
    constructor.
    trivial.
    intros s' Seq; subst s'.
    rewrite Req in Hp2; rewrite Req.

    assert (HH : s |||- (!even(rget s rroot) ** top) ** top).
    clear Hp; sconjMatch.
    linLogIntro; apply ptrPeven; assert (P : ptrP (k + rget s rroot)).
    eapply contElim1; eauto.
    destruct (ptrP_elim P) as [y Y]; destruct D; subst k;
      [apply (ptrP_intro y); auto | apply (ptrP_intro (pred y)); omega].

    linLogElim HH; clear HH; rename H1 into Ev.
    clear Hp Vhp' Vhp vs Req0 Req x Hp2 rootOk stOk edOk x0 x1 objOk.
    constructor.
    intros vs A rr Hp Irr.
    assert (I1 : In rroot rr).
    apply Irr; simpl; auto.
    assert (I2 : In a0 rr).
    apply Irr; simpl; auto.
    clear Irr.
    unfold mview in Hp; linLogElim Hp.
    destruct H3 as [rrOk [stOk [edOk [objOk [Req0 Vhp]]]]].
    exists (uhset vs (k + rget vs rroot) (rget vs a0)).
    constructor.
    apply gcGuarRefl.

    destruct (Req0 rroot) as [Req rootOk].
    assumption.
    destruct rootOk.

    elimtype False; eapply not_even_and_odd; eauto.

    clear Ev.
    hypReplace Vhp (objHpElim H Vhp).
    unfold okObjAt in Vhp; unfold okFieldAt in Vhp; linLogElim Vhp.
    unfold mview.
    linLogIntro.
    constructor.
    auto.
    split.
    apply stOk.
    split.
    apply edOk.
    split.
    apply objOk.
    split; trivial.
    destruct D; subst k.

    hypReplace Vhp (contUpdateSt Vhp (rget vs a0)).
    replace (Heap.uset (heapOf vs) (RFile.get (rfileOf s) rroot) (rget vs a0)) with
      (heapOf (uhset vs (0 + rget vs rroot) (rget vs a0))) in Vhp.
    unfold hpredTrue; rewrite <- (addRemoveEqv H).
    apply objHpIntro.
    unfold okObjAt; unfold okFieldAt.
    destruct (Req0 a0) as [ReqA rootOkA].
    assumption.

    linLogIntro.
    instantiate (1 := rget vs a0).
    rewrite (addRemoveEqv H); rewrite ReqA; assumption.
    instantiate (1 := x12).
    sconjMatch.
    unfold hpredTrue; rewrite (addRemoveEqv H); assumption.
    rewrite (addRemoveEqv H); assumption.
    rewrite Req.
    reflexivity.

    sconjLift2 Vhp; hypReplace Vhp (contUpdateSt Vhp (rget vs a0)).
    replace (Heap.uset (heapOf vs) (4 + RFile.get (rfileOf s) rroot) (rget vs a0)) with
      (heapOf (uhset vs (4 + rget vs rroot) (rget vs a0))) in Vhp.
    sconjMatch.
    unfold hpredTrue; rewrite <- (addRemoveEqv H).
    apply objHpIntro.
    unfold okObjAt; unfold okFieldAt.
    destruct (Req0 a0) as [ReqA rootOkA].
    assumption.

    linLogIntro.
    instantiate (1 := x11).
    rewrite (addRemoveEqv H); assumption.
    instantiate (1 := rget vs a0).
    apply sconjGLift2'.
    sconjMatch.
    unfold hpredTrue; rewrite (addRemoveEqv H); assumption.
    rewrite (addRemoveEqv H).
    rewrite ReqA; assumption.
    rewrite Req; reflexivity.

    simpl; eapply sconjHpSet; eauto.
    destruct (Req0 a0).
    assumption.
    rewrite H0.
    rewrite Req.
    destruct D; subst k.
    eapply contSetEqL; eauto.
    sconjLift2 Vhp; eapply contSetEqL; eauto.

    intros r _; stSimpl; reflexivity.

  Qed.


  Lemma write1ImplOk : iseqOkS phi (writePre' 0) (writeGuar' 0) (writekImpl 0).
  Proof.
    apply (writekImplOk 0); auto.
  Qed.

  Lemma write2ImplOk : iseqOkS phi (writePre' 4) (writeGuar' 4) (writekImpl 4).
  Proof.
    apply (writekImplOk 4); auto.
  Qed.

  Lemma allocRetPtrImplOk : iseqOkS phi allocRetPtrPre allocRetPtrGuar allocRetPtrImpl.
  Proof.
    iseqWeakSolver fail.

    intros s [fpn [fpv [fp Hp]]].
    linLogElim Hp.
    assert (Hp0 := Hp).
    sconjLiftH 2 Hp.
    rewrite (lwStep Hp); stSimpl; trivial.
    sconjLiftH 3 Hp.
    rewrite (lwStep Hp); stSimpl; trivial.
    sconjLiftH 1 Hp.
    rewrite (swStep Hp); stSimpl; trivial.
    hypReplace Hp (contUpdateSt Hp fpn).
    rewrite (addiuStep); stSimpl.
    sconjLiftH 2 Hp.
    rewrite (swStep Hp); stSimpl; trivial.
    hypReplace Hp (contUpdateSt Hp I.NULL).
    sconjLiftH 1 Hp.
    rewrite (swStep Hp); stSimpl; trivial.
    split; trivial.
    unfold allocRetPtrGuar.
    intros; subst.
    linLogElim H0.
    sconjLiftH 2 Hp0.
    sconjLiftH 2 H0.
    assert (Eq := contContEq Hp0 H0); subst.
    sconjLiftH 3 H0.
    sconjLiftH 3 Hp0.
    assert (Eq := contContEq Hp0 H0); subst.
    replace (I.NULL + 0) with I.NULL;
      [reflexivity | auto with arith].
  Qed.


  Lemma allocLoopImplOk : iseqOkS phi allocLoopPre allocLoopGuar allocLoopImpl.
  Proof.
    unfold phi.
    iseqWeakSolver fail;
      repeat (rewrite beq_false; [idtac | discriminate]);
	try (rewrite beq_true; reflexivity).

    intros s [fpn [fpv [fp Hp]]].
    rewrite addiuStep.
    sconjLiftH 1 Hp.
    rewrite (lwStep Hp); stSimpl; trivial.
    unfold doBtest.
    generalize Hp.
    compare fp 0; intro R; subst;
      [rewrite beq_true | rewrite (beq_false R)];
      clear Hp; intro; simpl.
    unfold allocRetPtrPre; unfold allocLoopGuar.
    split.
    exists fpn; exists fpv; exists 0.
    stSimpl.
    sconjStep' 2.
    stSimpl.
    intros s' G fpn' fpv' fp' Hp'.
    sconjLiftH 1 Hp'.
    assert (Eq := contContEq Hp Hp'); subst fp'.
    sconjLiftH 2 Hp'.
    destruct (G fpn' fpv' 0 Hp'). 

    unfold allocRetPtrPre; unfold allocLoopGuar; unfold allocRetPtrGuar.
    split.
    exists fpn; exists fpv; exists fp.
    stSimpl.
    sconjStep' 2.
    stSimpl.
    intros s' G fpn' fpv' fp' Hp'.
    sconjLiftH 1 Hp'.
    assert (Eq := contContEq Hp Hp'); subst fp'.
    sconjLiftH 2 Hp'.
    rewrite (beq_false R) in Hp'.
    hypReplace G (G _ _ _ Hp').
    subst s'.
    rewrite (beq_false R).
    reflexivity.
  Qed.

  Lemma allocImplOk : iseqOkS phi allocPre' (allocGuar' I.NULL) allocImpl.
  Proof.
    unfold phi.
    iseqWeakSolver fail;
      repeat (rewrite beq_false; [idtac | discriminate]);
	try (rewrite beq_true; reflexivity).
    unfold phi4.
    repeat (rewrite beq_false; [idtac | discriminate]);
      try (rewrite beq_true; reflexivity).

    intros.
    unfold markSweepGCGuar in H0.
    destruct H0 as [_ [_ [H0 _ ]]].
    apply H0.
    unfold calleeSaved.
    simpl; tauto.

    intros s [vs Hp].
    unfold mview in Hp; linLogElim Hp.
    destruct H3 as [rrOk [stOk [edOk [objOk [Req Vhp]]]]].
    clear rrOk.
    
    replace (4 + (4 + RFile.get (rfileOf s) rgcInfo)) 
      with (8 + RFile.get (rfileOf s) rgcInfo) in Hp;
      [sconjLiftH 6 Hp | auto with arith].
    rewrite (lwStep Hp); stSimpl; trivial.
    rename x1 into fp; rename x4 into objs; rename x5 into free; 
      rename x into ohpSt; rename x0 into ohpEnd; rename x2 into stBot; rename x3 into stBuff.
    simpl.
    compare fp 0; intro R; subst;
      [rewrite beq_true | rewrite (beq_false R)]; simpl.

  (* call gc *)
    apply callChain.
    destruct free.
    stSimpl.

    unfold markSweepGCPre.
    assert (markPreP (rset (rset (rset s v0 0) rcst0 (rget s ra)) ra ALLOCLOOP)
      objs ohpSt ohpEnd 0 stBot stBuff top top).
    unfold set_union in objOk.
    rewrite <- (app_nil_end objs) in objOk.
    unfold markPreP; stSimpl.
    repeat (split; trivial).
    destruct (Req rroot); trivial.
    left; reflexivity.
    elimSetIn.
    unfold contL.
    replace (4 + (4 + rget s rgcInfo)) 
      with (8 + rget s rgcInfo);
      [sconjLiftH 4 Hp | auto with arith].
    hypReplace Hp (sconjInsert Hp Vhp).
    assert (HpOld := Hp).
    sconjStep' 5.
    simpl in HpOld; linLogElim HpOld.
    apply (sconjImpl HpOld); clear HpOld; intros h' Hp'.
    apply aconjIntro; [apply topIntro | idtac]; trivial.
    unfold objHdrs in * . 
    apply (fasEqvSet _ _ Hp' (set_eqvSymm objOk)).

    split.
    exists objs; exists ohpSt; exists ohpEnd; 
      exists 0; exists stBot; exists stBuff.
    exact H.
    
    unfold markSweepGCGuar; stSimpl.
    intros s' [G [Gra [Gcs Grt]]].
    assert (markPreP (rset (rset (rset s v0 0) rcst0 (rget s ra)) ra ALLOCLOOP)
      objs ohpSt ohpEnd 0 stBot stBuff (objHp objs objs) top).
    clear G Hp.
    unfold markPreP in * .
    intuition.
    sconjStep' 5.
    hypReplace H4 (aconjER H4).
    apply aconjIntro; exact H4.

    assert (G' := (G _ _ _ _ _ _ _ H0)); clear H H0.
    destruct G' as [objs' [free' [fp' [_ H]]]].
    destruct H as [_ [_ [rtOk' [objOk' Hp']]]].
    unfold allocLoopPre.
    compare fp' 0; intro R; subst.
    linLogElim Hp'.
    replace (4 + (4 + rget s' rgcInfo)) 
      with (8 + rget s' rgcInfo) in Hp'; [idtac | auto with arith].
    assert (heapOf s' ||- (if beq_nat 0 0 then unit else 0 |->, (0 :: 0 :: nil)) **
      8 + rget s' rgcInfo |-> 0 ** top).
    rewrite beq_true.
    sconjStep' 4.
    apply unitIR; apply topIntro.
    split.
    exists 0; exists 0; exists 0.
    exact H.

    intros.
    hypReplace H0 (H0 0 0 0 H).
    rewrite beq_true in H0; tauto.

    destruct free'.
    simpl in Hp'; linLogElim Hp'.
    tauto.

    unfold freeList in Hp'; fold freeList in Hp'; linLogElim Hp'.
    subst a.
    assert (heapOf s' ||- (if beq_nat fp' 0 then unit else fp' |->, (x0 :: x :: nil)) **
      8 + rget s' rgcInfo |-> fp' ** top).
    rewrite (beq_false R).
    unfold contL.
    sconjStep' 4.
    split.
    exists x; exists x0; exists fp'.
    exact H.
    intros.
    hypReplace H0 (H0 x x0 fp' H).
    rewrite (beq_false R) in H0.
    subst s''.
    
    split.
    sconjLiftH 4 Hp.
    hypReplace Hp (sconjInsert Hp Vhp).
    
    clear vs H Req Vhp rtOk' H20 objOk'.
    intros vs A HpN.
    unfold mview in HpN.
    linLogElim HpN.
    assert (x1 = ohpSt).
    sconjLiftH 4 Hp.
    sconjLiftH 4 HpN.
    apply (contContEq HpN Hp).
    assert (x2 = ohpEnd).
    sconjLiftH 5 Hp.
    sconjLiftH 5 HpN.
    apply (contContEq HpN Hp).
    assert (x3 = 0).
    sconjLiftH 6 Hp.
    sconjLiftH 6 HpN.
    apply (contContEq HpN Hp).
    assert (x4 = stBot).
    sconjLiftH 7 Hp.
    sconjLiftH 7 HpN.
    apply (contContEq HpN Hp).
    assert (x5 = stBuff).
    sconjLiftH 8 Hp.
    sconjLiftH 8 HpN.
    apply (contContEq HpN Hp).
    subst x1 x2 x3 x4 x5.
    clear Hp objs objOk.
    
    destruct x7.
    destruct H3 as [rrOk [_ [_ [objOk [Req Vhp]]]]].
    unfold set_union in objOk.
    rewrite <- (app_nil_end x6) in objOk.
    assert (Hp0 : s |||- Heap.heap_eqv (heapOf vs) ** top).
    sconjStep' 2.
    assert (markPreP (rset (rset (rset s v0 0) rcst0 (rget s ra)) ra ALLOCLOOP)
      x6 ohpSt ohpEnd 0 stBot stBuff (objHp x6 x6) A).
    clear G Hp'.
    unfold markPreP.
    destruct (Req rroot).
    left; reflexivity.
    stSimpl.
    intuition.
    unfold contL.
    hypReplace HpN (sconjInsert HpN Vhp).
    sconjStep' 5.
    simpl in HpN; linLogElim HpN.
    apply (sconjImpl HpN); clear HpN; intros h' Hp'.
    apply aconjIntro;  trivial.
    unfold objHdrs in * . 
    apply (fasEqvSet _ _ Hp' (set_eqvSymm objOk)).
    hypReplace G (G _ _ _ _ _ _ _ H); clear H.
    destruct G as [objs [free [fp [Eqv H]]]].
    destruct H as [_ [_ [rtOk' [objOk' HpN']]]].
    clear HpN.

    unfold contL in HpN'.
    assert (fp' = fp).
    sconjLiftH 4 Hp'.
    sconjLiftH 4 HpN'.
    apply (contContEq Hp' HpN').
    subst fp'.
    destruct free.
    simpl in HpN'; linLogElim HpN'; tauto.
    unfold freeList in HpN'; fold freeList in HpN'; linLogElim HpN'.
    subst a.
    assert (x0 = x2).
    sconjLiftH 7 Hp'.
    sconjLiftH 8 HpN'.
    apply (contContEq Hp' HpN').
    assert (x = x1).
    sconjLiftH 8 Hp'.
    sconjLiftH 9 HpN'.
    apply (contContEq Hp' HpN').
    subst x1 x2.
    clear Hp' objs' free' H22.
    
    repeat (hypReplace HpN' (sconjAssocR HpN')).
    hypReplace HpN' (sconjComm HpN').
    destruct (sconjElim HpN') as [vh [h [J [Vhp' Hp]]]].
    hypReplace Vhp' (aconjEL Vhp').

    assert (HpS' : s' |||- Heap.heap_eqv vh ** (((((((((objHdrs (rangeMSObjs ohpSt ohpEnd) 0 **
                    buffStackSet stBot stBot stBuff nil) **
                   rget s' rgcInfo |-> ohpSt) **
                  4 + rget s' rgcInfo |-> ohpEnd) **
                 4 + (4 + rget s' rgcInfo) |-> fp) **
                4 + (4 + (4 + rget s' rgcInfo)) |-> stBot) **
               4 + (4 + (4 + (4 + rget s' rgcInfo))) |-> stBuff) ** A) **
             fp |-> x0) ** 4 + fp |-> x) ** freeList x free).
    apply sconjIntro with vh h; auto.
    unfold hpredTrue.
    apply heap_canonEqv'.
    assert (Hp1 : s' |||- Heap.heap_eqv vh ** top).
    sconjMatch.
    clear HpN'.
    exists (vh, rfileOf s').
    (*exists objs;*) exists fp.
    repeat split.

    clear Hp J h Gra Gcs  rrOk  R.
    intros objs0 h rr' Incl Vhp''.
    unfold minObjHp in Vhp''; unfold okObjHp in Vhp''; unfold rootReachable in Vhp''.
    linLogElim Vhp''.
    assert (vs |||- (((fun h : heap =>
      forall x : nat,
        set_In x objs0 -> reachable h (rget vs rroot) x) &&
    objHp objs0 objs0) && Heap.heap_eqv h) ** top).
    sconjMatch.
    apply aconjIntro; [hypReplace Hp (aconjEL Hp) | apply (aconjER Hp)].
    apply aconjIntro; [hypReplace Hp (aconjEL Hp) | apply (aconjER Hp)].
    unfold hpredTrue in * .
    intros x1 II.
    destruct (Hp _ II) as [v [Hv Hrch]].
    replace v with (rget vs rroot) in Hrch.
    apply Hrch.
    assert (Rroot : forall r, In r rr' -> r = rroot).
    intros r I.
    hypReplace Incl (Incl _ I).
    destruct Incl; [auto | tauto].
    unfold set_In in Hv.
    destruct (inMapExists Hv) as [y [XX YY]].
    hypReplace YY (Rroot _ YY); subst y; auto.
    clear Vhp''.
    sconjMatch.
    assert (Hp0' : s |||- Heap.heap_eqv vh ** top).
    eapply existsEqvHp.
    apply Hp1.
    apply Vhp'.
    intros x1 I1.
    generalize (Eqv _ I1).
    stSimpl.
    trivial.
    clear Eqv.
    destruct (Req rroot).
    left; reflexivity.
    assert (rtOk : rget vs rroot = rget s rroot);
      [apply H0 | clear H0].
    rename objs into objs'.
    rename x6 into objs.

    destruct (sconjElim H) as [ha [hb [J [Hpa Hpb]]]].
    assert (Hpa1 := aconjER Hpa).
    assert (Hpa0 := aconjEL Hpa).
    assert (Hpa2 := aconjEL Hpa0).
    assert (Hpa3 := aconjER Hpa0); clear Hpa0 Hpa.
    unfold hpredTrue in Hpa2.

    assert (forall x, set_In x objs0 -> set_In x objs').
    intros.
    hypReplace Hpa2 (Hpa2 _ H0).
    eapply reachableInObjHp.
    apply (sconjInsert Hp0' Vhp').
    instantiate (1 := (rget s rroot)).
    rewrite Grt in rtOk'.
    destruct rtOk'; trivial.
    elimtype False.
    compare (rget s rroot) x1; intro R; subst.
    assert (ha ||- objHp objs0 objs0 ** top).
    hypReplace Hpa3 (unitIR Hpa3).
    sconjMatch.
    apply (ptrNotAtom (inObjHpIsPtr H3 H0) H2).
    rewrite rtOk in Hpa2.
    apply (ptrNotAtom (reachableNotAtom R Hpa2) H2).
    assert (ha ||- objHp objs0 objs0 ** top).
    hypReplace Hpa3 (unitIR Hpa3).
    sconjMatch.
    apply (inObjHpIsPtr H2 H0).
    hypReplace Hpa2 (reachableEqv Hpa2 (heap_canonEqv' ha)).
    destruct (sconjElim Hp0) as [h1 [h2 [J0 [XX1 XX2]]]].
    apply reachableJoinLeft with h1 h2; auto.
    unfold hpredTrue in XX1.
    eapply reachableEqv.
    instantiate (1 := (heapOf vs)).
    rewrite <- rtOk.
    apply reachableJoinLeft with ha hb; auto.
    apply Heap.HeapMap.ABJoin.map_eqvTrans with (Heap.heap_canon h1).
    apply (heap_canonEqv' h1).
    apply Heap.HeapMap.ABJoin.map_eqvSymm; trivial.
    clear Hpa2 Req.
    assert (Heap.subHeap ha vh).
    intros l v Hl.
    assert (Heap.subHeap ha (heapOf s)).
    apply Heap.subHeapTrans with (heapOf vs).
    apply (Heap.heap_joinSub J).
    destruct (sconjElim Hp0) as [h1 [h2 [J0 [XX1 XX2]]]].
    unfold hpredTrue in XX1.
    eapply subHeapEqv.
    2: apply Heap.HeapMap.ABJoin.map_eqvSymm; apply XX1.
    eapply subHeapEqv.
    2: apply (heap_canonEqv' h1).
    apply (Heap.heap_joinSub J0).
    assert (Heap.subHeap vh (heapOf s)).
    destruct (sconjElim Hp0') as [h1 [h2 [J0 [XX1 XX2]]]].
    unfold hpredTrue in XX1.
    eapply subHeapEqv.
    2: apply Heap.HeapMap.ABJoin.map_eqvSymm; apply XX1.
    eapply subHeapEqv.
    2: apply (heap_canonEqv' h1).
    apply (Heap.heap_joinSub J0).
  
    destruct (set_In_dec l objs0).
    hypReplace H0 (H0 _ s0).
    hypReplace H2 (H2 _ _ Hl).
    hypReplace Vhp' (fasElim H0 Vhp').
    unfold okObjAt in Vhp'; unfold okFieldAt in Vhp'; linLogElim Vhp'.
    assert (Heap.get vh l = Some x1).
    apply (contGetEqL Vhp').
    hypReplace H3 (H3 _ _ H4).
    rewrite H2 in H3; injection H3; intro; subst; auto.
    
    destruct (set_In_dec l (List.map (fun x=>4+x) objs0)).
    destruct (set_map_elim s0) as [l' [Eq Hl']].
    hypReplace H0 (H0 _ Hl').
    hypReplace H2 (H2 _ _ Hl).
    hypReplace Vhp' (fasElim H0 Vhp').
    unfold okObjAt in Vhp'; unfold okFieldAt in Vhp'; linLogElim Vhp'.
    assert (Heap.get vh l = Some x2).
    subst l; sconjLiftH 1 Vhp'.
    apply (contGetEqL Vhp').
    hypReplace H3 (H3 _ _ H4).
    rewrite H2 in H3; injection H3; intro; subst; auto.
    unfold objHp in Hpa3.
    assert (ha ||- fas (fun x=>x |->__) objs0).
    eapply fasImpl.
    apply Hpa3.
    intros x1 h0 H5.
    unfold okObjAt in H5; unfold okFieldAt in H5; linLogElim H5.
    linLogIntro.
    instantiate (1 := x3).
    instantiate (1 := x2).
    apply H5.
    rewrite (objHpNotInDom n n0 H4) in Hl.
    discriminate Hl.
    apply subHeapImpl with ha; auto.
    
    clear Hp Eqv.
    intros.
    destruct H; try tauto; subst r.
    destruct (Req rroot).
    left; reflexivity.
    rewrite H.
    unfold rget in * .
    simpl; rewrite <- Grt. 
    reflexivity.

    clear Eqv Hp.
    unfold mview.
    linLogIntro.
    split.
    intro F; repeat (destruct F as [E | F]; try discriminate; auto).
    split.
    apply stOk.
    split.
    apply edOk.
    instantiate (1 := (set_add fp objs)).
    instantiate (1 := free).
    split.
    rewrite <- objOk'.
    split; intros; elimSetIn.
    split.
    intros.
    destruct H as [E | H].
    subst r; stSimpl.
    simpl.
    unfold RFile.get; unfold RFile.set; simpl.
    split.
    reflexivity.
    right; elimSetIn.
    destruct H as [E | H]; [subst r | tauto].
    stSimpl.
    simpl.
    unfold rget; unfold RFile.get; unfold RFile.set; simpl.
    split; trivial.
    destruct rtOk'.
    left; trivial.
    right.
    elimSetIn.
    simpl.
    assert (~ set_In fp objs).
    intro F.
    assert (s' |||- objHp objs objs ** fp |-> x0 ** top).
    hypReplace HpS' (sconjInsert HpS' Vhp').
    sconjStep' 9.
    hypReplace H (fasElimL F H).
    unfold okObjAt in H; unfold okFieldAt in H; linLogElim H.
    apply contContFalse with (h := heapOf s') (a := fp) (x := x1) (y :=x0) (A := top).
    sconjStep' 5.
    unfold objHp in * .
    apply fasIntro; auto.
    unfold okObjAt; unfold okFieldAt; linLogIntro.
    instantiate (1 := I.NULL).
    left; compute; repeat constructor.
    instantiate (1 := I.NULL).
    assert (ptrP fp).
    assert (s' |||- fp |-> x0 ** top).
    sconjLiftH 9 HpS'; sconjMatch; clear HpS'.
    destruct (sconjElim H0) as [hh [hhh [_ [Hhh _ ]]]].
    unfold cont in Hhh; unfold hpredTrue in Hhh.
    tauto.
    sconjLift 1.
    apply heapAddOk.
    sconjLift 1.
    apply heapAddOk.
    eapply fasImpl.
    apply Vhp'.
    unfold okObjAt; unfold okFieldAt; unfold okFieldVal.
    intros.
    linLogElim H1. 
    linLogIntro.
    instantiate (1 := x2).
    intuition.
    instantiate (1 := x3); trivial.
    intuition.
    trivial.
    intro F.
    linLogElim F.
    hypReplace HpS' (sconjInsert HpS' F).
    apply contContFalse with (h := heapOf s') (a := fp) (x := x1) (y :=x0) (A := top).
    smartAssocL HpS'.
    sconjStep' 1.
    sconjLiftH 9 HpS' .
    sconjStep' 2.
    constructor; assumption.
    intro F.
    linLogElim F.
    assert (s' |||- fp |-> x0 ** Heap.heap_eqv vh ** 4 + fp |-> x ** top).
    clear Hp1.
    sconjStep' 9.
    clear HpS'.
    hypReplace H1 (contUpdateSt H1 I.NULL).
    hypReplace H1 (heapEqvAdd H1).
    hypReplace H1 (sconjInsert H1 F).    
    apply contContFalse with (h := Heap.uset (heapOf s') fp I.NULL) (a := (4 + fp)) (x := x1) (y :=x) (A := top).
    sconjStep' 1.
    sconjLiftH 1 H1.
    sconjStep' 2.
    left; compute; repeat constructor.
    instantiate (1 := stBuff). 
    instantiate (1:=stBot). 
    instantiate (1:=x).
    sconjLiftH 5 HpS'.
    replace (4 + (4 + rget s' rgcInfo)) with (8 + rget s' rgcInfo) in HpS'; 
      [hypReplace HpS' (contUpdateSt HpS' x) | auto with arith].
    sconjLiftH 4 HpS'.
    hypReplace HpS' (contUpdateSt HpS' I.NULL).
    sconjLiftH 1 HpS'.
    hypReplace HpS' (contUpdateSt HpS' I.NULL).
    sconjStep' 4.
    simpl.
    apply heapEqvAdd.
    sconjLift 1.
    apply heapEqvAdd.
    sconjStep' 3.

    (* new: showing that the object is fresh *)
    stSimpl.
    intros B Vhp''; simpl.
    replace (S (S (S (S fp)))) with (4 + fp); [idtac | reflexivity].
    linLogElim Hp.

    assert (X : h ||- (!(ptrP fp) ** top) ** top).
    hypReplace Hp (sconjAssocR Hp); sconjLift8 Hp.
    sconjMatch.
    linLogIntro.
    eapply contElim1; eauto.

    linLogElim X; clear X; rename H1 into fpP.
    apply sconjComm; linLogIntro; apply sconjComm; linLogIntro.
    linLogElim HpS'.
    apply heapAddOk.
    apply sconjComm; apply heapAddOk.
    assumption.
    assumption.
    clear Vhp'; intro Vhp'; linLogElim Vhp'.
    hypReplace Hp (sconjInsert HpS' Vhp'); linLogElim Hp.
    assert (HX : s' |||- fp |-> x1 ** top ** top ** top ** top ** top ** top
      ** top ** top ** fp |-> x0 ** top).
    sconjMatch.
    linLogElim HX.
    destruct (ptrP_elim fpP) as [k E]; apply (ptrP_intro (S k)).
    omega.
    clear Vhp'; intro Vhp'; linLogElim Vhp'.
    assert (HpX : s' |||- Heap.heap_eqv vh ** top ** top ** top ** top ** top ** top
      ** top ** top ** fp |-> x0 ** 4 + fp |-> x ** top).
    sconjMatch.
    clear HpS'; linLogElim HpX.
    sconjLift2 HpX; hypReplace HpX (contUpdateSt HpX 1).
    hypReplace HpX (heapEqvAdd HpX).
    hypReplace HpX (sconjInsert HpX Vhp'); linLogElim HpX.

    elimtype False.
    assert (s |||- freeList 0 (a :: x7) ** top).
    sconjStep' 2.
    simpl in H; linLogElim H.
    tauto.

    unfold regPres; unfold I.presReg.
    clear Hp' Hp G H.
    intros r Hx; generalize (Gcs _ Hx); 
      destruct r; simpl; auto;
        destruct Hx as [H | [H | [H | [H | [
      H | [H | [H | [H | [H | H]]]]]]]]];
      try (discriminate H || 
        elimtype False; apply H). 

    elimtype False.
    unfold freeList in Hp; fold freeList in Hp; linLogElim Hp.
    tauto.

    (* no GC Call *)

    destruct free.
    
    elimtype False.
    unfold freeList in Hp; fold freeList in Hp; linLogElim Hp.
    tauto.

    unfold freeList in Hp; fold freeList in Hp; linLogElim Hp.
    subst a.
    assert (rset s v0 fp
      |||- fp |->, (x0 :: x :: nil) **
	8 + rget (rset s v0 fp) rgcInfo |-> fp ** top).
    linLogIntro; sconjStep' 5.
    split.
    exists x; exists x0; exists fp; assumption.
    intros s' G.

    split.
    clear H.
    sconjLiftH 3 Hp.
    hypReplace Hp (sconjInsert Hp Vhp).
    clear vs Req Vhp.
    intros vs A HpN.
    unfold mview in HpN.
    linLogElim HpN.
    assert (x1 = ohpSt).
    sconjLiftH 6 Hp.
    sconjLiftH 4 HpN.
    apply (contContEq HpN Hp).
    assert (x2 = ohpEnd).
    sconjLiftH 7 Hp.
    sconjLiftH 5 HpN.
    apply (contContEq HpN Hp).
    assert (x3 = fp).
    sconjLiftH 9 Hp.
    sconjLiftH 6 HpN.
    apply (contContEq HpN Hp).
    assert (x4 = stBot).
    sconjLiftH 10 Hp.
    sconjLiftH 7 HpN.
    apply (contContEq HpN Hp).
    assert (x5 = stBuff).
    sconjLiftH 11 Hp.
    sconjLiftH 8 HpN.
    apply (contContEq HpN Hp).
    subst x1 x2 x3 x4 x5.
    clear objOk.
    
    destruct x7.
    elimtype False.
    unfold freeList in HpN; fold freeList in HpN; linLogElim HpN.
    tauto.

    destruct H3 as [rrOk [_ [_ [objOk [Req Vhp]]]]].
    unfold freeList in HpN; fold freeList in HpN; linLogElim HpN.
    subst a.
    assert (x2 = x0).
    sconjLiftH 1 Hp.
    sconjLiftH 1 HpN.
    apply (contContEq HpN Hp).
    assert (x1 = x).
    sconjLiftH 2 Hp.
    sconjLiftH 2 HpN.
    apply (contContEq HpN Hp).
    subst x x0.
    clear Hp.
    assert (rset s v0 fp
      |||- fp |->, (x2 :: x1 :: nil) **
	8 + rget (rset s v0 fp) rgcInfo |-> fp ** top).
    stSimpl.
    replace (4 + (4 + RFile.get (rfileOf s) rgcInfo)) with (8 + rget s rgcInfo) in HpN; 
      [linLogIntro;  sconjStep' 7 | auto with arith].
    hypReplace G (G _ _ _ H).
    subst s'; clear H.
    clear free objs.
    rename x6 into objs; rename x7 into free.
    exists vs; exists fp.

    split.
    apply gcGuarRefl.
    split.

    unfold mview.
    linLogIntro.
    split.
    intro F; repeat (destruct F as [E | F]; try discriminate; auto).
    split.
    apply stOk.
    split.
    apply edOk.
    instantiate (1 := (set_add fp objs)).
    instantiate (1 := free).
    split.
    rewrite <- objOk.
    split; intros; elimSetIn.
    split.
    intros.
    destruct H as [E | H].
    subst r; stSimpl.
    simpl.
    unfold RFile.get; unfold RFile.set; simpl.
    split.
    reflexivity.
    right; elimSetIn.
    destruct H as [E | H]; [subst r | tauto].
    stSimpl.
    simpl.
    unfold rget; unfold RFile.get; unfold RFile.set; simpl.
    destruct (Req rroot).
    left; reflexivity.
    split; trivial.
    destruct H0; [left | right]; elimSetIn.
    simpl.
    assert (~ set_In fp objs).
    intro F.
    assert (s |||- objHp objs objs ** fp |-> x2 ** top).
    hypReplace HpN (sconjInsert HpN Vhp).
    sconjStep' 9.
    hypReplace H (fasElimL F H).
    unfold okObjAt in H; unfold okFieldAt in H; linLogElim H.
    apply contContFalse with (h := heapOf s) (a := fp) (x := x) (y :=x2) (A := top).
    sconjStep' 5.
    unfold objHp in * .
    apply fasIntro; auto.
    unfold okObjAt; unfold okFieldAt; linLogIntro.
    instantiate (1 := I.NULL).
    left; compute; repeat constructor.
    instantiate (1 := I.NULL).
    assert (ptrP fp).
    assert (s |||- fp |-> x2 ** top).
    sconjLiftH 1 HpN; sconjMatch; clear HpN.
    destruct (sconjElim H0) as [hh [hhh [_ [Hhh _ ]]]].
    unfold cont in Hhh; unfold hpredTrue in Hhh.
    tauto.
    sconjLift 1.
    apply heapAddOk.
    sconjLift 1.
    apply heapAddOk.
    eapply fasImpl.
    apply Vhp.
    unfold okObjAt; unfold okFieldAt; unfold okFieldVal.
    intros.
    linLogElim H1. 
    linLogIntro.
    instantiate (1 := x0).
    intuition.
    instantiate (1 := x3); trivial.
    intuition.
    trivial.
    intro F.
    linLogElim F.
    hypReplace HpN (sconjInsert HpN F).
    apply contContFalse with (h := heapOf s) (a := fp) (x := x) (y :=x2) (A := top).
    smartAssocL HpN.
    sconjStep' 1.
    sconjLiftH 1 HpN .
    sconjStep' 2.
    constructor; assumption.
    intro F.
    linLogElim F.
    assert (s |||- fp |-> x2 ** Heap.heap_eqv (heapOf vs) ** 4 + fp |-> x1 ** top).
    sconjStep' 9.
    clear HpN.
    hypReplace H1 (contUpdateSt H1 I.NULL).
    hypReplace H1 (heapEqvAdd H1).
    hypReplace H1 (sconjInsert H1 F).    
    apply contContFalse with (h := Heap.uset (heapOf s) fp I.NULL) (a := (4 + fp)) (x := x) (y :=x1) (A := top).
    clear F.
    sconjStep' 1.
    sconjLiftH 1 H1.
    sconjStep' 2.
    left; compute; repeat constructor.
    instantiate (1 := stBuff). 
    instantiate (1:=stBot). 
    instantiate (1:=x1).
    sconjLiftH 8 HpN.
    replace (4 + (4 + rget s rgcInfo)) with (8 + rget s rgcInfo) in HpN; 
      [hypReplace HpN (contUpdateSt HpN x1) | auto with arith].
    sconjLiftH 5 HpN.
    hypReplace HpN (contUpdateSt HpN I.NULL).
    sconjLiftH 1 HpN.
    hypReplace HpN (contUpdateSt HpN I.NULL).
    sconjStep' 4.
    simpl.
    apply heapEqvAdd.
    sconjLift 1.
    apply heapEqvAdd.
    sconjStep' 3.

    (* show the new object is fresh *)
    stSimpl.
    intros B Vhp''; simpl.
    replace (S (S (S (S fp)))) with (4 + fp); [idtac | reflexivity].

    assert (X : s |||- (!(ptrP fp) ** top) ** top).
    sconjLift2 HpN; sconjMatch; linLogIntro; eapply contElim1; eauto.

    linLogElim X; clear X; rename H1 into fpP.
    apply sconjComm; linLogIntro; apply sconjComm; linLogIntro.
    apply heapAddOk.
    apply sconjComm; apply heapAddOk.
    assumption.
    assumption.
    intro Vhp'; linLogElim Vhp'.
    hypReplace HpN (sconjInsert HpN Vhp'); linLogElim HpN.
    destruct (ptrP_elim fpP) as [k E]; apply (ptrP_intro (S k)).
    omega.
    intro Vhp'; linLogElim Vhp'.
    sconjLift2 HpN; hypReplace HpN (contUpdateSt HpN 1).
    hypReplace HpN (heapEqvAdd HpN).
    hypReplace HpN (sconjInsert HpN Vhp'); linLogElim HpN.

    unfold regPres; unfold I.presReg.
    hypReplace G (G _ _ _ H).
    subst s'.
    clear Hp H.
    calleeSavedOkSolver.
  Qed.

  Definition GCC1 f : option iseq :=
    (* return *)
    if beq_nat f RETURN then
      Some returnImpl
    else if beq_nat f RETURN1 then
      Some return1Impl
    else if beq_nat f RETURN2 then
      Some return2Impl
    else if beq_nat f RETURN3 then
      Some return3Impl
    else
      None.

  Definition GCC2 f :=
    (* stack *)
    if beq_nat f STACKEMPTY then
      Some isEmptyImpl
    else if beq_nat f STACKEMPTYT then
      Some isEmptyTImpl
    else if beq_nat f STACKPUSH then
      Some pushImpl
    else if beq_nat f STACKPOP then
      Some popImpl
    else if beq_nat f STACKLOOP then
      Some loopImpl
    (* markField *)
    else if beq_nat f MARKFIELD then
      Some markFieldImpl
    else 
      GCC1 f.

  Definition GCC3 f :=
    (* mark *)
    if beq_nat f MARK then
      Some markImpl
    else if beq_nat f MARKLOOP then
      Some markLoopImpl
    else if beq_nat f MARKLOOPBODY then
      Some markLoopBodyImpl
    else if beq_nat f MARKFIRST then
      Some markFirstImpl
    else if beq_nat f MARKSECOND then
      Some markSecondImpl
    else
      GCC2 f.
  
  Definition GCC4 f :=
    (* sweep *)
    if beq_nat f SWEEP then
      Some sweepImpl
    else if beq_nat f SWEEPLOOP then
      Some sweepLoopImpl
    else if beq_nat f SWEEPADD then
      Some sweepAddImpl
    else if beq_nat f SWEEPNEXT then
      Some sweepNextImpl
    else if beq_nat f SWEEPFIN then
      Some sweepFinImpl
    (* markSweepGC *)
    else if beq_nat f MARKSWEEPGC then
      Some markSweepGCImpl
    else if beq_nat f MARKSWEEPSWEEP then
      Some markSweepSweepImpl
    else
      GCC3 f.

  Definition GCC f :=
    if beq_nat f I.allocLbl then
      Some allocImpl
    else if beq_nat f I.read1Lbl then
      Some (readkImpl 0)
    else if beq_nat f I.read2Lbl then
      Some (readkImpl 4)
    else if beq_nat f I.write1Lbl then
      Some (writekImpl 0)
    else if beq_nat f I.write2Lbl then
      Some (writekImpl 4)
    else if beq_nat f ALLOCLOOP then
      Some allocLoopImpl
    else if beq_nat f ALLOCRETPTR then
      Some allocRetPtrImpl
    else
      GCC4 f.

  Ltac GCCokSolver :=
    match goal with 
      | Peq : (if beq_nat ?x ?y then _ else _ ) = SomeT (bspeci ?p ?g)  |- _ =>
	compare x y; intro R;
	  [subst x; rewrite beq_true in Peq; rewrite beq_true;
	    injection Peq; clear Peq; intros; subst p g |
	      rewrite (beq_false R); rewrite (beq_false R) in Peq; clear R]
    end.

  Notation chspec := (chspec bspec).

  Definition chspecSubset (P P' : chspec) :=
    forall f x, P f = SomeT x -> P' f = SomeT x.

  Lemma iseqOkSSubP : forall P P' p g i,
    chspecSubset P P' ->
    iseqOkS P p g i ->
    iseqOkS P' p g i.
  Proof.
    intros P P' p g i Hsub Hi.
    induction Hi;
      econstructor; auto; try apply (Hsub H); auto.
    apply IHHi.
    apply H.
  Qed.

  Lemma returnPSub :
    chspecSubset returnP phi.
  Proof.
    intros f x Hx.
    unfold returnP in Hx.
    discriminate.
  Qed.
 
  Ltac phiSubSolver :=
    match goal with
      | Hx : (if beq_nat ?x ?y then _ else _) = SomeT ?s |- 
	(if beq_nat ?y ?z then _ else _) = SomeT ?s =>	
	compare x y; intro R;
	  [subst y; rewrite beq_true in Hx;
	    repeat (rewrite beq_false; [idtac | discriminate]); phiSubSolver |
	      rewrite (beq_false R) in Hx; clear R; phiSubSolver]
      | Hx : (if beq_nat ?x ?y then _ else _) = SomeT ?s |-
	(if beq_nat ?y ?x then _ else _) = SomeT ?s =>
	compare x y; intro R;
	  [subst y; rewrite beq_true in Hx;
	    rewrite beq_true; trivial |
	      rewrite (beq_false R) in Hx; clear R; phiSubSolver]
      | Hx : SomeT _ = SomeT _ |- _ =>
	repeat (rewrite beq_false; [idtac | discriminate]); trivial; try phiSubSolver
      | |- (phi4 _) = _ => unfold phi4; phiSubSolver
      | |- (phi3 _) = _ => unfold phi3; phiSubSolver
      | |- (phi2 _) = _ => unfold phi2; phiSubSolver
      | |- (phi1 _) = _ => unfold phi1; phiSubSolver
      | |- (phi _) = _ => unfold phi; phiSubSolver
      | _ => idtac
    end.

  Ltac phiSubSolver' :=
    match goal with 
      | |- chspecSubset ?p1 ?p2 =>
	intros f x Hx; unfold p1 in Hx; phiSubSolver; discriminate
    end.

  Lemma stackPSub :
    chspecSubset stackP phi.
  Proof.
    phiSubSolver'.
  Qed.
 
  Lemma markFieldPSub :
    chspecSubset markFieldP phi.
  Proof.
    phiSubSolver'.
  Qed.
 
  Lemma markPSub :
    chspecSubset markP phi.
  Proof.
    phiSubSolver'.
  Qed.
 
  Lemma sweepPSub :
    chspecSubset sweepP phi.
  Proof.
    phiSubSolver'.
  Qed.

  Lemma markSweepPSub :
    chspecSubset markSweepP phi.
  Proof.
    phiSubSolver'.
  Qed.

  Ltac subPSimpl Hsub P:=
    apply iseqOkSSubP with P;
      [apply Hsub | idtac].

  Lemma GCCok : cdHeapOkSC phi GCC phi.
  Proof.
    intros x [p g] Peq.
    unfold phi in Peq; unfold GCC.

    GCCokSolver.
    exact allocImplOk.

    GCCokSolver.
    exact read1ImplOk.

    GCCokSolver.
    exact read2ImplOk.

    GCCokSolver.
    exact write1ImplOk.

    GCCokSolver.
    exact write2ImplOk.

    GCCokSolver.
    exact allocLoopImplOk.

    GCCokSolver.
    exact allocRetPtrImplOk.

    unfold phi4 in Peq; unfold GCC4.

    GCCokSolver.
    subPSimpl sweepPSub sweepP.
    exact sweepImplOk.

    GCCokSolver.
    subPSimpl sweepPSub sweepP.
    apply sweepLoopImplOk.

    GCCokSolver.
    subPSimpl sweepPSub sweepP.
    apply sweepAddImplOk.

    GCCokSolver.
    subPSimpl sweepPSub sweepP.
    apply sweepNextImplOk.

    GCCokSolver.
    subPSimpl sweepPSub sweepP.
    apply sweepFinImplOk.

    GCCokSolver.
    subPSimpl markSweepPSub markSweepP.
    apply markSweepGCImplOk.

    GCCokSolver.
    subPSimpl markSweepPSub markSweepP.
    apply markSweepSweepImplOk.

    unfold phi3 in Peq; unfold GCC3.

    GCCokSolver.
    subPSimpl markPSub markP.
    exact markImplOk.

    GCCokSolver.
    subPSimpl markPSub markP.
    exact markLoopImplOk.

    GCCokSolver.
    subPSimpl markPSub markP.
    exact markLoopBodyImplOk.

    GCCokSolver.
    subPSimpl markPSub markP.
    exact markFirstImplOk.

    GCCokSolver.
    subPSimpl markPSub markP.
    exact markSecondImplOk.

    unfold phi2 in Peq; unfold GCC2.

    GCCokSolver.
    subPSimpl stackPSub stackP.
    exact isEmptyImplOk.

    GCCokSolver.
    subPSimpl stackPSub stackP.
    exact isEmptyTImplOk.

    GCCokSolver.
    subPSimpl stackPSub stackP.
    exact pushImplOk.

    GCCokSolver.
    subPSimpl stackPSub stackP.
    exact popImplOk.

    GCCokSolver.
    subPSimpl stackPSub stackP.
    exact loopImplOk.

    GCCokSolver.
    subPSimpl markFieldPSub markFieldP.
    exact markFieldImplOk.

    unfold phi1 in Peq; unfold GCC1.

    GCCokSolver.
    subPSimpl returnPSub returnP.
    exact returnImplOk.

    GCCokSolver.
    subPSimpl returnPSub returnP.
    exact return1ImplOk.

    GCCokSolver.
    subPSimpl returnPSub returnP.
    exact return2ImplOk.

    GCCokSolver.
    subPSimpl returnPSub returnP.
    exact return3ImplOk.

    discriminate.
  Qed.

End MarkSweepGC.
