Require Import CheneyLemma.


Set Implicit Arguments.


(* code address constants *)
Definition SCANFIELD  := 30.
Definition FWDOBJ := 31.
Definition NOTTOSPACEPTR := 32.
Definition SCANPOINTER := 34.
Definition SCANNOCOPY := 35.
Definition SCANCOPIED := 36.
Definition RETURN := 37.


(* local register conventions *)
Definition rfree := a3.
Definition rtoStart := t5.
Definition rtoEnd := t6.

Notation rtemp := a2.
Notation temp0 := t8.
Notation temp1 := t1.


(* code block definitions *)

Definition fwdObj :=
  (* copy fields from src to free *)
  lw rtemp 0 a0;
  sw rtemp 0 rfree;
  lw rtemp WORDSIZE a0;
  sw rtemp WORDSIZE rfree;
  (* store forwarding pointer in src *)
  sw rfree 0 a0;
  (* return value of free *)
  addiu v0 rfree 0;
  (* increment free *)
  addiu rfree rfree OBJSIZE;
  jr ra.


Notation isNotSpacePtr :=
  (fun rstart rend =>
    andi1 v0 a0;
    bne v0 zero RETURN;;
    sltu v0 a0 rstart;
    bne v0 zero RETURN;;
    addiu v0 a0 1;
    sltu v0 rend v0;
    bne v0 zero RETURN;;
    addiu v0 zero 0;
    jr ra).

Definition isNotToSpacePtr := isNotSpacePtr rtoStart rtoEnd.


Definition scanPointer :=
  addiu rfree t4 0;
  beq v0 zero SCANNOCOPY;;
  addiu a0 temp1 0;
  jal FWDOBJ SCANCOPIED.

Definition scanNoCopy :=
  sw t2 0 temp0;
  addiu ra t7 0;
  jr ra.

Definition scanCopied :=
  sw v0 0 temp0;
  addiu ra t7 0;
  jr ra.

Definition scanField :=
  lw temp1 0 a0;           (* temp1 := [a0] *)
  andi1 rtemp temp1;       (* t := temp1 && 1 *)
  bne rtemp zero RETURN;;  (* if t <> 0 return *)
  addiu temp0 a0 0;        (* temp0 := a0 *)
  lw t2 0 temp1;           (* t2 := [temp1] *)
  addiu t7 ra 0;           (* t7 := ra *)
  addiu t4 rfree 0;        (* t4 := rfree *)
  addiu a0 t2 0;           (* a0 := t2 *)
  jal NOTTOSPACEPTR SCANPOINTER.


(* specs *)

Definition fwdObjPre s :=
  s |||- rget s a0 |->__ ** rget s rfree |->__ ** top.

Definition fwdObjGuar :=
  fun s s' => (forall A x y x' y',
    s |||- rget s a0 |->, (x::y::nil) ** 
      rget s rfree |->, (x'::y'::nil) ** A ->
	heapOf s' ||- rget s a0 |->, (rget s rfree::y::nil) ** 
	  rget s rfree |->, (x::y::nil) ** A) /\
	    rget s' rfree = OBJSIZE + rget s rfree /\
	    rget s' v0 = rget s rfree /\
	    regPres (regNotInList (rfree::rtemp::v0::nil)) s s'.


Definition isNotSpacePtrPre : spred := fun s => True.

Notation isNotSpacePtrGuar := (fun rstart rend s s' =>
  sbool (rget s' v0) (notSpacePtrP (rget s a0) (rget s rstart) (rget s rend)) /\
  heapOf s = heapOf s' /\ 
  regPres (fun r => r <> v0) s s').

Definition isNotToSpacePtrGuar := isNotSpacePtrGuar rtoStart rtoEnd.

Definition scanNoCopyPre s :=
  exists v, heapOf s ||- rget s temp0 |-> v ** top.

Definition scanNoCopyGuar s s' :=
  (forall v A, heapOf s ||- rget s temp0 |-> v ** A ->
    heapOf s' ||- rget s temp0 |-> rget s t2 ** A) /\
  rfileEqv s' (rset s ra (rget s t7)).


Definition scanCopiedPre s :=
  exists v, heapOf s ||- rget s temp0 |-> v ** top.

Definition scanCopiedGuar s s' :=
  (forall v A, heapOf s ||- rget s temp0 |-> v ** A ->
    heapOf s' ||- rget s temp0 |-> rget s v0 ** A) /\
  rfileEqv s' (rset s ra (rget s t7)).


Definition scanPointerPre s :=
  if beq_nat (rget s v0) 0 then
    scanNoCopyPre s
  else
    exists v, 
      heapOf s ||- rget s temp0 |-> v ** rget s temp1 |->__ ** rget s t4 |->__ ** top.

Definition scanPointerGuar s s' :=
  if beq_nat (rget s v0) 0 then
    scanNoCopyGuar (rset s rfree (rget s t4)) s'
  else
    exists s'',
      fwdObjGuar (rset (rset (rset s rfree (rget s t4)) a0
	(rget s temp1)) ra SCANCOPIED) s'' /\ 
      scanCopiedGuar s'' s'.

Definition scanFieldStOk phi W M F A s :=
  (set_card W) <= (set_card F) /\
  set_disj (set_union W M) (setRange (rget s rtoStart) (rget s rtoEnd)) /\
  set_eqv F (rangeObjs (rget s rfree) (rget s rtoEnd)) /\
  (forall x y, phi x = Some y -> 
    ~notSpacePtrP y (rget s rtoStart) (rget s rtoEnd)) /\
  rget s rtoStart <= rget s rfree <= rget s rtoEnd /\
  heapOf s ||- mapHp M phi ** objHp (set_union W M) W **
    (foralls x in F, x |->__) ** A.


Definition scanFieldPre s :=
  exists phi, exists W, exists M, exists F,
    scanFieldStOk phi W M F (okFieldAt (set_union W M) (rget s a0) ** top) s.

Definition scanFieldGuar s s' :=
  (forall phi W M F A,
    scanFieldStOk phi W M F (okFieldAt (set_union W M) (rget s a0) ** A) s ->
    (exists M'', exists W', exists F', exists G, exists phi',
      set_eqv W (set_union M'' W') /\
      set_eqv F (set_union G F') /\
      iso_set phi' M'' G /\
      (forall x, set_In x G -> Heap.get (heapOf s') (rget s a0) = Some x) /\
      aligned8 (rget s rfree) (rget s' rfree) /\
      (forall x, set_In x W' -> heapOf s' ||- objCopied x s x ** top) /\
      scanFieldStOk (map_union phi phi') W' (set_union M M'') F'
      (fieldRespMap (map_union phi phi') s (rget s a0) **
	(foralls x in G, objCopiedPhi phi' s x) ** A) s')) /\
    regPres 
    (regNotInList (rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s'.


(* code heap spec *)

Definition Pscanf : chspec := fun l => 
  if beq_nat l RETURN then
    SomeT (bspeci returnPre returnGuar)
  else if beq_nat l SCANPOINTER then
    SomeT (bspeci scanPointerPre scanPointerGuar)
  else if beq_nat l NOTTOSPACEPTR then
    SomeT (bspeci isNotSpacePtrPre isNotToSpacePtrGuar)
  else if beq_nat l SCANCOPIED then
    SomeT (bspeci scanCopiedPre scanCopiedGuar)
  else if beq_nat l SCANNOCOPY then
    SomeT (bspeci scanNoCopyPre scanNoCopyGuar)
  else if beq_nat l FWDOBJ then
    SomeT (bspeci fwdObjPre fwdObjGuar)
  else if beq_nat l SCANFIELD then
    SomeT (bspeci scanFieldPre scanFieldGuar)
  else
    NoneT _.



Ltac PscanfSimpl := 
  unfold Pscanf; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.


(* lemmas related to the scanField invariant *)


Lemma phiWEq : forall phi W M F P s,
  scanFieldStOk phi W M F P s ->
  forall x, set_In x W -> phi x = None.
Proof.
  unfold scanFieldStOk.
  intuition.
  assert (~set_In x M).
  intro.
  hypReplace H5 (sconjImplL H5 _ (mapHpElim1 H7 (phi:=phi))).
  linLogElim H5.
  sconjLift2 H5.
  hypReplace H5 (sconjImplL H5 _ (objHpElim H3 (s1 := set_union W M))).
  sconjLift2 H5.
  unfold okObjAt in H5; unfold okFieldAt in H5.
  linLogElim H5.
  refine (sconjifierL (mapHpElim2 H7 (phi:=phi)) H5).
Qed.


Lemma scanFieldPtrW : forall phi W M F A s,
  scanFieldStOk phi W M F A s ->
  forall x, set_In x W ->
    ptrP x.
Proof.
  unfold scanFieldStOk.
  intros.

  assert (hget s x <> None).
  intuition.
  sconjLift2 H7.
  hypReplace H7 (sconjImplL H7 _ (objHpElim H0 (s1 := _))).
  unfold okObjAt in H7; unfold okFieldAt in H7; linLogElim H7.
  assert (X := contGetEqL H7).
  unfold hget in H1.
  rewrite H1 in X; discriminate X.
  unfold hget in H1; unfold Heap.get in H1.
  assert (U := isPtrOk x).
  destruct (isPtr x).
  assumption.
  elimtype False; apply H1; reflexivity.
Qed.


Lemma scanFieldGuarPresRa : forall s s',
  scanFieldPre s -> 
  scanFieldGuar s s' -> 
  rget s ra = rget s' ra.
Proof.
  unfold scanFieldPre; unfold scanFieldGuar.
  intros s s' _ [_ Rp].
  rewrite (regPresEq Rp).
  reflexivity.
  regNotInListSolver.
Qed.


Lemma scanFieldStOkTop : forall phi W M F P P' s,
  scanFieldStOk phi W M F (P ** P') s ->
  scanFieldStOk phi W M F (P ** top) s.
Proof.
  unfold scanFieldStOk.
  intuition.
  clear H; rename H5 into H.
  eapply sconjImplR.  
  apply H.
  clear H; intros h H.
  eapply sconjImplR.
  apply H.
  clear h H; intros h H.
  eapply sconjImplR.
  apply H.
  clear h H; intros h H.
  eapply sconjImplR.
  apply H.
  clear h H; intros h H.
  linLogIntro.
Qed.


Lemma scanFieldStOkImpl : forall phi W M F P s ,
  scanFieldStOk phi W M F P s ->
  forall (P':hpred), (forall h, h ||- P -> h ||- P') ->
  scanFieldStOk phi W M F P' s.
Proof.
  unfold scanFieldStOk.
  intuition.
  apply (sconjImplR H5); clear H5; intros h H5.
  apply (sconjImplR H5); clear h H5; intros h H5.
  apply (sconjImplR H5); clear h H5; intros h H5.
  eauto.
Qed.


Lemma scanFieldStOkWMDisj : forall phi W M F P s,
  scanFieldStOk phi W M F P s -> set_disj W M.
Proof.
  destruct s; unfold scanFieldStOk; simpl.
  unfold mapHp; unfold objHp.
  intuition.
  clear H0 H H1 H2 H4 H6.
  linLogElim H5.
  clear H1.
  apply set_disjSymm.
  apply (fasFasDisj H5); clear h H5 P; intros x h H.
  linLogElim H; linLogIntro.

  match goal with
    | H : ?h ||- sconj _ _ |- ?h ||- sconj _ _ =>
      idtac
      (*
      apply (sconjImpl H); try (clear H); try (clear h); 
	let h' := fresh "h" in
	  (intros h' H; sconjMatch)*)
  end.


  sconjMatch.
  unfold okObjAt in H; unfold okFieldAt in H; linLogElim H.
  linLogIntro.  
  sconjMatch.
Qed.


Lemma scanFieldStOkFGDisj : forall phi W M F B G A s,
  scanFieldStOk phi W M F (fas B G ** A) s -> 
  (forall h x, h ||- B x -> h ||- x |->_ ** top) ->
  set_disj F G.
Proof.
  unfold scanFieldStOk.
  intros phi W M F B G A.
  destruct s.
  intros [_ [_ [_ [_ [_ H]]]]] Imp.
  eapply fasFasDisj.
  assert (Q := H).
  sconjLift3 Q; apply (sconjImplR Q); clear Q; intros h' Q.
  sconjLift3 Q; apply (sconjImplR Q); clear h' Q; intros h' Q.
  apply Q.
  simpl; intros h' x H'; apply (sconjImplR H'); intros; linLogIntro.
  assumption.
Qed.
Implicit Arguments scanFieldStOkFGDisj [phi W M F B G A s].


Add Morphism scanFieldStOk : scanFieldStOkMorphism.
  unfold scanFieldStOk.
  intros u s1 s1' s2 s2' s3 s3' A s s' E1 E2 E3 Seqv.
  repeat (rewrite <- (rgetEqv Seqv)).
  intuition.

  rewrite <- (set_cardMorph E1).
  rewrite <- (set_cardMorph E3).
  assumption.

  setoid_rewrite <- E1.
  setoid_rewrite <- E2.
  assumption.

  setoid_rewrite <- E3.
  assumption.
  
  clear H H1 H2 H4 H6.
  apply (heapEqvTrue (stateHeapEqv (state_eqvSymm Seqv))).
  apply (sconjImpl H5); clear H5; intros h H.
  unfold hpredTrue.
  setoid_rewrite <- E2.
  assumption.

  sconjMatch.
  unfold hpredTrue.
  setoid_rewrite <- E1.
  setoid_rewrite <- E2.
  assumption.
  unfold hpredTrue.
  setoid_rewrite <- E3.
  assumption.

Qed.


Lemma scanFieldStOkPhiEqv : forall phi W M F A s,
  scanFieldStOk phi W M F A s ->
  forall phi', mapEqv phi phi' ->
    scanFieldStOk phi' W M F A s.
Proof.
  unfold scanFieldStOk.
  intuition.
  assert (Q := H2 x y).
  rewrite (H3 x) in Q; auto.
  clear H H1 H2 H4 H6; rename H3 into E; rename H5 into H.
  apply (sconjImplL H); clear s H; intros h H.
  exact (mapHpPhiEqv H E).
Qed.


Notation heap_eqv := Heap.heap_eqv.

Lemma scanFieldStOkExtract : forall phi W M F A B s,
  scanFieldStOk phi W M F (A ** B) s ->
  exists h, scanFieldStOk phi W M F (A ** heap_eqv h) s /\ h ||- B.
Proof.
  unfold scanFieldStOk; intuition.
  sconjLift5of5 H5.
  destruct (sconjExtract H5) as [h [H' HB]].
  exists h.
  clear H5.
  intuition.
  clear H0 H H1 H2 H4 H6 HB.
  hypReplace H' (sconjComm H'); linLogElim H'.
  assumption.
Qed.


Lemma scanPhiDom : forall phi W M F A s,
  scanFieldStOk phi W M F A s ->
  forall x, ~set_In x M -> 
    phi x = None.
Proof.
  unfold scanFieldStOk.
  intros phi W M F A s H x NI.
  intuition.
  assert (heapOf s ||- (!(phi x = None) ** top) ** top).
  apply (sconjImpl H5); clear H; intros h H.
  hypReplace H (mapHpElim2 NI H).
  linLogIntro.
  linLogIntro.
  linLogElim H3.
  assumption.
Qed.


(* block verification *)

Lemma scanNoCopyOk : iseqOkS Pscanf scanNoCopyPre scanNoCopyGuar scanNoCopy.
Proof.
  unfold scanNoCopyPre; unfold scanNoCopyGuar; unfold scanNoCopy.
  iseqWeakSolver PscanfSimpl.
  intros s [v H].
  rewrite (swStep H); try reflexivity.
  rewrite addiuStep.
  constructor.
  trivial.
  stSimpl; simpl.
  intros; subst.
  stSimpl.
  constructor.
  intros v0 h0 H0.
  simpl.
  apply (contUpdateSt H0).
  intro r; destruct r; stSimpl; reflexivity.
Qed.


Lemma scanCopiedOk : iseqOkS Pscanf scanCopiedPre scanCopiedGuar scanCopied.
Proof.
  unfold scanCopiedPre; unfold scanCopiedGuar; unfold scanCopied.
  iseqWeakSolver PscanfSimpl.
  intros s [v H].
  rewrite (swStep H); [idtac | reflexivity | reflexivity].
  rewrite addiuStep.
  constructor.
  trivial.
  intros; subst.
  stSimpl; simpl.
  constructor.
  intros v0 h0 H0.
  apply (contUpdateSt H0).
  intro r; destruct r; stSimpl; reflexivity.
Qed.


Lemma scanPointerOk : iseqOkS Pscanf scanPointerPre scanPointerGuar scanPointer.
Proof.
  unfold scanPointerPre; unfold scanPointerGuar; unfold scanPointer.
  iseqWeakSolver PscanfSimpl.

  unfold fwdObjGuar.
  intros s s' _ [_ [_ [_ H]]].
  apply H.
  regNotInListSolver.

  intros s H.
  compare (rget s v0) 0.

  intro R; rewrite R in H; rewrite beq_true in H.
  rewrite addiuStep.
  stSimpl.
  rewrite bteqEq.
  rewrite R; rewrite beq_true.
  simpl.
  constructor.
  unfold scanNoCopyPre.
  stSimpl.
  simpl.
  auto.
  auto.
  
  intro NeqV0.
  rewrite beq_false in H.
  2:assumption.
  destruct H as [v H].
  rewrite addiuStep.
  stSimpl.
  rewrite bteqEq.
  rewrite beq_false.
  2:assumption.
  rewrite addiuStep.
  unfold fwdObjPre; unfold fwdObjGuar.
  stSimpl.
  replace (0 + rget s temp1) with (rget s temp1).
  2:reflexivity.
  replace (0 + rget s t4) with (rget s t4).
  2:reflexivity.
  constructor.
  constructor.
  sconjLift2 H.
  eapply sconjImpl.
  apply H.
  auto.
  clear H; intros h H.
  sconjLift2 H.
  apply (sconjImplR H); intros; linLogIntro.

  intros s' [G [Feq [V0eq Cso]]].
  linLogElim H.
  simpl in G.
  assert (Gx := G (rget s temp0 |-> v ** top) x x0 x1 x2); clear G.
  exists v.
  replace (rget s' temp0) with (rget s temp0).

  assert (X : ((rget s temp1 |-> rget s t4 ** S (S (S (S (rget s temp1)))) |-> x0) **
    (rget s t4 |-> x ** S (S (S (S (rget s t4)))) |-> x0) **
    rget s temp0 |-> v ** top) (Heap.heap_canon (heapOf s'))).
  apply Gx.
  clear Gx Cso V0eq Feq.
  sconjGLift3.
  linLogIntro.
  assumption.

  clear H Gx.
  sconjLift3 X.
  sconjMatch.

  rewrite (regPresEq Cso temp0).
  stSimpl; reflexivity.
  regNotInListSolver.

  intros.
  destruct H0 as [s'' [[G [Feq [V0eq Cso]]] SCG]].
  exists s''.
  rewrite Feq.
  intuition.
Qed.


Section INSP.

Variable rstart rend : rnum.

Variable rstartOk : rstart <> v0.
Variable rendOk : rend <> v0.

Lemma isNotSpacePtrOk : iseqOkS Pscanf isNotSpacePtrPre 
  (isNotSpacePtrGuar rstart rend) (isNotSpacePtr rstart rend).
Proof.
  iseqWeakSolver PscanfSimpl.

  unfold isNotSpacePtrPre.
  unfold returnPre; unfold returnGuar.
  intros s _.
  rewrite andi1Step.
  stSimpl.
  rewrite sltuStep.
  stSimpl.
  rewrite addiuStep.
  stSimpl.
  rewrite sltuStep.
  stSimpl.
  rewrite addiuStep.
  stSimpl.
  repeat (rewrite btneEq).
  repeat (rewrite beqSBoolEq).
  repeat (rewrite Bool.negb_elim).
  
  assert (R : isOdd (rget s a0) = true \/ isOdd (rget s a0) = false).
  destruct (isOdd (rget s a0)); auto.
  destruct R.

  rewrite H.
  constructor.
  trivial.
  intros; subst.
  stSimpl; simpl.
  intuition.
  left.
  unfold atomP.
  assert (Q := isOddOk (rget s a0)).
  rewrite H in Q; auto.

  unfold regPres; intros; stSimpl; reflexivity.

  rewrite H.
  assert (rget s a0 < rget s rstart \/ rget s a0 >= rget s rstart).
  omega.
  destruct H0.
  rewrite blt_true.
  2:assumption.
  constructor.
  trivial.
  intros; subst.
  stSimpl; simpl.
  intuition.
  right.
  intro.
  assert (X := setRangeElim H1).
  omega.

  unfold regPres; intros; stSimpl; reflexivity.

  rewrite blt_false.
  2:assumption.
  assert (rget s rend < 1 + rget s a0  \/ rget s rend >= 1 + rget s a0).
  omega.
  destruct H1.
  rewrite blt_true.
  2:assumption.
  constructor.
  trivial.
  intros; subst.
  stSimpl; simpl.
  constructor.
  2:intuition.
  right.
  intros Y.
  assert (X := setRangeElim Y).
  omega.

  unfold regPres; intros; stSimpl; reflexivity.

  rewrite blt_false.
  2:assumption.
  constructor.
  trivial.
  intros; subst.
  stSimpl; simpl.
  intuition.
  destruct H2.
  unfold atomP in H2.
  assert (Q := isOddOk (rget s a0)).  
  rewrite H in Q.
  eapply not_even_and_odd; eauto.
  apply H2; clear H2.
  apply setRangeIntro.
  omega.

  unfold regPres; intros; stSimpl; reflexivity.

Qed.

End INSP.


Lemma isNotToSpacePtrOk : iseqOkS Pscanf isNotSpacePtrPre isNotToSpacePtrGuar isNotToSpacePtr.
Proof.
  unfold isNotToSpacePtrGuar; unfold isNotToSpacePtr.
  apply isNotSpacePtrOk; discriminate.
Qed.


Lemma fwdObjOk : iseqOkS Pscanf fwdObjPre fwdObjGuar fwdObj.
Proof.
  unfold fwdObjPre; unfold fwdObjGuar; unfold fwdObj.
  iseqWeakSolver PscanfSimpl.

  intros s H.
  simpl in H.
  linLogElim H.
  assert (QQ := H).
  rewrite (lwStep H); try reflexivity.
  sconjLift3 H.
  rewrite (swStep H); try reflexivity.
  stSimpl.
  unfold WORDSIZE.
  assert (Z := contUpdateSt H x); clear H.
  assert (Q := Z); sconjLift3 Q.
  rewrite (lwStep Q); try reflexivity.
  clear Q.
  sconjLift4 Z.
  rewrite (swStep Z); try reflexivity.
  stSimpl.
  assert (Z2 := contUpdateSt Z x0).
  sconjLift3 Z2.
  rewrite (swStep Z2); try reflexivity.
  assert (ZQ := contUpdateSt Z (rget s rfree)).
  clear Z Z2.
  rewrite addiuStep.
  rewrite addiuStep.
  constructor.
  trivial.

  stSimpl.
  intros.
  subst s'.
  constructor.

  clear ZQ.
  simpl.
  intros.
  linLogIntro.
  linLogElim H.

  assert (R1 := contContEq QQ H).
  assert (x0 = y).
  sconjLift2 QQ; sconjLift2 H.
  exact (contContEq QQ H).
  assert (x1 = x').
  sconjLift3 QQ; sconjLift3 H.
  exact (contContEq QQ H).
  assert (x2 = y').
  sconjLift4 QQ; sconjLift4 H. 
  exact (contContEq QQ H).
  subst.

  sconjLift3 H.
  hypReplace H (contUpdateSt H x3).
  sconjLift4 H.
  hypReplace H (contUpdateSt H y).
  sconjLift3 H.
  hypReplace H (contUpdateSt H (rget s rfree)).
  apply (sconjImplR H); clear H; intros h H.
  sconjLift3 H.
  apply (sconjImplR H); clear h H; intros h H.
  sconjLift2 H.
  assumption.

  stSimpl.
  intuition.

  unfold regPres.
  intros; simpl in H.
  intuition.
  stSimpl.
  reflexivity.

Qed.




(* lower level version of scanFieldGuar that 
    should be easier to verify scanField using *)
Definition scanFieldGuar0 s s' :=
  (forall phi W M F A,
    scanFieldStOk phi W M F (okFieldAt (set_union W M) (rget s a0) ** A) s ->
    ((rget s rfree = rget s' rfree /\
      (forall x, set_In x W -> (heapOf s' ||- objCopied x s x ** top)) /\
      scanFieldStOk phi W M F 
      (fieldRespMap phi s (rget s a0) ** A) s') \/
     (8 + rget s rfree = rget s' rfree /\
       exists v, heapOf s ||- rget s a0 |-> v ** top /\ set_In v W /\
	 (forall x, set_In x (set_remove v W) -> (heapOf s' ||- objCopied x s x ** top)) /\
	 scanFieldStOk (map_add phi v (rget s rfree))
         (set_remove v W) (set_add v M) 
         (set_remove (rget s rfree) F) 
         (rget s a0 |-> rget s rfree ** 
	   objCopied v s (rget s rfree) ** A) s'))) /\
  regPres 
  (regNotInList (rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s'.


(* scanFieldGuar0 is stronger than scanFieldGuar *)
Lemma scanFieldGuar0Stronger : forall s s',
  scanFieldGuar0 s s' ->
  scanFieldGuar s s'.
Proof.
  intros s s' [X Rp].
  constructor.
  intros phi W M F A H.
  assert (Y := X _ _ _ _ _ H); clear X; rename H into UU.

  assert (RM : set_eqv (set_union M empty_set) M).
  setoid_replace (set_union M empty_set) with (set_union empty_set M).
  apply set_eqvUnionSymm.
  apply set_eqvSymm; apply set_eqvUnionEmptyUnit.

  destruct Y.

  destruct H as [FrEq H].
  exists empty_set; exists W; exists F; exists empty_set;
    exists (fun n:nat => None (A := nat)).
  constructor.
  apply set_eqvUnionEmptyUnit.
  constructor.
  apply set_eqvUnionEmptyUnit.
  constructor.
  unfold iso_set.
  simpl.
  intuition.
  discriminate H.
  discriminate H.
  constructor.
  simpl.
  intros; elimtype False; assumption.
  unfold scanFieldStOk in H; unfold scanFieldStOk.
  intuition.
  rewrite FrEq; apply aligned8Refl.
  
  setoid_rewrite RM.
  assumption.
  apply (H3 x y).
  unfold map_union in H4.
  destruct (phi x); assumption.
  assumption.

  clear H0 H H1 H5 H7.
  apply (sconjImpl H6); clear H6; intros h H.
  unfold hpredTrue.
  setoid_rewrite RM.
  rewrite <- hpredTrueEq.
  unfold mapHp; unfold mapHp in H.
  linLogElim H.
  linLogIntro.
  unfold map_union.
  intros x y Z.
  apply (H4 x y).
  destruct (phi x); assumption.
  eapply fasImpl.
  apply H.
  simpl.
  clear h H; intros x h H.
  linLogElim H.
  existsl x0.
  linLogIntro.
  unfold map_union; destruct (phi x); assumption.
  eauto.
  sconjMatch.
  unfold hpredTrue; setoid_rewrite RM; assumption.
  unfold fieldRespMap in Hp |- * .
  linLogElim Hp.
  destruct H1 as [v [Heq Meq]].
  linLogIntro.
  exists v; unfold map_union.
  constructor.
  assumption.
  unfold mapX in Meq |- * .
  destruct (isOdd v).
  eauto.
  rewrite Meq; reflexivity.
  assumption.
  assert (h ||- unit ** A).
  linLogIntro; assumption.
  sconjMatch.

  (* other case *)
  destruct H as [FrEq [v [vEq [vIn [Y0 H]]]]].
  exists (set_add v empty_set); exists (set_remove v W).
  exists (set_remove (rget s rfree) F); exists (set_add (rget s rfree) empty_set).
  exists (fun x => if beq_nat x v then Some (rget s rfree) else None).

  assert (Pvn := phiWEq UU _ vIn).
  assert (vptr := scanFieldPtrW UU _ vIn).

  constructor.
  simpl; apply set_eqvSymm; apply addRemoveEqv; assumption.
  constructor.

  assert (I : set_In (rget s rfree) F).
  unfold scanFieldStOk in UU; intuition.
  setoid_rewrite H1.
  clear RM H2 H.
  destruct (minInObjsRange (rget s rfree) (rget s rtoEnd)).
  assert (set_card W = 0).
  replace (set_card F) with 0 in H0.
  omega.
  assert (set_card (rangeObjs (rget s rfree) (rget s rtoEnd)) = 0).
  rewrite H; reflexivity.
  rewrite <- H2.
  apply set_cardMorph.
  apply set_eqvSymm; assumption.
  assert (Q := cardZeroIsEmpty_set W H2).
  assert (set_In v empty_set).
  setoid_rewrite <- Q; assumption.
  elimtype False; exact H4.
  assumption.

  simpl; apply set_eqvSymm; apply addRemoveEqv; assumption.

  constructor.
  unfold iso_set.
  simpl; intuition.
  subst; rewrite beq_true in H1; discriminate H1.
  compare x v; intro R.
  subst; left; reflexivity.
  rewrite (beq_false R) in H0.
  right; apply H0; reflexivity.
  compare x v; intro R.
  subst; rewrite beq_true in H0; injection H0; clear H0; intro; subst.
  left; reflexivity.
  rewrite beq_false in H0; [discriminate H0 | assumption].

  compare x v; intro R.
  compare x' v; intro R'.
  rewrite R; rewrite R'; reflexivity.
  rewrite beq_false in H1; [discriminate H1 | assumption].
  rewrite beq_false in H0; [discriminate H0 | assumption].
  exists v; rewrite H1; rewrite beq_true; reflexivity.
  2:assumption.

  constructor.
  simpl.
  intros x [Xeq | Fa].
  subst x.
  destruct H as [_ X].
  intuition.
  sconjLift4 H4.
  eapply contGetEqL; eauto.
  elimtype False; assumption.

  unfold scanFieldStOk in H |- * .

  assert (MvEq : set_eqv (set_union M (set_add v empty_set)) (set_add v M)).
  setoid_rewrite (set_eqvUnionSymm M (set_add v empty_set)).
  simpl; apply set_eqvRefl.

  intuition.

  rewrite <- FrEq; exists 1; omega.

  setoid_rewrite MvEq; assumption.

  clear H6 H H1 Rp.
  unfold map_union in H3.
  apply (H2 x y); unfold FnMap.map_add.
  compare v x; intro R.
  subst v; rewrite Pvn in H3.
  rewrite beq_true in H3; rewrite beq_true.
  assumption.
  rewrite beq_false in H3.
  rewrite beq_false.
  destruct (phi x); assumption.
  assumption.
  auto.
  assumption.

  clear Y0 UU FrEq H0 H H1 H2 H4.
  
  apply (sconjImpl H5); clear H5; intros h H.
  unfold hpredTrue; setoid_rewrite MvEq.
  rewrite <- hpredTrueEq; unfold mapHp in H |- * .
  linLogElim H.
  linLogIntro.
  intros.
  unfold map_union in H0.
  compare x v; intro R.
  rewrite R; apply set_add_intro1; reflexivity.
  apply (H2 x y).
  rewrite beq_false in H0.
  unfold FnMap.map_add.
  rewrite beq_false.
  destruct (phi x); assumption.
  auto.
  assumption.

  eapply fasImpl.
  apply H.
  clear h H; intros x h H.
  simpl in H; linLogElim H.
  existsl x0; linLogIntro.
  unfold FnMap.map_add in H3; unfold map_union.
  compare x v; intro R.
  subst v; rewrite beq_true; rewrite beq_true in H3.
  rewrite Pvn; assumption.
  rewrite beq_false; rewrite beq_false in H3.
  rewrite H3; reflexivity.
  auto.
  assumption.
  auto.
  eauto.

  sconjMatch.
  unfold hpredTrue; setoid_rewrite MvEq; assumption.

  unfold fieldRespMap.
  linLogIntro.
  exists v.

  constructor.
  assumption.

  unfold map_union; unfold mapX.
  rewrite beq_true.
  rewrite Pvn.
  replace (isOdd v) with false.
  reflexivity.
  assert (Ev := isOddOk v).
  destruct (isOdd v).
  clear Rp; elimtype False.
  eapply not_even_and_odd.
  2:apply Ev.
  apply ptrPeven.
  assert (X := isPtrOk v).
  assumption.
  reflexivity.

  assumption.

  apply fasIntro.
  hypReplace Hp (unitIR Hp).  
  sconjMatch.
  unfold objCopiedPhi; linLogIntro.
  rewrite beq_true.
  reflexivity.
  assumption.
  simpl; auto.
Qed.


Lemma scanFieldOk : iseqOkS Pscanf scanFieldPre scanFieldGuar scanField.
Proof.
  assert (iseqOkS Pscanf scanFieldPre scanFieldGuar0 scanField).
  unfold scanFieldPre; unfold scanFieldGuar; unfold scanField.
  iseqWeakSolver PscanfSimpl.

  (* ra is preserved *)
  intros s s' _ [_ [_ C]].
  rewrite C.
  reflexivity.
  discriminate.
  
  intros s [phi [W [M [F [Swf [Disj [Feq [PhiOk [FrG H]]]]]]]]].
  unfold okFieldAt in H.
  linLogElim H; sconjLift4 H.
  rewrite (lwStep H); try reflexivity.
  rewrite andi1Step.
  stSimpl.
  rewrite btneEq; rewrite beqSBoolEq; rewrite Bool.negb_elim.
  assert (OE := isOddOk x).
  destruct (isOdd x).

  (* field is not a pointer *)
  constructor.
  unfold returnPre; trivial.

  unfold returnGuar; unfold scanFieldGuar0.
  intros s' R.
  rewrite <- R; clear s' R.
  unfold scanFieldStOk.
  constructor.
  intros phi' W' M' F' A.
  intros [Swf' [Ephi' [Feq' [PhiOk' [FrG' H']]]]].

  left.
  stSimpl.
  intuition.

  rename x0 into x'; rename H4 into I.
  clear H0.
  unfold objCopied; unfold fieldCopied.
  sconjLift2 H'.
  assert (heapOf s ||- (x' |->__ ** top) ** top).
  assert (X := H').
  sconjMatch.
  hypReplace Hp (objHpElim I Hp).
  sconjMatch.
  unfold okObjAt in Hp; unfold okFieldAt in Hp; linLogElim Hp.
  linLogIntro; eauto.
  linLogElim H0.
  linLogIntro.
  sconjMatch.
  eauto.
  sconjLift2 H0; sconjMatch.
  assert (X := H').
  unfold okFieldAt in X; linLogElim X.
  assert (x0 = x).
  sconjLift4 X.
  apply (contContEq X H).
  subst.
  unfold fieldRespMap.
  linLogIntro.
  eauto.
  exists x; constructor.
  sconjMatch.

  assert (Q := isOddOk x).
  unfold mapX.
  destruct (isOdd x).
  reflexivity.
  elimtype False; eapply not_even_and_odd; eauto.

  unfold regPres.
  simpl.
  intuition.
  stSimpl.
  reflexivity.

  (* field is a pointer *)

  rewrite addiuStep.
  stSimpl.
  unfold okFieldVal in H10.
  destruct H10.
  unfold atomP in H0; elimtype False; eapply not_even_and_odd; eauto.

  destruct (set_union_elim H0); clear H0.

  (* x in W *)
  sconjLift3 H.
  assert (Q := objHpElimField _ H H1).
  unfold okFieldAt in Q.
  linLogElim Q.
  rewrite (lwStep Q).
  2:stSimpl; reflexivity.
  2:stSimpl; reflexivity.

  assert (NSP0 : notSpacePtrP x0 (rget s rtoStart) (rget s rtoEnd)).
  destruct H5.
  left; assumption.
  right.
  intro I.
  apply (set_disj_elim Disj _ H0 I).

  rewrite addiuStep.
  rewrite addiuStep.
  stSimpl; simpl.

  (* precondition *)
  constructor.
  unfold isNotSpacePtrPre.
  stSimpl.
  constructor.
  trivial.
  unfold isNotToSpacePtrGuar.
  stSimpl; simpl.
  intros s' [Bok [Heq Cok]].
  unfold scanPointerPre.
  unfold scanNoCopyPre.
  replace (rget s' temp0) with (rget s a0).
  rewrite <- Heq.
  clear Heq.
  compare (rget s' v0) 0.

  intro R.
  rewrite R in Bok; rewrite R; simpl in Bok.
  elimtype False; exact (Bok NSP0).

  intro Ne.
  rewrite beq_false.
  2:assumption.
  destruct (rget s' v0); simpl in Bok.
  elimtype False; apply Ne; reflexivity.
  clear Ne.
  exists x.
  sconjLift2 H.
  replace (rget s' temp1) with x.
  replace (rget s' t4) with (rget s rfree).
  apply (sconjImplR H); clear H; intros h0 H.
  clear Cok.

  assert (h0 ||- (x |->__ ** top) ** (rget s rfree |->__ ** top) ** top).
  sconjLift3 H; sconjLift2 H.
  apply (sconjImpl H); clear h0 H; intros h H.
  assert (X := objHpElim H1 H).
  apply (sconjImpl X).
  unfold okObjAt; unfold okFieldAt.
  intros h1 HH; linLogElim HH.
  linLogIntro; eauto.
  intros; linLogIntro.
  sconjMatch.

  assert (set_In (rget s rfree) F).

  assert (set_card F <> 0).
  assert (set_card W <> 0).
  intro Z.
  assert (T := cardZeroIsEmpty_set _ Z).
  exact (empty_set_elim (set_eqv_elim1 T _ H1)).
  omega.
  destruct (minInObjsRange (rget s rfree) (rget s rtoEnd)).
  rewrite H0 in Feq.
  rewrite (set_cardMorph Feq) in H.
  elimtype False; exact (H (empty_setCard)).
  hypReplace H0 (set_eqv_elim2 Feq _ H0).
  assumption.
  assert (HF := fasElim H Hp).
  sconjMatch.
  clear H Disj.
  hypReplace H0 (sconjAssocL H0).
  sconjLift3 H0; hypReplace H0 (sconjAssocL H0).
  sconjLift3 H0.
  sconjMatch.

  rewrite (regPresEq Cok t4); [reflexivity | discriminate].
  rewrite (regPresEq Cok temp1); [reflexivity | discriminate].
  rewrite (regPresEq Cok temp0); [reflexivity | discriminate].

  (* guarantee *)
  unfold isNotToSpacePtrGuar; unfold scanPointerGuar.
  stSimpl; simpl.

  intros s3 [s' [[SB [Heq Cso]] HB]].
  compare (rget s' v0) 0.

  intro R; rewrite R in SB; rewrite R in HB.
  rewrite beq_true in HB; simpl in SB.
  elimtype False; exact (SB NSP0).

  intro Ne.
  rewrite beq_false in HB.
  2:assumption.
  destruct (rget s' v0).
  elimtype False; apply Ne; reflexivity.
  clear Ne; simpl in SB; clear NSP0.

  generalize dependent HB.
  unfold fwdObjGuar; unfold scanCopiedGuar.
  stSimpl.
  assert (Rr := Cso temp1); simpl in Rr; rewrite <- Rr; clear Rr; stSimpl.
  assert (Rr := Cso t4); simpl in Rr; rewrite <- Rr; clear Rr; stSimpl.
  2:discriminate.
  2:discriminate.
  intros [s2 [[F3 [FrEq [v0Eq Cso']]] [Himp Req]]].
  unfold scanFieldGuar0.

  constructor.
  intros phi' W' M' F' A.
  intros H'.
  destruct H' as [WFle' [WMdisj' [F'eq [NSP' [[RTle RTle'] H']]]]].

  rewrite <- Heq in F3; clear Heq.

  unfold okFieldAt in H'.
  linLogElim H'.
  assert (x1 = x).
  sconjLift4 H'.
  sconjLift2 H.
  apply (contContEq H' H).
  subst.
  destruct H12.
  elimtype False; unfold atomP in H; eapply not_even_and_odd; eauto.
  destruct (set_union_elim H0) as [IW' | IM']; clear H0.

  destruct (minInObjsRange (rget s rfree) (rget s rtoEnd)).
  elimtype False.
  assert (set_card F' <> 0).
  assert (set_card W' <> 0).
  intro Z.
  assert (T := cardZeroIsEmpty_set _ Z).
  exact (empty_set_elim (set_eqv_elim1 T _ IW')).
  omega.
  rewrite H0 in F'eq.
  rewrite (set_cardMorph F'eq) in H2.
  elimtype False; exact (H2 (empty_setCard)).

  assert (IF' : set_In (rget s rfree) F').
  setoid_rewrite F'eq.
  assumption.

  assert (Hp1 : heapOf s ||- mapHp M' phi' **
        ((lexists v0, lexists v1, x |->, (v0::v1::nil)) **
	  objHp (set_union W' M') (set_remove x W')) **
        ((rget s rfree |->__) ** (foralls x in (set_remove (rget s rfree) F'), x |->__)) **
        rget s a0 |-> x ** A).
  apply (sconjImplR H'); clear H; intros h H.
  apply (sconjImpl H); clear h H; intros h H.
  hypReplace H (objHpElim IW' H).
  unfold okObjAt in H; unfold okFieldAt in H; linLogElim H.
  simpl; linLogIntro; eauto.
  sconjMatch.
  hypReplace Hp (fasElim IF' Hp); assumption.

  linLogElim Hp1; clear H' H.
  assert (Hp1' : heapOf s ||- 
    (mapHp M' phi' **
      objHp (set_union W' M') (set_remove x W') **
      (foralls x in set_remove (rget s rfree) F', x |->__) **
      rget s a0 |-> x ** A) **
    x |->, (x1 :: x2 :: nil) **
      rget s rfree |->, (x3::x4::nil)).
  linLogIntro.
  sconjStepX.

  rename Hp1 into Hp0.
  destruct (sconjExtract Hp1') as [h0 [Hp1 HA]]; clear Hp1'.
  hypReplace Hp1 (sconjAssocL (sconjComm Hp1)).
  assert (Hp2 := F3 _ _ _ _ _ Hp1); clear F3.

  generalize dependent Himp.
  rewrite (regPresEq Cso' temp0); stSimpl.
  rewrite (regPresEq Cso temp0); stSimpl.
  intro.
  2:discriminate.
  2:regNotInListSolver.

  sconjLift4 HA; hypReplace HA (sconjComm HA).
  destruct (sconjExtract HA) as [h1 [HA' HB]].
  hypReplace HA' (sconjComm HA'); clear HA.
  sconjLift3of3 Hp2; hypReplace Hp2 (sconjInsert Hp2 HA').
  hypReplace Hp2 (sconjAssocL Hp2).
  assert (Hp3 := Himp _ _ Hp2); clear Himp Hp2.

  (* now I've reflected the behavior into the invariants *)

  right.
  unfold OBJSIZE in FrEq; rewrite <- FrEq.
  constructor. 
  rewrite (Req rfree); stSimpl; reflexivity.

  exists x.
  constructor.
  clear Hp1 Q.
  sconjStepX.

  constructor.
  assumption.
  constructor.
  intros x' I'.
  unfold objCopied; unfold fieldCopied.
  assert (h1 ||- x' |->__ ** top).
  assert (h1 ||- (x' |->__ ** top) ** top).
  sconjLift2 HB; apply (sconjImpl HB).
  intros h H; hypReplace H (objHpElim I' H).
  apply (sconjImpl H); clear h H; intros h H.
  unfold okObjAt in H; unfold okFieldAt in H;
    linLogElim H; linLogIntro.
  eauto.
  linLogIntro.
  intros; linLogIntro.
  apply (sconjImplR (sconjAssocL H)); intros; linLogIntro.
  linLogElim H.
  hypReplace HA' (sconjComm HA').
  sconjLift3of3 Hp1; hypReplace Hp1 (sconjInsert Hp1 HA').
  hypReplace Hp1 (sconjInsert (sconjAssocL Hp1) H).
  sconjLift2 Hp3; hypReplace Hp3 (sconjInsert Hp3 H).
  linLogElim Hp1.
  linLogElim Hp3.
  clear Q.
  linLogIntro.
  sconjMatch.
  sconjMatch.
  sconjLift2 Hp1; sconjMatch.

  unfold scanFieldStOk.
  rewrite Req; rewrite Req; rewrite Req; stSimpl.
  rewrite FrEq; rewrite (regPresEq Cso').
  rewrite (regPresEq Cso').
  stSimpl.
  2:regNotInListSolver.
  2:regNotInListSolver.
  rewrite (regPresEq Cso).
  rewrite (regPresEq Cso).
  stSimpl.
  2:discriminate.
  2:discriminate.
  clear Cso Cso' Req.
  intuition.

  rewrite <- (set_card_remove _ _ IW') in WFle'.
  rewrite <- (set_card_remove _ _ IF') in WFle'.
  omega.

  setoid_rewrite <- (unionRemoveAddEqv _ _ M' IW').
  assumption.

  setoid_rewrite F'eq.
  apply rangeRemoveEqv.
  unfold FnMap.map_add in H3.
  compare x x5; intro R.
  subst x5; rewrite beq_true in H3.
  injection H3; intros; subst.
  destruct H4.
  assert (HH : s |||- (!(ptrP (rget s rfree)) ** top) ** top).
  sconjLift5 Hp0; apply (sconjImpl Hp0); intros hx HX.
  linLogIntro; exact (contElim1 HX).
  linLogIntro.
  linLogElim HH.
  eapply not_even_and_odd.
  eapply ptrPeven; apply H8.
  eauto.
  apply H4; clear H4.
  destruct (rangeObjsElim H0) as [Hz [k Hz']].
  eapply setRangeIntro.
  omega.

  rewrite beq_false in H3.
  apply (NSP' x5 y); assumption.
  assumption.

  destruct (rangeObjsElim H0) as [D [k D']].
  omega.

  rewrite v0Eq in Hp3.
  clear Swf Disj Feq PhiOk W M F phi H1 H5 Q SB FrEq v0Eq.
  sconjLift2 Hp3; hypReplace Hp3 (sconjInsert Hp3 HB).
  clear H H2 HA' HB Hp1 h0 h1.
  linLogElim Hp3; sconjLift2 Hp3; sconjGLift2.
  apply (sconjImpl Hp3); clear Hp3; intros h H.
  unfold hpredTrue; setoid_rewrite <- (unionRemoveAddEqv _ _ M' IW').
  assumption.
  linLogIntro.
  sconjStepX.
  sconjLift5of5 Hp.
  hypReplace Hp (sconjAssocR Hp).
  apply sconjComm; sconjMatch.
  unfold objCopied; unfold fieldCopied.
  linLogIntro.
  sconjStepX.
  eauto.
  sconjStepX.
  apply (mapHpIntro Hp).

  elimtype False.
  hypReplace H' (sconjImplL H' _ (mapHpElim1 IM' (phi:=phi'))).
  linLogElim H'.
  assert (R := contContEq H' Q).
  rewrite R in H6; rewrite R in H'; clear R x1.
  exact (NSP' _ _ H6 SB).

  (* regPres solving *)
  unfold regPres.
  simpl; intuition.
  rewrite Req.
  compare r ra; intro R.
  rewrite R; stSimpl.
  rewrite (regPresEq Cso'); stSimpl.
  rewrite (regPresEq Cso); stSimpl.
  reflexivity.
  discriminate.
  regNotInListSolver.
  stSimpl.
  rewrite (regPresEq Cso'); stSimpl.
  rewrite (regPresEq Cso); stSimpl.
  reflexivity.
  assumption.
  regNotInListSolver.

  (* final case *)
  sconjLift2 H.
  assert (Mok := sconjImplL H _ (mapHpElim1 H1 (phi:=phi))).
  linLogElim Mok.
  rewrite (lwStep Mok); try reflexivity.
  rewrite addiuStep.
  rewrite addiuStep.
  rewrite addiuStep.
  stSimpl.
  simpl.

  unfold isNotSpacePtrPre; unfold isNotToSpacePtrGuar.
  stSimpl.
  constructor.
  constructor.
  trivial.
  intros s' [Spv0 [Heq Cso]].
  unfold scanPointerPre.
  compare (rget s' v0) 0.

  intro v0Eq.
  rewrite v0Eq; rewrite v0Eq in Spv0; rewrite beq_true; simpl in Spv0.
  unfold scanNoCopyPre.
  rewrite <- Heq.
  rewrite (regPresEq Cso temp0).
  2:discriminate.
  stSimpl.
  exists x.
  sconjLift2 H.
  apply (sconjImplR H); intros; linLogIntro.

  (* case: rget s' v0 <> 0 *)
  intros Ne.
  elimtype False.
  destruct (rget s' v0).
  elimtype False; apply Ne; reflexivity.
  clear Ne; simpl in Spv0.
  clear Cso Heq.
  exact (PhiOk _ _  H5 Spv0).


  (* post condition *)
  simpl.
  unfold scanPointerGuar.
  intros s'' [s' [[SB [Heq Cso]] SPG]].
  generalize dependent SPG.
  unfold scanNoCopyGuar; unfold fwdObjGuar; unfold scanFieldGuar.
  stSimpl.
  rewrite (regPresEq Cso temp0); [stSimpl | discriminate].
  rewrite (regPresEq Cso temp1); [stSimpl | discriminate].
  rewrite (regPresEq Cso t4); [stSimpl | discriminate].
  rewrite (regPresEq Cso t7); [stSimpl | discriminate].

  generalize dependent SB.
  rewrite <- Heq; clear Heq.
  compare (rget s' v0) 0; intros Eq; unfold sbool.

  rewrite Eq; clear Eq.
  rewrite beq_true.
  intros NSP [Himp Req].
  unfold scanFieldGuar0.

  constructor.
  intros phi' W' M' F' A H'.
  left.

  constructor.
  rewrite Req; stSimpl; reflexivity.

  constructor.
  intros x' I'.
  unfold scanFieldStOk in H'; intuition.
  clear H H0 H2 H3 H6 H4 H7 H9 H11; rename H10 into H.
  unfold okFieldAt in H; linLogElim H; sconjLift4 H.
  assert (H'' : heapOf s ||- rget s a0 |-> x1 ** 
    (((lexists v1, lexists v2, x' |->, (v1::v2::nil)) ** top) ** top)).
  apply (sconjImplR H); clear H; intros h H.
  sconjLift2 H; apply (sconjImpl H); clear h H; intros h H.
  hypReplace H (objHpElim I' H).
  apply (sconjImpl H); clear h H.
  unfold okObjAt; unfold okFieldAt; intros h H; linLogElim H.
  simpl; linLogIntro; eauto.
  intros; linLogIntro.
  linLogIntro.
  linLogElim H''.
  assert (H2 := Himp _ _ H''); clear H Cso Mok PhiOk Feq Disj Himp.
  unfold objCopied; unfold fieldCopied.
  linLogIntro.
  sconjStepX.
  sconjStepX.
  sconjStepX.

  unfold scanFieldStOk in H' |- * .
  rewrite Req; rewrite Req; rewrite Req; clear Req; stSimpl.
  destruct s'' as [h2 rr]; simpl in Himp; simpl.
  generalize dependent Himp.
  rewrite (regPresEq Cso).
  rewrite (regPresEq Cso).
  rewrite (regPresEq Cso).
  stSimpl.
  2:discriminate.
  2:discriminate.
  2:discriminate.

  intro; clear Cso s'.
  intuition.
  unfold okFieldAt in H10.
  linLogElim H10.
  sconjLift2 Mok; sconjLift4 H10; assert (R := contContEq H10 Mok).
  rewrite R in H21; rewrite R in H10; clear x1 R.
  destruct H21.
  elimtype False; unfold atomP in H8; eapply not_even_and_odd; eauto.
  destruct (set_union_elim H8) as [I' | I']; clear H8.

  (* case: set_In x W'.   contradiction. *)
  elimtype False.
  sconjLift3 H10.
  assert (heapOf s ||- (okObjAt (set_union W' M') x ** top) ** top).
  apply (sconjImpl H10); clear H10 Mok H; intros h H.
  hypReplace H (objHpElim I' H).
  apply (sconjImplR H); intros; linLogIntro.
  linLogIntro.
  unfold okObjAt in H8; unfold okFieldAt in H8; linLogElim H8.
  sconjLift2 Mok.
  assert (R := contContEq H8 Mok); subst.
  apply NSP.
  destruct H16.
  left; assumption.
  right.
  intro Ni; exact (set_disj_elim H6 _ H12 Ni).

  assert (X := Himp _ _ H10); clear Himp.
  clear H0 H2 H3 H6 H4 H7 H9 H11 Swf Disj Feq PhiOk H1 H.
  sconjLift2 X; apply (sconjImplR X); clear X; intros h H.
  sconjLift2 H; apply (sconjImplR H); clear h H; intros h H.
  sconjLift2 H.
  apply (sconjImplR H); clear h H; intros h H.
  apply (sconjImplL H); clear h H; intros h H.
  unfold fieldRespMap.
  existsl x0; linLogIntro.
  exists x.
  constructor.
  apply (sconjImplR H10); intros; linLogIntro.
  unfold mapX.
  assert (O := isOddOk x).
  destruct (isOdd x).
  elimtype False; eapply not_even_and_odd; eauto.
  clear O.
  sconjLift2 H10.
  assert (heapOf s ||- ((lexists v, !(phi' x = Some v) ** x |-> v ** top) ** top)).
  apply (sconjImpl H10); clear h H H10; intros h H.
  hypReplace H (mapHpElim1 I' H).
  linLogElim H.
  existsl x1; linLogIntro.
  apply H.
  linLogIntro.
  linLogElim H0.
  sconjLift2 Mok.
  rewrite (contContEq Mok H0); assumption.
  assumption.

  (* regPres solving *)
  unfold regPres.
  simpl; intuition.
  rewrite Req.
  compare r ra; intro R.
  rewrite R; stSimpl.
  reflexivity.
  stSimpl.
  rewrite (regPresEq Cso); stSimpl.
  reflexivity.
  assumption.

  destruct (rget s' v0).
  elimtype False; apply Eq; reflexivity.
  clear Eq.
  intro U; elimtype False.
  exact (PhiOk _ _ H5 U).

  (* showing that the implementation spec matches the abstract spec *)
  apply (iseqOkWeak _ _ _ _ H); clear H.
  intros s H.
  constructor.
  assumption.
  intros s' H'.
  exact (scanFieldGuar0Stronger H').

Qed.

