Require Import MarkSweepGC.

(** ** Local specification map *)

Definition sweepP f :=
  if EqNat.beq_nat SWEEPLOOP f then
    SomeT (bspeci sweepLoopPre sweepLoopGuar)
    else if EqNat.beq_nat SWEEPADD f then
      SomeT (bspeci sweepAddPre sweepAddGuar)
      else if EqNat.beq_nat SWEEPNEXT f then
	SomeT (bspeci sweepNextPre sweepNextGuar)
	else if EqNat.beq_nat SWEEPFIN f then
	  SomeT (bspeci sweepFinPre sweepFinGuar)
	  else
	    NoneT _.

Ltac PsimplSweep := 
  unfold sweepP; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

(** ** Well-formedness Theorems *)

Theorem sweepImplOk : iseqOkS sweepP sweepPre sweepGuar sweepImpl.
Proof.
  unfold sweepImpl.
  iseqWeakSolver PsimplSweep.

(*   unfold sweepPre; unfold sweepGuar;  *)
(*     unfold sweepLoopPre; unfold sweepLoopGuar;  *)
(*       unfold sweepPreP; unfold sweepLoopInv. *)
  intros s [B [W [ohpSt [ohpEnd [freePtr [
    stBot [stBuff [Hst [Hed [Hroot [Hobjs
       Hh]]]]]]]]]]].
  
  assert (Hh' := sconjAssocL (sconjAssocL (
    sconjComm (sconjAssocR (sconjAssocR Hh))))).
  rewrite (lwStep Hh'); stSimpl; auto.
  hypReplace Hh' (sconjAssocL (sconjAssocL (
    sconjComm Hh'))).
  rewrite addiuStep; stSimpl; auto.
  rewrite (lwStep Hh'); stSimpl; auto.
  rewrite addiuStep; stSimpl; auto.
  clear Hh'.

  split.
  exists B; exists W; exists (rangeMSObjs ohpSt ohpEnd); exists empty_set;
    exists empty_set; exists ohpSt; exists freePtr;
      exists stBot; exists stBuff.
  unfold sweepLoopInv; stSimpl.
  rewrite <- set_unionEmpty.
  repeat (split;trivial).
  apply set_diffUnit.
  unfold rangeMSObjs.
  replace (7 + (4 + ohpSt) - ohpSt) with 11.
  compute; reflexivity.
  omega.
  exists 0;omega.
  intros a Ha; inversion_clear Ha.
  rewrite <- set_unionEmpty.
  apply (sconjImpl Hh); clear Hh; intros h Hh.
  apply bobjHdrsIntroGe; trivial.
  intros.
  assert (set_In x (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  destruct (rangeMSObjsElim _ _ _ H0) as [_ [k Hk]].
  omega.
  apply sconjAssocL; apply sconjAssocL; apply sconjAssocL.
  apply (sconjImplR (sconjAssocR (sconjAssocR (
    sconjAssocR Hh)))).
  clear h Hh; intros h Hh.
  simpl.
  linLogIntroOld.
  apply Hh.

  assert (Hh' := sconjAssocL (sconjAssocL (sconjComm 
      (sconjAssocR (sconjAssocR Hh))))).
  assert (exists tempSt, Heap.get (heapOf s) 
    (rget s rgcInfo) = Some tempSt /\ tempSt = ohpSt).
  exists ohpSt.
  split;trivial.
  eapply contGetEqL.
  apply Hh'.
  assert (exists tempEnd, Heap.get (heapOf s) 
    (4 + rget s rgcInfo) = Some tempEnd /\ tempEnd = ohpEnd).
  exists ohpEnd.
  split;trivial.
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hh'))).
  destruct H as [tempSt [Htst Hest]].
  subst ohpSt.
  destruct H0 as [tempEnd [Htend Heend]].
  subst ohpEnd.

  clear B W freePtr stBot stBuff
    Hst Hed Hroot Hobjs Hh Hh'.

  intros s' [G' [Gcall [Gcst Groot]]].
  split.
  intros.
  destruct H as [Hst [Hed [Hroot [Hobjs  Hh]]]].

  assert (Hh' := sconjAssocL (sconjAssocL (sconjComm 
    (sconjAssocR (sconjAssocR Hh))))).
  assert (tempSt = ohpSt).
  assert (Heap.get (heapOf s) (rget s rgcInfo) = Some ohpSt). 
  eapply contGetEqL.
  apply Hh'.
  rewrite H in Htst;inversion_clear Htst;reflexivity.
  assert (tempEnd = ohpEnd).
  assert (Heap.get (heapOf s) (4 + (rget s rgcInfo)) = Some ohpEnd). 
  eapply contGetEqL.
  apply (sconjAssocL (sconjAssocL (sconjComm Hh'))).  
  rewrite H0 in Htend;inversion_clear Htend;reflexivity.
  subst tempEnd; subst tempSt; clear Htst Htend.

  apply (G' B W (rangeMSObjs ohpSt ohpEnd) empty_set empty_set 
    ohpSt freePtr stBot stBuff);clear G'.
  unfold sweepLoopInv; stSimpl.
  rewrite <- set_unionEmpty.
  repeat (split;trivial).
  apply set_diffUnit.
  unfold rangeMSObjs.
  replace (7 + (4 + ohpSt) - ohpSt) with 11.
  compute; reflexivity.
  omega.
  exists 0;omega.
  intros a Ha; inversion_clear Ha.
  rewrite <- set_unionEmpty.
  apply (sconjImpl Hh); clear Hh; intros h Hh.
  apply bobjHdrsIntroGe; trivial.
  intros.
  assert (set_In x (rangeMSObjs ohpSt ohpEnd)).
  elimSetIn.
  destruct (rangeMSObjsElim _ _ _ H0) as [_ [k Hk]].
  omega.
  apply sconjAssocL; apply sconjAssocL; apply sconjAssocL.
  apply (sconjImplR (sconjAssocR (sconjAssocR (
    sconjAssocR Hh)))).
  clear h Hh; intros h Hh.
  simpl.
  linLogIntroOld.
  apply Hh.

  split.
  eapply calleeSavedOkTrans.
  2: apply Gcall.
  calleeSavedOkSolver.
  split;assumption.
Qed.

Theorem sweepFinImplOk : 
  iseqOkS sweepP sweepFinPre sweepFinGuar sweepFinImpl.
Proof.
  unfold sweepFinImpl.
  iseqWeakSolver PsimplSweep.

(*   unfold sweepFinPre; unfold sweepFinGuar;  *)
(*     unfold sweepFinPreP; unfold sweepPost. *)
  intros s [B [objs [free [ohpSt [freePtr [stBot [
    stBuff [Hst [Hed [Hroot [Hobjs [
      He Hh]]]]]]]]]]]].

  assert (Hh' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL (sconjAssocL 
	(sconjComm (sconjAssocR Hh)))))))))).
  
  rewrite (swStep Hh');stSimpl;auto.

  split;[tauto | idtac].
  intros s' Hs';subst s';stSimpl.

  clear B objs free ohpSt freePtr stBot stBuff 
    Hst Hed Hroot Hobjs He Hh Hh'.

  split. 
  intros.
  destruct H as [Hst [Hed [Hroot [Hobjs [He Hh]]]]].
  exists free; exists (rget s rfree).
  subst objs.
  repeat (split;trivial).
  
  assert (Hh' := sconjAssocL (sconjAssocL 
    (sconjComm (sconjAssocL (sconjAssocL 
      (sconjComm (sconjAssocL (sconjAssocL 
	(sconjComm (sconjAssocR Hh)))))))))).
  apply sconjAssocL; apply sconjComm; apply sconjAssocR.
  unfold contL.
  apply sconjAssocR; apply sconjComm; apply sconjAssocR;
    apply sconjAssocR; apply sconjComm; 
      apply sconjAssocR; apply sconjAssocR.
  simpl.
  eapply contUpdateSt; apply Hh'.

  split.
  calleeSavedOkSolver.
  split.
  reflexivity.
  split; reflexivity.
Qed.

Theorem sweepNextImplOk : 
  iseqOkS sweepP sweepNextPre sweepNextGuar sweepNextImpl.
Proof.
  unfold sweepNextImpl.
  iseqWeakSolver PsimplSweep.

(*   unfold sweepLoopPre; unfold sweepLoopGuar;  *)
(*     unfold sweepNextPre; unfold sweepNextGuar;  *)
(*       unfold sweepNextPreP; unfold sweepLoopInv. *)
  intros s [B [W [objs [free [swept [ohpSt [freePtr 
    [stBot [stBuff [Hst [Hed [Hroot [Hwf 
      [Hobjs [Hswpd [[k Hk] [Hsub [He Hh]]]]]]]]]]]]]]]]]].
  rewrite addiuStep; stSimpl.
  replace (7 + (12 + rget s rsweep)) with (19 + rget s rsweep);
    [split | omega].

  exists B; exists W; exists objs; exists free;
    exists swept; exists ohpSt; exists freePtr;
      exists stBot; exists stBuff.
  unfold sweepLoopInv; stSimpl.
  repeat (split; trivial).
  exists (S k); rewrite Hk; omega.

  clear B W objs free swept ohpSt freePtr stBot 
    stBuff Hst Hed Hroot Hwf Hobjs Hswpd 
    k Hk Hsub Hh He.
  intros s' [G [Gcall [Gcst1 Groot]]].
  
  split.
  intros.
  destruct H as [Hst [Hed [Hroot [Hwf 
      [Hobjs [Hswpd [[k Hk] [Hsub [He Hh]]]]]]]]].

  apply G with W objs free swept freePtr.
  unfold sweepLoopInv; stSimpl.
  repeat (split; trivial).
  exists (S k); rewrite Hk; omega.

  split.
  eapply calleeSavedOkTrans. 
  2: apply Gcall.
  calleeSavedOkSolver.
  split; assumption.
Qed.


Theorem sweepAddImplOk : 
  iseqOkS sweepP sweepAddPre sweepAddGuar sweepAddImpl.
Proof.
  unfold sweepAddImpl.
  iseqWeakSolver PsimplSweep.

(*   unfold sweepAddPre; unfold sweepAddGuar;  *)
(*     unfold sweepNextPre; unfold sweepNextGuar;  *)
(*       unfold sweepNextPreP; unfold sweepAddPreP. *)
  intros s [B [W [objs [free [swept [ohpSt [freePtr 
    [stBot [stBuff [Hst [Hed [Hroot [Hwf 
      [Hobjs [Hswpd [[k Hk] [Hsub [In [He Hh]]]]]]]]]]]]]]]]]]].
  hypReplace Hh (sconjAssocL (fasElimL  In (
    sconjAssocL (sconjComm (sconjAssocR (
      sconjAssocR (sconjAssocR (sconjAssocR Hh)))))))).
  unfold okObjAt in Hh; unfold okFieldAt in Hh.
  linLogElimOld Hh.
  rewrite (swStep (sconjAssocL (sconjComm Hh))); stSimpl; auto.
  rewrite addiuStep; stSimpl.

  split.
  exists B; exists (set_remove (rget s rsweep) W);
    exists objs; exists ((rget s rsweep) :: free);
      exists (rangeMSObjs ohpSt (19 + rget s rsweep));
	exists ohpSt; exists freePtr;
	  exists stBot; exists stBuff.

  unfold sweepNextPreP; stSimpl.
  replace (19 + rget s rsweep) with
    (12 + (7 + rget s rsweep)); [idtac | omega].
  rewrite <- (rangeMSObjsRemoveMax _ (7 + rget s rsweep) 
    (rget s rsweep) k Hk); [idtac | omega | omega].
  Opaque set_eqv.
  repeat (split; trivial).
  apply set_disj_intro.
  intros.
  destruct H0.
  subst x1.
  apply (set_remove_elim2 _ H).
  reflexivity.
  rewrite Hswpd in Hwf.
  apply (set_disj_elim Hwf x1); trivial.
  destruct (set_remove_elim _ _ _ H).
  apply H2.
  exists k; apply Hk.
  intros a Ha.
  elimSetIn.
  rewrite <- Hswpd.
  elimSetIn.
  rewrite <- He.
  Transparent set_eqv.
  split; intro; elimSetIn.

  apply sconjAssocL; apply sconjAssocL; apply sconjAssocL.
  apply sconjComm.
  assert (Hh' : (rget s rsweep |-> x **
    4 + rget s rsweep |-> (rget s rfree) **
      (foralls x in set_remove (rget s rsweep) W,
        (lexists y, !okFieldVal objs y ** x |-> y) **
        (lexists y, !okFieldVal objs y ** 4 + x |-> y)) ** 
      freeList (rget s rfree) free **
      (top && objHp B B) **
      bobjHdrs B (rget s rsweep) 0 1 **
      objHdrs (set_union W free) 0 **
      buffStackSet stBot stBot stBuff nil **
      rget s rgcInfo |->, (ohpSt :: rget s rhpEnd :: 
	freePtr :: stBot :: stBuff :: nil) ** top)
  (Heap.heap_canon (heapOf (uhset s (4 + rget s rsweep) 
    (rget s rfree))))).
  apply sconjComm; apply sconjAssocR.
  simpl.
  eapply contUpdateSt.
  apply (sconjAssocL (sconjComm Hh)).
  clear Hh.
  apply sconjAssocR.
  apply sconjComm.
  apply sconjAssocR.
  apply (sconjImpl (sconjAssocR (sconjAssocR (
    sconjAssocR (sconjAssocR Hh'))))); 
  clear Hh'; intros h Hh.
  apply sconjAssocL.
  apply (sconjImplL Hh).
  clear h Hh; intros h Hh.
  apply (sconjImplR (sconjAssocL (
    sconjComm (sconjAssocL Hh)))).
  clear h Hh; intros h Hh.
  simpl.
  existsl (rget s rfree).
  linLogIntroOld.
  omega.
  exists x.
  apply sconjAssocL; apply sconjComm; apply Hh.
  apply (sconjImplL (sconjAssocR 
    (sconjAssocR (sconjAssocR Hh))));
  clear h Hh; intros h Hh.
  apply (sconjImplL Hh).
  clear h Hh; intros h Hh.
  apply (sconjImplL Hh).
  clear h Hh; intros h Hh.
  assert (~ set_In (rget s rsweep) B).
  intro X.
  assert (Heap.get h (rget s rsweep - 4) = Some 1).
  hypReplace Hh (fasElimL  X Hh).
  simpl in Hh.
  rewrite blt_false in Hh.
  eapply contGetEqL; apply (sconjAssocL Hh).
  omega.
  assert (Heap.get h (rget s rsweep - 4) = Some 0).
  assert (I' : set_In (rget s rsweep) (set_union W free)).
  elimSetIn.
  hypReplace Hh (fasElimL  I' (sconjComm Hh)).
  eapply contGetEqL; apply (sconjAssocL Hh).
  rewrite H in H0; inversion_clear H0.
  apply (sconjImpl Hh); clear h Hh; intros h Hh.
  rewrite Hk.
  replace (12 + (ohpSt + 4 + 12 * k)) with
    (12 + (ohpSt + 4) + 12 * k); [idtac | omega].
  apply bobjHdrsStep; try (rewrite <- Hk); trivial.
  intros.
  assert (I : set_In x1 objs).
  elimSetIn.
  rewrite Hobjs in I.
  destruct (rangeMSObjsElim _ _ _ I).
  apply H2.
  unfold objHdrs in Hh; unfold objHdrs.
  eapply fasEqvSet.
  apply Hh. 
  split; intros; elimSetIn.

  clear B W objs free swept ohpSt freePtr stBot 
    stBuff Hst Hed Hroot Hwf Hobjs Hswpd 
    k Hk Hsub He Hh In H3 H6 x x0.
  intros s' [G [Gcall [Gcst1 Groot]]].
  
  split.
  intros.
  destruct H as [Hst [Hed [Hroot [Hwf 
      [Hobjs [Hswpd [[k Hk] [Hsub [In [He Hh]]]]]]]]]].

  apply G with (set_remove (rget s rsweep) W)
    objs ((rget s rsweep) :: free) 
    (rangeMSObjs ohpSt (19 + rget s rsweep)) freePtr; clear G.

  hypReplace Hh (sconjAssocL (fasElimL  In (
    sconjAssocL (sconjComm (sconjAssocR (sconjAssocR (
      sconjAssocR (sconjAssocR Hh)))))))).
  unfold okObjAt in Hh; unfold okFieldAt in Hh.
  linLogElimOld Hh.

  unfold sweepNextPreP; stSimpl.
  replace (19 + rget s rsweep) with
    (12 + (7 + rget s rsweep)); [idtac | omega].
  rewrite <- (rangeMSObjsRemoveMax _ (7 + rget s rsweep) 
    (rget s rsweep) k Hk); [idtac | omega | omega].
  Opaque set_eqv.
  repeat (split; trivial).
  apply set_disj_intro.
  intros.
  destruct H0.
  subst x1.
  apply (set_remove_elim2 _ H).
  reflexivity.
  rewrite Hswpd in Hwf.
  apply (set_disj_elim Hwf x1); trivial.
  destruct (set_remove_elim _ _ _ H).
  apply H2.
  exists k; apply Hk.
  intros a Ha.
  rewrite <- Hswpd.
  elimSetIn.
  rewrite <- He.
  Transparent set_eqv.
  split; intro; elimSetIn.
  apply sconjAssocL; apply sconjAssocL; apply sconjAssocL.
  apply sconjComm.
  assert (Hh' : (rget s rsweep |-> x **
    4 + rget s rsweep |-> (rget s rfree) **
      (foralls x in set_remove (rget s rsweep) W,
        (lexists y, !okFieldVal objs y ** x |-> y) **
        (lexists y, !okFieldVal objs y ** 4 + x |-> y)) **
      freeList (rget s rfree) free **
      (A && objHp B B) **
      bobjHdrs B (rget s rsweep) 0 1 **
      objHdrs (set_union W free) 0 **
      buffStackSet stBot stBot stBuff nil **
      rget s rgcInfo |->, (ohpSt :: rget s rhpEnd :: 
	freePtr :: stBot :: stBuff :: nil) ** A')
  (Heap.heap_canon (heapOf (uhset s (4 + rget s rsweep) 
    (rget s rfree))))).
  apply sconjComm; apply sconjAssocR.
  simpl.
  eapply contUpdateSt.
  apply (sconjAssocL (sconjComm Hh)).
  clear Hh.
  apply sconjAssocR.
  apply sconjComm.
  apply sconjAssocR.
  apply (sconjImpl (sconjAssocR (sconjAssocR (
    sconjAssocR (sconjAssocR Hh'))))); 
  clear Hh'; intros h Hh.
  apply sconjAssocL.
  apply (sconjImplL Hh).
  clear h Hh; intros h Hh.
  apply (sconjImplR (sconjAssocL (
    sconjComm (sconjAssocL Hh)))).
  clear h Hh; intros h Hh.
  simpl.
  existsl (rget s rfree).
  linLogIntroOld.
  omega.
  existsl x.
  apply sconjAssocL; apply sconjComm; apply Hh.
  apply (sconjImplL (sconjAssocR 
    (sconjAssocR (sconjAssocR Hh))));
  clear h Hh; intros h Hh.
  apply (sconjImplL Hh).
  clear h Hh; intros h Hh.
  apply (sconjImplL Hh).
  clear h Hh; intros h Hh.
  assert (~ set_In (rget s rsweep) B).
  intro X.
  assert (Heap.get h (rget s rsweep - 4) = Some 1).
  hypReplace Hh (fasElimL  X Hh).
  simpl in Hh.
  rewrite blt_false in Hh.
  eapply contGetEqL; apply (sconjAssocL Hh).
  omega.
  assert (Heap.get h (rget s rsweep - 4) = Some 0).
  assert (I' : set_In (rget s rsweep) (set_union W free)).
  elimSetIn.
  hypReplace Hh (fasElimL  I' (sconjComm Hh)).
  eapply contGetEqL; apply (sconjAssocL Hh).
  rewrite H in H0; inversion_clear H0.
  apply (sconjImpl Hh); clear h Hh; intros h Hh.
  rewrite Hk.
  replace (12 + (ohpSt + 4 + 12 * k)) with
    (12 + (ohpSt + 4) + 12 * k); [idtac | omega].
  apply bobjHdrsStep; try (rewrite <- Hk); trivial.
  intros.
  assert (I : set_In x1 objs).
  elimSetIn.
  rewrite Hobjs in I.
  destruct (rangeMSObjsElim _ _ _ I).
  apply H2.
  unfold objHdrs in Hh; unfold objHdrs.
  eapply fasEqvSet.
  apply Hh. 
  split; intro; elimSetIn.

  split.
  eapply calleeSavedOkTrans. 
  2: apply Gcall.
  calleeSavedOkSolver.
  split; assumption.
Qed.

Theorem sweepLoopImplOk : 
  iseqOkS sweepP sweepLoopPre sweepLoopGuar sweepLoopImpl.
Proof.
  unfold sweepLoopImpl.
  iseqWeakSolver PsimplSweep.

  intros s [B [W [objs [free [swept [ohpSt [freePtr 
    [stBot [stBuff [Hst [Hed [Hroot [Hwf 
      [Hobjs [Hswpd [[k Hk] [Hsub [He Hh]]]]]]]]]]]]]]]]]].
  rewrite addiuStep; stSimpl.
  rewrite sltuStep; stSimpl.
  doBtestSimpl.
  
  (* case 1 : hpEnd < sweep + 8, 
	sweep is not in objs,
	to sweepFin *)
  split.
  exists B; exists objs; exists free;
    exists ohpSt; exists freePtr;
      exists stBot; exists stBuff.
  assert (set_subset objs swept).
  subst objs; subst swept.
  apply rangeMSObjsSubSet.
  omega.
  assert (set_diff W swept = empty_set).
  apply set_diffSuperSet.
  apply set_subsetTrans with objs; trivial.
  intros a A.
  elimSetIn.
  assert (Bx : forall x, set_In x B -> x < (rget s rsweep)).
  intros.
  assert (set_In x objs).
  elimSetIn.
  rewrite Hobjs in H2.
  destruct (rangeMSObjsElim _ _ _ H2) as [X _].
  omega.
  assert (W = empty_set).
  rewrite Hwf in H0; inversion_clear H0.
  reflexivity.
  subst W; clear H H0.
  rewrite <- set_unionEmpty in He.
  repeat (split;trivial).
  eapply sconjImpl. 
  apply (sconjAssocR Hh).
  clear Hh; intros h Hh.
  simpl in Hh.
  unfold hpredTrue; setoid_rewrite <- He; rewrite <- hpredTrueEq.
  apply objHdrsMerge.
  apply (sconjImplL Hh); clear h Hh; intros h Hh.
  apply bobjHdrsElim with 1 (rget s rsweep); trivial.
  clear Hh; intros h Hh.
  apply sconjAssocL.
  apply (sconjImplR (sconjAssocR Hh)).
  clear h Hh; intros h Hh.
  apply (sconjImplR Hh);
    clear h Hh; intros h Hh.
  apply (unitEL Hh).

  clear B W objs free swept ohpSt freePtr stBot stBuff 
    Hst Hed Hroot Hwf Hobjs Hswpd k Hk Hsub He Hh.
  intros s' [G [Gcall [Gcst1 Groot]]].
  split.
  intros.
  apply G with objs free freePtr.

  destruct H as [Hst [Hed [Hroot [Hwf 
      [Hobjs [Hswpd [[k Hk] [Hsub [He Hh]]]]]]]]].

  assert (set_subset objs swept).
  subst objs; subst swept.
  apply rangeMSObjsSubSet.
  omega.
  assert (set_diff W swept = empty_set).
  apply set_diffSuperSet.
  apply set_subsetTrans with objs; trivial.
  intros a Aa.
  elimSetIn.
  assert (Bx : forall x, set_In x B -> x < (rget s rsweep)).
  intros.
  assert (set_In x objs).
  elimSetIn.
  rewrite Hobjs in H2.
  destruct (rangeMSObjsElim _ _ _ H2) as [X _].
  omega.
  assert (W = empty_set).
  rewrite Hwf in H0; inversion_clear H0.
  reflexivity.
  subst W; clear H H0.
  rewrite <- set_unionEmpty in He.
  repeat (split;trivial).
  eapply sconjImpl. 
  apply (sconjAssocR Hh).
  clear Hh; intros h Hh.
  simpl in Hh.
  unfold hpredTrue; setoid_rewrite <- He; rewrite <- hpredTrueEq.
  apply objHdrsMerge.
  apply (sconjImplL Hh); clear h Hh; intros h Hh.
  apply bobjHdrsElim with 1 (rget s rsweep); trivial.
  clear Hh; intros h Hh.
  apply sconjAssocL.
  apply (sconjImplR (sconjAssocR Hh)).
  clear h Hh; intros h Hh.
  apply (sconjImplR Hh);
    clear h Hh; intros h Hh.
  apply (unitEL Hh).

  split.
  eapply calleeSavedOkTrans. 
  2: apply Gcall.
  calleeSavedOkSolver.
  split; assumption.

  (* case 2 : hpEnd >= sweep + 8, 
	sweep is in objs *)

  rewrite addiuStep; stSimpl.
  rewrite subuStep; stSimpl.
  assert (set_In (rget s rsweep) objs).
  rewrite Hobjs.
  apply rangeMSObjsIntro with k; trivial.
  omega.
  assert (NI : ~ set_In (rget s rsweep) swept).
  rewrite Hswpd.
  intro I.
  destruct (rangeMSObjsElim _ _ _ I).
  omega.
 
  assert (In : heapOf s ||- (rget s rsweep - 4) |-> 0 ** top \/
    heapOf s ||- (rget s rsweep - 4) |-> 1 ** top).
  destruct (He (rget s rsweep)) as [_ Hx].
  destruct (set_union_elim (Hx H)) as [In | In]; 
    clear H Hx.
  destruct (set_union_elim In) as [In' | In']; 
    clear In.
  right.
  unfold objHdrs in Hh; unfold bobjHdrs in Hh.
  assert (Hh' := sconjAssocL (fasElimL  In' Hh)).
  simpl in Hh'.
  rewrite blt_false in Hh'.
  apply (sconjImplR Hh').
  intros; linLogIntroOld.
  omega.
  left.
  assert (In : set_In (rget s rsweep) (set_union W free)).
  elimSetIn.
  unfold objHdrs in Hh; unfold bobjHdrs in Hh.
  assert (Hh' := sconjAssocL (fasElimL  In 
    (sconjAssocL (sconjComm Hh)))).
  apply (sconjImplR Hh').
  intros; linLogIntroOld.
  hypReplace In (Hsub _ In).
  rewrite Hswpd in In.
  elimtype False.
  destruct (rangeMSObjsElim _ _ _ In).
  omega.
  
  split.
  
  destruct In as [Inw | Inb].

  (* case 2.1 : sweep in W *)
  rewrite (lwStep Inw); stSimpl; auto.
  unfold doBtest.
  rewrite beq_true; auto.
  exists B; exists W; exists objs; exists free;
    exists swept; exists ohpSt; exists freePtr;
	exists stBot; exists stBuff.
  repeat (split;trivial).
  exists k; apply Hk.
  assert (hget s (rget s rsweep - 4) = Some 0). 
  unfold hget; eapply contGetEqL; apply Inw.
  destruct (He (rget s rsweep)) as [_ Hx].
  destruct (set_union_elim (Hx H)) as [In | In]; 
    clear H Hx.
  destruct (set_union_elim In) as [In' | In']; 
    clear In.
  unfold objHdrs in Hh; unfold bobjHdrs in Hh.
  assert (Hh' := sconjAssocL (fasElimL  In' Hh)).
  simpl in Hh'.
  rewrite blt_false in Hh'.
  assert (hget s (rget s rsweep - 4) = Some 1). 
  unfold hget; eapply contGetEqL; apply Hh'.
  rewrite H0 in H; inversion_clear H.
  omega.
  apply In'.
  hypReplace In (Hsub _ In).
  rewrite Hswpd in In.
  elimtype False.
  destruct (rangeMSObjsElim _ _ _ In).
  omega.

  (* case 2.2 sweep in B *)
  rewrite (lwStep Inb); stSimpl; auto.
  unfold doBtest.
  rewrite beq_false; auto.
  rewrite (swStep Inb); stSimpl; auto.
  exists B; exists W; exists objs; exists free;
    exists (rangeMSObjs ohpSt (19 + rget s rsweep));
      exists ohpSt; exists freePtr;
	exists stBot; exists stBuff.
  assert (NI' : ~ set_In (rget s rsweep) W).
  intro I'.
  assert (hget s (rget s rsweep - 4) = Some 1).
  unfold hget.
  eapply contGetEqL; apply Inb.
  assert (hget s (rget s rsweep - 4) = Some 0).
  unfold hget.
  assert (I'' : set_In (rget s rsweep) 
    (set_union W free)).
  elimSetIn.
  assert (Hh'' := sconjAssocL (fasElimL  I'' 
    (sconjAssocL (sconjComm Hh)))).
  eapply contGetEqL; apply Hh''.
  rewrite H0 in H1; inversion_clear H1.
  assert (In : set_In (rget s rsweep) B).
  destruct (He (rget s rsweep)) as [_ Hx].
  destruct (set_union_elim (Hx H)) as [In | In]; 
    clear H Hx.
  destruct (set_union_elim In) as [In' | In']; 
    clear In.
  apply In'.
  assert (set_In (rget s rsweep) (set_union W free)).
  elimSetIn.
  unfold objHdrs in Hh; unfold bobjHdrs in Hh.
  assert (Hh' := sconjAssocL (fasElimL  H 
    (sconjAssocL (sconjComm Hh)))).
  assert (hget s (rget s rsweep - 4) = Some 1). 
  unfold hget; eapply contGetEqL; apply Inb.
  assert (hget s (rget s rsweep - 4) = Some 0). 
  unfold hget; eapply contGetEqL; apply Hh'.
  rewrite H0 in H1; inversion_clear H1.
  hypReplace In (Hsub _ In).
  rewrite Hswpd in In.
  elimtype False.
  destruct (rangeMSObjsElim _ _ _ In).
  omega.
  repeat (split;trivial).
  apply set_disj_intro.
  intros.
  eapply set_disj_elim.
  unfold set_disj; apply Hwf.
  apply H0.
  replace (19 + rget s rsweep) with
    (12 + (7 + rget s rsweep)) in H1.
  rewrite <- (rangeMSObjsRemoveMax _ (7 + rget s rsweep) 
    (rget s rsweep) k Hk) in H1.
  destruct H1.
  subst x; elimtype False; auto.
  rewrite Hswpd; apply H1.
  omega.
  omega.
  omega.
  exists k; apply Hk.
  eapply set_subsetTrans.
  apply Hsub.
  rewrite Hswpd; apply rangeMSObjsSubSet.
  omega.
  unfold objHdrs.
  simpl.
  assert (Hh' : Heap.uset (heapOf s) (rget s rsweep - 4) 0 ||-
    rget s rsweep - 4 |-> 0 **
      (foralls x in set_remove (rget s rsweep) B,
	x - 4 |-> (if blt_nat x (rget s rsweep) then 0 else 1)) **
      (foralls x in set_union W free, x - 4 |-> 0) **
      buffStackSet stBot stBot stBuff nil **
      rget s rgcInfo |->, (ohpSt :: rget s rhpEnd :: 
	freePtr :: stBot :: stBuff :: nil) ** top **
      objHp objs W **
      freeList (rget s rfree) free ** top && objHp B B).
  eapply contUpdateSt.
  apply (sconjAssocL (fasElimL  In Hh)).
  apply (sconjImplL (sconjAssocR Hh')).
  clear Hh' Hh; intros h Hh.
  unfold hpredTrue; setoid_replace B with ((rget s rsweep) :: 
    (set_remove (rget s rsweep) B)).
  apply addRemoveEqv; assumption.
  rewrite <- hpredTrueEq.
  unfold bobjHdrs.
  apply fasIntro.
  apply (sconjImpl Hh); clear h Hh; intros h Hh.
  stSimpl.
  rewrite blt_true; trivial.
  omega.
  stSimpl.
  rewrite Hk.
  replace (S (S (S (S (S (S (S (S (S (S (S (S 
    (ohpSt + 4 + 12 * k))))))))))))) with 
  (12 + (ohpSt + 4) + 12 * k).
  rewrite <- Hk.
  assert (Z := bobjHdrsStep); unfold bobjHdrs in Z.
  apply Z; clear Z.
  intro X.
  destruct (set_remove_elim _ _ _ X).
  apply H0; rewrite Hk; reflexivity.
  intros.
  destruct (set_remove_elim _ _ _ H0) as [_ X].
  assert (I : set_In x objs).
  elimSetIn.
  rewrite Hobjs in I.
  destruct (rangeMSObjsElim _ _ _ I).
  apply H2.
  rewrite <- Hk; apply Hh.
  omega.
  intro X.
  destruct (set_remove_elim _ _ _ X).
  apply H0; reflexivity.

  clear B W objs free swept ohpSt freePtr stBot stBuff 
    Hst Hed Hroot Hwf Hobjs Hswpd k Hk Hsub He Hh 
    NI H.

  destruct In as [Inw | Inb].

  (* case 2.1 : sweep in W *)
  rewrite (lwStep Inw); stSimpl; auto.
  unfold doBtest.
  rewrite beq_true; auto.
  intros s' [G [Gcall [Gcst1 Groot]]].
  
  split.
  intros.
  destruct H as [Hst [Hed [Hroot [Hwf 
    [Hobjs [Hswpd [[k Hk] [Hsub [He Hh]]]]]]]]].

  apply G with W objs free swept freePtr; clear G.

  assert (set_In (rget s rsweep) objs).
  rewrite Hobjs.
  apply rangeMSObjsIntro with k; trivial.
  omega.
  assert (NI : ~ set_In (rget s rsweep) swept).
  rewrite Hswpd.
  intro I.
  destruct (rangeMSObjsElim _ _ _ I).
  omega.

  repeat (split;trivial).
  exists k; apply Hk.
  assert (hget s (rget s rsweep - 4) = Some 0). 
  unfold hget; eapply contGetEqL; apply Inw.
  destruct (He (rget s rsweep)) as [_ Hx].
  destruct (set_union_elim (Hx H)) as [In | In]; 
    clear H Hx.
  destruct (set_union_elim In) as [In' | In']; 
    clear In.
  unfold objHdrs in Hh; unfold bobjHdrs in Hh.
  assert (Hh' := sconjAssocL (fasElimL  In' Hh)).
  simpl in Hh'.
  rewrite blt_false in Hh'.
  assert (hget s (rget s rsweep - 4) = Some 1). 
  unfold hget; eapply contGetEqL; apply Hh'.
  rewrite H0 in H; inversion_clear H.
  omega.
  apply In'.
  hypReplace In (Hsub _ In).
  rewrite Hswpd in In.
  elimtype False.
  destruct (rangeMSObjsElim _ _ _ In).
  omega.

  split.
  eapply calleeSavedOkTrans. 
  2: apply Gcall.
  calleeSavedOkSolver.
  split; assumption.

  (* case 2.2 sweep in B *)
  rewrite (lwStep Inb); stSimpl; auto.
  unfold doBtest.
  rewrite beq_false; auto.
  rewrite (swStep Inb); stSimpl; auto.
  intros s' [G [Gcall [Gcst1 Groot]]].
  
  split.
  intros.
  destruct H as [Hst [Hed [Hroot [Hwf 
    [Hobjs [Hswpd [[k Hk] [Hsub [He Hh]]]]]]]]].

  apply G with W objs free 
    (rangeMSObjs ohpSt (19 + rget s rsweep))
    freePtr; clear G.

  assert (set_In (rget s rsweep) objs).
  rewrite Hobjs.
  apply rangeMSObjsIntro with k; trivial.
  omega.
  assert (NI : ~ set_In (rget s rsweep) swept).
  rewrite Hswpd.
  intro I.
  destruct (rangeMSObjsElim _ _ _ I).
  omega.

  assert (NI' : ~ set_In (rget s rsweep) W).
  intro I'.
  assert (hget s (rget s rsweep - 4) = Some 1).
  unfold hget.
  eapply contGetEqL; apply Inb.
  assert (hget s (rget s rsweep - 4) = Some 0).
  unfold hget.
  assert (I'' : set_In (rget s rsweep) 
    (set_union W free)).
  elimSetIn.
  assert (Hh'' := sconjAssocL (fasElimL  I'' 
    (sconjAssocL (sconjComm Hh)))).
  eapply contGetEqL; apply Hh''.
  rewrite H0 in H1; inversion_clear H1.
  assert (In : set_In (rget s rsweep) B).
  destruct (He (rget s rsweep)) as [_ Hx].
  destruct (set_union_elim (Hx H)) as [In | In]; 
    clear H Hx.
  destruct (set_union_elim In) as [In' | In']; 
    clear In.
  apply In'.
  assert (set_In (rget s rsweep) (set_union W free)).
  elimSetIn.
  unfold objHdrs in Hh; unfold bobjHdrs in Hh.
  assert (Hh' := sconjAssocL (fasElimL  H 
    (sconjAssocL (sconjComm Hh)))).
  assert (hget s (rget s rsweep - 4) = Some 1). 
  unfold hget; eapply contGetEqL; apply Inb.
  assert (hget s (rget s rsweep - 4) = Some 0). 
  unfold hget; eapply contGetEqL; apply Hh'.
  rewrite H0 in H1; inversion_clear H1.
  hypReplace In (Hsub _ In).
  rewrite Hswpd in In.
  elimtype False.
  destruct (rangeMSObjsElim _ _ _ In).
  omega.
  repeat (split;trivial).
  apply set_disj_intro.
  intros.
  eapply set_disj_elim.
  unfold set_disj; apply Hwf.
  apply H0.
  replace (19 + rget s rsweep) with
    (12 + (7 + rget s rsweep)) in H1.
  rewrite <- (rangeMSObjsRemoveMax _ (7 + rget s rsweep) 
    (rget s rsweep) k Hk) in H1.
  destruct H1.
  subst x; elimtype False; auto.
  rewrite Hswpd; apply H1.
  omega.
  omega.
  omega.
  exists k; apply Hk.
  eapply set_subsetTrans.
  apply Hsub.
  rewrite Hswpd; apply rangeMSObjsSubSet.
  omega.
  unfold objHdrs.
  simpl.
  assert (Hh' : Heap.uset (heapOf s) (rget s rsweep - 4) 0 ||-
    rget s rsweep - 4 |-> 0 **
      (foralls x in set_remove (rget s rsweep) B,
	x - 4 |-> (if blt_nat x (rget s rsweep) then 0 else 1)) **
      (foralls x in set_union W free, x - 4 |-> 0) **
      buffStackSet stBot stBot stBuff nil **
      rget s rgcInfo |->, (ohpSt :: rget s rhpEnd :: 
	freePtr :: stBot :: stBuff :: nil) ** A' **
      objHp objs W **
      freeList (rget s rfree) free ** A && objHp B B).
  eapply contUpdateSt.
  apply (sconjAssocL (fasElimL  In Hh)).
  apply (sconjImplL (sconjAssocR Hh')).
  clear Hh' Hh; intros h Hh.
  unfold bobjHdrs.
  apply fasEqvSet with ((rget s rsweep) :: 
    (set_remove (rget s rsweep) B)).
  apply fasIntro.
  apply (sconjImpl Hh); clear h Hh; intros h Hh.
  stSimpl.
  rewrite blt_true; trivial.
  omega.
  stSimpl.
  rewrite Hk.
  replace (S (S (S (S (S (S (S (S (S (S (S (S 
    (ohpSt + 4 + 12 * k))))))))))))) with 
  (12 + (ohpSt + 4) + 12 * k).

  rewrite <- Hk.
  assert (Z := bobjHdrsStep); unfold bobjHdrs in Z.
  apply Z; clear Z.
  intro X.
  destruct (set_remove_elim _ _ _ X).
  apply H0; rewrite Hk; reflexivity.
  intros.
  destruct (set_remove_elim _ _ _ H0) as [_ X].
  assert (I : set_In x objs).
  elimSetIn.
  rewrite Hobjs in I.
  destruct (rangeMSObjsElim _ _ _ I).
  apply H2.
  rewrite <- Hk; apply Hh.
  omega.
  intro X.
  destruct (set_remove_elim _ _ _ X).
  apply H0; reflexivity.
  apply addRemoveEqv; assumption.

  split.
  eapply calleeSavedOkTrans. 
  2: apply Gcall.
  calleeSavedOkSolver.
  split; assumption.
Qed.
